//
// NameContext.C -- Implementation of a CORBA-like naming service.
//

#include "NameContext.h"

// import templates

#include <rts/HPCxx_Sync.cc>
#include <rts/HPCxx_GlobalPtr.cc>
#include <rts/hpcxx_register.cc>
#include <HPCxx_GlobalRefImpl.cc>
#include <invoke.cc>


// class variables

int NameContextStub::bindID_s = -1;
int NameContextStub::resolveID_s = -1;

// Construct the ORB and associated name server
NameContextImpl::NameContextImpl(int& beginID)
{
  if (NameContextStub::bindID_s < 0) {
    beginID = NameContextStub::registerMemberFunctions(beginID);
  }
}

// Binds the object to the given name.
int NameContextImpl::bind(const char* objectName, CORBA_Object_var obj)
{
  table_m[objectName] = obj;
  return 0;
}


// Binds the object to the given name.
int NameContextImpl::bind(const char* objectName, CORBA_Object_ptr obj_ptr)
{
  CORBA_Object_var obj(obj_ptr);
  table_m[objectName] = obj;
  return 0;
}


// Retrieves the object previously bound to the given name.
CORBA_Object_var NameContextImpl::resolve(const char* objectName)
{
  map<const char*, CORBA_Object_var, less<const char*> >::iterator itr;

  // table_m.find() doesn't work in a separate thread so just do a linear
  // search for now.

  for (itr = table_m.begin(); itr != table_m.end(); itr++) {
    if (strcmp(objectName, (*itr).first) == 0) {
      return (*itr).second;
    }
  }
  CORBA_Object_var co;
  return co;
}


// Binds the object to the given name.
int NameContextStub::bind(const char* objectName, CORBA_Object_ptr obj_ptr)
{
  int i;        // dummy return argument
  CORBA_Object_var obj(obj_ptr);
  hpcxx_minvoke((HPCxx_GlobalPtr<NameContextImpl> *) this, i,
		 bindID_s, objectName, obj);
  return 0;
}


// Retrieves the object previously bound to the given name.
CORBA_Object_var NameContextStub::resolve(const char* objectName)
{
  CORBA_Object_var co;
  hpcxx_minvoke((HPCxx_GlobalPtr<NameContextImpl> *) this, co, resolveID_s,
		objectName);
  return co;
}


// Registers member functions
int NameContextStub::registerMemberFunctions(int beginID)
{
  int (NameContextImpl::* pfbind) (const char*, CORBA_Object_var);
  pfbind = &NameContextImpl::bind;
  bindID_s    = hpcxx_register(pfbind, beginID++);

  resolveID_s = hpcxx_register(NameContextImpl::resolve, beginID++);
  return beginID;
}


//
// explicit instantiations
//

template class HPCxx_GlobalRefImpl<CORBA_Object_var>;

template class HPCxx_Sync<CORBA_Object_var>;

template class HPCxx_GlobalPtr<HPCxx_Sync<CORBA_Object_var> >;

template class HPCxx_GlobalRefImpl<HPCxx_Sync<CORBA_Object_var> >;

template class HPCxx_GlobalRefImpl<int>;

template class HPCxx_GlobalPtr<HPCxx_Sync<int> >;

template class HPCxx_GlobalRefImpl<HPCxx_Sync<int> >;

template hpcxx_id_t hpcxx_register(
    CORBA_Object_var (NameContextImpl::*)(const char*),int
  );

template hpcxx_id_t hpcxx_register(
    int (NameContextImpl::*)(const char*,CORBA_Object_var),int
  );

template int hpcxx_minvoke(
    HPCxx_GlobalPtr<NameContextImpl>*,CORBA_Object_var&,int,const char*&
  );

template int hpcxx_minvoke(
    HPCxx_GlobalPtr<NameContextImpl>*,int&,int,const char*,CORBA_Object_var
  );

template void hpcxx_pack(
    HPCxx_Buffer&,HPCxx_GlobalPtr<HPCxx_Sync<int> >*,int
  );

template void hpcxx_pack(
    HPCxx_Buffer&,HPCxx_GlobalPtr<HPCxx_Sync<CORBA_Object_var> >*,int
  );

template void hpcxx_pack(HPCxx_Buffer&,HPCxx_Sync<CORBA_Object_var>*,int);

template void hpcxx_pack(HPCxx_Buffer&,HPCxx_Sync<int>*,int);

template int HPCxx_Type<HPCxx_Sync<CORBA_Object_var> >::registrationID(void);

template int HPCxx_Type<HPCxx_Sync<int> >::registrationID(void);

template int HPCxx_Type<CORBA_Object_var>::registrationID(void);
