//
// NameContext.h -- Category declarations for a CORBA-like naming service.
//

#include <CORBA/CORBA.h>
#include <map.h>

#ifndef _NAME_SERVICE_H
#define _NAME_SERVICE_H

class NameContext;
typedef NameContext* NameContext_ptr;   // not sure what this should be


//-----------------------------------------------------------------------------
// NameContext - Interface for a naming service which binds names to objects
// for later retrieval.
//-----------------------------------------------------------------------------
class NameContext : public CORBA_Object {
 public:

  // Binds the object to the given name.
  virtual int bind(const char* objectName, CORBA_Object_ptr obj_ptr) = 0;

  // Retrieves the object previously bound to the given name.
  virtual CORBA_Object_var resolve(const char* objectName) = 0;

  virtual ~NameContext() { }
};


//-----------------------------------------------------------------------------
// NameContextImpl - Implementation class for a naming service.
//-----------------------------------------------------------------------------
class NameContextImpl : public NameContext {
 public:

  // Binds the object to the given name.
  virtual int bind(const char* objectName, CORBA_Object_var obj);

  // Binds the object to the given name.
  virtual int bind(const char* objectName, CORBA_Object_ptr obj_ptr);

  // Retrieves the object previously bound to the given name.
  virtual CORBA_Object_var resolve(const char* objectName);

 private:

  // map to hold the {name, value} pairs
  map<const char*, CORBA_Object_var, less<const char*> > table_m;

  // Constructor.
  NameContextImpl(int& beginID);

  friend class ORB;
};


//-----------------------------------------------------------------------------
// NameContextStub - Client stub class for a naming service.
//-----------------------------------------------------------------------------
class NameContextStub : public NameContext,
			public HPCxx_GlobalPtr<NameContextImpl> {
 public:

  // Binds the object to the given name.
  virtual int bind(const char* objectName, CORBA_Object_ptr obj_ptr);

  // Retrieves the object previously bound to the given name.
  virtual CORBA_Object_var resolve(const char* objectName);

  // Registers member functions
  static int registerMemberFunctions(int beginID);

 private:
  
  static int bindID_s;
  static int resolveID_s;

  // Constructs a naming service within the given context
  NameContextStub(const HPCxx_GlobalPtr<NameContextImpl> & gp)
                 : HPCxx_GlobalPtr<NameContextImpl>(gp) { }

  friend class ORB;		// needed to construct a stub
  friend class NameContextImpl;
};

// marshall/unmarshall prototypes

void hpcxx_pack(HPCxx_Buffer &buf, NameContext_ptr* nc, int count);
void hpcxx_unpack(HPCxx_Buffer &buf, NameContext_ptr* nc, int count);

#endif	// _NAME_SERVICE_H
