// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: hpcxx_rtsImpl.C,v 1.2 1998/06/24 21:43:58 khalsa Exp $
#define HPCXX_MAX_CHILDREN 100
#include <unistd.h>
#include <stream.h>
#include <libgen.h>
#include <nexus.h>
#include <assert.h>
#include <hpcxx_rts.h>
#include <hpcxx_rtsImpl.h>

#ifndef HPCXX_NO_THREADS
HPCxx_Context           *hpcxx_currentContext;
#endif
HPCxx_ContextID         *hpcxx_currentContextID;
static int hpcxx_currentPosition = -1;
static nexus_startpoint_t     hpcxx_startpoint_master; 

static nexus_endpoint_t  hpcxx_nexus_endpoint_;
static nexus_node_t     *hpcxx_nexus_nodes_;
static int               hpcxx_nexus_numNodes_;
static bool              hpcxx_nexus_is_master_context_ = (bool)0;
static HPCxx_ContextID **hpcxx_contextIDs_;


nexus_startpoint_t *hpcxx_get_sp(){ return &hpcxx_startpoint_master;}

//
// This table *must* be created in the same order as the consts
// declared in the header file!!!
//
nexus_handler_t hpcxx_nexusHandlerTable[HPCXX_NUM_HANDLERS] = {
  {NEXUS_HANDLER_TYPE_NON_THREADED,
   (nexus_handler_func_t)hpcxx_initHandler},
  {NEXUS_HANDLER_TYPE_NON_THREADED,
   (nexus_handler_func_t)hpcxx_barrierHandler},
  {NEXUS_HANDLER_TYPE_NON_THREADED,
   (nexus_handler_func_t)hpcxx_barrierOutHandler},
  {NEXUS_HANDLER_TYPE_NON_THREADED,
   (nexus_handler_func_t)hpcxx_invokeGFHandler},
  {NEXUS_HANDLER_TYPE_NON_THREADED,
   (nexus_handler_func_t)hpcxx_returnHandler},
  {NEXUS_HANDLER_TYPE_NON_THREADED,
   (nexus_handler_func_t)hpcxx_testHandler},
  {NEXUS_HANDLER_TYPE_NON_THREADED,
   (nexus_handler_func_t)hpcxx_invokeMFHandler},
  {NEXUS_HANDLER_TYPE_NON_THREADED,
   (nexus_handler_func_t)hpcxx_readHandler},
  {NEXUS_HANDLER_TYPE_NON_THREADED,
   (nexus_handler_func_t)hpcxx_writeHandler}
};

// a function called by the master context to send copies of
// the startpoints to each of the other contexts.
static void sendContextIDs(){
    int i, j, n;
    n = hpcxx_nexus_numNodes_;
    nexus_startpoint_t *theSPs = new nexus_startpoint_t[n];
    hpcxx_currentPosition = 0;
    nexus_buffer_t buffer;
    for(i = 0; i < n; i++)
         nexus_startpoint_copy(&(theSPs[i]), hpcxx_contextIDs_[i]);
    nexus_buffer_init(&buffer, +nexus_sizeof_int(1)+ 
                               nexus_sizeof_startpoint(theSPs,n),
                      0);
    nexus_put_int(&buffer, &n, 1);
    nexus_put_startpoint_transfer(&buffer, theSPs, n);

    for ( j = 1; j < hpcxx_nexus_numNodes_; j++){
        nexus_send_rsr(&buffer, hpcxx_contextIDs_[j] , HPCXX_INIT_HANDLER, 
                  NEXUS_FALSE, NEXUS_FALSE);
    }
}

static HPCxx_Sync<int> init_sync;

// Nexus Hanlder to bootstrap HPC++ by distributing the contextIDs
void hpcxx_initHandler(nexus_endpoint_t *endpoint,
                       nexus_buffer_t *buffer,
                       nexus_bool_t called_from_non_threaded_handler)
{
   int i, n;
   nexus_get_int(buffer, &n, 1);
   hpcxx_nexus_numNodes_ = n;
   nexus_startpoint_t *theSPs = new nexus_startpoint_t[n];
   nexus_get_startpoint(buffer, theSPs, n);
   hpcxx_contextIDs_ = new HPCxx_ContextID *[hpcxx_nexus_numNodes_];
   for(i = 0; i < n; i++) hpcxx_contextIDs_[i] = &(theSPs[i]);
   for(i = 0; i < n; i++)
     if (nexus_startpoint_equal(hpcxx_currentContextID, hpcxx_contextIDs_[i]))
       hpcxx_currentPosition = i;
   init_sync.write(n);
}

//
// This function is currently unimplemented yet defined by
// Nexus.  We'll fudge it be seeing whether or not the
// number of nodes returned by nexus_init is > 0
//
int nexus_master_context()
{
  return(hpcxx_nexus_is_master_context_);
}
// special nexus barrier code.

void hpcxx_send_unblock(nexus_startpoint_t *n, int i){
  nexus_buffer_t buffer;
  nexus_startpoint_t sp;
  nexus_startpoint_copy(&sp, n);
  nexus_buffer_init(&buffer, nexus_sizeof_int(1)+ 
		    nexus_sizeof_startpoint(&sp,1),
		    0);
  nexus_put_int(&buffer, &i, 1);
  nexus_put_startpoint_transfer(&buffer, &sp, 1);
  nexus_send_rsr(&buffer, hpcxx_contextIDs_[i] , HPCXX_BARRIER_IN_HANDLER, 
		 NEXUS_FALSE, NEXUS_FALSE);
  nexus_buffer_destroy(&buffer);
}

static HPCxx_SyncQ<int> from_parent;
static HPCxx_SyncQ<int> from_children[HPCXX_MAX_CHILDREN];
static nexus_startpoint_t parent;
static int child_id;

static void notify_parent(){
  nexus_buffer_t buffer2;
  nexus_buffer_init(&buffer2, nexus_sizeof_int(1), 0);
  nexus_put_int(&buffer2, &child_id, 1);
  //nexus_startpoint_t s;
  //nexus_startpoint_copy(&s, &parent);
  nexus_send_rsr(&buffer2, &parent, HPCXX_BARRIER_OUT_HANDLER,
                     NEXUS_FALSE, NEXUS_FALSE);
  nexus_buffer_destroy(&buffer2);
}

void hpcxx_barrierHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler)
{
  nexus_get_int(buffer, &child_id, 1);
  nexus_get_startpoint(buffer, &parent, 1);
  from_parent.write(child_id);
}

void hpcxx_barrierOutHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler)
{
  int x;
  nexus_get_int(buffer, &x,1);
  from_children[x].write(x);
}


void hpcxx_barrier(){
   int i,j;
   if(nexus_master_context()){
      for(i = 1; i < hpcxx_nexus_numNodes_; i++) 
	hpcxx_send_unblock( hpcxx_currentContextID, i);
      if(hpcxx_nexus_numNodes_ >=  HPCXX_MAX_CHILDREN)
	printf("too many nodes: HPCXX_MAX_CHILDREN=%d\n", HPCXX_MAX_CHILDREN);
      
      for(i = 1; i < hpcxx_nexus_numNodes_; i++) {
	from_children[i].read(j);
      }
      for(i = 1; i < hpcxx_nexus_numNodes_; i++) 
	hpcxx_send_unblock( hpcxx_currentContextID, i);
   } else {
      from_parent.read(j);
      notify_parent();
      from_parent.read(j);
      }
}
// end of barrier code.
void hpcxx_typeInit();

int hpcxx_init(int &argc, char **&argv, HPCxx_Group *&group) {
  
  hpcxx_typeInit();
  cout << "in init" << endl;
  nexus_init(&argc,
			 &argv,
			 "NEXUS_ARGS",
			 "nx",
			 NULL,
			 NULL,
			 NULL,
			 NULL,
			 &hpcxx_nexus_nodes_,
			 &hpcxx_nexus_numNodes_);


  nexus_start_nonblocking();
  //nexus_printf("after start non-blocking\n");

  //  printf("starting up hpc++.  num contexts = %d\n", hpcxx_nexus_numNodes_);
  // Nexus currently doesn't implement nexus_master_context();
  // so fudge it.  hpcxx_nexus_numNodes is only > 0 on master context.
  //
  if (hpcxx_nexus_numNodes_) hpcxx_nexus_is_master_context_ = true;

  // FIXME
  // need to propagate nodes/group to the other non-master contexts

  if (nexus_master_context()) {
	hpcxx_contextIDs_ = new HPCxx_ContextID *[hpcxx_nexus_numNodes_];

	for (int i = 0; i < hpcxx_nexus_numNodes_; i++) {
	  hpcxx_contextIDs_[i] = &(hpcxx_nexus_nodes_[i].startpoint);
	 }
	hpcxx_contextIDs_[0] = &hpcxx_startpoint_master; // added by dbg
        group = new HPCxx_Group(hpcxx_contextIDs_, 
				hpcxx_nexus_numNodes_, 
				HPCXX_GEN_LOCAL_GROUPID, "master");
	group->setLocalIndex(0);
	hpcxx_currentContextID = hpcxx_contextIDs_[0];
        sendContextIDs();
  } else {
    int x;
    hpcxx_currentContextID = &hpcxx_startpoint_master; // added by dbg
    init_sync.read(x);
    group = new HPCxx_Group(hpcxx_contextIDs_, 
			    hpcxx_nexus_numNodes_, 
			    HPCXX_GEN_LOCAL_GROUPID, "child");
    group->setLocalIndex(hpcxx_currentPosition);
  }

  hpcxx_barrier();
  return(HPCXX_SUCCESS);
}

int hpcxx_exit(HPCxx_Group*& group) 
{
  nexus_shutdown_nonexiting();
  _exit(1);
  return(0);
}




//***************************************************************************
// Nexus specific routines that must be implemented by the library
//

int NexusBoot(nexus_startpoint_t *sp)
{
  nexus_endpointattr_t attr;
  nexus_endpointattr_init(&attr);
  nexus_endpointattr_set_handler_table(&attr, hpcxx_nexusHandlerTable,
					HPCXX_NUM_HANDLERS);
  nexus_endpoint_init(&hpcxx_nexus_endpoint_, &attr);

  nexus_startpoint_bind(sp, &hpcxx_nexus_endpoint_);
  nexus_startpoint_bind(&hpcxx_startpoint_master, &hpcxx_nexus_endpoint_);
  nexus_endpointattr_destroy(&attr); 

  return NEXUS_SUCCESS;
}

int NexusAcquiredAsNode(nexus_startpoint_t *sp)
{
  return(NexusBoot(sp));
}

void NexusExit()
{
  // FIXME - should HPC++ need to do anything to shutdown here?
}

// two prototypes that should be in an include.
int hpcxx_runRF(void * gp, void *p, int funcID, HPCxx_Buffer &b);
int hpcxx_runResult(void * sync, int funcID, HPCxx_Buffer &b);
int hpcxx_runRead(void * sync, void *q, int typeID, int count, int offset);
int hpcxx_runWrite(void * sync, void *q, int typeID, int count, int offset,
HPCxx_Buffer &b);

// Nexus Handler for Remote (global) Function invocation

void hpcxx_invokeGFHandler(nexus_endpoint_t *endpoint,
                           nexus_buffer_t *buffer,
                           nexus_bool_t called_from_non_threaded_handler)
{
  int funcID, offset;
  nexus_startpoint_t reply_done_sp;

 // unpack the global pointer (without calling hpcxx_unpack), probably
 // because the type is not known at this point

 nexus_get_startpoint(buffer, &reply_done_sp, 1);
 nexus_get_int(buffer, &offset, 1);

 nexus_get_int(buffer, &funcID, 1);
 HPCxx_BufferImpl bi(*buffer);
 HPCxx_Buffer b(bi);
 void *q = (void *) nexus_endpoint_get_user_pointer(endpoint);
 // q = q+offset;  this is wrong.  the offset must be added to typed pointer.
 hpcxx_runRF((void *) &reply_done_sp, q,  funcID, b);
 nexus_startpoint_destroy(&reply_done_sp);  // shridhar added this ...
}

void hpcxx_returnHandler(nexus_endpoint_t *endpoint,
                           nexus_buffer_t *buffer,
                           nexus_bool_t called_from_non_threaded_handler)
{
 int funcID;
 void *d = (void *) nexus_endpoint_get_user_pointer(endpoint);
 nexus_get_int(buffer, &funcID, 1);
 HPCxx_BufferImpl bi(*buffer);
 HPCxx_Buffer b(bi);
 hpcxx_runResult(d, funcID, b);  
}


// Nexus Handler for Remote Member Function invocation
void hpcxx_invokeMFHandler(nexus_endpoint_t *endpoint,
                           nexus_buffer_t *buffer,
                           nexus_bool_t called_from_non_threaded_handler)
{
  assert(0);
  //  nexus_printf("in invoke handler\n");  
  int funcID;
  nexus_startpoint_t reply_done_sp;
  nexus_get_startpoint(buffer, &reply_done_sp, 1);
  //  int offset, gp_offset;
  //  nexus_get_int(buffer, &offset, 1);
  nexus_get_int(buffer, &funcID, 1);
  //  nexus_printf("in invoke handler id=%d\n", funcID);  
  HPCxx_BufferImpl bi(*buffer);
  HPCxx_Buffer b(bi);
  void *q = (void *) nexus_endpoint_get_user_pointer(endpoint);
  // q = q+offset;  this is wrong.  the offset must be added to typed pointer.
  hpcxx_runRF((void *) &reply_done_sp, q,  funcID, b);  
  //  DONE = 1;
}


// Nexus Handler for GlobalPointer reads
void hpcxx_readHandler(nexus_endpoint_t *endpoint,
				  nexus_buffer_t *buffer,
				  nexus_bool_t called_from_non_threaded_handler)
{
 int count, offset;
 int typeID;
 nexus_startpoint_t reply_done_sp;
 nexus_get_startpoint(buffer, &reply_done_sp, 1);
 nexus_get_int(buffer, &typeID, 1);
 nexus_get_int(buffer, &count, 1);
 nexus_get_int(buffer, &offset, 1);
 void *q = (void *) nexus_endpoint_get_user_pointer(endpoint);
 hpcxx_runRead((void *) &reply_done_sp, q, typeID, count, offset);
}

// Nexus Handler for GlobalPointer writes
void hpcxx_writeHandler(nexus_endpoint_t *endpoint,
				   nexus_buffer_t *buffer,
				   nexus_bool_t called_from_non_threaded_handler)
{
 int count, offset;
 int typeID;
 nexus_startpoint_t reply_done_sp;
 nexus_get_startpoint(buffer, &reply_done_sp, 1);
 nexus_get_int(buffer, &typeID, 1);
 nexus_get_int(buffer, &count, 1);
 nexus_get_int(buffer, &offset, 1);
 HPCxx_BufferImpl bi(*buffer);
 HPCxx_Buffer hpbuff(bi);
 void *q = (void *) nexus_endpoint_get_user_pointer(endpoint);
 hpcxx_runWrite((void *) &reply_done_sp, q, typeID, count, offset, hpbuff);
}


void hpcxx_poll()
{
  nexus_poll();
}

void *hpcxx_malloc(size_t size)
{
  return(nexus_malloc(size));
}

void hpcxx_free(void *ptr)
{
  if (ptr) nexus_free(ptr);
}

void hpcxx_msg(const char *msg)
{
  nexus_printf((char *)msg);
}

void hpcxx_sendGlobalFunction(HPCxx_Buffer &b, HPCxx_ContextID &target){
      nexus_send_rsr(b.impl_.buffer_, &target, HPCXX_INVOKE_GF_HANDLER, 
                  NEXUS_FALSE, NEXUS_FALSE);
}
void hpcxx_returnGlobalFunction(HPCxx_Buffer &b, void *target){
      nexus_send_rsr(b.impl_.buffer_, (HPCxx_ContextID *) target,
                    HPCXX_RETURN_HANDLER,  NEXUS_FALSE, NEXUS_FALSE);
}


void hpcxx_sendMemberFunction(HPCxx_Buffer &b, HPCxx_ContextID &target){
  //  nexus_printf("send\n");
  nexus_send_rsr(b.impl_.buffer_, &target, HPCXX_INVOKE_MF_HANDLER, 
		 NEXUS_FALSE, NEXUS_FALSE);
  //  nexus_printf("after send\n");  
}



void hpcxx_testHandler(nexus_endpoint_t *endpoint,
                           nexus_buffer_t *buffer,
                           nexus_bool_t called_from_non_threaded_handler)
{
    int i; 
	
    /* extracting int from buffer */
    nexus_get_int(buffer, &i, 1);

    /* printing received value */
    nexus_printf("Slave received %d\n", i);

    //    DONE =1;
}


