// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: HPCxx_GlobalRefImpl.cc,v 1.3 1998/06/30 21:55:28 rasmussn Exp $

#include <HPCxx_GlobalRefImpl.h>
#include <hpcxx_rtsImpl.h>
#ifndef HPCXX_NO_THREADS
#include <hpcxx_rts.h>
#endif

//The classes below are used to provide type registration key ids.
// Default Constructor
template <class C>
int HPCxx_Type<C>::registrationID()
{
  return C::registrationID();
}

template<class T> 
HPCxx_GlobalRefImpl<T>::HPCxx_GlobalRefImpl() 
  : localPtr_(NULL)
{
#ifdef HPCXX_NO_THREADS
  contextID_ = hpcxx_currentContextID; 
#else 
  // hpcxx_currentContext is NULL if hpcxx_init hasn't been called yet
  if (hpcxx_currentContext != NULL) {
    contextID_ = hpcxx_currentContext->getContextID();
  } else {
    contextID_ = NULL;
  }
#endif 
  GlobalEndpoint = NULL;
  offset_ = 0;
}

template<class T>
HPCxx_GlobalRefImpl<T>::HPCxx_GlobalRefImpl(T *ptr) 
  : localPtr_(ptr)
{
     nexus_endpointattr_t EpAttr ;
     GlobalEndpoint = new nexus_endpoint_t;
     contextID_ = new nexus_startpoint_t;
     nexus_endpointattr_init(&EpAttr);
     nexus_endpointattr_set_handler_table(&EpAttr,
			hpcxx_nexusHandlerTable,
                   sizeof(hpcxx_nexusHandlerTable)/sizeof(nexus_handler_t));
     nexus_endpoint_init(GlobalEndpoint, &EpAttr);
     nexus_endpointattr_destroy(&EpAttr);
     nexus_endpoint_set_user_pointer(GlobalEndpoint, (void *) ptr);
     nexus_startpoint_bind(contextID_, GlobalEndpoint);
     offset_ = 0;
}

template<class T> 
HPCxx_GlobalRefImpl<T>::HPCxx_GlobalRefImpl(T *ptr, HPCxx_ContextIDImpl *id)
  : localPtr_(ptr), contextID_(id)
{
 // contextID_ = new nexus_startpoint_t;
 // nexus_startpoint_copy(contextID_, id);
  GlobalEndpoint = NULL;
  offset_ = 0;
}

// Copy contructor
template<class T> HPCxx_GlobalRefImpl<T>::HPCxx_GlobalRefImpl(const HPCxx_GlobalRefImpl<T> &gr) 
  : localPtr_(gr.localPtr_), contextID_(gr.contextID_)
{
  //contextID_ = new nexus_startpoint_t;
  //nexus_startpoint_copy(contextID_, gr.contextID_);
  GlobalEndpoint = gr.GlobalEndpoint;
  offset_ = gr.offset_;
}

// Destructor  
template<class T>
HPCxx_GlobalRefImpl<T>::~HPCxx_GlobalRefImpl()
{
  // don't want to destroy the endpoint just because a global ptr goes
  // out of scope.
  // if(GlobalEndpoint)
  //   nexus_endpoint_destroy(GlobalEndpoint);
}
template<class T>
void HPCxx_GlobalRefImpl<T>::destroy() const
{
   if(GlobalEndpoint)
     nexus_endpoint_destroy(GlobalEndpoint);
}

// cast a GlobalRefImpl to type T
template<class T>
HPCxx_GlobalRefImpl<T>::operator T() const
{
  if (!isLocal()) {
	T tmp;
	read(&tmp);
	return(tmp);
  }
  return(*localPtr_);
}
/*
template<class T>
const T HPCxx_GlobalRefImpl<T>::operator*() const
{
  if (!isLocal()) {
	T tmp;
	read(&tmp);
	return(tmp);
  }
  return(*localPtr_);
}
*/
// Cast Operators
template<class T> T* HPCxx_GlobalRefImpl<T>::getPointer() const
{
  if (!isLocal()) {
	return(NULL);
  }
  return(localPtr_);
}
template<class T> HPCxx_GlobalRefImpl<T>::operator HPCxx_ContextIDImpl &() const
{
  return(*contextID_);
}

template<class T> HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::
     operator=(const HPCxx_GlobalRefImpl<T> &gr)
{
  if (this == &gr) return *this; // check for self assigment
  localPtr_  = gr.localPtr_;
  contextID_ = gr.contextID_; // new nexus_startpoint_t;
  //if(gr.contextID_ == NULL) printf("bad assignment of global pointers\n");
  // nexus_startpoint_copy(contextID_, gr.contextID_);
  GlobalEndpoint = NULL;
  offset_ = gr.offset_;
  
  return *this;
}

template<class T> HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::operator=
         (const T *ptr)
{
     nexus_endpointattr_t EpAttr ;
     GlobalEndpoint = new nexus_endpoint_t;
     contextID_ = new nexus_startpoint_t;
     nexus_endpointattr_init(&EpAttr);
     nexus_endpointattr_set_handler_table(&EpAttr,
			hpcxx_nexusHandlerTable,
                   sizeof(hpcxx_nexusHandlerTable)/sizeof(nexus_handler_t));
     nexus_endpoint_init(GlobalEndpoint, &EpAttr);
     nexus_endpointattr_destroy(&EpAttr);
     nexus_endpoint_set_user_pointer(GlobalEndpoint, (void *) ptr);
     nexus_startpoint_bind(contextID_, GlobalEndpoint);
     offset_ = 0;
     return *this;
}

// this one does not really belong to a "reference type" object.   it
// should really be part of the pointer impl, but we don't have a pointer
// impl.
template<class T> HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::operator
   =(T *ptr)
{
     nexus_endpointattr_t EpAttr ;
     GlobalEndpoint = new nexus_endpoint_t;
     contextID_ = new nexus_startpoint_t;
     nexus_endpointattr_init(&EpAttr);
     nexus_endpointattr_set_handler_table(&EpAttr,
			hpcxx_nexusHandlerTable,
                   sizeof(hpcxx_nexusHandlerTable)/sizeof(nexus_handler_t));
     nexus_endpoint_init(GlobalEndpoint, &EpAttr);
     nexus_endpointattr_destroy(&EpAttr);
     nexus_endpoint_set_user_pointer(GlobalEndpoint, (void *) ptr);
     nexus_startpoint_bind(contextID_, GlobalEndpoint);
     offset_ = 0;
     return *this;
}

template<class T> HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::operator=
  (T data)
{
  write(&data, 1);
  return *this;
  
}

// Test the equality of two GlobalPtr's
template<class T> int HPCxx_GlobalRefImpl<T>::operator==(const HPCxx_GlobalRefImpl<T>& gr)
{
  return(hpcxx_contextid_equal(contextID_, gr.contextID_) &&
		 offset_ == gr.offset_);
}


template<class T> int HPCxx_GlobalRefImpl<T>::operator==
                     (const HPCxx_GlobalRefImpl<T>& gr) const
{
  return(hpcxx_contextid_equal(contextID_, gr.contextID_) &&
		 offset_ == gr.offset_);
}


template<class T> int HPCxx_GlobalRefImpl<T>::operator!=(const HPCxx_GlobalRefImpl<T>& gr)
{
  return(!hpcxx_contextid_equal(contextID_, gr.contextID_) ||
		 !(offset_ == gr.offset_));
}


template<class T> int HPCxx_GlobalRefImpl<T>::operator!=(const HPCxx_GlobalRefImpl<T>& gr) const
{
  return(!hpcxx_contextid_equal(contextID_, gr.contextID_) ||
		 !(offset_ == gr.offset_));
}


// FIXME!!! Pointer arithmetic really needs to go through rsr's so that
// the artithmetic occurs in the context in which the pointer is defined
// so that the correct size of the pointer can be determined. 
// (i.e. sizeof(T *) in the local context may not equal sizeof(T *) in
// the remote context

// GlobalPtr Arithmetic
// prefix ++
template<class T> HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::operator++()
{
  localPtr_++;
  offset_++;
  return *this;
}

// postfix ++.  note. this one is never used! the job belongs to
// global pointer.
template<class T> HPCxx_GlobalRefImpl<T> HPCxx_GlobalRefImpl<T>::operator++(int)
{
  HPCxx_GlobalRefImpl<T> tmp(*this);
  localPtr_++;
  offset_++;
  return tmp;
}

// prefix --
template<class T> HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::operator--()
{
  localPtr_--;
  offset_--;
  return *this;
}

// postfix --
template<class T> HPCxx_GlobalRefImpl<T> HPCxx_GlobalRefImpl<T>::operator--(int)
{
  HPCxx_GlobalRefImpl<T> tmp(*this);
  localPtr_--;
  offset_--;
  return tmp;
}

template<class T> HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::operator+=(int inc)
{
  localPtr_ += inc;
  offset_ += inc;
  return *this;
}

template<class T> HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::operator-=(int dec)
{
  localPtr_ -= dec;
  offset_ -= dec;
  return *this;
}

template<class T> HPCxx_GlobalRefImpl<T> HPCxx_GlobalRefImpl<T>::operator+(int inc)
{
  HPCxx_GlobalRefImpl<T> tmp(*this);
  tmp.localPtr_ += inc;
  tmp.offset_ += inc;
  return tmp;
}
template<class T> HPCxx_GlobalRefImpl<T> HPCxx_GlobalRefImpl<T>::incr(int inc)
{
  HPCxx_GlobalRefImpl<T> tmp(*this);
  tmp.localPtr_ += inc;
  tmp.offset_ += inc;
  return tmp;
}


template<class T> HPCxx_GlobalRefImpl<T> HPCxx_GlobalRefImpl<T>::operator-(int dec)
{
  HPCxx_GlobalRefImpl<T> tmp(*this);
  tmp.localPtr_ -= dec;
  tmp.offset_ -= dec;
  return tmp;
}
  
template<class T> int HPCxx_GlobalRefImpl<T>::operator<(const HPCxx_GlobalRefImpl<T>& gr)
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 < GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ < gr.localPtr_);
}


template<class T> int HPCxx_GlobalRefImpl<T>::operator<(const HPCxx_GlobalRefImpl<T>& gr) const
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 < GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ < gr.localPtr_);
}


template<class T> int HPCxx_GlobalRefImpl<T>::operator>(const HPCxx_GlobalRefImpl<T>& gr)
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 > GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ > gr.localPtr_);
}


template<class T> int HPCxx_GlobalRefImpl<T>::operator>(const HPCxx_GlobalRefImpl<T>& gr) const
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 > GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ > gr.localPtr_);
}


template<class T> int HPCxx_GlobalRefImpl<T>::operator<=(const HPCxx_GlobalRefImpl<T>& gr)
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 <= GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ <= gr.localPtr_);
}


template<class T> int HPCxx_GlobalRefImpl<T>::operator<=(const HPCxx_GlobalRefImpl<T>& gr) const
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 <= GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ <= gr.localPtr_);
}


template<class T> int HPCxx_GlobalRefImpl<T>::operator>=(const HPCxx_GlobalRefImpl<T>& gr)
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 >= GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ >= gr.localPtr_);
}


template<class T> int HPCxx_GlobalRefImpl<T>::operator>=(const HPCxx_GlobalRefImpl<T>& gr) const
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 >= GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ >= gr.localPtr_);
}


template<class T> T HPCxx_GlobalRefImpl<T>::operator[](int index)
{
  if (!isLocal()) {
	T tmp;
	read(&tmp, 1, index);
	return(tmp);
  } else {
	return(*(localPtr_ + index));
  }
}
/*
template<class T> const T HPCxx_GlobalRefImpl<T>::operator[](int index) const
{
  if (!isLocal()) {
	T &tmp
	read(&tmp, 1, index);
  } else {
	return(*(localPtr_ + index));
  }
}
*/  
// Block data transfer
template<class T> void HPCxx_GlobalRefImpl<T>::read(T *ptr, int count,
    int offset) const
{
   // set up an end point for the sync var.
     nexus_endpointattr_t EpAttr ;
     nexus_endpoint_t ep;
     nexus_startpoint_t  sp2;
     nexus_endpointattr_init(&EpAttr);
     nexus_endpointattr_set_handler_table(&EpAttr,
			hpcxx_nexusHandlerTable,
                   sizeof(hpcxx_nexusHandlerTable)/sizeof(nexus_handler_t));
     nexus_endpoint_init(&ep, &EpAttr);
     nexus_endpointattr_destroy(&EpAttr);
     HPCxx_Sync<int> result_sync;
     nexus_endpoint_set_user_pointer(&ep, (void *) &result_sync);
     nexus_startpoint_bind(&sp2, &ep);
     // created a global ptr to result sync
     result_sync.setBufferAddr((void *)ptr);
     nexus_buffer_t buffer;
     nexus_buffer_init(&buffer, +nexus_sizeof_int(3)+
                               nexus_sizeof_startpoint(&sp2,1),
                      0);
     nexus_put_startpoint_transfer(&buffer, &sp2, 1);
     int typeID = HPCxx_Type<T>::registrationID();
     nexus_put_int(&buffer, &typeID, 1);
     nexus_put_int(&buffer, &count, 1);
     offset += offset_;
     nexus_put_int(&buffer, &offset, 1);
     nexus_send_rsr(&buffer, contextID_ , HPCXX_READ_HANDLER,
                  NEXUS_FALSE, NEXUS_FALSE);
     int x;
     result_sync.read(x);
     nexus_buffer_destroy(&buffer);
     nexus_endpoint_destroy(&ep);


}



template<class T> void HPCxx_GlobalRefImpl<T>::write(T *ptr, int count, int offset)
{
   // set up an end point for the sync var.
     nexus_endpointattr_t EpAttr ;
     nexus_endpoint_t ep;
     nexus_startpoint_t  sp2;
     nexus_endpointattr_init(&EpAttr);
     nexus_endpointattr_set_handler_table(&EpAttr,
			hpcxx_nexusHandlerTable,
                   sizeof(hpcxx_nexusHandlerTable)/sizeof(nexus_handler_t));
     nexus_endpoint_init(&ep, &EpAttr);
     //nexus_endpointattr_destroy(&EpAttr);
     HPCxx_Sync<int> result_sync;
     result_sync.setBufferAddr((void *)NULL);   
     nexus_endpoint_set_user_pointer(&ep, (void *) &result_sync);
     nexus_startpoint_bind(&sp2, &ep);
     HPCxx_Buffer buffer(nexus_sizeof_int(3)+ nexus_sizeof_startpoint(&sp2,1)
                         + count*sizeof(T));
     nexus_put_startpoint_transfer(buffer.impl_.buffer_, &sp2, 1);
     int typeID = HPCxx_Type<T>::registrationID();
     nexus_put_int(buffer.impl_.buffer_, &typeID, 1);
     nexus_put_int(buffer.impl_.buffer_, &count, 1);
     offset += offset_;
     nexus_put_int(buffer.impl_.buffer_, &offset, 1);
     hpcxx_pack(buffer, ptr, count);
     //printf("sending type=%d count=%d offset = %d val = %d\n",
     //         typeID, count, offset, *ptr);
     nexus_send_rsr(buffer.impl_.buffer_, contextID_ , HPCXX_WRITE_HANDLER,
                  NEXUS_FALSE, NEXUS_FALSE);
     int x;
     result_sync.read(x);
     nexus_buffer_destroy(buffer.impl_.buffer_);
     nexus_endpoint_destroy(&ep);
}
  
// Test to see if the pointer is local to our context
template<class T> bool HPCxx_GlobalRefImpl<T>::isLocal() const
{
  return((bool)((GlobalEndpoint != NULL) && nexus_startpoint_to_current_context(contextID_)));
}

// Test to see if the pointer is local to our context
// ptr is a valid reference only if it is local
template<class T> bool HPCxx_GlobalRefImpl<T>::isLocal(T *&ptr) const
{
  ptr = localPtr_;
  return((bool)isLocal());
}

