// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: HPCxx_MutexImpl.C,v 1.2 1998/07/09 23:53:48 sameer Exp $

#include <HPCxx_Profile.h>
#include <HPCxx_MutexImpl.h>

HPCxx_MutexImpl::HPCxx_MutexImpl()
{
  TAU_PROFILE("HPCxx_MutexImpl::HPCxx_MutexImpl()", "void ()", TAU_HPCXX);
  nexus_mutex_init(&mutex_, NULL);
}

HPCxx_MutexImpl::HPCxx_MutexImpl(const HPCxx_MutexImpl& mutex)
  : mutex_(mutex.mutex_)
{
  TAU_PROFILE("HPCxx_MutexImpl::HPCxx_MutexImpl()", "void (HPCxx_MutexImpl& )",
    TAU_HPCXX);
}

HPCxx_MutexImpl& HPCxx_MutexImpl::operator=(const HPCxx_MutexImpl& mutex)
{
  TAU_PROFILE("HPCxx_MutexImpl::operator=()", 
    "HPCxx_MutexImpl& (HPCxx_MutexImpl& )", TAU_HPCXX);
  if (this == &mutex) return *this; // check for self-assignment

  nexus_mutex_destroy(&mutex_); // get rid of the old mutex;
  mutex_ = mutex.mutex_;

  return *this;
}

HPCxx_MutexImpl::~HPCxx_MutexImpl()
{
  TAU_PROFILE("HPCxx_MutexImpl::~HPCxx_MutexImpl()", "void ()", TAU_HPCXX);
  nexus_mutex_destroy(&mutex_);
}

int HPCxx_MutexImpl::lock()
{
  TAU_PROFILE("HPCxx_MutexImpl::lock()", "int ()", TAU_HPCXX);
  return(nexus_mutex_lock(&mutex_));
}

int HPCxx_MutexImpl::unlock()
{
  TAU_PROFILE("HPCxx_MutexImpl::unlock()", "int ()", TAU_HPCXX);
  return(nexus_mutex_unlock(&mutex_));
}
