// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: HPCxx_ThreadImpl.h,v 1.1 1998/06/23 00:03:27 ejohnson Exp $

#ifndef __HPCXX_THREADIMPL_H_
#define __HPCXX_THREADIMPL_H_

#include <nexus.h>
#include <hpcxx_types.h>
#include <HPCxx_Runnable.h>
#include <HPCxx_CondImpl.h>


class HPCxx_Thread;

class HPCxx_ThreadImpl {
private:
  nexus_thread_t thread_;
  HPCxx_MutexImpl joinmutex;
  HPCxx_CondImpl joincond;
  int alive_;
  HPCxx_MutexImpl suspendmutex;
  HPCxx_CondImpl suspendcond;
  int suspended_;
  HPCxx_Runnable* runnable_;

  static void* start_(void* rhs);

public:
  HPCxx_ThreadImpl(void);
  HPCxx_ThreadImpl(const HPCxx_ThreadImpl& thread);
  HPCxx_ThreadImpl(HPCxx_Runnable *runnable);
  HPCxx_ThreadImpl& operator=(const HPCxx_ThreadImpl& thread);
  
  virtual ~HPCxx_ThreadImpl(void);

  virtual void run(void);
  
  // Starts execution of the thread and calls the run() method
  void start(HPCxx_Thread& thread);

  // Ceases execution of the thread.  Status is not currently used.
  static void stop(void *status);

  // Yields the processesor to another thread
  static void yield(void);

  // Sleep for milliseconds,nanoseconds.  Will be rounded up to the
  // resolution of the system timer
  static void sleep(long milliseconds, long nanoseconds = 0);

  // Suspends execution of thread until resume is called
  void suspend(void);
  
  // Resume a suspended thread
  void resume(void);
  
  // Check to see if thread has been started but has not yet died
  bool isAlive(void);
  
  static HPCxx_Thread *currentThread(void);

  // Returns the numeric ID of the thread
  static int getThreadID(void);

  // Join waits for the completion of the specified thread,
  // by definition join returns as soon as the thread is considered
  // not alive.
  void join(int milliseconds, int nanoseconds = 0);

  // Returns the current priority of the thread, where
  // HPCXX_MIN_THREAD_PRIORITY <= getPriority() <= HPCXX_MAX_THREAD_PRIORITY
  int getPriority(void);

  // Set the new priority for a thread.  Priority is clamped to
  // (HPCXX_MIN_THREAD_PRIORITY, HPCXX_MAX_THREAD_PRIORITY)
  int setPriority(int priority);

};

#endif
