// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: hpcxx_invokeImpl.cc,v 1.3 1998/08/08 00:00:35 rasmussn Exp $

#include <hpcxx_invoke.h>
#include <hpcxx_register.h>
#include <HPCxx_Buffer.h>


void hpcxx_pack(HPCxx_Buffer& buffer, HPCxx_GenericPtr** r, int count) {

  int sz = count*nexus_sizeof_startpoint(r[0]->contextIDImpl(), 1)
	    + 2*count*sizeof(int) + 32;
  if(nexus_check_buffer_size(buffer.impl_.buffer_, sz, sz, 0, 0) < 0) {
    cout << "global pointer pack failed, buffer size failure" << endl;
  }
  for (int i = 0; i < count; i++) {
    int offset = r[i]->offset();
    int type   = r[i]->type();
    nexus_startpoint_t *sp = new nexus_startpoint_t;
    nexus_startpoint_copy(sp, r[i]->contextIDImpl());
    nexus_put_startpoint_transfer(buffer.impl_.buffer_ , sp, 1);
    nexus_put_int(buffer.impl_.buffer_, &offset, 1);
    nexus_put_int(buffer.impl_.buffer_, &type, 1);
  }
}


void hpcxx_unpack(HPCxx_Buffer& buffer, HPCxx_GenericPtr** r, int count) {

  for (int i = 0; i < count; i++) {
    int offset;
    int type;

    // Since we are unpacking to pointers, first we need to instantiate the
    // correct type of global pointer.

    HPCxx_ContextIDImpl* contextID = new nexus_startpoint_t;     
    nexus_get_startpoint(buffer.impl_.buffer_, contextID, 1); 
    nexus_get_int(buffer.impl_.buffer_, &offset, 1);  
    nexus_get_int(buffer.impl_.buffer_, &type, 1);  

    // create the correct global pointer type

    switch (type) {
      case HPCXX_TYPE_INT:
	r[i] = new HPCxx_GlobalPtr<int>();
        break;
      case HPCXX_TYPE_FLOAT:
	r[i] = new HPCxx_GlobalPtr<float>();
	break;
      case HPCXX_TYPE_DOUBLE:
	r[i] = new HPCxx_GlobalPtr<double>();
	break;
      default:
	r[i] = NULL;
	cerr << "hpcxx_unpack of HPCxx_GenericPtr failed: type = "
	     << type << "is unknown" << endl;
    }
    if (r[i] != NULL) r[i]->initialize(NULL, offset, contextID);
  }
}


template <class T>
void hpcxx_pack(HPCxx_Buffer &buffer, HPCxx_GlobalPtr<T> *r, int count) {
  int sz = count*nexus_sizeof_startpoint((*(r[0])).contextID_, 1)
	    + count*sizeof(int) + 32;
  if(nexus_check_buffer_size(buffer.impl_.buffer_, sz, sz, 0, 0) < 0) {
    cout << "global pointer pack failed" << endl;
  }
  for (int i = 0; i < count; i++) {
    HPCxx_GlobalRefImpl<T> &gr = *(r[i]);
    nexus_startpoint_t *sp = new nexus_startpoint_t;
    nexus_startpoint_copy(sp,  gr.contextID_);
    nexus_put_startpoint_transfer(buffer.impl_.buffer_ , sp, 1);
    nexus_put_int(buffer.impl_.buffer_, &(gr.offset_), 1);
  }
}


template <class T>
void hpcxx_unpack(HPCxx_Buffer &buffer, HPCxx_GlobalPtr<T> *r, int count) {
  for (int i = 0; i < count; i++) {
    HPCxx_GlobalRefImpl<T> &gr = *(r[i]);
    gr.contextID_ = new nexus_startpoint_t;     
    nexus_get_startpoint(buffer.impl_.buffer_,gr.contextID_, 1); 
    nexus_get_int(buffer.impl_.buffer_, &(gr.offset_), 1 );  
    gr.localPtr_ = NULL;
    gr.GlobalEndpoint = NULL;
  }
}


template <class T>
void hpcxx_pack(HPCxx_Buffer &buffer, HPCxx_Sync<T> *r, int count) {
  for (int i = 0; i < count; i++) {
    T x = r[i];		// i am fairly certain this is wrong! dbg.
    nexus_startpoint_t sp ;
    HPCxx_GlobalRefImpl<T> gpi(&x);
    nexus_startpoint_copy(&sp,  gpi.contextID_);
    nexus_put_startpoint_transfer(buffer.impl_.buffer_ , &sp, 1);
    nexus_put_int(buffer.impl_.buffer_, &gpi.offset_, 1);
  }
}
