// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: hpcxx_rtsImpl.C,v 1.7 1998/08/07 20:08:12 sameer Exp $
#define HPCXX_MAX_CHILDREN 100
#include <HPCxx_Profile.h>
#include <unistd.h>
#include <iostream.h>
#include <libgen.h>
#include <pthread.h>
#include <nexus.h>
#include <string.h>

#include <hpcxx_rts.h>
#include <hpcxx_rtsImpl.h>
#include <HPCxx_Prototype.h>

// create a global map for mapping function prototypes
HPCxx_IdProtoMap hpcxx_prototypeMap;


HPCxx_Context           *hpcxx_currentContext;
HPCxx_ContextID         *hpcxx_currentContextID = NULL;
static int hpcxx_currentPosition = -1;
static nexus_startpoint_t     hpcxx_startpoint_master; 

static nexus_endpoint_t  hpcxx_nexus_endpoint_;
static nexus_node_t     *hpcxx_nexus_nodes_;
static int               hpcxx_nexus_numNodes_;
static bool              hpcxx_nexus_is_master_context_ = (bool)0;
static HPCxx_ContextID **hpcxx_contextIDs_;

nexus_endpoint_t *hpcxx_get_ep(){ return &hpcxx_nexus_endpoint_; }
nexus_startpoint_t *hpcxx_get_sp(){ return &hpcxx_startpoint_master;}

static const char* hpcxx_basename(const char* path)
{
  if (path == NULL) return NULL;
  else if ((strrchr(path, '/')) == NULL) return path;
  else return (strrchr(path, '/') + 1);
}

//
// This table *must* be created in the same order as the consts
// declared in the header file!!!
//
nexus_handler_t hpcxx_nexusHandlerTable[HPCXX_NUM_HANDLERS] = {
  {NEXUS_HANDLER_TYPE_THREADED,
   (nexus_handler_func_t)hpcxx_initHandler},
  {NEXUS_HANDLER_TYPE_THREADED,
   (nexus_handler_func_t)hpcxx_invokeGFHandler},
  {NEXUS_HANDLER_TYPE_THREADED,
   (nexus_handler_func_t)hpcxx_invokeMFHandler},
  {NEXUS_HANDLER_TYPE_THREADED,
   (nexus_handler_func_t)hpcxx_readHandler},
  {NEXUS_HANDLER_TYPE_THREADED,
   (nexus_handler_func_t)hpcxx_writeHandler},
  {NEXUS_HANDLER_TYPE_THREADED,
   (nexus_handler_func_t)hpcxx_ackHandler},
  {NEXUS_HANDLER_TYPE_THREADED,
   (nexus_handler_func_t)hpcxx_returnHandler},
  {NEXUS_HANDLER_TYPE_THREADED,
   (nexus_handler_func_t)hpcxx_barrierHandler},
  {NEXUS_HANDLER_TYPE_THREADED,
   (nexus_handler_func_t)hpcxx_barrierOutHandler},
  {NEXUS_HANDLER_TYPE_THREADED,
   (nexus_handler_func_t)hpcxx_lc_bufferMoveHandler},
  {NEXUS_HANDLER_TYPE_THREADED,
   (nexus_handler_func_t)hpcxx_rc_bufferMoveHandler},
  {NEXUS_HANDLER_TYPE_THREADED,
   (nexus_handler_func_t)hpcxx_p_bufferMoveHandler},
  {NEXUS_HANDLER_TYPE_THREADED,
   (nexus_handler_func_t)receive_handler},
};

// a function called by the master context to send copies of
// the startpoints to each of the other contexts.
static void sendContextIDs(){
    TAU_PROFILE("sendContextIDs()", "void ()", TAU_HPCXX);
    int i, j, n;
    n = hpcxx_nexus_numNodes_;
    nexus_startpoint_t *theSPs = new nexus_startpoint_t[n];
    hpcxx_currentPosition = 0;
    nexus_buffer_t buffer;
    for(i = 0; i < n; i++)
         nexus_startpoint_copy(&(theSPs[i]), hpcxx_contextIDs_[i]);
    nexus_buffer_init(&buffer, +nexus_sizeof_int(1)+ 
                               nexus_sizeof_startpoint(theSPs,n),
                      0);
    nexus_put_int(&buffer, &n, 1);
    nexus_put_startpoint_transfer(&buffer, theSPs, n);

    for ( j = 1; j < hpcxx_nexus_numNodes_; j++){
        nexus_send_rsr(&buffer, hpcxx_contextIDs_[j] , HPCXX_INIT_HANDLER, 
                  NEXUS_FALSE, NEXUS_FALSE);
    }
}

static HPCxx_Sync<int> init_sync;

// Nexus Hanlder to bootstrap HPC++ by distributing the contextIDs
void hpcxx_initHandler(nexus_endpoint_t *endpoint,
                       nexus_buffer_t *buffer,
                       nexus_bool_t called_from_non_threaded_handler)
{
   TAU_PROFILE("hpcxx_initHandler()", 
     "void (nexus_endpoint_t *, nexus_buffer_t *, nexus_bool_t)", TAU_HPCXX);
   int i, n;
   nexus_get_int(buffer, &n, 1);
   hpcxx_nexus_numNodes_ = n;
   nexus_startpoint_t *theSPs = new nexus_startpoint_t[n];
   nexus_get_startpoint(buffer, theSPs, n);
   hpcxx_contextIDs_ = new HPCxx_ContextID *[hpcxx_nexus_numNodes_];
   for(i = 0; i < n; i++) hpcxx_contextIDs_[i] = &(theSPs[i]);
   for(i = 0; i < n; i++)
      if (nexus_startpoint_equal(hpcxx_currentContextID, hpcxx_contextIDs_[i]))
            hpcxx_currentPosition = i;
   init_sync.write(n);
}


//
// This function is currently unimplemented yet defined by
// Nexus.  We'll fudge it be seeing whether or not the
// number of nodes returned by nexus_init is > 0
//
int nexus_master_context()
{
  return(hpcxx_nexus_is_master_context_);
}
// special nexus barrier code.

void hpxx_send_unblock(nexus_startpoint_t *n, int i){
    TAU_PROFILE("hpcxx_send_unblock()", "void (nexus_startpoint_t *, int)", 
      TAU_HPCXX);
    nexus_buffer_t buffer;
    nexus_startpoint_t sp;
    nexus_startpoint_copy(&sp, n);
    nexus_buffer_init(&buffer, nexus_sizeof_int(1)+ 
                               nexus_sizeof_startpoint(&sp,1),
                      0);
    nexus_put_int(&buffer, &i, 1);
    nexus_put_startpoint_transfer(&buffer, &sp, 1);
    nexus_send_rsr(&buffer, hpcxx_contextIDs_[i] , HPCXX_BARRIER_IN_HANDLER, 
                  NEXUS_FALSE, NEXUS_FALSE);
    nexus_buffer_destroy(&buffer);
}

 
static HPCxx_SyncQ<int> from_parent;
static HPCxx_SyncQ<int> from_children[HPCXX_MAX_CHILDREN];
static nexus_startpoint_t parent;
static int child_id;

static void notify_parent(){
  TAU_PROFILE("notify_parent()", "void ()", TAU_HPCXX);
  nexus_buffer_t buffer2;
  nexus_buffer_init(&buffer2, nexus_sizeof_int(1), 0);
  nexus_put_int(&buffer2, &child_id, 1);
  //nexus_startpoint_t s;
  //nexus_startpoint_copy(&s, &parent);
  nexus_send_rsr(&buffer2, &parent, HPCXX_BARRIER_OUT_HANDLER,
                     NEXUS_FALSE, NEXUS_FALSE);
  nexus_buffer_destroy(&buffer2);
}
 
void hpcxx_barrierHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler)
{
     TAU_PROFILE("hpcxx_barrierHandler()", 
       "void (nexus_endpoint_t *, nexus_buffer_t *, nexus_bool_t )", TAU_HPCXX);
     nexus_get_int(buffer, &child_id,1);
     nexus_get_startpoint(buffer, &parent, 1);
     from_parent.write(child_id);
}

void hpcxx_barrierOutHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler)
{
    TAU_PROFILE("hpcxx_barrierOutHandler()", 
      "void (nexus_endpoint_t *, nexus_buffer_t *, nexus_bool_t )", TAU_HPCXX);
    int x;
    nexus_get_int(buffer, &x,1);
    from_children[x].write(x);
}

void hpcxx_barrier(){
   TAU_PROFILE("hpcxx_barrier()", "void ()", TAU_HPCXX);
   int i, j;
   if(nexus_master_context()){
      for(i = 1; i < hpcxx_nexus_numNodes_; i++) 
             hpxx_send_unblock( hpcxx_currentContextID, i);
      if(hpcxx_nexus_numNodes_ >=  HPCXX_MAX_CHILDREN)
         printf("too many nodes: HPCXX_MAX_CHILDREN=%d\n", HPCXX_MAX_CHILDREN);

      for(i = 1; i < hpcxx_nexus_numNodes_; i++) from_children[i].read(j);
      
      for(i = 1; i < hpcxx_nexus_numNodes_; i++) 
             hpxx_send_unblock( hpcxx_currentContextID, i);
      }
   else{
      from_parent.read(j);
      notify_parent();
      from_parent.read(j);
      }
}
// end of barrier code.

void hpcxx_typeInit();

void init_thread_key();
int hpcxx_initAgent(int &argc, char **&argv)
{
  TAU_PROFILE("hpcxx_initAgent()", "int (int, char **)", TAU_HPCXX);
  hpcxx_typeInit();
  init_thread_key();
  nexus_init(&argc,
                         &argv,
                         "NEXUS_ARGS",
                         "nx",
                         NULL,
                         NULL,
                         NULL,
                         NULL,
                         &hpcxx_nexus_nodes_,
                         &hpcxx_nexus_numNodes_);

  // Start SPMD style
  nexus_start_nonblocking();
  //nexus_start();
  hpcxx_currentContextID = &hpcxx_startpoint_master; // added by dbg
  hpcxx_currentContext =  new HPCxx_Context(hpcxx_currentContextID,
                                                  true,
                                                  "child_context");
   return 0;
}


/* BEGIN NEW STUFF for FAULT TOLERANCE */

int nexus_fault_callback(void *user_arg, int fault_code) {
  cerr << "nexus got a fatal error, errno = " << fault_code << endl;
  cerr << nexus_fault_strings[fault_code] << endl;
  return 0;  // means cleanup and continue
}

int user_arg_ = 0;

/* END NEW STUFF for FAULT TOLERANCE */


int hpcxx_init(int &argc, char **&argv, HPCxx_Group *&group)
{
  TAU_PROFILE("hpcxx_init()", "int (int, char **, HPCxx_Group)", TAU_HPCXX);
  TAU_PROFILE_INIT(argc, argv);
  hpcxx_typeInit();
  init_thread_key();
  nexus_init(&argc,
			 &argv,
			 "NEXUS_ARGS",
			 "nx",
			 NULL,
			 NULL,
			 NULL,
			 NULL,
			 &hpcxx_nexus_nodes_,
			 &hpcxx_nexus_numNodes_);

 
  // Start SPMD style 
  nexus_start_nonblocking();

  // NEW STUFF for FAULT TOLERANCE
  nexus_enable_fault_tolerance(nexus_fault_callback, (void *)&user_arg_);

  //nexus_start();

  //printf("starting up hpc++.  num contexts = %d\n", hpcxx_nexus_numNodes_);
  // Nexus currently doesn't implement nexus_master_context();
  // so fudge it.  hpcxx_nexus_numNodes is only > 0 on master context.
  //
  if (hpcxx_nexus_numNodes_) hpcxx_nexus_is_master_context_ = true;

  // FIXME
  // need to propagate nodes/group to the other non-master contexts

  if (nexus_master_context()) {

	hpcxx_contextIDs_ = new HPCxx_ContextID *[hpcxx_nexus_numNodes_];

	for(int i = 0; i < hpcxx_nexus_numNodes_; i++) {
	  hpcxx_contextIDs_[i] = &(hpcxx_nexus_nodes_[i].startpoint);
	 }
	hpcxx_contextIDs_[0] = &hpcxx_startpoint_master; // added by dbg
	group = new HPCxx_Group(hpcxx_contextIDs_, 
					hpcxx_nexus_numNodes_, 
					HPCXX_GEN_LOCAL_GROUPID,
					hpcxx_basename(argv[0]));
        group->setLocalIndex(0);
	hpcxx_currentContext =  new HPCxx_Context(hpcxx_contextIDs_[0],
			                          true,
						  "master_context");
	hpcxx_currentContextID = hpcxx_contextIDs_[0];
        sendContextIDs();
  } else{
        int x;
	hpcxx_currentContextID = &hpcxx_startpoint_master; // added by dbg
        init_sync.read(x);
	group = new HPCxx_Group(hpcxx_contextIDs_, 
				hpcxx_nexus_numNodes_, 
				HPCXX_GEN_LOCAL_GROUPID,
				hpcxx_basename("child"));
        group->setLocalIndex(hpcxx_currentPosition);
	hpcxx_currentContext =  new HPCxx_Context(hpcxx_currentContextID,
			                          true,
						  "child_context");
     }
  TAU_PROFILE_SET_NODE(0);
  TAU_PROFILE_SET_CONTEXT(group->getLocalIndex());
  hpcxx_barrier();
  return(HPCXX_SUCCESS);
}


// Thomas Stuckey suggested this ...
int hpcxx_exit()
{
  TAU_PROFILE("hpcxx_exit()", "int ()", TAU_HPCXX);
  nexus_startpoint_destroy(&hpcxx_startpoint_master);
  nexus_shutdown_nonexiting();
  return(0);
}

int hpcxx_exit(HPCxx_Group*& group)
{
  TAU_PROFILE("hpcxx_exit()", "int (HPCxx_Group *)", TAU_DEFAULT);
  // Only the master context should shutdown nexus
  // the other contexts should simply exit.
  hpcxx_barrier();
  return hpcxx_exit();
}

// 5/4/98 : We've made an executive decision that createContext creates
// the process and returns a handle.  node_counter is keeping us from
// repeating a "id" parameter to node_acquire (which is going away shortly
// anyway) so that we don't reuse anything.  We don't know if this would
// cause a problem, but it's better safe than sorry.
static int node_counter = 1;
HPCxx_ContextID *hpcxx_createContext(char *machine, char *exec_path){
  TAU_PROFILE("hpcxx_createContext()", "HPCxx_ContextID * (char *, char *)",
    TAU_HPCXX);
  nexus_node_t *nodes;
  int n_nodes;
  HPCxx_Mutex l;
  l.lock();
  nexus_node_acquire(machine, node_counter,1, "/", exec_path, &nodes, &n_nodes);
  node_counter++;
  l.unlock();
  //printf("nodes = %d\n", n_nodes);
  if(nodes == NULL){
    printf("null node array in makeContext\n");
    return NULL;
  }
  HPCxx_ContextID *x = &(nodes[0].startpoint);
  return x;
}

// hpcxx_destroyContext : Destroys a contextID and therefore a startpoint
// BDT 4/23/98 (Happy Anniversary, Wanda!)
int hpcxx_destroyContextID(HPCxx_ContextID *context)
{
  TAU_PROFILE("hpcxx_destroyContextID()", "int (HPCxx_ContextID *)", TAU_HPCXX);
  int retval;

  if(context==NULL) return 0;
  retval=nexus_startpoint_destroy((nexus_startpoint_t *)context);
  return retval;
}


void (*userCallback)(void *)=NULL;
void *userCallbackArg=NULL;

// hpcxx_AllowAttach : This function must be called in order to allow
// other HPC++ contexts to attach to a running context which calls it.
// The port argument is used if it contains a non - null value, otherwise
// a port will be chosen from the available ports at random.  You probably
// don't want this, since you need to know which port you are using
// in order to have other contexts attach to it...
int hpcxx_allowAttach(unsigned short attach_port, void (*callback)(void *), void *arg)
{
  TAU_PROFILE("hpcxx_allowAttach()", 
    "int (unsigned short, void (*)(void *), void *)", TAU_HPCXX);
  char *attach_host;
  int retval;
 
  retval = nexus_allow_attach(&attach_port, &attach_host, attach_requested, NULL);
  if(retval) return -1;
  else {
    userCallback=callback;
    userCallbackArg=arg;
    return (int)attach_port;
  }
}

// attach_requested : This func handles the binding of the remote
// startpoint to the local endpoint.  Returns 0 if successful, non-
// zero otherwise, though what these non-zero values mean is beyond
// me at this point, as there's no documentation for this stuff anywhere. 8(
static int attach_requested(void *arg, char *url, nexus_startpoint_t *sp)
{
  TAU_PROFILE("attach_requested()", 
    "int (void *, char *, nexus_startpoint_t *)", TAU_HPCXX);
  int retval;

  retval=nexus_startpoint_bind(sp, &hpcxx_nexus_endpoint_);
  if(userCallback) userCallback(userCallbackArg);
  return retval;
}

// hpcxx_grabContext : makes an attempt to attach to the context given
// by a host name and port number.  This information is concatenated
// together to form a string of the following form:
//
// x-nexus://host:port
//
// Return a new HPCxx_ContextID (which is actually just a nexus_startpoint_t)
// on success, NULL on failure.
HPCxx_ContextID *hpcxx_grabContext(char *host, int port)
{
  TAU_PROFILE("hpcxx_grabContext()", 
    "HPCxx_ContextID * (char *, int)", TAU_HPCXX);
  int retval;
  char xurl[256];
  nexus_startpoint_t *newID;

  newID=(nexus_startpoint_t *)malloc(sizeof(nexus_startpoint_t));
  sprintf(xurl, "x-nexus://%s:%d", host, port);
  retval=nexus_attach(xurl, newID);
  if(retval) {
    free(newID);
    return NULL;
  }
  else {
    return (HPCxx_ContextID *)newID;
  }
}

void hpcxx_poll()
{
  TAU_PROFILE("hpcxx_poll()", "void ()", TAU_HPCXX);
  nexus_poll();
}

void *hpcxx_malloc(size_t size)
{
  TAU_PROFILE("hpcxx_malloc()", "void * (size_t)", TAU_HPCXX);
  return(nexus_malloc(size));
}

void hpcxx_free(void *ptr)
{
  TAU_PROFILE("hpcxx_free()", "void (void *)", TAU_HPCXX);
  if (ptr) nexus_free(ptr);
}

// Return the number of nodes in this computation
int hpcxx_getNumNodes()
{
  TAU_PROFILE("hpcxx_getNumNodes()", "int ()", TAU_HPCXX);
  return(hpcxx_nexus_numNodes_);
}

// Get a list of all nodes
// User must 'delete []' the null terminated node array
HPCxx_GlobalPtr<HPCxx_Node> *hpcxx_getNodes()
{
  // FIXME - not implemented
  return((HPCxx_GlobalPtr<HPCxx_Node> *)0);
}

//
// Nexus specific routines that must be implemented by the library
//

int NexusBoot(nexus_startpoint_t *sp)
{
  TAU_PROFILE("NexusBoot()", "int (nexus_startpoint_t *)", TAU_HPCXX);
  nexus_endpointattr_t attr;
  nexus_endpointattr_init(&attr);
  nexus_endpointattr_set_handler_table(&attr, hpcxx_nexusHandlerTable,
					HPCXX_NUM_HANDLERS);
  nexus_endpoint_init(&hpcxx_nexus_endpoint_, &attr);

  nexus_startpoint_bind(sp, &hpcxx_nexus_endpoint_);
  nexus_startpoint_bind(&hpcxx_startpoint_master, &hpcxx_nexus_endpoint_);
  nexus_endpointattr_destroy(&attr); 

  return NEXUS_SUCCESS;
}

int NexusAcquiredAsNode(nexus_startpoint_t *sp)
{
  return(NexusBoot(sp));
}

void NexusExit()
{
  // FIXME - should HPC++ need to do anything to shutdown here?
}

// Print out an error in the master context only
void hpcxx_msg(const char *msg)
{
  nexus_printf((char *)msg);
}

// Abort the hpcxx computation.  hpcxx_abort() will call hpcxx_exit()
void hpcxx_abort(int)
{
  nexus_abort();
}

bool hpcxx_contextid_equal(HPCxx_ContextID *id1, HPCxx_ContextID *id2)
{
  return((bool)nexus_startpoint_equal(id1, id2));
}


// two prototypes that should be in an include.
int hpcxx_runRF(void * gp, void *p, int funcID, HPCxx_Buffer &b);
int hpcxx_runResult(void * sync, int funcID, HPCxx_Buffer &b);
int hpcxx_runRead(void * sync, void *q, int typeID, int count, int offset);
int hpcxx_runWrite(void * sync, void *q, int typeID, int count, int offset,
HPCxx_Buffer &b);

// Nexus Handler for Remote (global) Function invocation

void hpcxx_invokeGFHandler(nexus_endpoint_t *endpoint,
                           nexus_buffer_t *buffer,
                           nexus_bool_t called_from_non_threaded_handler)
{
 TAU_PROFILE("hpcxx_invokeGFHandler()", 
   "void (nexus_endpoint_t *, nexus_buffer_t *, nexus_bool_t)", TAU_HPCXX);
 int funcID, offset;
 nexus_startpoint_t reply_done_sp;

 // unpack the global pointer (without calling hpcxx_unpack), probably
 // because the type is not known at this point

 nexus_get_startpoint(buffer, &reply_done_sp, 1);
 nexus_get_int(buffer, &offset, 1);

 nexus_get_int(buffer, &funcID, 1);
 HPCxx_BufferImpl bi(*buffer);
 HPCxx_Buffer b(bi);
 void *q = (void *) nexus_endpoint_get_user_pointer(endpoint);
 // q = q+offset;  this is wrong.  the offset must be added to typed pointer.
 hpcxx_runRF((void *) &reply_done_sp, q,  funcID, b);
 nexus_startpoint_destroy(&reply_done_sp);  // shridhar added this ...
}

void hpcxx_returnHandler(nexus_endpoint_t *endpoint,
                           nexus_buffer_t *buffer,
                           nexus_bool_t called_from_non_threaded_handler)
{
 TAU_PROFILE("hpcxx_returnHandler()", 
   "void (nexus_endpoint_t *, nexus_buffer_t *, nexus_bool_t )", TAU_HPCXX);
 int funcID;
 void *d = (void *) nexus_endpoint_get_user_pointer(endpoint);
 nexus_get_int(buffer, &funcID, 1);
 HPCxx_BufferImpl bi(*buffer);
 HPCxx_Buffer b(bi);
 hpcxx_runResult(d, funcID, b);  
}

// Nexus Handler for Remote Member Function invocation
void hpcxx_invokeMFHandler(nexus_endpoint_t *endpoint,
                           nexus_buffer_t *buffer,
                           nexus_bool_t called_from_non_threaded_handler)
{
// FIXME
}

// Nexus Handler for GlobalPointer reads
void hpcxx_readHandler(nexus_endpoint_t *endpoint,
				  nexus_buffer_t *buffer,
				  nexus_bool_t called_from_non_threaded_handler)
{
 TAU_PROFILE("hpcxx_readHandler()", 
   "void (nexus_endpoint_t *, nexus_buffer_t *, nexus_bool_t )", TAU_HPCXX);
 int count, offset;
 int typeID;
 nexus_startpoint_t reply_done_sp;
 nexus_get_startpoint(buffer, &reply_done_sp, 1);
 nexus_get_int(buffer, &typeID, 1);
 nexus_get_int(buffer, &count, 1);
 nexus_get_int(buffer, &offset, 1);
 void *q = (void *) nexus_endpoint_get_user_pointer(endpoint);
 hpcxx_runRead((void *) &reply_done_sp, q, typeID, count, offset);
}

// Nexus Handler for GlobalPointer writes
void hpcxx_writeHandler(nexus_endpoint_t *endpoint,
				   nexus_buffer_t *buffer,
				   nexus_bool_t called_from_non_threaded_handler)
{
 TAU_PROFILE("hpcxx_writeHandler()", 
   "void (nexus_endpoint_t *, nexus_buffer_t *, nexus_bool_t )", TAU_HPCXX);
 int count, offset;
 int typeID;
 nexus_startpoint_t reply_done_sp;
 nexus_get_startpoint(buffer, &reply_done_sp, 1);
 nexus_get_int(buffer, &typeID, 1);
 nexus_get_int(buffer, &count, 1);
 nexus_get_int(buffer, &offset, 1);
 HPCxx_BufferImpl bi(*buffer);
 HPCxx_Buffer hpbuff(bi);
 void *q = (void *) nexus_endpoint_get_user_pointer(endpoint);
 hpcxx_runWrite((void *) &reply_done_sp, q, typeID, count, offset, hpbuff);
}

// Nexus Handler to acknowledge a received global communication
void hpcxx_ackHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler)
{
// FIXME
}

// the following is a pack routine for a buffer block object
// it is never to be called because a buffer block object should
// not move.  BUT this is needed by the aix C++ compiler.

void hpcxx_pack(HPCxx_Buffer &, HPCxx_BufferBlock *, int){
}

void hpcxx_lc_bufferMoveHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler){
 TAU_PROFILE("hpcxx_lc_bufferMoveHandler()", 
   "void (nexus_endpoint_t *, nexus_buffer_t *, nexus_bool_t )", TAU_HPCXX);
 HPCxx_BufferImpl bi(*buffer);
 HPCxx_Buffer *b = new HPCxx_Buffer(bi);
 HPCxx_BufferBlock * p = 
   (HPCxx_BufferBlock *) nexus_endpoint_get_user_pointer(endpoint);
   p->lcBuffer = b;
   int i = 1;
   (p->lcLock).write(i);
   i = p->lcLock2;
}

void hpcxx_sendToBufferBlockLC(HPCxx_Buffer &b, HPCxx_ContextID &target){
  TAU_PROFILE("hpcxx_sendToBufferBlockLC()", 
    "void (HPCxx_Buffer, HPCxx_ContextID )", TAU_HPCXX);
     nexus_send_rsr(b.impl_.buffer_, &target, HPCXX_LCBUFF_MOVE_HANDLER,
                  NEXUS_FALSE, NEXUS_FALSE);
}

void hpcxx_rc_bufferMoveHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler){
 TAU_PROFILE("hpcxx_rc_bufferMoveHandler()", 
   "void (nexus_endpoint_t *, nexus_buffer_t *, nexus_bool_t )", TAU_HPCXX);
 HPCxx_BufferImpl bi(*buffer);
 HPCxx_Buffer *b = new HPCxx_Buffer(bi);
 HPCxx_BufferBlock * p = 
   (HPCxx_BufferBlock *) nexus_endpoint_get_user_pointer(endpoint);
   p->rcBuffer = b;
   int i = 1;
   (p->rcLock).write(i);
   i = p->rcLock2;

}

void hpcxx_sendToBufferBlockRC(HPCxx_Buffer &b, HPCxx_ContextID & target){
  TAU_PROFILE("hpcxx_sendToBufferBlockRC()", 
    "void (HPCxx_Buffer, HPCxx_ContextID )", TAU_HPCXX);
     nexus_send_rsr(b.impl_.buffer_, &target, HPCXX_RCBUFF_MOVE_HANDLER,
                  NEXUS_FALSE, NEXUS_FALSE);
}

void hpcxx_p_bufferMoveHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler){
 TAU_PROFILE("hpcxx_p_bufferMoveHandler()", 
   "void (nexus_endpoint_t *, nexus_buffer_t *, nexus_bool_t )", TAU_HPCXX);
 HPCxx_BufferImpl bi(*buffer);
 HPCxx_Buffer *b = new HPCxx_Buffer(bi);
 HPCxx_BufferBlock * p = 
   (HPCxx_BufferBlock *) nexus_endpoint_get_user_pointer(endpoint);
   p->pBuffer = b;
   int i = 1;
   (p->pLock).write(i);
   i = p->pLock2;
}

void hpcxx_sendToBufferBlockP(HPCxx_Buffer &b,HPCxx_ContextID & target){
  TAU_PROFILE("hpcxx_sendToBufferBlockP()", 
    "void (HPCxx_Buffer, HPCxx_ContextID )", TAU_HPCXX);
     nexus_send_rsr(b.impl_.buffer_, &target, HPCXX_PBUFF_MOVE_HANDLER,
                  NEXUS_FALSE, NEXUS_FALSE);
}
void hpcxx_sendGlobalFunction(HPCxx_Buffer &b, HPCxx_ContextID &target){
  TAU_PROFILE("hpcxx_sendGlobalFunction()", 
    "void (HPCxx_Buffer, HPCxx_ContextID)", TAU_HPCXX);
      nexus_send_rsr(b.impl_.buffer_, &target, HPCXX_INVOKE_GF_HANDLER, 
                  NEXUS_FALSE, NEXUS_FALSE);
}
void hpcxx_returnGlobalFunction(HPCxx_Buffer &b, void *target){
  TAU_PROFILE("hpcxx_returnGlobalFunction()", "void (HPCxx_Buffer, void *)", 
    TAU_HPCXX);
      nexus_send_rsr(b.impl_.buffer_, (HPCxx_ContextID *) target,
                    HPCXX_RETURN_HANDLER,  NEXUS_FALSE, NEXUS_FALSE);
}

void receive_handler(nexus_endpoint_t *endpoint,
					 nexus_buffer_t *buffer,
					 nexus_bool_t is_non_threaded_handler)
{
}
