// HPC++ Run Time Sytem Library (nexus)
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: hpcxx_rtsImpl.h,v 1.1 1998/06/23 00:03:36 ejohnson Exp $

#ifndef __HPCXX_RTSIMPL_H_
#define __HPCXX_RTSIMPL_H_

#include "HPCxx_Buffer.h"// for hpcxx_pack and hpcxx_unpack routines

const int HPCXX_INIT_HANDLER       = 0;
const int HPCXX_INVOKE_GF_HANDLER  = 1;
const int HPCXX_INVOKE_MF_HANDLER  = 2;
const int HPCXX_READ_HANDLER       = 3;
const int HPCXX_WRITE_HANDLER      = 4;
const int HPCXX_ACK_HANDLER        = 5;
const int HPCXX_RETURN_HANDLER     = 6;
const int HPCXX_BARRIER_IN_HANDLER = 7;
const int HPCXX_BARRIER_OUT_HANDLER= 8;
const int HPCXX_LCBUFF_MOVE_HANDLER= 9;
const int HPCXX_RCBUFF_MOVE_HANDLER= 10;
const int HPCXX_PBUFF_MOVE_HANDLER = 11;
const int RECEIVE_HANDLER          = 12;
const int HPCXX_NUM_HANDLERS       = 13;

extern nexus_handler_t hpcxx_nexusHandlerTable[HPCXX_NUM_HANDLERS];
nexus_endpoint_t *hpcxx_get_ep();
nexus_startpoint_t *hpcxx_get_sp();

// addition by Thomas Stuckey to exit cleanly from an HPC++ program
int hpcxx_exit();

// User defined handler to receive messages from Java program
void receive_handler(nexus_endpoint_t *endpoint, 
			    	nexus_buffer_t *buffer, 
			    	nexus_bool_t is_non_threaded_handler);

// Nexus Handler to bootstrap HPC++ by distributing the contextIDs
void hpcxx_initHandler(nexus_endpoint_t *endpoint,
                       nexus_buffer_t *buffer,
                       nexus_bool_t called_from_non_threaded_handler);

// Nexus Handler for Remote (global) Function invocation
void hpcxx_invokeGFHandler(nexus_endpoint_t *endpoint,
                           nexus_buffer_t *buffer,
                           nexus_bool_t called_from_non_threaded_handler);

// Nexus Handler for Remote Member Function invocation
void hpcxx_invokeMFHandler(nexus_endpoint_t *endpoint,
                           nexus_buffer_t *buffer,
                           nexus_bool_t called_from_non_threaded_handler);

// Nexus Handler for GlobalPointer reads
void hpcxx_readHandler(nexus_endpoint_t *endpoint,
				  nexus_buffer_t *buffer,
				  nexus_bool_t called_from_non_threaded_handler);

// Nexus Handler for GlobalPointer writes
void hpcxx_writeHandler(nexus_endpoint_t *endpoint,
				   nexus_buffer_t *buffer,
				   nexus_bool_t called_from_non_threaded_handler);

// Nexus Handler to acknowledge a received global communication
void hpcxx_ackHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler);

void hpcxx_returnHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler);

void hpcxx_barrierHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler);

void hpcxx_barrierOutHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler);

void hpcxx_lc_bufferMoveHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler);

void hpcxx_rc_bufferMoveHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler);

void hpcxx_p_bufferMoveHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler);

void hpcxx_sendGlobalFunction(HPCxx_Buffer &b, HPCxx_ContextID &target);
void hpcxx_returnGlobalFunction(HPCxx_Buffer &b, void *target);
bool hpcxx_contextid_equal(HPCxx_ContextID *id1, HPCxx_ContextID *id2);
void hpcxx_abort(int);
void hpcxx_msg(const char *msg);
void hpcxx_sendToBufferBlockRC(HPCxx_Buffer &b,HPCxx_ContextID & target);
void hpcxx_sendToBufferBlockLC(HPCxx_Buffer &b,HPCxx_ContextID & target);
void hpcxx_sendToBufferBlockP(HPCxx_Buffer &b, HPCxx_ContextID & target);

void hpcxx_pack(HPCxx_Buffer &, HPCxx_BufferBlock *, int);

int hpcxx_initAgent(int &argc, char **&argv);
HPCxx_ContextID * hpcxx_createContext(char *machine, char *exec_path);
int hpcxx_destroyContextID(HPCxx_ContextID *context);
int hpcxx_allowAttach(unsigned short attach_port, void (*callback)(void *), void *arg);
static int attach_requested(void *arg, char *url, nexus_startpoint_t *sp);
HPCxx_ContextID *hpcxx_grabContext(char *host, int port);

#endif

