// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: HPCxx_BufferImpl.h,v 1.1 1998/06/23 00:04:04 ejohnson Exp $

#ifndef __HPCXX_BUFFERIMPL_H_
#define __HPCXX_BUFFERIMPL_H_

#include <hpcxx_types.h>

class HPCxx_Buffer;

class HPCxx_BufferImpl {

public:
  HPCxx_BufferImpl(size_t bufSize, void* buffer = (void*) 0);
  ~HPCxx_BufferImpl(void);

  void*  buffer(void);
  void   copyBuffer(HPCxx_Buffer& buf, size_t size);
  void   reuse(size_t offset = 0);
  void   setAllocatedSize(size_t size);
  size_t size(void);

  int pack(const float *data, int count);
  int pack(const double *data, int count);
  int pack(const short *data, int count);
  int pack(const u_short *data, int count);
  int pack(const int *data, int count);
  int pack(const u_int *data, int count);
  int pack(const long *data, int count);
  int pack(const u_long *data, int count);
  int pack(const char *data, int count);
  int pack(const u_char *data, int count);

  void unpack(float *data, int count);
  void unpack(double *data, int count);
  void unpack(short *data, int count);
  void unpack(u_short *data, int count);
  void unpack(int *data, int count);
  void unpack(u_int *data, int count);
  void unpack(long *data, int count);
  void unpack(u_long *data, int count);
  void unpack(char *data, int count);
  void unpack(u_char *data, int count);

private:
  size_t   offset_;		// current offset into buffer
  size_t   size_;		// size of the buffer 
  char*    buffer_;		// address of buffer memory
  bool     doOwnBuffer_;	// ownership of buffer for freeing purposes

  void reallocIfNecessary(size_t size);
  int  copyToBuffer(const void* data, size_t size);
  void copyFromBuffer(void* data, size_t size);

};

#endif
