// HPC++ Run Time Sytem Library Extreme! Computing Lab - Computer
// Science Department Copyright (c) 1997 Indiana University Authors:
// Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: HPCxx_GlobalRefImpl.cc,v 1.2 1998/07/02 23:27:18 rasmussn Exp $

#include <hpcxx_rtsImpl.h>
#include <hpcxx_invokeImpl.h>
#include <HPCxx_GlobalPtr.h>
#include <HPCxx_GlobalRefImpl.h>
#include <HPCxx_CommCoord.h>
#include <HPCxx_Comm.h>
#include <hpcxx_rts.cc>			// for hpcxx_sizeof


//-----------------------------------------------------------------------------
// The pack and unpack routines for a global pointer are located here
// as the GlobalRefImpl class actually contains the data for the global
// pointer.
//-----------------------------------------------------------------------------


//---- hpcxx_pack -------------------------------------------------------------
// Packs an array of global pointers into the provided buffer.
//
// buffer -- the buffer which will receive the global pointers
// r      -- the global pointer to pack
// count  -- the number of global pointers to pack
//-----------------------------------------------------------------------------
template < class T >
void hpcxx_pack(HPCxx_Buffer& buffer, HPCxx_GlobalPtr<T>* r, int count)
{
  for (int i = 0; i < count; i++) {
    // pack the contextID and then the local pointer
    buffer.pack(&(r[i].globalRef_.contextID_), 1);
    buffer.pack((const unsigned long*) &(r[i].globalRef_.localPtr_), 1);
  }
}


//---- hpcxx_unpack -----------------------------------------------------------
// Unpacks an array of global pointers from the provided buffer.
//
// buffer -- the buffer which contains the global pointers
// r      -- the global pointer to receive the data
// count  -- the number of global pointers to unpack
//-----------------------------------------------------------------------------
template < class T >
void hpcxx_unpack(HPCxx_Buffer &buffer, HPCxx_GlobalPtr<T> *r, int count)
{
  for (int i = 0 ; i < count; i++) {
    buffer.unpack(&(r[i].globalRef_.contextID_), 1);
    buffer.unpack((unsigned long*) &(r[i].globalRef_.localPtr_), 1);
  }
}

//The classes below are used to provide type registration key ids.
// Default Constructor
template < class C>
int HPCxx_Type<C>::registrationID(void)
{
  return C::registrationID();
}

template<class T> 
HPCxx_GlobalRefImpl<T>::HPCxx_GlobalRefImpl(T* ptr, HPCxx_ContextIDImpl* id)
  : localPtr_(ptr)
{
  if (id == NULL) {
    if (hpcxx_currentContext == NULL) {
      // hpcxx_init may not have been called yet
      contextID_ = -1;
    } else {
      id = hpcxx_currentContext->getContextID();
      contextID_ = *id;
    }
  }

  offset_ = 0;
}

// Copy contructor
template<class T> HPCxx_GlobalRefImpl<T>::HPCxx_GlobalRefImpl(const HPCxx_GlobalRefImpl<T> &gr) 
  : localPtr_(gr.localPtr_), contextID_(gr.contextID_)
{
  offset_ = gr.offset_;
}

// Destructor  
template<class T>
HPCxx_GlobalRefImpl<T>::~HPCxx_GlobalRefImpl(void)
{
}

// convert a GlobalRefImpl to type T
template<class T>
HPCxx_GlobalRefImpl<T>::operator T() const
{
  if (!isLocal()) {
	T tmp;
	read(&tmp);
	return(tmp);
  }
  return(*localPtr_);
}

/**********
template<class T>
const T HPCxx_GlobalRefImpl<T>::operator*() const
{
  if (!isLocal()) {
	T tmp;
	read(&tmp);
	return(tmp);
  }
  return(*localPtr_);
}
**********/

// Cast Operators
template<class T> T* HPCxx_GlobalRefImpl<T>::getPointer() const
{
  if (!isLocal()) {
	return(NULL);
  }
  return(localPtr_);
}

template<class T> HPCxx_GlobalRefImpl<T>::operator HPCxx_ContextIDImpl& () const
{
  // The cast is used to get around const complaints from the compiler,
  // I think due to contextID_ being a private member variable.  Should
  // probably have this return a const HPCxx_ContextIDImpl&.

  return (HPCxx_ContextIDImpl&) contextID_;
}

template<class T> HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::
     operator=(const HPCxx_GlobalRefImpl<T> &gr)
{
  if (this == &gr) return *this;	// check for self assigment
  localPtr_  = gr.localPtr_;
  contextID_ = gr.contextID_;
  offset_ = gr.offset_;

  return *this;
}

template<class T>
HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::operator=(const T* ptr)
{
  // Cast used to get around const ackwardness, should this function even
  // be here?
  localPtr_ = (T*) ptr;
  contextID_ = *hpcxx_currentContext->getContextID();
  offset_ = 0;

  return *this;
}

// this one does not really belong to a "reference type" object.   it
// should really be part of the pointer impl, but we don't have a pointer
// impl.
template<class T>
HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::operator=(T* ptr)
{
  localPtr_ = ptr;
  contextID_ = *hpcxx_currentContext->getContextID();
  offset_ = 0;

  return *this;
}

template<class T> HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::operator=
  (T data)
{
  write(&data, 1);
  return *this;
  
}

// Test the equality of two GlobalPtr's
template<class T>
int HPCxx_GlobalRefImpl<T>::operator==(const HPCxx_GlobalRefImpl<T>& gr)
{
  return(hpcxx_contextid_equal(contextID_, gr.contextID_) &&
		 offset_ == gr.offset_);
}


template<class T>
int HPCxx_GlobalRefImpl<T>::operator==(const HPCxx_GlobalRefImpl<T>& gr) const
{
  return(hpcxx_contextid_equal(contextID_, gr.contextID_) &&
		 offset_ == gr.offset_);
}


template<class T>
int HPCxx_GlobalRefImpl<T>::operator!=(const HPCxx_GlobalRefImpl<T>& gr)
{
  return(!hpcxx_contextid_equal(contextID_, gr.contextID_) ||
		 !(offset_ == gr.offset_));
}


template<class T>
int HPCxx_GlobalRefImpl<T>::operator!=(const HPCxx_GlobalRefImpl<T>& gr) const
{
  return(!hpcxx_contextid_equal(contextID_, gr.contextID_) ||
		 !(offset_ == gr.offset_));
}


// FIXME!!! Pointer arithmetic really needs to go through rsr's so that
// the artithmetic occurs in the context in which the pointer is defined
// so that the correct size of the pointer can be determined. 
// (i.e. sizeof(T *) in the local context may not equal sizeof(T *) in
// the remote context

// GlobalPtr Arithmetic
// prefix ++
template<class T>
HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::operator++()
{
  localPtr_++;
  offset_++;
  return *this;
}

// postfix ++.  note. this one is never used! the job belongs to
// global pointer.
template<class T>
HPCxx_GlobalRefImpl<T> HPCxx_GlobalRefImpl<T>::operator++(int)
{
  HPCxx_GlobalRefImpl<T> tmp(*this);
  localPtr_++;
  offset_++;
  return tmp;
}

// prefix --
template<class T>
HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::operator--()
{
  localPtr_--;
  offset_--;
  return *this;
}

// postfix --
template<class T>
HPCxx_GlobalRefImpl<T> HPCxx_GlobalRefImpl<T>::operator--(int)
{
  HPCxx_GlobalRefImpl<T> tmp(*this);
  localPtr_--;
  offset_--;
  return tmp;
}

template<class T>
HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::operator+=(int inc)
{
  localPtr_ += inc;
  offset_ += inc;
  return *this;
}

template<class T>
HPCxx_GlobalRefImpl<T>& HPCxx_GlobalRefImpl<T>::operator-=(int dec)
{
  localPtr_ -= dec;
  offset_ -= dec;
  return *this;
}

template<class T>
HPCxx_GlobalRefImpl<T> HPCxx_GlobalRefImpl<T>::operator+(int inc)
{
  HPCxx_GlobalRefImpl<T> tmp(*this);
  tmp.localPtr_ += inc;
  tmp.offset_ += inc;
  return tmp;
}

template<class T>
HPCxx_GlobalRefImpl<T> HPCxx_GlobalRefImpl<T>::incr(int inc)
{
  HPCxx_GlobalRefImpl<T> tmp(*this);
  tmp.localPtr_ += inc;
  tmp.offset_ += inc;
  return tmp;
}

template<class T>
HPCxx_GlobalRefImpl<T> HPCxx_GlobalRefImpl<T>::operator-(int dec)
{
  HPCxx_GlobalRefImpl<T> tmp(*this);
  tmp.localPtr_ -= dec;
  tmp.offset_ -= dec;
  return tmp;
}
  
template<class T> int HPCxx_GlobalRefImpl<T>::operator<(const HPCxx_GlobalRefImpl<T>& gr)
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 < GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ < gr.localPtr_);
}

template<class T>
int HPCxx_GlobalRefImpl<T>::operator<(const HPCxx_GlobalRefImpl<T>& gr) const
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 < GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ < gr.localPtr_);
}

template<class T>
int HPCxx_GlobalRefImpl<T>::operator>(const HPCxx_GlobalRefImpl<T>& gr)
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 > GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ > gr.localPtr_);
}

template<class T>
int HPCxx_GlobalRefImpl<T>::operator>(const HPCxx_GlobalRefImpl<T>& gr) const
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 > GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ > gr.localPtr_);
}

template<class T>
int HPCxx_GlobalRefImpl<T>::operator<=(const HPCxx_GlobalRefImpl<T>& gr)
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 <= GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ <= gr.localPtr_);
}

template<class T>
int HPCxx_GlobalRefImpl<T>::operator<=(const HPCxx_GlobalRefImpl<T>& gr) const
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 <= GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ <= gr.localPtr_);
}

template<class T>
int HPCxx_GlobalRefImpl<T>::operator>=(const HPCxx_GlobalRefImpl<T>& gr)
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 >= GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ >= gr.localPtr_);
}

template<class T>
int HPCxx_GlobalRefImpl<T>::operator>=(const HPCxx_GlobalRefImpl<T>& gr) const
{
  // Contexts must be the same
  if (!hpcxx_contextid_equal(contextID_, gr.contextID_)) {
	hpcxx_msg("Operands of (GP1 >= GP2) must share a context");
	hpcxx_abort(HPCXX_FAILURE);
  }
  return(localPtr_ >= gr.localPtr_);
}

template<class T>
T HPCxx_GlobalRefImpl<T>::operator[](int index)
{
  if (!isLocal()) {
	T tmp;
	read(&tmp, 1, index);
	return(tmp);
  } else {
	return(*(localPtr_ + index));
  }
}

/*****
template<class T>
const T HPCxx_GlobalRefImpl<T>::operator[](int index) const
{
  if (!isLocal()) {
    T& tmp;
    read(&tmp, 1, index);
  } else {
    return(*(localPtr_ + index));
  }
}
******/


//---- read -------------------------------------------------------------------
// Gets remote data and copies it into the provided buffer, ptr.
//-----------------------------------------------------------------------------
template<class T>
void HPCxx_GlobalRefImpl<T>::read(T* ptr, int count, int offset) const
{
  HPCxx_ContextID sender = *hpcxx_currentContext->getContextID();

  // get the packed size of T in bytes
  size_t sizeT = hpcxx_sizeof(ptr[0]);
  int byteCount = count*sizeT;
  int byteOffset = offset*sizeT;

  // Check for purely local communication.  If local, this is inefficient
  // (it should have been just a local memory reference).

  if (isLocal()) {
    if (sizeT == sizeof(T)) {
      memcpy(ptr, localPtr_ + byteOffset, byteCount);
    } else {
      // FIXME
    }
    return;
  }

  // Get communicator objects to make the get request and to receive the
  // returned data.

  HPCxx_CommCoord& coord = hpcxx_commCoordinator();
  HPCxx_Comm* requester = coord.communicator(contextID_, AskForDataMsg, true);
  HPCxx_Comm* receiver = coord.communicator(contextID_, EndOfFetchMsg);

  // The receiver will be sent the same message id that the requester
  // used, so get it and set it.  Also, prepost the receive to ensure that
  // a buffer will be ready to receive the get, once it is requested.

  receiver -> setMessageID( requester->messageID() );
  receiver -> setSender(sender);
  receiver -> recv( (void*) ptr, byteCount );

  // now pack the message buffer with parameters and send the request

  HPCxx_Buffer& reqBuffer = requester->messageBuffer();
  reqBuffer.pack((const unsigned long*)&localPtr_, 1);
  reqBuffer.pack(&byteCount, 1);
  reqBuffer.pack(&byteOffset, 1);

  requester -> setSender(sender);
  requester -> send();

  // release control of the comm objects back to the coordinator
  coord.releaseToPosted(requester);
  coord.releaseToPreposted(receiver);

  // Wait until the receive is completed (the request is guaranteed to have
  // completed once the receive finishes).
  receiver -> wait();
}


template<class T>
void HPCxx_GlobalRefImpl<T>::write(T* ptr, int count, int offset)
{
  // get the packed size of T in bytes
  size_t sizeT = hpcxx_sizeof(ptr[0]);
  int byteCount = count*sizeT;
  int byteOffset = offset*sizeT;

  // Check for purely local communication.  If local, this is inefficient
  // (it should have been just a local memory reference).

  if (isLocal()) {
    if (sizeT == sizeof(T)) {
      memcpy(localPtr_ + byteOffset, ptr, byteCount);
    } else {
      // FIXME
    }
    return;
  }

  // get communicator objects to make the put message

  HPCxx_CommCoord& coord = hpcxx_commCoordinator();
  HPCxx_Comm* comm = coord.communicator(contextID_, EagerPutMsg);
  HPCxx_Buffer& buffer = comm->messageBuffer();

  // pack the global pointer, the count, offset, and data

  pack(buffer);
  buffer.pack(&byteCount);
  buffer.pack(&byteOffset);
  buffer.pack((char*)ptr, byteCount);

  // send the message
  comm -> setSender( *hpcxx_currentContext->getContextID() );
  comm -> send();

  // release control of the comm object back to the coordinator
  coord.releaseToPosted(comm);
  
  // wait until transfer buffer is released
  //  comm->wait();
}
  

//---- isLocal ----------------------------------------------------------------
// Returns true if the local pointer is local to our context, false otherwise.
//-----------------------------------------------------------------------------
template<class T>
bool HPCxx_GlobalRefImpl<T>::isLocal(void) const
{
  return (contextID_ == *hpcxx_currentContext->getContextID());
}


// Test to see if the pointer is local to our context
// ptr is a valid reference only if it is local
template<class T>
bool HPCxx_GlobalRefImpl<T>::isLocal(T*& ptr) const
{
  ptr = localPtr_;
  return(isLocal());
}


//---- pack -------------------------------------------------------------------
// Packs this object into the provided buffer.
//-----------------------------------------------------------------------------
template<class T>
void HPCxx_GlobalRefImpl<T>::pack(HPCxx_Buffer& buffer)
{
  // pack the contextID and then the local pointer
  buffer.pack(&contextID_);
  buffer.pack((const unsigned long*) &localPtr_);
}


//---- unpack -----------------------------------------------------------------
// Unpacks this object from the provided buffer.
//-----------------------------------------------------------------------------
template<class T>
void HPCxx_GlobalRefImpl<T>::unpack(HPCxx_Buffer& buffer)
{
  // unpack the contextID and then the local pointer
  buffer.unpack(&contextID_);
  buffer.unpack((unsigned long*) &localPtr_);
}


