// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: HPCxx_GlobalRefImpl.h,v 1.1 1998/06/23 00:04:25 ejohnson Exp $

#ifndef __HPCXX_GLOBALREFIMPL_H_
#define __HPCXX_GLOBALREFIMPL_H_

#include <hpcxx_types.h>
#include <HPCxx_Buffer.h>

template <class T>
class HPCxx_GlobalRefImpl {
public:
  T*                    localPtr_;	// pointer to the local data
  HPCxx_ContextIDImpl   contextID_;	// context to which the pointer belongs
  int                   offset_;

public:
  HPCxx_GlobalRefImpl(T* ptr = (T*) 0,
		      HPCxx_ContextIDImpl* id = (HPCxx_ContextIDImpl*) 0);
  HPCxx_GlobalRefImpl(const HPCxx_GlobalRefImpl<T>& gr);
  
  virtual ~HPCxx_GlobalRefImpl(void);

  // Deref a GlobalRefImpl to type T
  // T operator*();
  //const T operator*() const;

  // Conversion Operators
  operator T () const; 
  operator HPCxx_ContextIDImpl &() const;
  T * getPointer() const;

  HPCxx_GlobalRefImpl<T>& operator=(const HPCxx_GlobalRefImpl<T> &gr);
  HPCxx_GlobalRefImpl<T>& operator=(const T *ptr);
  HPCxx_GlobalRefImpl<T>& operator=(T *ptr);
  HPCxx_GlobalRefImpl<T>& operator=(T data);

  // Test the equality of two GlobalPtr's
  int operator==(const HPCxx_GlobalRefImpl<T>& gr);
  int operator==(const HPCxx_GlobalRefImpl<T>& gr) const;
  int operator!=(const HPCxx_GlobalRefImpl<T>& gr);
  int operator!=(const HPCxx_GlobalRefImpl<T>& gr) const;

  // GlobalPtr Arithmetic
  HPCxx_GlobalRefImpl<T>& operator++();         // prefix ++
  HPCxx_GlobalRefImpl<T> operator++(int);	// postfix ++

  HPCxx_GlobalRefImpl<T>& operator--();         // prefix --
  HPCxx_GlobalRefImpl<T> operator--(int);	// postfix --

  HPCxx_GlobalRefImpl<T>& operator+=(int inc);
  HPCxx_GlobalRefImpl<T>& operator-=(int dec);
  HPCxx_GlobalRefImpl<T> operator+(int inc);
  HPCxx_GlobalRefImpl<T> operator-(int dec);
  HPCxx_GlobalRefImpl<T> incr(int dec);
  
  int operator<(const HPCxx_GlobalRefImpl<T>& gr);
  int operator<(const HPCxx_GlobalRefImpl<T>& gr) const;
  int operator>(const HPCxx_GlobalRefImpl<T>& gr);
  int operator>(const HPCxx_GlobalRefImpl<T>& gr) const;
  int operator<=(const HPCxx_GlobalRefImpl<T>& gr);
  int operator<=(const HPCxx_GlobalRefImpl<T>& gr) const;
  int operator>=(const HPCxx_GlobalRefImpl<T>& gr);
  int operator>=(const HPCxx_GlobalRefImpl<T>& gr) const;

  T operator[](int index);
  // const T operator[](int index) const;
  
  // Block data transfer
  void read(T *ptr, int count = 1, int offset = 0) const;
  void write(T *ptr, int count = 1, int offset = 0);
  
  // Test to see if the pointer is local to our context
  bool isLocal() const;

  // Test to see if the pointer is local to our context
  // ptr is a valid reference only if it is local
  bool isLocal(T *&ptr) const;

  // pack and unpack this object
  void pack(HPCxx_Buffer& buffer);
  void unpack(HPCxx_Buffer& buffer);

};


//-----------------------------------------------------------------------------
// The classes below are used to provide type registration key ids.
//-----------------------------------------------------------------------------

template < class C>
class HPCxx_Type {
  public:
     HPCxx_Type(void) { }
     static int registrationID();
};

class HPCxx_Type<int> {
  public:
    static int registrationID(void) { return HPCXX_INT_KEY; }
};

class HPCxx_Type<float> {
  public:
    static int registrationID(void) { return HPCXX_FLOAT_KEY; }
};

class HPCxx_Type<double> {
  public:
    static int registrationID(void) { return HPCXX_DOUBLE_KEY; }
};

class HPCxx_Type<char> {
  public:
    static int registrationID(void) { return HPCXX_CHAR_KEY; }
};

class HPCxx_Type<u_char> {
  public:
    static int registrationID(void) { return HPCXX_U_CHAR_KEY; }
};

class HPCxx_Type<u_int> {
  public:
    static int registrationID(void) { return HPCXX_U_INT_KEY; }
};

class HPCxx_Type<short> {
  public:
    static int registrationID(void) { return HPCXX_SHORT_KEY; }
};

class HPCxx_Type<long> {
  public:
    static int registrationID(void) { return HPCXX_LONG_KEY; }
};

class HPCxx_Type<u_short> {
  public:
    static int registrationID(void) { return HPCXX_U_SHORT_KEY; }
};

class HPCxx_Type<u_long> {
  public:
    static int registrationID(void) { return HPCXX_U_LONG_KEY;}
};

#endif
