// HPC++ Run Time Sytem Library (nexus)
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: hpcxx_rtsImpl.h,v 1.1 1998/06/23 00:04:48 ejohnson Exp $

#ifndef __HPCXX_RTSIMPL_H_
#define __HPCXX_RTSIMPL_H_

#include "HPCxx_Buffer.h"	// for hpcxx_pack and hpcxx_unpack routines
#include <HPCxx_CommCoord.h>
#include <HPCxx_Context.h>

extern HPCxx_Context* hpcxx_currentContext;

// Handler for RMI (Remote Method Invocation)
void hpcxx_invokeRMIHandler(HPCxx_Comm& comm);

// Handler for RFI (Remote (global) Function Invocation)
void hpcxx_invokeGFHandler(HPCxx_Comm& comm);

// Handler for return of RFI (Remote Function Invocation)
void hpcxx_returnHandler(HPCxx_Comm& comm);

// Handlers for barrier fan in and fan out
void hpcxx_barrierFanIn(HPCxx_Comm& comm);
void hpcxx_barrierFanOut(HPCxx_Comm& comm);

// Handlers for global pointer read and write
void hpcxx_readHandler(HPCxx_Comm& comm);
void hpcxx_writeHandler(HPCxx_Comm& comm);


/******

extern tulip_handler_t hpcxx_tulipHandlerTable[HPCXX_NUM_HANDLERS];

// Tulip Handler to bootstrap HPC++ by distributing the contextIDs
void hpcxx_initHandler(tulip_endpoint_t *endpoint,
                       void *buffer,
                       bool called_from_non_threaded_handler);

// Tulip Handler for Remote Member Function invocation
void hpcxx_invokeMFHandler(tulip_endpoint_t *endpoint,
                           void* buffer,
                           bool called_from_non_threaded_handler);

// Tulip Handler for GlobalPointer reads
void hpcxx_readHandler(tulip_endpoint_t *endpoint,
				  void* buffer,
				  bool called_from_non_threaded_handler);

// Tulip Handler for GlobalPointer writes
void hpcxx_writeHandler(tulip_endpoint_t *endpoint,
				   void* buffer,
				   bool called_from_non_threaded_handler);

// Tulip Handler to acknowledge a received global communication
void hpcxx_ackHandler(tulip_endpoint_t *endpoint,
                      void* buffer,
                      bool called_from_non_threaded_handler);

void hpcxx_returnHandler(tulip_endpoint_t *endpoint,
                      void* buffer,
                      bool called_from_non_threaded_handler);

void hpcxx_barrierHandler(tulip_endpoint_t *endpoint,
                      void* buffer,
                      bool called_from_non_threaded_handler);

void hpcxx_barrierOutHandler(tulip_endpoint_t *endpoint,
                      void* buffer,
                      bool called_from_non_threaded_handler);

void hpcxx_lc_bufferMoveHandler(tulip_endpoint_t *endpoint,
                      void* buffer,
                      bool called_from_non_threaded_handler);

void hpcxx_rc_bufferMoveHandler(tulip_endpoint_t *endpoint,
                      void* buffer,
                      bool called_from_non_threaded_handler);

void hpcxx_p_bufferMoveHandler(tulip_endpoint_t *endpoint,
                      void* buffer,
                      bool called_from_non_threaded_handler);

******/


void hpcxx_sendGlobalFunction(HPCxx_Buffer &b, HPCxx_ContextID &target);
void hpcxx_returnGlobalFunction(HPCxx_Buffer &b, void* target);
bool hpcxx_contextid_equal(HPCxx_ContextID id1, HPCxx_ContextID id2);
void hpcxx_abort(int);
void hpcxx_msg(const char *msg);
void hpcxx_sendToBufferBlockRC(HPCxx_Buffer &b,HPCxx_ContextID & target);
void hpcxx_sendToBufferBlockLC(HPCxx_Buffer &b,HPCxx_ContextID & target);
void hpcxx_sendToBufferBlockP(HPCxx_Buffer &b, HPCxx_ContextID & target);

hpcxx_id_t hpcxx_threadID(void);
HPCxx_CommCoord& hpcxx_commCoordinator(void);

int hpcxx_initAgent(int &argc, char **&argv);
HPCxx_ContextID * hpcxx_makeContext(char *machine, char *exec_path);

#endif

