// HPC++ Run Time Sytem Library (nexus)
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: hpcxx_rtsImpl.h,v 1.2 1998/06/24 21:55:39 khalsa Exp $

#ifndef __HPCXX_RTSIMPL_H_
#define __HPCXX_RTSIMPL_H_

#include "HPCxx_Buffer.h"// for hpcxx_pack and hpcxx_unpack routines
#include "HPCxx_Group.h"

const int HPCXX_INIT_HANDLER       = 0;
const int HPCXX_BARRIER_IN_HANDLER = 1;
const int HPCXX_BARRIER_OUT_HANDLER= 2;
const int HPCXX_INVOKE_GF_HANDLER  = 3;
const int HPCXX_RETURN_HANDLER     = 4;
const int HPCXX_TEST_HANDLER       = 5;
const int HPCXX_INVOKE_MF_HANDLER  = 6;
const int HPCXX_READ_HANDLER       = 7;
const int HPCXX_WRITE_HANDLER      = 8;

const int HPCXX_NUM_HANDLERS       = 9;

extern nexus_handler_t hpcxx_nexusHandlerTable[HPCXX_NUM_HANDLERS];
nexus_endpoint_t *hpcxx_get_ep();
nexus_startpoint_t *hpcxx_get_sp();
extern HPCxx_ContextID         *hpcxx_currentContextID;

void hpcxx_pack(HPCxx_Buffer &, HPCxx_BufferBlock *, int);

int hpcxx_init(int &argc, char **&argv, HPCxx_Group*&);
int hpcxx_exit(HPCxx_Group*& group);
void hpcxx_barrier();

// Nexus Hanlder to bootstrap HPC++ by distributing the contextIDs
void hpcxx_initHandler(nexus_endpoint_t *endpoint,
                       nexus_buffer_t *buffer,
                       nexus_bool_t called_from_non_threaded_handler);

void hpcxx_barrierHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler);

void hpcxx_barrierOutHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler);

 // Nexus Handler for Remote (global) Function invocation
void hpcxx_invokeGFHandler(nexus_endpoint_t *endpoint,
                           nexus_buffer_t *buffer,
                           nexus_bool_t called_from_non_threaded_handler);


void hpcxx_returnHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler);

void hpcxx_testHandler(nexus_endpoint_t *endpoint,
                      nexus_buffer_t *buffer,
                      nexus_bool_t called_from_non_threaded_handler);

void hpcxx_invokeMFHandler(nexus_endpoint_t *endpoint,
                           nexus_buffer_t *buffer,
                           nexus_bool_t called_from_non_threaded_handler);


// Nexus Handler for GlobalPointer reads
void hpcxx_readHandler(nexus_endpoint_t *endpoint,
		       nexus_buffer_t *buffer,
		       nexus_bool_t called_from_non_threaded_handler);

// Nexus Handler for GlobalPointer writes
void hpcxx_writeHandler(nexus_endpoint_t *endpoint,
			nexus_buffer_t *buffer,
			nexus_bool_t called_from_non_threaded_handler);

void hpcxx_sendMemberFunction(HPCxx_Buffer &b, HPCxx_ContextID &target);
void hpcxx_sendGlobalFunction(HPCxx_Buffer &b, HPCxx_ContextID &target);
void hpcxx_returnGlobalFunction(HPCxx_Buffer &b, void *target);
bool hpcxx_contextid_equal(HPCxx_ContextID *id1, HPCxx_ContextID *id2);
void hpcxx_abort(int);
void hpcxx_msg(const char *msg);

#endif
