#include <hpcxx_rts.h>
#include <hpcxx_register.h>
#include <HPCxx_GlobalPtr.h>
#include <HPCxx_Prototype.h>
#include <stdarg.h>
#include <map.h>


map<int, HPCxx_GenericPtr*> gpMap;
static int GPFindID = 1001;


// WARNING this buffer size thing is a problem, you have to be able to
// allocate the buffer before you know the size it needs to be.  One
// could parse the args first to find the size, right now just punt
// and make it a constant size.
#define INVOKE_BUF_SIZE 512


static inline int packVarArg(va_list& ap, HPCxx_Buffer& buf, int type)
{
  int error = 0;

  switch(type) {
    case HPCXX_TYPE_VOID: {
      break;	// nothing to pack
    }
    case HPCXX_TYPE_INT: {
      int* iVal = va_arg(ap, int*);
      hpcxx_pack(buf, iVal, 1);
      break;
    }
    case HPCXX_TYPE_FLOAT: {
      float* rVal = va_arg(ap, float*);
      hpcxx_pack(buf, rVal, 1);
      break;
    }
    case HPCXX_TYPE_DOUBLE: {
      double* dVal = va_arg(ap, double*);
      hpcxx_pack(buf, dVal, 1);
      break;
    }
    default:
      fprintf(stderr, "packVarArg: unknown arg type %d\n", type);
      error = 1;
  }
  return error;
}


extern "C" {

// C prototypes
int iargc_(void);
void getarg_(int*, char*, int);
HPCxx_GenericPtr* GlobalPtr_find(int id);


void hpcxx_init_(HPCxx_Group** group, int* error)
{
  // get argc and argv from fortran functions iargc_ and getarg_

  int i;
  int argsize = 1024;
  int argc = iargc_() + 1;
  char** argv = (char **) malloc( argc * sizeof(char*) );    

  for (i = 0; i < argc; i++) {
    argv[i] = (char*) malloc(argsize + 1);
    getarg_(&i, argv[i], argsize);

    // trim trailing blanks
    char* p = argv[i] + argsize - 1;
    while (p >= argv[i]) {
      if (*p != ' ' && *p != '\0') {
	p[1] = '\0';
	break;
      }
      p--;
    }
  }

  *error = hpcxx_init(argc, argv, *group);

  // register the global pointer find function
  GPFindID = hpcxx_register(GlobalPtr_find, GPFindID);

  for (i = 0; i < argc; i++) {
    free(argv[i]);
  }
  free(argv);
}


void hpcxx_exit_(HPCxx_Group** group, int* error)
{
  *error = hpcxx_exit(*group);
}


void hpcxx_barrier_(int* error)
{
  hpcxx_barrier();
  *error = 0;
 }


HPCxx_GenericPtr* GlobalPtr_find(int id)
{
  return gpMap[id];
}


void globalptr_find_(HPCxx_GenericPtr** gp,
		     HPCxx_ContextID** context, int* id, int* error)
{
  // return gp from local map if context is local

  if (hpcxx_contextid_equal(*context, hpcxx_currentContext->getContextID())) {
    *gp = gpMap[*id];
    return;
  }

  *error = hpcxx_invoke(**context, *gp, GPFindID, *id);
}


// can't use this from fortran
void hpcxx_register_rmfi_(int* rmfi_id, int* error)
{
  *error = 0;
} 


void hpcxx_invoke_rmfi_( HPCxx_ContextID** target,
			 int* rmfi_id, int* error, ...)
{
  va_list ap;
  va_start(ap, error);

  // WARNING this buffer size thing is a problem, you have to be able to
  // allocate the buffer before you know the size it needs to be.  One
  // could parse the args first to find the size, right now just punt
  // and make it a constant size.
  HPCxx_Buffer buf(INVOKE_BUF_SIZE);

  // get function prototype/selector using the id

  HPCxx_Prototype& proto = hpcxx_prototypeMap[*rmfi_id];
  int nArgs = proto.numberOfArgs();

  *error = 0;
  switch(proto.returnType()) {
    case HPCXX_TYPE_INT: {
      HPCxx_Sync<int> result_sync;
      HPCxx_GlobalPtr<HPCxx_Sync<int> > gp_to_res_sync(&result_sync);
      hpcxx_pack(buf, &gp_to_res_sync, 1);
      hpcxx_pack(buf, rmfi_id, 1);

      int* result = va_arg(ap, int*);

      // pack the args, invoke the function, and wait for the results

      for (int i = 0; i < nArgs; i++) packVarArg(ap, buf, proto.argTypeN(i));
      hpcxx_sendGlobalFunction(buf, **target); 
      result_sync.read(*result);
      break;
    }
    case HPCXX_TYPE_FLOAT: {
      HPCxx_Sync<float> result_sync;
      HPCxx_GlobalPtr<HPCxx_Sync<float> > gp_to_res_sync(&result_sync);
      hpcxx_pack(buf, &gp_to_res_sync, 1);
      hpcxx_pack(buf, rmfi_id, 1);

      float* result = va_arg(ap, float*);

      // pack the args, invoke the function, and wait for the results

      for (int i = 0; i < nArgs; i++) packVarArg(ap, buf, proto.argTypeN(i));
      hpcxx_sendGlobalFunction(buf, **target); 
      result_sync.read(*result);
      break;
    }
    case HPCXX_TYPE_DOUBLE: {
      HPCxx_Sync<double> result_sync;
      HPCxx_GlobalPtr<HPCxx_Sync<double> > gp_to_res_sync(&result_sync);
      hpcxx_pack(buf, &gp_to_res_sync, 1);
      hpcxx_pack(buf, rmfi_id, 1);

      double* result = va_arg(ap, double*);

      // pack the args, invoke the function, and wait for the results

      for (int i = 0; i < nArgs; i++) packVarArg(ap, buf, proto.argTypeN(i));
      hpcxx_sendGlobalFunction(buf, **target); 
      result_sync.read(*result);
      break;
    }
    default: *error = 1;
  }
  va_end(ap);
}


void hpcxx_is_master_context_(HPCxx_Context** context, int* flag, int* error)
{
  *flag = ( (*context)->isMasterContext() ) ? 1 : 0;
  *error = 0;
}


void hpcxx_number_of_contexts_(HPCxx_Group** group, int* number, int* error)
{
  *number = (*group)->getNumThreads();
  *error = 0;
}


void hpcxx_my_context_number_(HPCxx_Context** context)
{
}


/****************

void hpcxx_create_context_(HPCxx_ContextID** context, char* machine,
                           char* exec_path, int* error)
{
  *error = 0;
  *context = hpcxx_createContext(machine, exec_path);
  if (*context == NULL) {
    *error = 1;
  }
}


void hpcxx_destroy_context_(HPCxx_ContextID** context, int* error)
{
  *error = hpcxx_destroyContextID(*context);
}

******************/


int* globalptr_new_(HPCxx_GenericPtr** gp,
		    int* type, void* ptr, int* id, int* error)
{
  *error = 0;
  switch (*type) {
    case HPCXX_TYPE_INT: {
      *gp = new HPCxx_GlobalPtr<int>((int*) ptr);
      break;
    }
    case HPCXX_TYPE_FLOAT: {
      *gp = new HPCxx_GlobalPtr<float>((float*) ptr);
      break;
    }
    case HPCXX_TYPE_DOUBLE: {
      *gp = new HPCxx_GlobalPtr<double>((double*) ptr);
      break;
    }
    default:
      *gp = NULL;
      *error = 1;
  }
  gpMap[*id] = *gp;
  return id;
}


void globalptr_delete_(HPCxx_GenericPtr** gp, int* error)
{
  *error = 0;
  delete *gp;
}


void globalptr_read_(HPCxx_GenericPtr** gp,
		     void* buffer, int* count, int* offset, int* error)
{
  *error = 0;

  switch ( (*gp)->type() ) {
    case HPCXX_TYPE_INT: {
      ((HPCxx_GlobalPtr<int>*) *gp)->read((int*) buffer, *count, *offset);
      break;
    }
    case HPCXX_TYPE_FLOAT: {
      ((HPCxx_GlobalPtr<float>*) *gp)->read((float*) buffer, *count, *offset);
      break;
    }
    case HPCXX_TYPE_DOUBLE: {
      ((HPCxx_GlobalPtr<double>*) *gp)->read((double*)buffer, *count, *offset);
      break;
    }
    default:
      fprintf(stderr, "globalptr_read_: unknown type = %d\n", (*gp)->type());
      *error = 1;
  }
}


void globalptr_write_(HPCxx_GenericPtr** gp,
		      void* buffer, int* count, int* offset, int* error)
{
  *error = 0;

  switch ( (*gp)->type() ) {
    case HPCXX_TYPE_INT: {
      ((HPCxx_GlobalPtr<int>*) *gp)->write((int*) buffer, *count, *offset);
      break;
    }
    case HPCXX_TYPE_FLOAT: {
      ((HPCxx_GlobalPtr<float>*) *gp)->write((float*) buffer, *count, *offset);
      break;
    }
    case HPCXX_TYPE_DOUBLE: {
      ((HPCxx_GlobalPtr<double>*) *gp)->write((double*)buffer,*count,*offset);
      break;
    }
    default:
      fprintf(stderr, "globalptr_write_: unknown type = %d\n", (*gp)->type());
      *error = 1;
  }
}


}; // extern "C"


//
// force the necessary template instantiations to happen
//

#include <HPCxx_Sync.cc>
#include <hpcxx_invoke.cc>
#include <hpcxx_invokeImpl.cc>
#include <hpcxx_register.cc>
#include <HPCxx_GlobalPtr.cc>
#include <HPCxx_GlobalRefImpl.cc>

template class map<int, HPCxx_GenericPtr*>;

template class HPCxx_Sync<int>;
template class HPCxx_Sync<float>;
template class HPCxx_Sync<double>;
template class HPCxx_GlobalPtr<int>;
template class HPCxx_GlobalPtr<float>;
template class HPCxx_GlobalPtr<double>;
template class HPCxx_GlobalRefImpl<int>;
template class HPCxx_GlobalRefImpl<float>;
template class HPCxx_GlobalRefImpl<double>;
template class HPCxx_GlobalPtr<HPCxx_Sync<int> >;
template class HPCxx_GlobalPtr<HPCxx_Sync<float> >;
template class HPCxx_GlobalPtr<HPCxx_Sync<double> >;
template class HPCxx_GlobalRefImpl<HPCxx_Sync<int> >;
template class HPCxx_GlobalRefImpl<HPCxx_Sync<float> >;
template class HPCxx_GlobalRefImpl<HPCxx_Sync<double> >;

template void hpcxx_pack(HPCxx_Buffer&,HPCxx_Sync<int>*,int);
template void hpcxx_pack(HPCxx_Buffer&,HPCxx_Sync<float>*,int);
template void hpcxx_pack(HPCxx_Buffer&,HPCxx_Sync<double>*,int);
template void hpcxx_pack(HPCxx_Buffer&,HPCxx_GlobalPtr<int>*,int);
template void
hpcxx_pack(HPCxx_Buffer&,HPCxx_GlobalPtr<HPCxx_Sync<int> >*,int);
template void
hpcxx_pack(HPCxx_Buffer&,HPCxx_GlobalPtr<HPCxx_Sync<float> >*,int);
template void
hpcxx_pack(HPCxx_Buffer&,HPCxx_GlobalPtr<HPCxx_Sync<double> >*,int);

template hpcxx_id_t HPCxx_Type<HPCxx_Sync<int> >::registrationID();
template hpcxx_id_t HPCxx_Type<HPCxx_Sync<float> >::registrationID();
template hpcxx_id_t HPCxx_Type<HPCxx_Sync<double> >::registrationID();

template hpcxx_id_t hpcxx_register(HPCxx_GenericPtr* (*)(int),int);
template int hpcxx_invoke(HPCxx_ContextID&,HPCxx_GenericPtr*&,int,int&);

template class HPCxx_Sync<HPCxx_GenericPtr*>;
template class HPCxx_Type<HPCxx_Sync<HPCxx_GenericPtr*> >;
template class HPCxx_GlobalPtr<HPCxx_Sync<HPCxx_GenericPtr*> >;
template class HPCxx_GlobalRefImpl<HPCxx_GenericPtr*>;
template class HPCxx_GlobalRefImpl<HPCxx_Sync<HPCxx_GenericPtr*> >;

template void
hpcxx_pack(HPCxx_Buffer&,HPCxx_GlobalPtr<HPCxx_Sync<HPCxx_GenericPtr*> >*,int);
template void hpcxx_pack(HPCxx_Buffer&,HPCxx_Sync<HPCxx_GenericPtr*>*,int);


int HPCxx_Type<HPCxx_GenericPtr*>::registrationID(void)
{
  return HPCxx_GenericPtr::registrationID();
}
