//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: hpcxx_registerFTN.cc,v 1.1 1998/08/08 00:14:19 rasmussn Exp $
//##########################################################################

#include <hpcxx_rts.h>
#include <hpcxx_registerFTN.h>
#include <hpcxx_rtsImpl.h>
#include <HPCxx_Buffer.h>
#include <HPCxx_RFITable.h>


//
// Remote global function pointer class
//
template <class R>
class HPCxx_RFFPointerZeroArgs : public HPCxx_RFITableEntry {
private:
  R (*fp_)();
public:
  HPCxx_RFFPointerZeroArgs(R fp(void)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer& b) {
      R r;
      r = (*fp_)();
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1); 
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void* sync, HPCxx_Buffer& b) {
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R>* rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
   }
};

template <class R, class A1>
class HPCxx_RFFPointerOneArg : public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1);

public:
  HPCxx_RFFPointerOneArg(R fp(A1)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b) {
      A1 a1;
      R r;
      hpcxx_unpack(b, &a1, 1);
      r = (* (R (*)(A1*)) fp_)(&a1);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1); 
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

template <class R, class A1, class A2>
class HPCxx_RFFPointerTwoArgs: public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1, A2);

public:
  HPCxx_RFFPointerTwoArgs(R fp(A1, A2)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b ){
      A1 a1;
      A2 a2;
      R r;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      r = (* (R (*)(A1*, A2*)) fp_)(&a1, &a2);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

template <class R, class A1, class A2, class A3>
class HPCxx_RFFPointerThreeArgs: public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1, A2, A3);

public:
  HPCxx_RFFPointerThreeArgs(R fp(A1, A2, A3)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b ){
      A1 a1;
      A2 a2;
      A3 a3;
      R r;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      r = (* (R (*)(A1*, A2*, A3*)) fp_)(&a1, &a2, &a3);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

template <class R, class A1, class A2, class A3, class A4>
class HPCxx_RFFPointerFourArgs: public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1, A2, A3, A4);

public:
  HPCxx_RFFPointerFourArgs(R fp(A1, A2, A3, A4)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b ){
       A1 a1;
       A2 a2;
       A3 a3;
       A4 a4;
      R r;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      hpcxx_unpack(b, &a4, 1);
      r = (* (R (*)(A1*, A2*, A3*, A4*)) fp_)(&a1, &a2, &a3, &a4);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

template <class R, class A1, class A2, class A3, class A4, class A5>
class HPCxx_RFFPointerFiveArgs: public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1, A2, A3, A4, A5);

public:
  HPCxx_RFFPointerFiveArgs(R fp(A1, A2, A3, A4, A5)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b ){
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      A5 a5;
      R r;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      hpcxx_unpack(b, &a4, 1);
      hpcxx_unpack(b, &a5, 1);
      r = (* (R (*)(A1*, A2*, A3*, A4*, A5*)) fp_)(&a1, &a2, &a3, &a4, &a5);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

template <class R, class A1, class A2, class A3, class A4, class A5, class A6>
class HPCxx_RFFPointerSixArgs: public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1, A2, A3, A4, A5, A6);

public:
  HPCxx_RFFPointerSixArgs(R fp(A1, A2, A3, A4, A5, A6)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b ){
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      A5 a5;
      A6 a6;
      R r;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      hpcxx_unpack(b, &a4, 1);
      hpcxx_unpack(b, &a5, 1);
      hpcxx_unpack(b, &a6, 1);
      r = (* (R (*)(A1*, A2*, A3*, A4*, A5*, A6*)) fp_)
						(&a1, &a2, &a3, &a4, &a5, &a6);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

template <class R, class A1, class A2, class A3, class A4, class A5, class A6,
          class A7>
class HPCxx_RFFPointerSevenArgs: public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1, A2, A3, A4, A5, A6, A7);

public:
  HPCxx_RFFPointerSevenArgs(R fp(A1, A2, A3, A4, A5, A6, A7)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b ){
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      A5 a5;
      A6 a6;
      A7 a7;
      R r;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      hpcxx_unpack(b, &a4, 1);
      hpcxx_unpack(b, &a5, 1);
      hpcxx_unpack(b, &a6, 1);
      hpcxx_unpack(b, &a7, 1);
      r = (* (R (*)(A1*, A2*, A3*, A4*, A5*, A6*, A7*)) fp_)
					(&a1, &a2, &a3, &a4, &a5, &a6, &a7);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

template <class R, class A1, class A2, class A3, class A4, class A5, class A6,
          class A7, class A8>
class HPCxx_RFFPointerEightArgs: public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1, A2, A3, A4, A5, A6, A7, A8);

public:
  HPCxx_RFFPointerEightArgs(R fp(A1, A2, A3, A4, A5, A6, A7, A8)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b ){
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      A5 a5;
      A6 a6;
      A7 a7;
      A8 a8;
      R r;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      hpcxx_unpack(b, &a4, 1);
      hpcxx_unpack(b, &a5, 1);
      hpcxx_unpack(b, &a6, 1);
      hpcxx_unpack(b, &a7, 1);
      hpcxx_unpack(b, &a8, 1);
      r = (* (R (*)(A1*, A2*, A3*, A4*, A5*, A6*, A7*, A8*)) fp_)
				(&a1, &a2, &a3, &a4, &a5, &a6, &a7, &a8);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};


//-----------------------------------------------------------------------------
// Register a global function that takes no arguments.
//-----------------------------------------------------------------------------
template<class R>
hpcxx_id_t hpcxx_registerFTN(R (*f) (void), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
	        new HPCxx_RFFPointerZeroArgs<R>(f), registerID);
}


//-----------------------------------------------------------------------------
// Register a global function that takes one argument.
//-----------------------------------------------------------------------------
template<class R, class A1>
hpcxx_id_t hpcxx_registerFTN(R (*f) (A1), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
                new HPCxx_RFFPointerOneArg<R,A1>(f), registerID);
}


//-----------------------------------------------------------------------------
// Register a global function that takes two arguments.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2>
hpcxx_id_t hpcxx_registerFTN(R (*f) (A1, A2), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
                new HPCxx_RFFPointerTwoArgs<R,A1,A2>(f), registerID);
}


//-----------------------------------------------------------------------------
// Register a global function that takes three arguments.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3>
hpcxx_id_t hpcxx_registerFTN(R (*f) (A1, A2, A3), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
                new HPCxx_RFFPointerThreeArgs<R,A1,A2,A3>(f), registerID);
}


//-----------------------------------------------------------------------------
// Register a global function that takes four arguments.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4>
hpcxx_id_t hpcxx_registerFTN(R (*f) (A1, A2, A3, A4), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
                new HPCxx_RFFPointerFourArgs<R,A1,A2,A3,A4>(f), registerID);
}

//-----------------------------------------------------------------------------
// Register a global function that takes five arguments.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4, class A5>
hpcxx_id_t hpcxx_registerFTN( R (*f) (A1, A2, A3, A4, A5),
			      hpcxx_id_t registerID )
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
                new HPCxx_RFFPointerFiveArgs<R,A1,A2,A3,A4,A5>(f), registerID);
}


//-----------------------------------------------------------------------------
// Register a global function that takes six arguments.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4, class A5, class A6>
hpcxx_id_t hpcxx_registerFTN( R (*f) (A1, A2, A3, A4, A5, A6),
			      hpcxx_id_t registerID )
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
              new HPCxx_RFFPointerSixArgs<R,A1,A2,A3,A4,A5,A6>(f), registerID);
}


//-----------------------------------------------------------------------------
// Register a global function that takes seven arguments.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2,
         class A3, class A4, class A5, class A6, class A7>
hpcxx_id_t hpcxx_registerFTN( R (*f) (A1, A2, A3, A4, A5, A6, A7),
			   hpcxx_id_t registerID )
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
     new HPCxx_RFFPointerSevenArgs<R,A1,A2,A3,A4,A5,A6,A7>(f), registerID);
}


//-----------------------------------------------------------------------------
// Register a global function that takes eight arguments.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2,
         class A3, class A4, class A5, class A6, class A7, class A8>
hpcxx_id_t hpcxx_registerFTN( R (*f) (A1, A2, A3, A4, A5, A6, A7, A8),
			   hpcxx_id_t registerID )
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
     new HPCxx_RFFPointerEightArgs<R,A1,A2,A3,A4,A5,A6,A7,A8>(f), registerID);
}
