//#----------------------------------------------------------------------------
//# Remote.h - Declarations for the Remote category of classes.
//#
//# Copyright (c) 1997, Los Alamos National Laboratory
//# 
//# http://www.acl.lanl.gov/Tulip
//# 
//# Pete Beckman, Craig Rasmussen
//#
//#----------------------------------------------------------------------------
//# $Id: Remote.h,v 1.1 1998/06/23 00:05:46 ejohnson Exp $
//#----------------------------------------------------------------------------

#ifndef __HPCXX_REMOTE_H_
#define __HPCXX_REMOTE_H_

//#include "Utility/bstring.h"


//#----------------------------------------------------------------------------
// <summary> Abstract base class for Remote Objects </summary>
//
// Abstract base class for Remote objects.  Derived classes can use multiple
// inheritance to mix this class in with existing classes.
//#----------------------------------------------------------------------------
class Remote
{
  public: virtual ~Remote(void) { }
  //# java has empty interface for this class?
};


//#----------------------------------------------------------------------------
// Abstract base class for Registry objects.
//#----------------------------------------------------------------------------
class Registry : public Remote
{
//#  public: virtual void bind(const string& name, Remote* pRemote) = 0;
//#  public: virtual vector<string&> list(void) = 0;
//#  public: virtual Remote* lookup(const string& name) = 0;
};


//#----------------------------------------------------------------------------
// Concrete class implementing the Registry interface for obtaining remote
// objects.  This class also pretends to be an ORB by running threads for
// communication between remote objects.
//#----------------------------------------------------------------------------
//
class ParallelRegistry : public Registry		// regp
{
  // flag to allow exiting of polling loop
  static bool fExitPoll;

  // Constructs a registry and runs a number of contexts.
  // <dd><li> nContexts -- the number of contexts to run
  public: ParallelRegistry(int nContexts);
  public: ~ParallelRegistry(void);

  //#  public: void bind(const string& name, Remote* pRemote);

  // Causes all worker contexts to exit their polling loop and terminate.
  public: static void exitPolling(void)		{ fExitPoll = true; }

  //#  public: vector<string&> list(void);
  //#  public: Remote* lookup(const string& name);

  // Returns true if context should exit polling loop.
  public: static bool shouldExitPolling(void)	{ return fExitPoll; }
};


//#----------------------------------------------------------------------------
// Factory class to create ParallelRegistry objects.
//#----------------------------------------------------------------------------
class LocateParallelRegistry
{
  // Returns the registry object.  If one isn't currently running, a registry
  // will be created.
  // <dd><li> port -- the port to use (currently not implemented)
  // <dd><li> nContexts -- the number of contexts to run
  public: static Registry* createRegistry(int port, int nContexts);
//#  public: static Registry* getRegistry(int port, const string& host);
};


//#----------------------------------------------------------------------------
// Base class for shared/distributed objects.
//#----------------------------------------------------------------------------
class RemoteObject : public Remote
{
  int iOpCode;			//# code for operation to perform

 public:

  static int Default_CMD;	//# op code for default operation


  // Creates a RemoteObject.
  RemoteObject(void);

  // Destroys the remote object.  Note, this method replaces a destructor.
  virtual void destroy(void);

  // Returns the context this object is running in.
  int getContext(void);

  // Returns the number of contexts.
  int numberOfContexts(void);

  // Returns the operation code that this object is supposed to perform.
  int opCode(void)			{ return iOpCode; }

  // Performs the requested remote action as defined by the operation code.
  virtual void performAction(void);

  // Performs the requested remote action.
  // <dd><li> sender -- the sender of the action
  // <dd><li> ro -- a pointer to the remote object
  static void performAction(int sender, RemoteObject* ro);

  // Performs the requested remote action.  This is a stub method called by
  // clients wishing to intiate a remote action.
  // <dd><li> opCode -- the operation code of the action to perform
  //                    (0 is default).
  // <dd><li> synchronous -- if true, returns only after operation has
  //                         completed, returns immediately otherwise.
  virtual void remoteAction(int opCode = 0, bool synchronous = true);
};


#endif	// __HPCXX_REMOTE_H_

