/*****************************************************************************
 * hashtable.h
 * 
 * Copyright (c) 1997, Los Alamos National Laboratory
 *****************************************************************************/

/*****************************************************************************
 * Declaration of the Hashtable object for the tulip framework.
 *
 * @author Craig Rasmussen
 * @version 1.1 23 Oct 1997
 *****************************************************************************/

#include "tulip.h"

struct entry {
  struct entry* next;		/* next entry in the chain */
  tulip_key_t key;		/* access key */
  tulip_LocalPointer_t value;	/* value associated with the key */
};

struct Hashtable {
  int shared;                   /* true (=1) if allocated in shared memory */
  int hashSize;			/* size of the pointer table */
  struct entry** table;		/* pointer table */
};


#ifdef __cplusplus
extern "C" {
#endif

/*****************************************************************************
 * Initializes the hash table.
 *
 * @param pht a pointer to the hash table
 * @param hashSize the size of the hash table
 * @param shared == true (1) if allocated in shared memory
 *****************************************************************************/
void tulip_HashtableInit(struct Hashtable* pht, int hashSize, int shared);

/*****************************************************************************
 * Destroys the hash table.
 *
 * @param pht a pointer to the hash table
 *****************************************************************************/
void tulip_HashtableDestroy(struct Hashtable* pht);

/*****************************************************************************
 * Gets the object associated with the specified key in the hashtable.
 *
 * @param pht a pointer to the hash table
 * @param key the specified key
 * @returns the element for the key or NULL if the key is not defined in
 *          the hash table.
 *****************************************************************************/
tulip_LocalPointer_t tulip_HashtableGet(struct Hashtable* pht,
					tulip_key_t key);

/*****************************************************************************
 * Puts the object associated with the specified key in the hashtable.
 *
 * @param pht a pointer to the hash table
 * @param key the specified key
 * @param value the specified element to store in the table
 * @returns the old value of the key or NULL if it did not have one
 *****************************************************************************/
tulip_LocalPointer_t tulip_HashtablePut(struct Hashtable* pht, tulip_key_t key,
					tulip_LocalPointer_t value);


#ifdef __cplusplus
}
#endif

