/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/* Written by Pete Beckman (by modifying Bernd Mohr's symmetry version) */
/* 6/22/94 */

/*==============================================================*/
/* sgilib : SGI run-time support library			*/
/*==============================================================*/

#include <sys/types.h>
#include <sys/prctl.h>
/*#include <ulocks.h>*/
#include <mutex.h>
#include <task.h>
#include <stdio.h>
#include <sys/sysmp.h>
#include <unistd.h>
 
#include "tulip.h"
#include "hashtable.h"
/* #include "profiler_defs.h" */
#include "Profile/Profiler.h"

/* function prototypes */

// find out where to include this prototype
extern "C" {
int pcxx_InitParam (int *argc, char *argv[]);
}
int SMRTSSetup(void);
int SMRTSFork(int argc, char **argv);
int SMRTSForkFinish(void);

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define CACHE_LINE 1024         /* size of a cache line */

extern pcxx_PARAM_t pcxx_param;

/* these numbers have to be the same as they are in KernelCom.c for 
   shared memory machines.  MAX_PROC has to be >= MyTotalNodes */
#define MAX_PROC 128

/*****************************/
/* Structs                   */
/*****************************/
struct ArgcArgv {
  int i;
  int argc;
  char **argv;
};

/*****************************************************************************
 * Shared parameter list used in performing remote action requests. 
 *****************************************************************************/
struct RA_params {
  char*  buffer;                /* pointer to buffer containing data */
  int    length;                /* length of data buffer */
  tulip_ACK_t* ackhdl;          /* ACK when finished */
  int sender;                   /* sending context */
  tulip_key_t rsrKey;		/* key used to select remote actions */
};
 
/*****************************************************************************
 * An array [0..NumberOfContexts-1] of pointers to a remote action parameter
 * list. Each context allocates its own array.
 *****************************************************************************/
struct RA_params **MyRAQueue;

/*****************************************************************************
 * Dispatch table for storing function pointers and associated keys to be
 * used to dispatch remote action calls.
 *****************************************************************************/
Hashtable* pDispatchTable;

/*****************************************************************************
 * Table for storing registered global pointers and associated keys.
 *****************************************************************************/
Hashtable* pGPTable;

struct msg_params {
  char * address;
  int    length;
};

/* largest message buffer size before shared alloc needed */
#define NO_ALLOC_SIZE  (CACHE_LINE - sizeof(RA_params)) /* FIX THIS SIZE */

void AlmostReady(void *);  /* Parallel function executed before calling user */

/* these numbers have to be the same as they are in KernelCom.c for 
   shared memory machines.  */
#define MAXBUFF 4096

static int tulip_SharedMemoryInitialized = 0;
int tulip_MsgBlockSize = -1;
int left_child, right_child, parent_id, num_children;
void Poll();

/*******************************************************************/
/* Shared data structures */
/*******************************************************************/
long *BarrierVector;
barrier_t *SGI_Gen_Barr;       /* general-purpose barrier */
barrier_t *SGI_Done_Barr;      /* Done with everything barrier */
char *BroadcastBuff;           /* a place to store broacast values */
char *ReduceBuff[MAX_PROC];    /* a place to store reduction values */
struct RA_params ***GlobalRATable; /* after init, read-only */

int *SharedHandle;         /* Name of "Arena" created by usinit() */
int MyTotalNodes = -1;     /* number of processors */
int MyNodeNumber = -1;     /* who I am */

void * tulip_SharedMalloc(int size) {
  TAU_PROFILE("tulip_SharedMalloc()", "void* (int)", TAU_MESSAGE);

  if (tulip_SharedMemoryInitialized)
    return (void *) usmalloc(size, SharedHandle);
  else
    return (void *) malloc(size);
}

void tulip_SharedFree(void *p) {
  TAU_PROFILE("tulip_SharedFree()", "void (void*)", TAU_MESSAGE);

  if (tulip_SharedMemoryInitialized)
    usfree(p,SharedHandle);
  else
    free(p);
}

/*****************************************************************************
 * Initializes the Tulip run time system so that it is in an usable/running
 * state. The arguments will be parsed and any tulip specific arguments will
 * be removed.
 *
 * pArgc (inout) : the number of arguments, on output, the new number of args
 * argv  (inout) : array of arguments, on output, the new argument list
 *****************************************************************************/
void tulip_Init(int* pArgc, char* argv[])
{
  pcxx_InitParam(pArgc, argv);
  SMRTSSetup();
  SMRTSFork(*pArgc, argv);
}

/*****************************************************************************
 * Shutsdown the Tulip run time system by freeing system resources.  After
 * calling this function, the RTS will be in an unusable/nonrunning state.
 *****************************************************************************/
void tulip_Shutdown(void)
{
  SMRTSForkFinish();
}

/*
 @ SMRTSSetup() : set up to run on the SGI
 */

int SMRTSSetup(void)
{
  char fname[1024];
  int i;

  /* allocate processors */
  if (pcxx_param.numproc >= sysmp(MP_NAPROCS)){
    MyTotalNodes = sysmp(MP_NAPROCS);
  } else {
    MyTotalNodes = pcxx_param.numproc;
  }

  MyNodeNumber = MyTotalNodes - 1;
  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_ENTER, pcxx_param.numproc);

  usconfig(CONF_INITUSERS,20);              /* Up to 20 processors */
  usconfig(CONF_INITSIZE,1024*1024*128);    /* 128 Meg of shared data */
  usconfig(CONF_ARENATYPE,US_SHAREDONLY);   /* unlink memmap after open */

  /* Open unique memmap filename */
  sprintf(fname,"/usr/tmp/pC++%d.memmap",getpid());
  SharedHandle = (int*) usinit(fname);

  /* initialize SGI barriers */
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_SETUP_INIT, 0);
  SGI_Gen_Barr = new_barrier(SharedHandle);
  init_barrier(SGI_Gen_Barr);
  SGI_Done_Barr = new_barrier(SharedHandle);
  init_barrier(SGI_Done_Barr);

  tulip_SharedMemoryInitialized = 1;

  /* Allocate & init our homebrew polling Barrier area */
  BarrierVector = (long*)
    usmalloc(sizeof(unsigned long)*CACHE_LINE*MAX_PROC,SharedHandle);
  for (i=0; i<CACHE_LINE*MAX_PROC; i++) BarrierVector[i]=0;

  /* Allocate Global list of RA queues */
  GlobalRATable = (RA_params***)
    usmalloc(sizeof(RA_params**)*MAX_PROC, SharedHandle);
  for (i=0; i<MAX_PROC; i++) {
    GlobalRATable[i] = NULL;
  }

  /* allocate Shared Broadcast Buffer */
  BroadcastBuff = (char *) usmalloc(MAXBUFF,SharedHandle);

  /* allocate shared locks and reduction buffer */
  if(MyTotalNodes > MAX_PROC) {
    fprintf(stderr,"Error in sgilib.c: MAX_PROC < number of processors.\n");
    exit(0);
  }
  for(i = 0; i < MyTotalNodes; i++) {
    ReduceBuff[i] = (char*) usmalloc(MAXBUFF,SharedHandle);
  }

  /* construct table for registering/storing global pointers */
  pGPTable = (Hashtable*) tulip_SharedMalloc(sizeof(*pGPTable));
  tulip_HashtableInit(pGPTable, 16, TRUE);

  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_EXIT, MyTotalNodes);
  return(0);
}

void tulip_InternalBarrier(barrier_t *b)
{
/*printf("%d: starting Barrier (%d)\n",MyNodeNumber,MyTotalNodes);*/
/*  PROFILE_EXIT(PROFILE_Barrier); */
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);
  barrier(b,MyTotalNodes);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
/*  PROFILE_EXIT(PROFILE_Barrier); */
/*printf("%d: Ending Barrier (%d)\n",MyNodeNumber,MyTotalNodes);*/
}

/*
 @ SMRTSFork : start parallel processing
 */

int SMRTSFork(int argc, char **argv)
{
  int k;
  struct ArgcArgv *acav;

  /* fork */
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_ENTER, 0);

  for (k=0; k<MyTotalNodes-1; k++) {
    acav = (struct ArgcArgv *) tulip_SharedMalloc(sizeof(struct ArgcArgv));
    acav->i=k;
    acav->argc=argc;
    acav->argv=argv;
    sproc(AlmostReady, 0, (void *) acav);
  }

  /* Go do AlmostReady() myself! */
  acav = (struct ArgcArgv *) tulip_SharedMalloc(sizeof(struct ArgcArgv));
  acav->i=MyTotalNodes-1;
  acav->argc=argc;
  acav->argv=argv;
  AlmostReady( (void *) acav ); 

#ifdef SPMD_MODE
  return SMRTSForkFinish();
#else
  /* don't forget to call SMRTSForkFinish() after other contexts finish */
  return 0;
#endif
}

int SMRTSForkFinish(void)
{
  tulip_InternalBarrier(SGI_Done_Barr);
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_EXIT, 0);
  pcxx_EvClose ();

  // delete the global pointer table
  tulip_HashtableDestroy(pGPTable);

  return(0);
}

/*
 @ SMRTSNewProc : more book-keeping for newly-created process
 */

int SMRTSNewProc(int num)
{
  /* record my ID */
  MyNodeNumber = num;
  tulip_InitClocks ();
  pcxx_EvInit (pcxx_param.tracefile);
  return(MyNodeNumber);
}

/***********************************************************************/
/* Pete's super-fast barrier */
/***********************************************************************/
void tulip_Barrier()
{
#define LNx 0
#define CNx 256
#define RNx 512

/*  PROFILE_ENTER(PROFILE_Barrier); */
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);

  /* Is a previous barrier still executing? */
  /* Hey! A show stopper! */
/*  while ( (BarrierVector[MyNodeNumber*CACHE_LINE+LNx] != 0) ||
          (BarrierVector[MyNodeNumber*CACHE_LINE+CNx] != 0) ||
          (BarrierVector[MyNodeNumber*CACHE_LINE+RNx] != 0) )
    printf("%d: Hey, slow down! %d %d %d\n",MyNodeNumber,
          (BarrierVector[MyNodeNumber*CACHE_LINE+LNx]),
          (BarrierVector[MyNodeNumber*CACHE_LINE+CNx]),
          (BarrierVector[MyNodeNumber*CACHE_LINE+RNx]));*/

  /* if there is only one processor, simply return.  S.Y. 8/9/95 */
  if(num_children == 0 && MyNodeNumber == 0) {
    PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
/*    PROFILE_EXIT(PROFILE_Barrier); */
    return;
  }

  /* Start fan-in, leaves bubble up... */
  if (num_children == 0) {
    if (MyNodeNumber & 1) /* Odd numbered leaf */
      BarrierVector[parent_id*CACHE_LINE+LNx]=1;
    else 
      BarrierVector[parent_id*CACHE_LINE+RNx]=1;
  }

  /* I am not a leaf. Leaf me alone!!! */
  else
    {
      /* Wait for fan-in from children */
      if (num_children == 1)
	while  (BarrierVector[MyNodeNumber*CACHE_LINE+LNx] == 0)
	  tulip_Poll();
      else
	while ((BarrierVector[MyNodeNumber*CACHE_LINE+LNx] == 0) || 
	       (BarrierVector[MyNodeNumber*CACHE_LINE+RNx] == 0))
	  tulip_Poll();

      /* OK, I have received all my children */
      /* propagate fan-in to my parent */
      if (MyNodeNumber != 0) {
	if (MyNodeNumber & 1)  /* Odd numbered leaf */
	  BarrierVector[parent_id*CACHE_LINE+LNx]=1;
	else 
	  BarrierVector[parent_id*CACHE_LINE+RNx]=1;
      }
      else {
	/* Hey!! the root has received the fan-in message, start fan-out */
	/* reset my slots*/
	BarrierVector[MyNodeNumber*CACHE_LINE+LNx] =  0;
	BarrierVector[MyNodeNumber*CACHE_LINE+CNx] =  0;
	BarrierVector[MyNodeNumber*CACHE_LINE+RNx] =  0;

	/* Propagate to children */
	if (num_children >= 1) BarrierVector[ left_child*CACHE_LINE+CNx] = -1;
	if (num_children == 2) BarrierVector[right_child*CACHE_LINE+CNx] = -1;
        PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
/*        PROFILE_EXIT(PROFILE_Barrier); */
	return;
      }
    }

  /* Ok, I have received any required fan-in msgs, and have propagated them */
  /* now wait of fan-out message */

  while (BarrierVector[MyNodeNumber*CACHE_LINE+CNx] != -1)
    tulip_Poll();
    
  /* Reset my variables */
  BarrierVector[MyNodeNumber*CACHE_LINE+LNx] =  0;
  BarrierVector[MyNodeNumber*CACHE_LINE+CNx] =  0;
  BarrierVector[MyNodeNumber*CACHE_LINE+RNx] =  0;

  /* Ok, I got the fan-out message, propagate it */
  if (num_children >= 1) BarrierVector[ left_child*CACHE_LINE+CNx] = -1;
  if (num_children == 2) BarrierVector[right_child*CACHE_LINE+CNx] = -1;
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
/*  PROFILE_EXIT(PROFILE_Barrier); */
}


/*
 @ tulip_BroadcastBytes : sends messages from 0 to all other nodes
 */

void tulip_BroadcastBytes(int flag, int length, void* buffer)
{
  int i;
  char *c, *p;

  /* This barrier is needed to prevent BroadcastBuff from being overwritten
     when two consecutive tulip_BroadcastBytes are called */
  tulip_Barrier();

  if ( length > MAXBUFF ) 
    tulip_Message("Broadcast buffer too small!  Increase buffer.",TULIP_FATAL);

  c = BroadcastBuff; p = (char *) buffer;

  if (flag == 0) for (i=0; i<length; i++) c[i] = p[i];
  tulip_Barrier();
  if (flag != 0) for (i=0; i<length; i++) p[i] = c[i];
}


/*
 @ SMRTSDealloc : deallocate shared memory pointers
 */

void SMRTSDealloc(void* ptr)
{
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_ENTER, (long) ptr);
  usfree(ptr,SharedHandle);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_EXIT, 0);
}

void SMRTSWaitToEnd(void) {
  if ((int) MyNodeNumber != (MyTotalNodes-1)) {
    tulip_InternalBarrier(SGI_Done_Barr);
  }
}

void AlmostReady(void *x)
{
  int i, num, argc;
  char **argv;
  struct ArgcArgv *acav;

  acav = (struct ArgcArgv *) x;
  num  = acav->i;
  argc = acav->argc;
  argv = acav->argv;

  SMRTSNewProc(num);
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_ENTER, 0);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);

  /* Allocate MY RemoteAction area */
  MyRAQueue = (RA_params**)
    usmalloc(sizeof(struct RA_params*)*CACHE_LINE*MAX_PROC, SharedHandle);
  for (i=0; i<CACHE_LINE*MAX_PROC; i++) {
    MyRAQueue[i] = NULL;             /* touch to own */
  }

  /* Register My RA area with Global Table */
  GlobalRATable[MyNodeNumber] = MyRAQueue;

  /* INITIALIZE BARRIER VARIABLES */
  left_child = (MyTotalNodes > 2*MyNodeNumber+1)? 
                                   2*MyNodeNumber+1: 0;
  right_child = (MyTotalNodes > 2*MyNodeNumber+2)? 
                                   2*MyNodeNumber+2: 0;
  if (left_child && right_child) num_children = 2;
  else if(left_child || right_child) num_children = 1;
  else num_children = 0;
  parent_id = (MyNodeNumber-1)/2;

  /* construct the remote-action dispatch table */
  pDispatchTable = (Hashtable*) malloc(sizeof(*pDispatchTable));
  tulip_HashtableInit(pDispatchTable, 16, FALSE);

  tulip_InternalBarrier(SGI_Gen_Barr);  /* sync up before Polling, etc */
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
  tulip_ParallelMain(argc,argv);  /* Transfer control to compiler/user code */
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_EXIT, 0);

  /* free local memory */
  tulip_HashtableDestroy(pDispatchTable);

  SMRTSWaitToEnd();
  pcxx_EvClose ();
}

/*********************************************************************/
/* Check if a "handle" has completed (non-blocking) */
void  tulip_Abort(void)
{
  printf("Exiting!!!!! ERROR\n");
  exit(1);
}

/*********************************************************************/
int tulip_Get(char *dst, tulip_GlobalPointer_t src, int len, 
	      tulip_ACK_t *ackhandle)
{
  /* Why call tulip_Get on a shared memory machine? */
  memcpy(dst, tulip_ToLocalPointer(&src), len);

  tulip_SetACK(ackhandle);

  return 0;
}  

/*********************************************************************/
int tulip_Put(tulip_GlobalPointer_t dst, char *src, int len, 
	      tulip_ACK_t *ackhandle)
{
  /* Why call tulip_Put on a shared memory machine? */
  memcpy(tulip_ToLocalPointer(&dst), src, len);

  tulip_SetACK(ackhandle);

  return 0;
}  

/*****************************************************************************
 * Check for RemoteAction requests
 *****************************************************************************/
void tulip_Poll(void)
{
  int i, length;
  struct RA_params* tmpParams;

  TAU_PROFILE("tulip_Poll()", "void (void)", TAU_MESSAGE);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_POLL_ENTER, 0);

/* Scan through RemoteAction area looking for work */
  for (i=0; i<=MyTotalNodes; i++) {
    if ( MyRAQueue[i*CACHE_LINE] !=  NULL ) {
      /* Hey, found some work! */

      TAU_PROFILE_TIMER(timer, "Call_tulip_RemoteActionHandlerTimer",
			"in tulip_Poll()", TAU_MESSAGE);
      TAU_PROFILE_START(timer); 

      tmpParams = MyRAQueue[i*CACHE_LINE];

      if (tmpParams->rsrKey == NULL) {
	tulip_RemoteActionHandler(tmpParams->sender, tmpParams->buffer,
				  tmpParams->length);
        length = tmpParams->length;
      } else {
	tulip_RemoteActionHandlerKeyN(tmpParams->sender, tmpParams->rsrKey,
				      tmpParams->buffer, tmpParams->length);
        length = tmpParams->length + strlen(tmpParams->rsrKey) + 1;
      }

      /* Send acknowledgement (through ACK handle) */
      if (tmpParams->ackhdl != NULL) {
        tulip_SetACK(tmpParams->ackhdl);
      }

      /* OK, clean up! */

      if (length > NO_ALLOC_SIZE) {
        /* free buffers */
        tulip_SharedFree(tmpParams->buffer);
      }

      /* Clean out slot */
      MyRAQueue[i*CACHE_LINE] = NULL;

      TAU_PROFILE_STOP(timer); 
    }
  }
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_POLL_EXIT, 0);
}

/*****************************************************************************
 * Sends RemoteAction request to remote processor.
 *
 * remote_node (input) : the receiving context of the remote action
 * buffer      (input) : data buffer, could be additional parameters
 * len         (input) : the size of the data buffer
 * ackhandle   (input) : acknowlegement handle (no ack sent if NULL)
 *****************************************************************************/
int tulip_RemoteAction(int remote_node, char *buffer, int len, 
		       tulip_ACK_t *ackhandle)
{
  struct RA_params **rmtslot;
  struct RA_params *tmpParams;
  char *tmpbuf;

  TAU_PROFILE("tulip_RemoteAction()", "int ()", TAU_MESSAGE);
  TAU_PROFILE_TIMER(timer, "BeginActionBarrier",
                    "in tulip_RemoteAction", TAU_MESSAGE);

  rmtslot = &( (GlobalRATable[remote_node])[MyNodeNumber*CACHE_LINE]);
 
  /* check to see if a slot is available */
  TAU_PROFILE_START(timer); 
  while (rmtslot[0] != NULL) {
    /* no slot is available. poll and wait */
    tulip_Poll();
  }
  TAU_PROFILE_STOP(timer); 

  /* OK, the slot is free! */

  tulip_ClearACK(ackhandle);
 
  /* store the parameter list within the (active) message */
  tmpParams = (struct RA_params*) (rmtslot + 1);

  if (len > NO_ALLOC_SIZE) {
    /* allocate buffer for remote action request, if needed */
    tmpbuf = (char *) tulip_SharedMalloc(len);
  } else {
    /* store the parameter list within the (active) message also */
    tmpbuf = (char *) tmpParams + sizeof(RA_params);
  }

  /* copy buffer into shared memory */
  memcpy(tmpbuf, buffer, len);

  /* install parameters */
  tmpParams->buffer    = tmpbuf;
  tmpParams->sender    = MyNodeNumber;
  tmpParams->length    = len;
  tmpParams->ackhdl    = ackhandle;
  tmpParams->rsrKey    = NULL;
 
  /* OK, now the pointer tmpParams is ready to pass to the remote node */

  /* fill remote slot */
  *rmtslot = tmpParams;
    
  return 0;
}


/*****************************************************************************
 * Sends a RemoteAction request to a remote processor with a rmi/rsr key.
 *
 * receiver  (input) : the receiving context of the remote action
 * key       (input) : a key used to select a particular remote action
 * lptr      (input) : local pointer to data buffer or additional parameters
 * nBytes    (input) : the size of the data buffer
 * ackHandle (input) : acknowlegement handle (no ack sent if NULL)
 *****************************************************************************/
void tulip_RemoteActionKeyN(tulip_context_t receiver, tulip_key_t key,
			    tulip_LocalPointer_t lptr, size_t nBytes,
			    tulip_ACK_t* ackHandle)
{
  int		     len;
  struct RA_params** rmtSlot;
  struct RA_params*  tmpParams;
  char*		     tmpBuf;
  tulip_key_t	     tmpKey;

  TAU_PROFILE("tulip_RemoteActionKeyN()", "int ()", TAU_MESSAGE);
  TAU_PROFILE_TIMER(timer, "BeginActionBarrier",
                    "in tulip_RemoteActionKeyN", TAU_MESSAGE);

  rmtSlot = &( (GlobalRATable[receiver])[MyNodeNumber*CACHE_LINE]);
 
  /* check to see if a slot is available */
  TAU_PROFILE_START(timer); 
  while (rmtSlot[0] != NULL) {
    /* no slot is available. poll and wait */
    tulip_Poll();
  }
  TAU_PROFILE_STOP(timer); 

  /* OK, the slot is free! */

  tulip_ClearACK(ackHandle);
 
  /* store the parameter list within the (active) message */
  tmpParams = (struct RA_params*) (rmtSlot + 1);

  len = 1 + strlen(key) + nBytes;
  if (len > NO_ALLOC_SIZE) {
    /* allocate buffer for remote action request, if needed */
    tmpBuf = (char*) tulip_SharedMalloc(len);
  } else {
    /* store the parameter list within the (active) message also */
    tmpBuf = (char*) tmpParams + sizeof(RA_params);
  }
  tmpKey = tmpBuf + nBytes;

  /* copy local buffers into shared memory */
  memcpy(tmpBuf, lptr, nBytes);
  strcpy(tmpKey, key);

  /* install parameters */
  tmpParams->buffer    = tmpBuf;
  tmpParams->sender    = MyNodeNumber;
  tmpParams->length    = nBytes;
  tmpParams->ackhdl    = ackHandle;
  tmpParams->rsrKey    = tmpKey;
 
  /* OK, now the pointer tmpParams is ready to pass to the remote node */

  /* fill remote slot */
  *rmtSlot = tmpParams;
}


/*****************************************************************************
 * Services RemoteAction requests by calling the appropriate function.
 *
 * sender (input) the sending context of the remote action
 * key    (input) a key used to select a particular remote action
 * lptr   (input) a local pointer to data buffer, could be additional params
 * nBytes (input) the size of the data buffer
 *****************************************************************************/
void tulip_RemoteActionHandlerKeyN(tulip_context_t sender, tulip_key_t key,
				   tulip_LocalPointer_t lptr, size_t nBytes)
{
  tulip_rsrN_t rsr = (tulip_rsrN_t) tulip_HashtableGet(pDispatchTable, key);
  if (rsr == NULL) return;
  rsr(sender, lptr, nBytes);
}


/* ----------------------------------------------------------------------------
 * Registers a function so that it can be called remotely.
 *
 * key  (input) : a key used to select the function
 * rsr  (input) : the function to register
 * ------------------------------------------------------------------------- */
void tulip_RegisterRSRNHandler(tulip_key_t key, tulip_rsrN_t rsr)
{
  tulip_HashtablePut(pDispatchTable, key, rsr);
}

/* ----------------------------------------------------------------------------
 * Makes a global pointer.  The calling function is responsible for freeing
 * the global pointer.
 *
 * key    (input) : a registration key used to identify the global pointer
 * nBytes (input) : the size of the buffer to allocate
 * returns : the global pointer (NULL if error)
 * ------------------------------------------------------------------------- */
tulip_GlobalPointer_t* tulip_MakeGlobalPointer(tulip_key_t key, size_t nBytes)
{
  tulip_GlobalPointer_t* gPtr = (tulip_GlobalPointer_t*)
                                tulip_SharedMalloc(sizeof(*gPtr));
  tulip_LocalPointer_t lPtr = (tulip_LocalPointer_t)
                              tulip_SharedMalloc(nBytes);
  tulip_SetLocalPointer(gPtr, lPtr);
  tulip_SetContext(gPtr, tulip_MyContext());
  tulip_HashtablePut(pGPTable, key, gPtr);
  return gPtr;
}

/* ----------------------------------------------------------------------------
 * Returns the global pointer associated with the given key (NULL if
 * not found).
 *
 * key    (input) : the registration key by which the global pointer is known
 * ------------------------------------------------------------------------- */
tulip_GlobalPointer_t* tulip_GetGlobalPointer(tulip_key_t key)
{
  return (tulip_GlobalPointer_t*) tulip_HashtableGet(pGPTable, key);
}

/* ----------------------------------------------------------------------------
 * Registers a global pointer by associating it with the given key/name.
 *
 * key  (input) : a registration key used to identify the global pointer
 * gPtr (input) : the global pointer to register
 * ------------------------------------------------------------------------- */
void tulip_RegisterGlobalPointer(tulip_key_t key, tulip_GlobalPointer_t* gPtr)
{
  tulip_HashtablePut(pGPTable, key, gPtr);
}

