#include <stdio.h>
#include <hpcxx_rts.h>
#include "tulip.h"
#include "Remote.h"
#include "Profile/Profiler.h"

#define ARRAY_SIZE 4096

class RemoteTest : public RemoteObject
//-----------------------------------------------------------------------------
// This class is used to test the allocation and remote invocation of a
// shared/distributed object.
//-----------------------------------------------------------------------------
{
 public:

  int* viArray;		// an array of integers, one for each processor

  RemoteTest(void);
  virtual ~RemoteTest(void);

  void initArray(void);
  virtual void performAction(void);
};


RemoteTest::RemoteTest(void)
//-----------------------------------------------------------------------------
// Constructor.  Allocated memory for the array.
//-----------------------------------------------------------------------------
{
  viArray = (int*) tulip_SharedMalloc( ARRAY_SIZE*sizeof(int) );
}


RemoteTest::~RemoteTest(void)
//-----------------------------------------------------------------------------
// Destructor.  Free allocated memory.
//-----------------------------------------------------------------------------
{
  tulip_SharedFree(viArray);
  viArray = NULL;
}


void RemoteTest::initArray(void)
//-----------------------------------------------------------------------------
// Initialize the shared array.
//-----------------------------------------------------------------------------
{
  for (int i = 0; i < ARRAY_SIZE; i++) {
    viArray[i] = -1;
  }
}


void RemoteTest::performAction(void)
//-----------------------------------------------------------------------------
// Performs an action.
//-----------------------------------------------------------------------------
{
  TAU_PROFILE("RemoteTest::performAction()", "void (void)", TAU_MESSAGE);
  int id = getContext();
  int iMySize = ARRAY_SIZE / numberOfContexts();
  int iStart = id * iMySize;

  for (int i = iStart; i < iStart + iMySize; i++) {
    viArray[i] = id;
  }
}


main(int argc, char* argv[])
{
  // startup the remote processes
  Registry* pRegistry = LocateParallelRegistry::createRegistry(0, 2);

  TAU_PROFILE("main()", "int(int,char**)", TAU_DEFAULT);

  RemoteTest* pRemote = new (HPCXX_NEW) RemoteTest();
  pRemote->initArray();         // touch so master owns array
//  pRegistry -> bind("RemoteTest", pRemote);

  // have all slave contexts do some work

  for (int i = 0; i < 1000; i++) {
    pRemote -> remoteAction(RemoteObject::Default_CMD, false);
  }

//  printf("\n");
//  for (int i = 0; i < 8; i++) {
//    printf("context %d:    pRemote->viArray[%d] = %2d\n",
//	   pRemote->getContext(), i, pRemote->viArray[i]);
//  }
//  printf("\n");

  // delete remote objects
  pRemote -> destroy();

  // shut down registry
  delete pRegistry;
}

