//-----------------------------------------------------------------------------
// testhashtable.cpp -- code to test the tulip hashtable in hashtable.c
//
// Copyright (c) 1997, Los Alamos National Laboratory
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// @author Craig Rasmussen
// @version 1.1 23 Oct 1997
//-----------------------------------------------------------------------------

#include "hashtable.h"
#include <stdio.h>
#include <string.h>
#include <assert.h>

#define FALSE 0
#define TRUE 1

int main(int argc, char* argv[])
{
  char* sz1 = "item 1";
  char* sz2 = "item 2";
  char* sz3 = "item 3";

  struct Hashtable* pht = (struct Hashtable*) malloc(sizeof(*pht));

  // test tulip_HashtableInit

  tulip_HashtableInit(NULL, 0, FALSE);		// shouldn't crash
  tulip_HashtableInit(pht, 1, FALSE);
  assert(pht->hashSize > 1);

  tulip_HashtableDestroy(pht);
  pht = (struct Hashtable*) malloc(sizeof(*pht));
  tulip_HashtableInit(pht, 16, TRUE);
  assert(pht->hashSize == 16);
  
  // test NULL puts

  assert(tulip_HashtablePut(NULL, "string", sz1) == NULL);
  assert(tulip_HashtablePut(pht, NULL, sz1) == NULL);
  assert(tulip_HashtablePut(pht, "string", NULL) == NULL);

  // test puts and gets

  tulip_HashtablePut(pht, "1", sz1);
  tulip_HashtablePut(pht, "  xx  yy 3", sz3);
  tulip_HashtablePut(pht, "2", sz2);

  char* szItem = (char*) tulip_HashtableGet(pht, "2");
  assert(strcmp(szItem, sz2) == 0);
  szItem = (char*) tulip_HashtableGet(pht, "  xx  yy 3");
  assert(strcmp(szItem, sz3) == 0);
  szItem = (char*) tulip_HashtableGet(pht, "not found");
  assert(szItem == NULL);
  szItem = (char*) tulip_HashtableGet(pht, "1");
  assert(strcmp(szItem, sz1) == 0);

  // test NULL gets

  assert(tulip_HashtableGet(NULL, "1") == NULL);
  assert(tulip_HashtableGet(pht, NULL) == NULL);

  tulip_HashtableDestroy(pht);
  assert(pht->table == NULL);

  printf("\n  Testing of Hashtable object completed successfully.\n\n");
}
