// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: HPCxx_ThreadImpl.C,v 1.1 1998/06/23 00:08:24 ejohnson Exp $

#include <HPCxx_ThreadImpl.h>
#include <HPCxx_Thread.h>
#include <HPCxx_Runnable.h>
#include <unistd.h>

HPCxx_ThreadImpl::HPCxx_ThreadImpl(void) : runnable_(NULL)
{
  isAlive_ = false;
}
 
HPCxx_ThreadImpl::HPCxx_ThreadImpl(const HPCxx_ThreadImpl& thread)
  : runnable_(thread.runnable_)
{
  isAlive_ = false;
}
 
HPCxx_ThreadImpl::HPCxx_ThreadImpl(HPCxx_Runnable *runnable)
  : runnable_(runnable)
{
  isAlive_ = false;
}

//
// What is this supposed to do?  WARNING, look into what happens when
// both threads are deleted.  How does this thread know when the other
// is finished.  Can you start the new thread?
// I think this method should go away.
//
HPCxx_ThreadImpl& HPCxx_ThreadImpl::operator=(const HPCxx_ThreadImpl& thread)
{
  if (this == &thread) return(*this); // self-assignment

  isAlive_ = thread.isAlive_;	// race condition?
  runnable_ = thread.runnable_;

  return(*this);
}

HPCxx_ThreadImpl::~HPCxx_ThreadImpl()
{
  delete runnable_;
}

// Should this method be protected?  Can threads be rerun?  I think
// one should have to start and not run, and restart can happen.
void HPCxx_ThreadImpl::run()
{
  if (runnable_) runnable_->run();
}

void HPCxx_ThreadImpl::start(HPCxx_Thread &thread)
{
  if (isAlive_)	return;

  isAlive_ = true;
  run();
  isAlive_ = false;
}

void HPCxx_ThreadImpl::sleep(long milliseconds, long nanoseconds)
{
  usleep(milliseconds*1000 + (int)(nanoseconds/1000 + 0.5));
}
