// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: HPCxx_ThreadImpl.h,v 1.2 1998/06/26 19:14:43 khalsa Exp $

#ifndef __HPCXX_THREADIMPL_H_
#define __HPCXX_THREADIMPL_H_

#include <unistd.h>		// for NULL
#include <hpcxx_types.h>
#include <HPCxx_Runnable.h>

class HPCxx_Thread;

class HPCxx_ThreadImpl {

public:
  HPCxx_ThreadImpl(void);
  HPCxx_ThreadImpl(const HPCxx_ThreadImpl& thread);
  HPCxx_ThreadImpl(HPCxx_Runnable *runnable);
  HPCxx_ThreadImpl& operator=(const HPCxx_ThreadImpl& thread);
  
  virtual ~HPCxx_ThreadImpl(void);

  virtual void run(void);
  
  // Starts execution of this thread and calls the run() method.
  void start(HPCxx_Thread& thread);

  // Ceases execution of this thread.  Status is not currently used.
  void stop(void *status) { }

  // Causes the currently executing thread object to temporarily pause and
  // allow other threads to execute.
  static void yield(void) { }

  // Causes the currently executing thread to sleep (cease executing) for the
  // specified number of milliseconds plus the specified number of nanoseconds.
  // The sleep period will be rounded up to the resolution of the system timer.
  static void sleep(long milliseconds, long nanoseconds = 0);

  // Suspends execution of this thread until resume is called.
  void suspend(void) { }
  
  // Resume a suspended thread.
  void resume(void) { }
  
  // Check to see if thread has been started but has not yet died.
  bool isAlive(void) { return isAlive_; }
  
  // Returns a reference to the currently executing thread object.
  static HPCxx_Thread* currentThread(void) { return NULL; }

  // Returns the numeric ID of the currently executing thread object.
  static int getThreadID(void) { return 0; }

  // Join waits for the completion of this thread.  By definition join returns
  // as soon as this thread is considered not alive.
  void join(int milliseconds, int nanoseconds = 0) { }

  // Returns this thread's priority, where
  // HPCXX_MIN_THREAD_PRIORITY <= getPriority() <= HPCXX_MAX_THREAD_PRIORITY
  int getPriority(void) { return 0; }

  // Changes the priority of this thread.  Priority is clamped to
  // (HPCXX_MIN_THREAD_PRIORITY, HPCXX_MAX_THREAD_PRIORITY)
  int setPriority(int priority) { return 0; }

private:
  HPCxx_Runnable *runnable_;
  bool isAlive_;

  void* start_(void* rhs);

};

#endif


