// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: HPCxx_CondImpl.C,v 1.1 1998/06/23 00:08:46 ejohnson Exp $

#include <HPCxx_CondImpl.h>

HPCxx_CondImpl::HPCxx_CondImpl(void)
{
  pthread_condattr_init(&attr_);
  pthread_cond_init(&cond_, &attr_);
}

HPCxx_CondImpl::HPCxx_CondImpl(const HPCxx_CondImpl& cond) : cond_(cond.cond_)
{
}

HPCxx_CondImpl& HPCxx_CondImpl::operator=(const HPCxx_CondImpl& cond)
{
  if (this == &cond) return *this; // self-assignment
  //((HPCXX_Cond&) *this) = cond;  // Copy any data in HPCxx_Cond
  
  pthread_cond_destroy(&cond_);
  cond_ = cond.cond_;

  return *this;
}

HPCxx_CondImpl::~HPCxx_CondImpl(void)
{
  pthread_cond_destroy(&cond_);
  pthread_condattr_destroy(&attr_);
}

int HPCxx_CondImpl::wait(HPCxx_MutexImpl &mutex)
{
  return(pthread_cond_wait(&cond_, &(mutex.mutex_)));
}

int HPCxx_CondImpl::signal(void)
{
  return(pthread_cond_signal(&cond_));
}

int HPCxx_CondImpl::broadcast(void)
{
  return(pthread_cond_broadcast(&cond_));
}
