// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: HPCxx_ThreadImpl.C,v 1.1 1998/06/23 00:09:00 ejohnson Exp $

#include <HPCxx_ThreadImpl.h>
#include <HPCxx_Thread.h>
#include <HPCxx_Runnable.h>

HPCxx_ThreadImpl::HPCxx_ThreadImpl(void) : runnable_((HPCxx_Runnable *)0)
{
  is_alive = false;
}
 
HPCxx_ThreadImpl::HPCxx_ThreadImpl(const HPCxx_ThreadImpl& thread)
  : thread_(thread.thread_), runnable_(thread.runnable_)
{
  is_alive = false;
}
 
HPCxx_ThreadImpl::HPCxx_ThreadImpl(HPCxx_Runnable *runnable)
  : runnable_(runnable)
{
  is_alive = false;
}

//
// What is this supposed to do?  WARNING, look into what happens when
// both threads are deleted.  How does this thread know when the other
// is finished.  Can you start the new thread?
// I think this method should go away.
//
HPCxx_ThreadImpl& HPCxx_ThreadImpl::operator=(const HPCxx_ThreadImpl& thread)
{
  if (this == &thread) return(*this); // self-assignment

  is_alive = thread.is_alive;	// race condition?
  thread_ = thread.thread_;
  runnable_ = thread.runnable_;

  return(*this);
}

HPCxx_ThreadImpl::~HPCxx_ThreadImpl()
{
  is_alive = false;
  delete runnable_;
}

// Should this method be protected?  Can threads be rerun?  I think
// one should have to start and not run, and restart can happen.
void HPCxx_ThreadImpl::run()
{
  if (runnable_) runnable_->run();
}

void *HPCxx_ThreadImpl::start_(void *rhs)
{
  HPCxx_Thread *thread = (HPCxx_Thread *)rhs;
  thread->run();
  thread->impl_.is_alive = false;

  return((void *)0);
}

// Can threads be restarted while alive?  I think not.
void HPCxx_ThreadImpl::start(HPCxx_Thread &thread)
{
  // WARNING, this needs to be protected by a lock.
  if (is_alive) {
    return;
  } else {
    is_alive = true;
  }
  pthread_create(&thread_, NULL, start_, (void *)&thread);
}

void HPCxx_ThreadImpl::stop(void *status)
{
  is_alive = false;
  pthread_cancel(thread_);
}

void HPCxx_ThreadImpl::yield()
{
  sched_yield();
}

void HPCxx_ThreadImpl::sleep(long milliseconds, long nanoseconds)
{
  // FIXME!!!!!!!! SGI specfic
  //  sginap(milliseconds);
  //  nexus_usleep(milliseconds*1000 + (nanoseconds/1000 + 0.5));
}

void HPCxx_ThreadImpl::suspend()
{
  // FIXME - not implemented
}
  
void HPCxx_ThreadImpl::resume()
{
  // FIXME - not implemented
}
  
bool HPCxx_ThreadImpl::isAlive()
{
  return(is_alive);
}
  
HPCxx_Thread* HPCxx_ThreadImpl::currentThread()
{
  return NULL;
}

// Returns the numeric ID of the thread
int HPCxx_ThreadImpl::getThreadID()
{
  return(pthread_self());
}

void HPCxx_ThreadImpl::join(int milliseconds, int nanoseconds)
{
  pthread_join(thread_, NULL);
}

int HPCxx_ThreadImpl::getPriority()
{
  // FIXME - not implemented
  return(0);
}

int HPCxx_ThreadImpl::setPriority(int priority)
{
  // FIXME - not implemented
  return(0);
}
