// HPC++ Run Time Sytem Library
// Extreme! Computing Lab - Computer Science Department
// Copyright (c) 1997 Indiana University
// Authors: Pete Beckman, Dennis Gannon, Todd Green, and Liz Johnson
//
// $Id: HPCxx_ThreadImpl.h,v 1.1 1998/06/23 00:09:01 ejohnson Exp $

#ifndef __HPCXX_THREADIMPL_H_
#define __HPCXX_THREADIMPL_H_

#include <pthread.h>
#include <hpcxx_types.h>
#include <HPCxx_Runnable.h>

class HPCxx_Thread;

class HPCxx_ThreadImpl {
private:
  pthread_t thread_;
  HPCxx_Runnable *runnable_;
  bool is_alive;

  static void *start_(void *rhs);

public:
  HPCxx_ThreadImpl(void);
  HPCxx_ThreadImpl(const HPCxx_ThreadImpl& thread);
  HPCxx_ThreadImpl(HPCxx_Runnable *runnable);
  HPCxx_ThreadImpl& operator=(const HPCxx_ThreadImpl& thread);
  
  virtual ~HPCxx_ThreadImpl(void);

  virtual void run(void);
  
  // Starts execution of this thread and calls the run() method.
  void start(HPCxx_Thread& thread);

  // Ceases execution of this thread.  Status is not currently used.
  void stop(void *status);

  // Causes the currently executing thread object to temporarily pause and
  // allow other threads to execute.
  static void yield(void);

  // Causes the currently executing thread to sleep (cease executing) for the
  // specified number of milliseconds plus the specified number of nanoseconds.
  // The sleep period will be rounded up to the resolution of the system timer.
  static void sleep(long milliseconds, long nanoseconds = 0);

  // Suspends execution of this thread until resume is called.
  void suspend(void);
  
  // Resume a suspended thread.
  void resume(void);
  
  // Check to see if thread has been started but has not yet died.
  bool isAlive(void);
  
  // Returns a reference to the currently executing thread object.
  static HPCxx_Thread *currentThread(void);

  // Returns the numeric ID of the currently executing thread object.
  static int getThreadID(void);

  // Join waits for the completion of this thread.  By definition join returns
  // as soon as this thread is considered not alive.
  void join(int milliseconds, int nanoseconds = 0);

  // Returns this thread's priority, where
  // HPCXX_MIN_THREAD_PRIORITY <= getPriority() <= HPCXX_MAX_THREAD_PRIORITY
  int getPriority(void);

  // Changes the priority of this thread.  Priority is clamped to
  // (HPCXX_MIN_THREAD_PRIORITY, HPCXX_MAX_THREAD_PRIORITY)
  int setPriority(int priority);

};

#endif
