#include "Tulip_Timers.h"
#include <stdio.h>
#include <stdlib.h>
#include "Tulip_Thread.h"
#include "Tulip_Context.h"
#include "Tulip_Cond.h"

// This example was adapted from the book "Pthreads Programming"

#define TCOUNT 40
#define WATCH_COUNT 4

int count = 0;
Tulip_Mutex count_mutex;
Tulip_Cond count_threshold_cv;
Tulip_ThreadPtr th[3];

class inc_count: public Tulip_Thread{
  Tulip_Cond *c;
  int my_num;
public:
  inc_count(int id, char *name ):  Tulip_Thread(name,2*default_stack_size){
    my_num = id;
  }
  void run(){
    for (int i=0; i < TCOUNT; i++) {
      count_mutex.lock();
      count++;
      BEGIN_PRINT
	cout << "inc_count: Thread "<< my_num << "has a old count of " << count -1 << " and a new count of " << count << "\n";
      END_PRINT
      if (count == WATCH_COUNT) {
	count_threshold_cv.signal();
      }
      count_mutex.unlock();
    }
    BEGIN_PRINT
      cerr << "Thread "<< my_num << "exiting\n";
    END_PRINT
  }

};

class watch_count: public Tulip_Thread{
  int my_num;
public:
  watch_count(int id, char *name ):  Tulip_Thread(name,2*default_stack_size){
    my_num = id;
  }
  void run(){

    count_mutex.lock();
    while (count <= WATCH_COUNT) {
      count_threshold_cv.wait(count_mutex);
      BEGIN_PRINT
	cerr << "watch_count: Thread "<< my_num << "has a count of " << count<< "\n";
      END_PRINT
    }
    count_mutex.unlock();
    BEGIN_PRINT
      cerr << "Watching Thread exiting\n";
    END_PRINT
    
  }
};


int main(int argc, char *argv[]){

  int cpus = (argc > 1) ? atoi(argv[1]) : 1;
  int debug = (argc > 3) ? atoi(argv[3]) : 0;
  int numThreads = (argc > 2) ? atoi(argv[2]) : 20;

  BEGIN_PRINT
  cerr << "About to wait on a CSem variable\n";
  END_PRINT
  double start = Tulip_StartTimer();
  systemContext.addNCpus(cpus-1);
  th[0] = new watch_count(0,"Thread 0");
  systemContext.add(th[0]);

  th[1] = new inc_count(0,"Thread 1");
  systemContext.add(th[1]);

  th[2] = new inc_count(1,"Thread 2");
  systemContext.add(th[2]);
  systemContext.startVerbose();
  join(*th[0]);
  join(*th[1]);
  join(*th[2]);
  systemContext.wait();

  double delta = Tulip_StopTimer(start);
  BEGIN_PRINT
    cerr << "Took " << delta << " seconds\n";
  END_PRINT

} /* main() */


