#include <stdio.h>
#include <stdlib.h>
#include "Tulip_Thread.h"
#include "Tulip_Context.h"
#include "Tulip_ThreadPipe.h"
#include "Tulip_Timers.h"


//
// A bounded pipe into which you can put two
// items before you block adding more
//
Tulip_BoundedPipeList<int> int_pipe(2);

/////////////////////////////////////////////////////////////////////////////
// Producer
/////////////////////////////////////////////////////////////////////////////
class Produce : public Tulip_Thread {
public:
  Produce(char *name, int limit) : Tulip_Thread(name) {
    limit_ = limit;
  }

  virtual void run() {
    for (int i = 0; i < limit_; i++ ) {
      int_pipe.put(i);
      if ( limit_ < 100 ) {
	cerr <<  getName() << i << "\n";
      }
    }
  }
private:
  int limit_;
};

/////////////////////////////////////////////////////////////////////////////
// Consumer
/////////////////////////////////////////////////////////////////////////////

class Consume : public Tulip_Thread {
public:
  Consume(char *name,int limit) : Tulip_Thread(name) {
    limit_ = limit;
  }

  virtual void run() {
    while(1) {
      int i;
      int_pipe.get(i);
      if ( limit_ < 100 ) {
	cerr << getName() << i << "\n";
      }
    }
  }
private:
  int limit_;
};


/////////////////////////////////////////////////////////////////////////////
// Main routines
/////////////////////////////////////////////////////////////////////////////

void main(int argc, char** argv)
{
  int cpus = (argc > 1) ? atoi(argv[1]) : 1;
  int debug = (argc > 2) ? atoi(argv[2]) : 1;

  fprintf(stderr,"About to start scheduling..\n");
  double start = Tulip_StartTimer();
  systemContext.addNCpus(cpus-1);
  Tulip_Context::add(new Produce("produce", 50));
  Tulip_Context::add(new Consume("consume", 50));
  systemContext.wait();
  double delta = Tulip_StopTimer(start);

  fprintf(stderr,"Took %f seconds\n", delta);
  fprintf(stderr,"Back in main, about to exit\n");
}
