#include <stream.h>
#include "Tulip_Context.h"
#include "Tulip_Barrier.h"
#include "Tulip_Timers.h"
#include "Tulip_iolock.h"

/* This example was created by Jean-Francois Mehaut */

class Sum : public Tulip_Thread {
public:
    Sum(int, int, int *) ;
    virtual void run() ;
private:
    int  inf  ;
    int  sup  ;
    int  *res ;  
};


Sum::Sum(int x, int y, int *r) : Tulip_Thread ("")
{
    inf = x ;
    sup = y ;
    res = r ; 
}


void Sum::run()
{
 Tulip_Thread *tid1, *tid2 ;
 int             r1, r2 ;

 if (inf == sup)
   {
    *res = inf ; 
   }
  else
    {
     tid1 = new Sum (inf, (inf + sup)/2, &r1) ;
     Tulip_Context::add (tid1) ;

     tid2 = new Sum (((inf + sup)/2)+1, sup, &r2) ;
     Tulip_Context::add (tid2) ;

     Tulip_Thread::join (*tid1) ;
     Tulip_Thread::join (*tid2) ;  
     delete tid1;
     delete tid2;

     *res = r1 + r2 ; 
    }
}




main(int argc, char** argv)
{
  int cpus = (argc > 1) ? atoi(argv[1]) : 1;

  int limit    = -1 ;
  int result        ;
  Tulip_Thread *tid ;

  systemContext.addNCpus(cpus-1);
  while (limit != 0)     {
    cout << "Enter a rather small integer (0 to quit) : " ;
    cin >> limit ;

    if (limit != 0)      {
      double start = Tulip_StartTimer();
      
      tid = new Sum (1, limit, &result) ;
      Tulip_Context::add (tid) ;
      
      Tulip_Thread::join (*tid) ;
      delete tid;
      cout << "Summing 1.." << limit << " = " << result << "\n" ;
      
      double eltime = Tulip_StopTimer(start);
      cerr << lock << "Total time elapsed in secs: " << eltime << "\n"<< unlock; 
    }
  }
  systemContext.wait();
}






