#include <stream.h>
#include "Tulip_Context.h"
#include "Tulip_Timers.h"
#include "Tulip_Semaphore.h"
#include "Tulip_ThreadPipe.h"
#include "Tulip_iolock.h"
#include <limits.h>

#include <sys/time.h>

#include "Types.h"
#include "Map.h"
#include "Worker.h"

// System variables

int                  nb_workers    ;
Tulip_Semaphore      mutex (1)     ;
Tulip_PipeList<Node> tsp_queue     ;

 
// Application variables

DistTab_t    Map     ;
int          minimum ;

extern void  GenerateMap   (int) ;
extern void  PrintMap      ()    ;

int present (int city, int hops, Path_t path)
{
 unsigned int i ;
 
 for (i=0; i<hops; i++)
   if (path [i] == city) return 1 ;
 return 0 ;
}

void tsp (int hops, int len, Path_t path,  int num_worker)
{
 int i ;
 int city, me, dist ;
 
 if (len >= minimum)
  {
   return ;
  }
 
 if (hops == Map.NrTowns)
  {
   mutex.acquire () ;
 
   if (len < minimum)
     {
       minimum = len ;
       cout << lock << "worker[" << num_worker << "] finds path len = " << len << " :"  << flush << unlock;

       for (i=0; i < Map.NrTowns; i++) {
	 cout << lock << path [i] << " " << flush << unlock;
       }
       cout << lock << "\n" << flush << unlock;
     }
 
    mutex.release () ;
  }
 else
  {
   me = path [hops-1] ;
 
   for (i=0; i < Map.NrTowns; i++)
    {
     city = Map.dst [me][i].ToCity ;
     if (!present (city, hops, path))
      {
       path [hops] = city ;
       dist = Map.dst[me][i].dist ;
       tsp (hops+1, len+dist, path, num_worker) ;
      }
    }
  }
}

void distributor (int hops, int len, Path_t path,  DistTab_t distance)
{
 Node n ;
 int i ;
 int me, city, dist ;
 
 if (hops == MAXHOPS)
  {
   n.len = len ;
   for (i=0; i<hops;i++)
     n.path[i] = path[i] ;

   tsp_queue.put (n) ;
  }
 else
  {
   me = path [hops-1] ;
   for (i=0;i<Map.NrTowns;i++)
    {
     city = Map.dst[me][i].ToCity ;
     if (!present(city,hops, path))
      {
       path [hops] = city ;
       dist = Map.dst[me][i].dist ;
       distributor (hops+1, len+dist, path, Map) ;       
      }
    }
  } 
}

void GenerateJobs () 
{
 Path_t path ;
 
 path [0] = 0 ;
 distributor (1, 0, path, Map) ; 
}


main(int argc, char** argv)
{

 if (argc != 5)
   {
     cout << lock << "Usage: " << argv [0] << " <nb_cpus> <nb_threads > <nb_cities> <seed for random generator> \n";
    exit (1) ;
   }

  int cpus = atoi(argv[1]) ;
  nb_workers         =  atoi (argv [2]) ;
  Map.NrTowns        =  atoi (argv [3]) ;
  int seed           =  atoi (argv [4]) ;
  minimum            =  INT_MAX         ;
  Tulip_ThreadPtr *tids = (Tulip_ThreadPtr *) malloc (nb_workers * sizeof (Tulip_Thread)) ;

  systemContext.addNCpus(cpus-1);
  GenerateMap (seed) ;
  /* PrintMap    ()  ;*/
  double start = Tulip_StartTimer(); 

  GenerateJobs () ;

  for (int i=0; i < nb_workers; i++)
    {
      // Create the worker threads
      tids [i] = new Worker (i) ;
      tids[i]->affinity(i%cpus);
      Tulip_Context::add (tids [i]) ;
    }

  for (i=0; i < nb_workers; i++)
    {
      // Wait the worker terminations
      Tulip_Thread::join (*tids [i]) ;
    }
  systemContext.wait();
  double eltime = Tulip_StopTimer(start);
  cout << lock << "Total time elapsed in secs: " << eltime << "\n" << flush << unlock;
  cout << lock << "Final " << flush << unlock;

  return (0);
}







