/*=======================================================================
 $Log: TulipGenericBarrier.h,v $
 Revision 1.1  1998/07/27 22:04:29  shobana
 central barrier implemented using fetchop instructions

 Revision 1.1  1998/06/23 00:12:07  ejohnson
 moving directory hpcxx_rts up

 Revision 2.0  1998/05/22 16:54:25  suvas
 New Release (using main as a thread)

 Revision 1.3  1998/05/22 16:21:42  suvas
 Regression Test on linux and sunos.

 Revision 1.2  1998/05/12 23:14:12  suvas
 This may look like C code, but it is really C++

 Copyright (c) 1998 Los Alamos National Laboratory
 Copyright (c) 1997 University of Colorado, Boulder

 Suvas Vajracharya and Dirk Grunwald

 $Id: TulipGenericBarrier.h,v 1.1 1998/07/27 22:04:29 shobana Exp $
 $Date: 1998/07/27 22:04:29 $
 $Author: shobana $
 ========================================================================*/
#ifndef TulipBarrier_h
#define TulipBarrier_h

#ifdef __GNUG__
#  pragma interface
#  include <bool.h>
#endif
//
//	TulipBarrier.h
//	The barrier is initially closed.
// 
#include "Tulip_Config.h"
#include <Tulip_Context.h>
#include <Tulip_Mutex.h>
#include <Tulip_Monitor.h>
#include <Tulip_iolock.h>
#include <vector>
template <class Storage> 
class Tulip_BarrierTemplate : public Tulip_Monitor {

protected:
    int pHeight;
    int pCount;
    Storage container;    
    
private:
    //
    // This is executed in the context of a 
    // different thread than the one attempting to do the locking
							       //

    virtual bool acquire_by_upcall(Tulip_Thread *byWho)
      {
	bool blocked = false;
	mutex.lock();
	pCount++;
//	BEGIN_PRINT
//	  cout << "pCount = " << pCount << "," << pHeight << "\n";
//	END_PRINT
	if (pCount < pHeight) {
	  container.push_back( byWho );
	  blocked = true;
	}
	else {
	  release();
	}
	mutex.unlock();
	return( blocked );
      }
    
public:
    
    Tulip_BarrierTemplate(int height = 0) {
      pHeight = height;
      pCount = 0;
      mutex.unlock();
    }
    virtual ~Tulip_BarrierTemplate() {
      //
      // Put in exception check for deleting a 
      // Tulip_Barrier with blocked threads
    }

    void rendezvous() {
      mutex.lock();
      if ( pCount+1 == pHeight) {
	release();
	mutex.unlock();
      } else {
	mutex.unlock();
	// We may need to block - have this done by a third party..
	Tulip_Context::reservation_upcall( this );
      }
    }
    virtual void release()
    {
//      cerr << lock << " RELEASE!!!!\n" << flush << unlock;  
      Tulip_Thread *p = 0;
      while (container.size() > 0) {
	p = container.front();
	Tulip_Context::add(p);
	container.erase(container.begin());
      }
      pCount = 0;
    }

    //
    // lowerBarrier forces all waiting jobs to continue
    //
    void lower() {
      mutex.lock();
      release();
      mutex.unlock();
    }
    //
    // Changing the height of the barrier can cause people to continue
    // if it is lowered below the current pCount.
    //
    int height() {
      return(pHeight);
    }
    void height(int h) {
      mutex.lock();
      if (pCount >= h) {
	release();
      }
      else {
	cout <<"here setting the count";
	pHeight = h;
      }
      mutex.unlock();
    }
    //
    // how many tasks are waiting?
    //
    int count() {
      return( pCount );
    }
};

//
// Create a default instance of the template for most applications to
// use -- the Tulip_Semaphore, which uses a STL vector type for
// internal storage.
//
typedef vector<Tulip_ThreadPtr> Tulip_BarrierThreadVector;
typedef Tulip_BarrierTemplate<Tulip_BarrierThreadVector> Tulip_Barrier;
#endif /* Tulip_Barrier_h */
