/*=======================================================================
 $Log: Tulip_CSem.h,v $
 Revision 1.2  1998/06/23 14:14:18  suvas
 Fixed the problem of the system exiting while there
 are still threads blocked on synchronization primitives.

 Revision 2.2  1998/06/23 00:37:10  ejohnson
 moving directory hpcxx_rts up

 Revision 2.0  1998/05/22 16:54:26  suvas
 New Release (using main as a thread)

 Revision 1.2  1998/05/12 23:14:14  suvas
 This may look like C code, but it is really C++

 Copyright (c) 1998 Los Alamos National Laboratory
 Copyright (c) 1997 University of Colorado, Boulder

 Suvas Vajracharya and Dirk Grunwald

 ========================================================================*/
#ifndef _Tulip_CSem_h_
#define _Tulip_CSem_h_

#ifdef __GNUG__
#  pragma interface
#  include <bool.h>
#endif

#include "Tulip_Monitor.h"
#include <vector>

template <class Storage>
class Tulip_CSemTemplate : public Tulip_Monitor {

protected:
  int pCount;
  int height;
  Storage container;

private:
  //
  // This is executed in the context of a 
  // different thread than the one attempting to do the locking
  //
  virtual bool acquire_by_upcall(Tulip_Thread *byWho)
  {
    bool blocked = false;
    mutex.lock();
    if (pCount < height) {
      Tulip_Context::GlobalEventsCounter+=1;
      container.push_back( byWho );
      blocked = true;
    }
    mutex.unlock();
    return( blocked );
  }   
 
  void release()
    {
      Tulip_Thread *p = 0;
      while (!container.empty()) {
	p = container.front();
	Tulip_Context::add(p);
	Tulip_Context::GlobalEventsCounter-=1;
	container.erase(container.begin());
      }
    }

public:

  Tulip_CSemTemplate(int limit=0)
  {
    height = limit;
    pCount = 0;
    mutex.unlock();
  }
  virtual ~Tulip_CSemTemplate() {
    //
    // Put in exception check for deleting a 
    // Tulip_CSem with blocked threads
  }

  void wait() {
    mutex.lock();
    if (pCount < height) {
      mutex.unlock();
      // We may need to block - have this done by a third party..
      Tulip_Context::reservation_upcall( this );
    }
    else {
      mutex.unlock();
    }
    pCount = 0;
  }

  int get_bcount() {
    return pCount;
  }
  void set_height(int d) {
    height = d;
  }
  void raise_height(int d) {
    height += d;
  }
  void incr(){
    mutex.lock();
    pCount += 1;
    //#ifdef MIRACLE_VERBOSE
    BEGIN_PRINT
      cerr << "Finished:pCount,height = " << pCount << "," << height <<"\n" << flush;
    END_PRINT
      //#endif
    if (pCount >= height) {
      release();
    }
    mutex.unlock();
  };
  int operator+=(int d) 
  {
      mutex.lock();
      pCount += d;
      //#ifdef MIRACLE_VERBOSE
      BEGIN_PRINT
	cerr << "pCount,height,num waiting = " << pCount << "," << height <<"," << container.size() << "\n";
      END_PRINT
      //#endif
      if (pCount >= height) {
	release();
      }
      mutex.unlock();
      return pCount;
  }
};

#include "Tulip_CSem-interface.h"

#endif
