/*=======================================================================
 $Log: Tulip_Cond.h,v $
 Revision 1.2  1998/06/23 14:14:18  suvas
 Fixed the problem of the system exiting while there
 are still threads blocked on synchronization primitives.

 Revision 2.2  1998/06/23 00:37:11  ejohnson
 moving directory hpcxx_rts up

 Revision 2.0  1998/05/22 16:54:26  suvas
 New Release (using main as a thread)

 Revision 1.2  1998/05/12 23:14:16  suvas
 This may look like C code, but it is really C++

 Copyright (c) 1998 Los Alamos National Laboratory
 Copyright (c) 1997 University of Colorado, Boulder

 Suvas Vajracharya and Dirk Grunwald

 $Id: Tulip_Cond.h,v 1.2 1998/06/23 14:14:18 suvas Exp $
 $Date: 1998/06/23 14:14:18 $
 $Author: suvas $
 ========================================================================*/
#ifndef _Tulip_Cond_h_
#define _Tulip_Cond_h_

#ifdef __GNUG__
#  pragma interface
#  include <bool.h>
#endif

#include "Tulip_Monitor.h"
#include "Tulip_Thread.h"
#include <vector>

template <class Storage>
class Tulip_CondTemplate : public Tulip_Monitor {

protected:
  Storage container;

private:
  //
  // This is executed in the context of a 
  // different thread than the one attempting to do the locking
  //
  Tulip_Mutex *u_mutex;
  virtual bool acquire_by_upcall(Tulip_Thread *byWho)
  {
    mutex.lock();
    Tulip_Context::GlobalEventsCounter+=1;
    container.push_back( byWho );
    u_mutex->unlock();
    mutex.unlock();
    return true;
  }   
    
public:
  Tulip_CondTemplate(void ) {
    mutex.unlock();
  }

  //#--------------------------------------------------------------------------
  // Suspends the calling thread until another thread releases the wait
  // condition.  The calling thread is required to lock the given mutex
  // variable before calling wait.  The mutex is unlocked when the wait
  // state begins so that another thread can acquire the mutex before
  // releasing the wait condition.
  //#--------------------------------------------------------------------------

  void wait(Tulip_Mutex& user_mutex) {
    u_mutex = &user_mutex;
    Tulip_Context::reservation_upcall( this);
    user_mutex.lock();
  }

  //#--------------------------------------------------------------------------
  // Releases one of the potentially many threads waiting on the condition.
  // The calling thread must first acquire a lock on the mutex object
  // (that was used in the call to wait()) before calling this method.
  //#--------------------------------------------------------------------------
  bool signal(void) {
    bool releasedOne = false;
    Tulip_ThreadPtr p;
    mutex.lock();
    if (!container.empty()) {
      p = container.front();
      Tulip_Context::add(p);
      Tulip_Context::GlobalEventsCounter-=1;
      container.erase(container.begin());
      releasedOne = true;
    }
    mutex.unlock();
    return releasedOne;
  }

  //#--------------------------------------------------------------------------
  // Releases all of the threads waiting on the condition.  The calling thread
  // must first acquire a lock on the mutex object (that was used in the call
  // to wait()) before calling this method.
  //#--------------------------------------------------------------------------
  int broadcast(void){
    Tulip_Thread *p;
    mutex.lock();
    while (!container.empty()) {
      p = container.front();
      Tulip_Context::add(p);
      Tulip_Context::GlobalEventsCounter-=1;
      container.erase(container.begin());
    }
    mutex.unlock();
  }

  virtual ~Tulip_CondTemplate() {
    //
    // Put in exception check for deleting a 
    // Tulip_Cond with blocked threads
  }
};

typedef vector<Tulip_ThreadPtr> Tulip_CondThreadVector;
//template <class Storage> class Tulip_CondTemplate;
typedef Tulip_CondTemplate<Tulip_CondThreadVector> Tulip_Cond;

#endif
