/*=======================================================================
 $Log: Tulip_Context.h,v $
 Revision 1.2  1998/06/23 14:14:19  suvas
 Fixed the problem of the system exiting while there
 are still threads blocked on synchronization primitives.

 Revision 1.1  1998/06/23 00:12:14  ejohnson
 moving directory hpcxx_rts up

 Revision 2.0  1998/05/22 16:54:27  suvas
 New Release (using main as a thread)

 Revision 1.5  1998/05/21 22:18:49  suvas
 Changed thread model to make "main" a thread. Fixed cpu affinity.

 Revision 1.2  1998/05/12 23:14:18  suvas
 This may look like C code, but it is really C++

 Copyright (c) 1998 Los Alamos National Laboratory
 Copyright (c) 1997 University of Colorado, Boulder

 Suvas Vajracharya and Dirk Grunwald

 $Id: Tulip_Context.h,v 1.2 1998/06/23 14:14:19 suvas Exp $
 $Date: 1998/06/23 14:14:19 $
 $Author: suvas $
 ========================================================================*/
//-*-c++-*-
#ifndef _Tulip_Context_h_
#define _Tulip_Context_h_

#ifdef __GNUG__
#  pragma interface
#endif

#include <assert.h>
#include <iostream.h>
#include "Tulip_Config.h"
#include "Context.h"
#include "Tulip_Thread.h"
#ifdef MIRACLE
#include "Tulip_Iterate.h"
#endif
#include "Tulip_MainThread.h"
#include "Tulip_SpinBarrier.h"
#include "Tulip_SpinFetchAndOp.h"
#include "Tulip_Helpers.h"
#include "Tulip_iolock.h"

#define SYSTEM_STACK_SIZE 4*8192

class Tulip_CSem;
//class Tulip_Iterate;
//typedef Tulip_Iterate * Tulip_IteratePtr;
//typedef list<Tulip_IteratePtr>::iterator  listIterator;
class Tulip_AcquireByUpcall;
class Tulip_Context;
typedef Tulip_Context *Tulip_ContextPtr;
extern Tulip_Context systemContext;
#define MYID (Tulip_Context::system())->Me()
#define BEGIN_PRINT cerr<<lock<<"P"<<Tulip_Context::Me()<<": ";
#define END_PRINT ;cerr<<flush<<unlock; 

#define BEGIN_DEBUG_PRINT if (Tulip_Context::debug) { cerr<<lock<<"P"<<Tulip_Context::Me()<<": " ;
#define END_DEBUG_PRINT ;cerr<<flush<<unlock; } 



class Tulip_Context {

public:
  /* don't move or delete this */
  long pid;
  static int waveFront;
  static void addNCpus(int);
  static void _call_schedule(Tulip_ContextPtr sysContext);
  void wait() {
    loopback_pointer = &systemContext;
    systemContext.schedule_private();
  }
  static Tulip_SpinBarrier barrier;
  void Initialize();
  static void startVerbose() {debug =1;};
  static void endVerbose() {debug =0;};
  static Tulip_FetchAndOp GlobalEventsCounter;  
  void operator=(const Tulip_Context &x){};
  static Tulip_PrivateContext loopback_pointer;
  static int numcpus;
  static int numThreadsScheduled;
  Tulip_Context();
  void schedule(int _numcpus);
  void suspend(Tulip_ThreadPtr);
  void resume(list<Tulip_ThreadPtr>::iterator);
  void schedule_private();
  static Tulip_ContextPtr getContextPointer() {
    return (Tulip_ContextPtr)&loopback_pointer->pid;
  }
  static int Me() {
    return ((Tulip_Context::system())->me);
  }
  static void SetMe(int me) {
    (Tulip_Context::system())->me = me;
  }
  static Tulip_ContextPtr *Contexts;
  static int debug;

  static Tulip_PrivateContext system() {
    return loopback_pointer;
  }
  static Tulip_Thread *thread() {
    return  loopback_pointer -> current_thread;
  }

  static void addToCpu(int cpu, Tulip_Thread *thread) {
    int affin = thread->affinity();
    if (affin > -1) {
      if (affin >= numcpus) {
	BEGIN_PRINT
	  cerr << "Tulip_Context: Error, can't add to a non-existent CPU\n";
        END_PRINT
      }
      Contexts[affin]->affinityThreadQueue.add(thread);
    }
    else {
      if (affin >= numcpus) {
	BEGIN_PRINT
	  cerr << "Tulip_Context: Error, can't add to a non-existent CPU\n";
        END_PRINT
      }
      Contexts[cpu] -> threadQueue.add(thread);
    }
    GlobalThreadCounter+=1;
  }

  static void add(Tulip_Thread *thread) {
    int affin = thread->affinity();
    if (affin > -1) {
      if (affin >= numcpus) {
	BEGIN_PRINT
	  cerr << "Tulip_Context: Error, can't add to a non-existent CPU\n";
        END_PRINT
      }
      BEGIN_DEBUG_PRINT
	cerr << "Adding " << thread->getName() << " to proc " << affin << "\n";
      END_DEBUG_PRINT
      Contexts[affin]->affinityThreadQueue.add(thread);
    }
    else {
      BEGIN_DEBUG_PRINT
	cerr << "Adding " << thread->getName() << " to proc " << Me() << "\n";
      END_DEBUG_PRINT
      loopback_pointer -> threadQueue.add(thread);
    }
    GlobalThreadCounter+=1;
  }


#ifdef MIRACLE
  static void blockingEvaluate();
  static Tulip_IteratePtr removeIterate();
  static void addIterate(Tulip_IteratePtr iterate);
  static void addIterateNoIncr(Tulip_IteratePtr iterate);
  static void registerIterates(Tulip_IteratePtr *iterateArray, int num);
  static void evaluate();
  static Tulip_IteratePtr removeNonRegistered();
  static void releaseReadyIterates();
  static void addReadyIterate(Tulip_IteratePtr iterate);
  Tulip_IteratePtr removeIterate_internal();
  void addIterate_internal(Tulip_IteratePtr);
#endif

  static int size() {
    return loopback_pointer -> size_internal();
  }

  static void yield() {
    loopback_pointer -> yield_internal(loopback_pointer -> current_thread);
  }

  static void reservation_upcall(Tulip_AcquireByUpcall *resource) {
    loopback_pointer -> reservation_upcall_internal(resource);
  }



protected:
  //
  // Data specific to this instance
  //
  int me;
  Tulip_NonPremptiveContext syscontext;
  Tulip_Thread *current_thread;
  void setCurrentThread(Tulip_ThreadPtr th) {
    current_thread=th;
  } 

  LockedPriorityQueue<Tulip_RunnablePtr> threadQueue;
  LockedPriorityQueue<Tulip_RunnablePtr> affinityThreadQueue;

#ifdef MIRACLE
  LockedPriorityQueue<Tulip_RunnablePtr> iterateQueue;
  static list<Tulip_IteratePtr> ready_iterate_list;
  static list<Tulip_IteratePtr> not_ready_iterate_list;
  static Tulip_Mutex ready_iterate_list_lock;
  static Tulip_Mutex not_ready_iterate_list_lock;
#endif

  Tulip_Thread *yielding_thread;

  list<Tulip_ThreadPtr> thread_bag;
  Tulip_Mutex thread_bag_lock;
  //
  // Class data
  //
protected:
  static int GivingUpCounter;
  static int GivingUpGeneration;
  static Tulip_Mutex GivingUpLock;
  static Tulip_FetchAndOp GlobalThreadCounter;  
#ifdef MIRACLE
  static Tulip_FetchAndOp GlobalIterateCounter;  
  static Tulip_CSem GlobalRegistered;  
  static Tulip_CSem GlobalFinished;
#endif
  static int nextID;
  static Tulip_Mutex nextIDLock;
  //
  // Internal functions
  //
  friend class Tulip_Thread;
  friend class Tulip_Iterate;

  static void kill_thread(Tulip_Thread *thread);


  Tulip_Thread* removeThread_internal();
  int size_internal();
  void yield_internal(Tulip_Thread *);

  //
  // Reservation_upcall is for synchronization primitives
  //
  void reservation_upcall_internal(Tulip_AcquireByUpcall *resource);
  static void reservation_helper(Tulip_Context* who);
  Tulip_AcquireByUpcall *upcall_resource;
  Tulip_Thread* upcall_blocking_thread;
private:
  static Tulip_ThreadPtr removeThread();
  Tulip_ThreadPtr affinityThreadRemove() {
    Tulip_ThreadPtr thr=NULL;
    if (thr = (Tulip_ThreadPtr)affinityThreadQueue.remove()) {
      BEGIN_DEBUG_PRINT
      cerr << "Removing an affinity thread "<< thr->getName() << "\n";
      END_DEBUG_PRINT
      if (thr->state != ALIVE) {
	return (Tulip_ThreadPtr)affinityThreadQueue.remove();
      }
      else {
	GlobalThreadCounter-=1;
      }
    }
    return thr;

  }
};




extern void Tulip_Init();

struct __Tulip_Init_hack_t 
{ 
  __Tulip_Init_hack_t () { Tulip_Init (); } 
}; 
inline
void yield() {
  Tulip_Context::yield();
}
typedef Tulip_Context *Tulip_ContextPtr;
#ifdef MIRACLE
#include "Tulip_IterateContext.h"
#endif
#endif


