/*=======================================================================
 $Log: Tulip_Semaphore.h,v $
 Revision 1.2  1998/06/23 14:14:20  suvas
 Fixed the problem of the system exiting while there
 are still threads blocked on synchronization primitives.

 Revision 2.1  1998/06/23 00:37:22  ejohnson
 moving directory hpcxx_rts up

 Revision 2.0  1998/05/22 16:54:31  suvas
 New Release (using main as a thread)

 Revision 1.2  1998/05/12 23:14:24  suvas
 This may look like C code, but it is really C++

 Copyright (c) 1998 Los Alamos National Laboratory
 Copyright (c) 1997 University of Colorado, Boulder

 Suvas Vajracharya and Dirk Grunwald

 $Id: Tulip_Semaphore.h,v 1.2 1998/06/23 14:14:20 suvas Exp $
 $Date: 1998/06/23 14:14:20 $
 $Author: suvas $
 ========================================================================*/
#ifndef _Tulip_Semaphore_h_
#define _Tulip_Semaphore_h_

#ifdef __GNUG__
#  pragma interface
#  include <bool.h>
#endif

#include "Tulip_Monitor.h"
#ifdef __sgi
#include <vector.h>
#else
#include <vector>
#endif


template <class Storage>
class Tulip_SemaphoreTemplate : public Tulip_Monitor {

protected:
  int pCount;
  Storage container;

private:
  //
  // This is executed in the context of a 
  // different thread than the one attempting to do the locking
  //
  virtual bool acquire_by_upcall(Tulip_Thread *byWho)
  {
    bool blocked = false;
    mutex.lock();
    pCount--;
    if (pCount < 0) {
      Tulip_Context::GlobalEventsCounter+=1;
      container.push_back( byWho );
      blocked = true;
    };
    mutex.unlock();
    return( blocked );
  }
    
public:

  Tulip_SemaphoreTemplate(int count = 1)
  {
    pCount = count;
    mutex.unlock();
  }

  virtual ~Tulip_SemaphoreTemplate() {
    //
    // Put in exception check for deleting a 
    // Tulip_Semaphore with blocked threads
  }

  virtual void acquire() {
    mutex.lock();
    if (pCount < 1) {
      //
      // We may need to block - have this done by a third party..
      //
      mutex.unlock();
      Tulip_Context::reservation_upcall( this );
    }
    else {
      pCount--;
      mutex.unlock();
    }

  }

  virtual void release()
  {
    Tulip_Thread *p = 0;
    mutex.lock();
    if (pCount < 0) {
      p = container.front();
      container.erase(container.begin());
    }
    pCount++;
    mutex.unlock();

    if (p != 0) {
	Tulip_Context::add(p);
	Tulip_Context::GlobalEventsCounter-=1;
    }
  }

  
  //
  // Attempt to lock the resource without blocking.
  // 
  virtual bool tryAcquire()
  {
    int gotIt = TRUE;

    mutex.lock();
    if (pCount < 1) {
	gotIt = FALSE;
    }
    else {
	pCount--;
    }
    mutex.unlock();
    return( gotIt );
  }

  virtual unsigned size() {
    return container.size();
  }

  //
  //	You should not change the count when threads are blocked
  //
  int get_count() {
    return pCount;
  }

  void set_count(int count) {
    mutex.lock();
    pCount = count;
    mutex.unlock();
  }

  void increment_count(int increment) {
    mutex.lock();
    pCount += increment;
    mutex.unlock();
  }

  int isEmpty() {
    return container.size() == 0;
  }
};

//
// Create a default instance of the template for most applications to
// use -- the Tulip_Semaphore, which uses a STL vector type for
// internal storage.
//
typedef vector<Tulip_ThreadPtr> Tulip_SemaphoreThreadVector;
typedef Tulip_SemaphoreTemplate<Tulip_SemaphoreThreadVector> Tulip_Semaphore;


#endif
