/*=======================================================================
 $Log: Tulip_SpinBarrier.h,v $
 Revision 1.1  1998/06/23 00:12:25  ejohnson
 moving directory hpcxx_rts up

 Revision 2.0  1998/05/22 16:54:31  suvas
 New Release (using main as a thread)

 Revision 1.2  1998/05/12 23:14:25  suvas
 This may look like C code, but it is really C++

 Copyright (c) 1998 Los Alamos National Laboratory
 Copyright (c) 1997 University of Colorado, Boulder

 Suvas Vajracharya and Dirk Grunwald

 $Id: Tulip_SpinBarrier.h,v 1.1 1998/06/23 00:12:25 ejohnson Exp $
 $Date: 1998/06/23 00:12:25 $
 $Author: ejohnson $
 ========================================================================*/
#ifndef Tulip_SpinBarrier_h
#define Tulip_SpinBarrier_h

#ifdef __GNUG__
#  pragma interface
#endif

//
//	Implement a UNIX process barrier
//
#include <Tulip_Config.h>
#include <Tulip_Mutex.h>
#include <assert.h>

class Tulip_SpinBarrier : public Tulip_Mutex {
 protected:
    int pHeight;
    volatile int pCount;
    volatile int generation;
    int pLoops;
    int pMaxLoops;

public:
     Tulip_SpinBarrier( int h, int loops = DefaultTulipSpinBarrierLoops, int max = 0 );
    ~Tulip_SpinBarrier();

    //
    // Normal entry. Everyone calls rendezvous()
    //
    int rendezvous();
    //
    // lowerBarrier forces all waiting jobs to continue
    //
    void lower();

    //
    // Changing the height of the barrier can cause people to continue
    // if it is lowered below the current pCount.
    //
    int height();
    void height(int h);

    //
    // loops() is the numbr of times a barrier spins before doing a getpid()
    // to release the UNIX cpu. maxLoopss() the maximum number of getpids()
    // to do before returning from the barrier (i.e. for barrier with timeout)
    //
    int loops();
    void loops(int h);
    int maxLoops();
    void maxLoops(int h);

    int count();
};




inline int
Tulip_SpinBarrier::height()
{
    return(pHeight);
}

inline int
Tulip_SpinBarrier::loops()
{
    return(pLoops);
}

inline void
Tulip_SpinBarrier::loops(int h)
{
    pLoops = h;
}

inline int
Tulip_SpinBarrier::maxLoops()
{
    return(pMaxLoops);
}

inline void
Tulip_SpinBarrier::maxLoops(int h)
{
    pMaxLoops = h;
}

inline int
Tulip_SpinBarrier::count()
{
    return(pCount);
}

inline
Tulip_SpinBarrier::~Tulip_SpinBarrier()
{
    assert( pCount == 0 );
}


#endif /* Tulip_SpinBarrier */
