/*=======================================================================
 $Log: Tulip_Sync.h,v $
 Revision 1.2  1998/06/23 14:14:20  suvas
 Fixed the problem of the system exiting while there
 are still threads blocked on synchronization primitives.

 Revision 2.1  1998/06/23 00:37:25  ejohnson
 moving directory hpcxx_rts up

 Revision 2.0  1998/05/22 16:54:32  suvas
 New Release (using main as a thread)

 Revision 1.2  1998/05/12 23:14:27  suvas
 This may look like C code, but it is really C++

 Copyright (c) 1998 Los Alamos National Laboratory
 Copyright (c) 1997 University of Colorado, Boulder

 Suvas Vajracharya and Dirk Grunwald

 $Id: Tulip_Sync.h,v 1.2 1998/06/23 14:14:20 suvas Exp $
 $Date: 1998/06/23 14:14:20 $
 $Author: suvas $
 ========================================================================*/
#ifndef _Tulip_Sync_h_
#define _Tulip_Sync_h_

#ifdef __GNUG__
#  pragma interface
#  include <bool.h>
#endif

#include "Tulip_Monitor.h"
#include "Tulip_Thread.h"
#include <vector>

template <class T, class Storage>
class Tulip_SyncTemplate : public Tulip_Monitor {

protected:
  T data;
  Storage container;
  bool written;

private:
  //
  // This is executed in the context of a 
  // different thread than the one attempting to do the locking
  //
  virtual bool acquire_by_upcall(Tulip_Thread *byWho)
  {
    bool blocked = false;
    if (!written) {
      mutex.lock();
      Tulip_Context::GlobalEventsCounter+=1;
      container.push_back( byWho );
      blocked = true;
      mutex.unlock();
    }

    return( blocked );
  }   
  virtual void release()
  {
    Tulip_Thread *p = 0;
    mutex.lock();
    while (!container.empty()) {
      p = container.front();
      container.erase(container.begin());
      Tulip_Context::add(p);
      Tulip_Context::GlobalEventsCounter-=1;
    }
    mutex.unlock();
  }
public:

  Tulip_SyncTemplate()
  {
    written = FALSE;
    mutex.unlock();
  }

  virtual ~Tulip_SyncTemplate() {
    //
    // Put in exception check for deleting a 
    // Tulip_Sync with blocked threads
  }
  T& operator=(const T& value) {
    data = value;
    return data;
  }
  T& read() {
    if (!written) {
      Tulip_Context::reservation_upcall( this );      
    }
    return data;
  }
  operator T() {
    if (!written) {
      Tulip_Context::reservation_upcall( this );      
    }
    return data;
   }
  void reset() {
    written = FALSE;
    mutex.unlock();
  }
    
  void write(const T &value) {
    data = value;
    written = TRUE;
    release();
  }
  bool peek() {
    return written;
  }
    
};

typedef vector<Tulip_ThreadPtr> Tulip_SyncThreadVector;

template <class Contained>
class Tulip_Sync: public Tulip_SyncTemplate<Contained, Tulip_SyncThreadVector> 
{
    //
    // Empty
    //
};
#endif
