/*=======================================================================
 $Log: Tulip_Thread.h,v $
 Revision 1.1  1998/06/23 00:12:29  ejohnson
 moving directory hpcxx_rts up

 Revision 2.1  1998/06/15 13:11:16  suvas
 *** empty log message ***

 Revision 1.3  1998/05/12 23:14:28  suvas
 This may look like C code, but it is really C++

 Copyright (c) 1998 Los Alamos National Laboratory
 Copyright (c) 1997 University of Colorado, Boulder

 Suvas Vajracharya and Dirk Grunwald

 $Id: Tulip_Thread.h,v 1.1 1998/06/23 00:12:29 ejohnson Exp $
 $Date: 1998/06/23 00:12:29 $
 $Author: ejohnson $
 ========================================================================*/
//-*-c++-*-
#ifndef _Tulip_Thread_h_
#define _Tulip_Thread_h_

#ifdef __GNUG__
#  pragma interface
#endif

#include <list>
#include "Context.h"
#include "Tulip_Config.h"
#include "Tulip_Mutex.h"
#include "Tulip_Helpers.h"
#include "Tulip_Runnable.h"

#include <sys/types.h>
class Tulip_Thread;
typedef Tulip_Thread *Tulip_ThreadPtr;

class Tulip_Thread : public Tulip_Runnable
{
  friend class Tulip_Context;
public:
  #define NAME_SIZE 40
  bool isAlive();
  void suspend();
  void resume();

  int getThreadID() {
    return threadID;
  }
  virtual void run() = 0;
  // Starts execution of the thread and calls the run() method
  void start();

  // Ceases execution of the thread.  Status is not currently used.
  void stop(void *status);

  // Yields the processesor to another thread
  virtual void yield();

  // Sleep for milliseconds,nanoseconds
  // Will be rounded up to the resolution of the sytem timer
  static void sleep(long milliseconds, long nanoseconds = 0);

  // Join waits for the completion of the specified thread,
  // by definition join returns as soon as the thread is considered
  // not alive
  static void join(const Tulip_Thread &withThread);
  virtual ~Tulip_Thread();
  char *getName() const {
    return (char *)_name;
  }

protected:
  //
  //
  Tulip_NonPremptiveContext context;
  Tulip_U64 *stack_malloc_at;
  Tulip_U64 *stack_end_at;

  Tulip_Thread(char *name);
  Tulip_Thread(char *name, unsigned long stacksize);
  Tulip_Thread(char *name, Tulip_U64 *stack, unsigned long stacksize);

  void _thread_init();
  static void _startoff(Tulip_Thread*);

  char *setName(const char *name)  {
    if (strlen(name) < NAME_SIZE)
      strcpy(_name,name);
    else
      strncpy(_name,name,NAME_SIZE-1);
    return (char *) _name;
  }

#define default_stack_size 8192
  int _stacksize;

private:
  char _name[NAME_SIZE];
  void execute() {
    run();
    state = STOPPED;
  }
  void execute(Tulip_NonPremptiveContext *);
  list<Tulip_ThreadPtr>::iterator suspendedAt;
  int threadID;
  Type theType();
  static Tulip_Mutex numThreadsLock;
  static int numThreadsCreated;
};

inline 
Type Tulip_Thread::theType() {
  return ThreadType;
}
inline bool Tulip_Thread::isAlive()
{
  if (state==ALIVE)
    return TRUE;
  else 
    return FALSE;
}
inline void Tulip_Thread::start()
{
  state = ALIVE;
  run();
}
inline
void Tulip_Thread::stop(void *status)
{
  state = STOPPED;
}


#endif
