/*=======================================================================
 $Log: Context.h,v $
 Revision 1.1  1998/06/23 00:12:52  ejohnson
 moving directory hpcxx_rts up

 Revision 2.0  1998/05/22 16:54:46  suvas
 New Release (using main as a thread)

 Revision 1.2  1998/05/12 23:15:07  suvas
 This may look like C code, but it is really C++

 Copyright (c) 1998 Los Alamos National Laboratory
 Copyright (c) 1997 University of Colorado, Boulder

 Suvas Vajracharya and Dirk Grunwald

 $Id: Context.h,v 1.1 1998/06/23 00:12:52 ejohnson Exp $
 $Date: 1998/06/23 00:12:52 $
 $Author: ejohnson $
 ========================================================================*/
#ifndef _Context_h_
#define _Context_h_

#ifdef __cplusplus
extern "C" {
#endif

typedef unsigned long Tulip_U64;

typedef void (*Tulip_voidfuncptr)();

typedef struct {
  Tulip_voidfuncptr func;
  void *arg;
} Tulip_Closure;

typedef struct {
  /*
   * This contains the closure for the continuation used
   * to resurect this context. This must be the first field
   * in this struct and in the struct for all resumption routines
   * (i.e., the premptive case as well)
   */

  Tulip_voidfuncptr resume;

  /*
   * Information for this non-premptive context
   */
  Tulip_U64 R9;  /* S0 */
  Tulip_U64 R10; /* S1 */
  Tulip_U64 R11; /* S2 */
  Tulip_U64 R12; /* S3 */
  Tulip_U64 R13; /* S4 */
  Tulip_U64 R14; /* S5 */
  Tulip_U64 R15; /* FP */
  Tulip_U64 R26; /* RA */
  Tulip_U64 R29; /* GP */
  Tulip_U64 R30; /* SP */

  Tulip_U64 F2;
  Tulip_U64 F3;
  Tulip_U64 F4;
  Tulip_U64 F5;
  Tulip_U64 F6;
  Tulip_U64 F7;
  Tulip_U64 F8;
  Tulip_U64 F9;

  /*
   * Full register state for a premptive context switch
   */
  
  Tulip_U64 R0;
  Tulip_U64 R1;
  Tulip_U64 R2;
  Tulip_U64 R3;
  Tulip_U64 R4;
  Tulip_U64 R5;
  Tulip_U64 R6;
  Tulip_U64 R7;
  Tulip_U64 R8;

  Tulip_U64 R16;
  Tulip_U64 R17;
  Tulip_U64 R18;
  Tulip_U64 R19;

  Tulip_U64 R20;
  Tulip_U64 R21;
  Tulip_U64 R22;
  Tulip_U64 R23;
  Tulip_U64 R24;
  Tulip_U64 R25;

  Tulip_U64 R27;
  Tulip_U64 R28;

  /*
   * Floating point registers
   */
  Tulip_U64 F0;
  Tulip_U64 F1;

  Tulip_U64 F10;
  Tulip_U64 F11;
  Tulip_U64 F12;
  Tulip_U64 F13;
  Tulip_U64 F14;
  Tulip_U64 F15;
  Tulip_U64 F16;
  Tulip_U64 F17;
  Tulip_U64 F18;
  Tulip_U64 F19;
  Tulip_U64 F20;
  Tulip_U64 F21;
  Tulip_U64 F22;
  Tulip_U64 F23;
  Tulip_U64 F24;
  Tulip_U64 F25;
  Tulip_U64 F26;
  Tulip_U64 F27;
  Tulip_U64 F28;
  Tulip_U64 F29;
  Tulip_U64 F30;

} Tulip_NonPremptiveContext;
extern "C" {
void 
Tulip_SaveNP(Tulip_NonPremptiveContext *resurrect,
	     Tulip_voidfuncptr *hook_func,
	     void* hook_data,
	     Tulip_NonPremptiveContext *saveto);

void 
Tulip_RestoreNP(Tulip_NonPremptiveContext *resurrect,
	     Tulip_voidfuncptr *hook_func,
	     void* hook_data);
}
#define Tulip_SwitchTo(to, hookf, hookd, from) \
  { Tulip_SaveNP(to, hookf, hookd, from); }


/*
 * Initialize the context to use the specified
 * stack address, and initialize such that execution
 * starts at 'routine' with initial argument 'argument'
 */
extern void
Tulip_InitNP(Tulip_NonPremptiveContext *context,
	     Tulip_U64 *stack,
	     void *routine,
	     void *argument);

extern void
Tulip_SetResume(Tulip_NonPremptiveContext *context,
		void *routine);

#define Tulip_SetNonPremptiveResume(X) Tulip_SetResume(X, (void *) &Tulip_RestoreNP);

extern
void Tulip_spinlock(unsigned long *, unsigned long);

extern
long Tulip_spintrylock(unsigned long *, unsigned long);

#ifdef __cplusplus
}
#endif

#endif
