/*=======================================================================
 $Log: Context.h,v $
 Revision 1.1  1998/06/23 00:13:22  ejohnson
 moving directory hpcxx_rts up

 Revision 2.0  1998/05/22 16:55:03  suvas
 New Release (using main as a thread)

 Revision 1.2  1998/05/12 23:15:41  suvas
 This may look like C code, but it is really C++

 Copyright (c) 1998 Los Alamos National Laboratory
 Copyright (c) 1997 University of Colorado, Boulder

 Suvas Vajracharya and Dirk Grunwald

 $Id: Context.h,v 1.1 1998/06/23 00:13:22 ejohnson Exp $
 $Date: 1998/06/23 00:13:22 $
 $Author: ejohnson $
 ========================================================================*/
#ifndef _Context_h_
#define _Context_h_

typedef unsigned long Tulip_U64;

typedef void (*Tulip_voidfuncptr)();

typedef struct {
  Tulip_voidfuncptr func;
  void *arg;
} Tulip_Closure;

typedef struct {
  /*
   * This contains the closure for the continuation used
   * to resurect this context. This must be the first field
   * in this struct and in the struct for all resumption routines
   * (i.e., the premptive case as well)
   */

  Tulip_voidfuncptr resume;

  Tulip_U64 R16;  /* S0 */
  Tulip_U64 R17;  /* S1 */
  Tulip_U64 R18;  /* S2 */
  Tulip_U64 R19;  /* S3 */
  Tulip_U64 R20;  /* S4 */
  Tulip_U64 R21;  /* S5 */
  Tulip_U64 R22;  /* S6 */
  Tulip_U64 R23;  /* S7 */
  Tulip_U64 R24;  /* S8 */
  Tulip_U64 R28;  /* GP */

  Tulip_U64 R29;  /* SP */
  Tulip_U64 R30;  /* FP */
  Tulip_U64 R31;  /* RA */

  Tulip_U64 F20;
  Tulip_U64 F22;
  Tulip_U64 F24;
  Tulip_U64 F26;
  Tulip_U64 F28;
  Tulip_U64 F30;


} Tulip_NonPremptiveContext;

extern "C" {
void Tulip_spinlock(unsigned long *, unsigned long);
long Tulip_spintrylock(unsigned long *, unsigned long);
void 
Tulip_SaveNP(Tulip_NonPremptiveContext *resurrect,
	     Tulip_voidfuncptr *hook_func,
	     void* hook_data,
	     Tulip_NonPremptiveContext *saveto);

void 
Tulip_RestoreNP(Tulip_NonPremptiveContext *resurrect,
	     Tulip_voidfuncptr *hook_func,
	     void* hook_data);
}

#define Tulip_SwitchTo(to, hookf, hookd, from) \
  { Tulip_SaveNP(to, hookf, hookd, from); }


/*
 * Initialize the context to use the specified
 * stack address, and initialize such that execution
 * starts at 'routine' with initial argument 'argument'
 */
extern void
Tulip_InitNP(Tulip_NonPremptiveContext *context,
	     Tulip_U64 *stack,
	     void *routine,
	     void *argument);

extern void
Tulip_SetResume(Tulip_NonPremptiveContext *context,
		void *routine);

#define Tulip_SetNonPremptiveResume(X) Tulip_SetResume(X, (void *) &Tulip_RestoreNP);




#endif
