#ifdef __GNUG__
#  pragma interface
#endif
#ifndef _Context_h_
#define _Context_h_

#include <Tulip_Config.h>

typedef unsigned long Tulip_U64;

typedef void (*Tulip_voidfuncptr)();

typedef struct {
  Tulip_voidfuncptr func;
  void *arg;
} Tulip_Closure;

class Tulip_NonPremptiveContext {
 public:
  /*
   * This contains the closure for the continuation used
   * to resurect this context. This must be the first field
   * in this struct and in the struct for all resumption routines
   * (i.e., the premptive case as well)
   */

  Tulip_voidfuncptr resume;

  Tulip_U64 R1;	/* SP */
  Tulip_U64 R2;	/* PC */
  Tulip_U64 R3;	/* out0 */
  Tulip_U64 R4;	/* out1 */
  void print_state();

};
// void Tulip_spinlock(unsigned long *, unsigned long);
// long Tulip_spintrylock(unsigned long *, unsigned long);

extern "C" {
void Tulip_spinlock(unsigned long *lock,unsigned long);
unsigned long Tulip_spintrylock(unsigned long *lock,unsigned long);
void 
Tulip_SaveNP(Tulip_NonPremptiveContext *resurrect,
	     Tulip_voidfuncptr *hook_func,
	     void* hook_data,
	     Tulip_NonPremptiveContext *saveto);

void 
Tulip_RestoreNP(Tulip_NonPremptiveContext *resurrect,
	     Tulip_voidfuncptr *hook_func,
	     void* hook_data);
}

#define Tulip_SwitchTo(to, hookf, hookd, from) \
  { Tulip_SaveNP(to, hookf, hookd, from); }


/*
 * Initialize the context to use the specified
 * stack address, and initialize such that execution
 * starts at 'routine' with initial argument 'argument'
 */
extern void
Tulip_InitNP(Tulip_NonPremptiveContext *context,
	     Tulip_U64 *stack,
	     void *routine,
	     void *argument);

extern void
Tulip_SetResume(Tulip_NonPremptiveContext *context,
		void *routine);

#define Tulip_SetNonPremptiveResume(X) Tulip_SetResume(X, (void *) &Tulip_RestoreNP);

#define SpinLockFree 0L
typedef unsigned long SpinLockType;
typedef volatile unsigned long VolatileSpinLockType;

void Tulip_spinlock(unsigned long *lock,unsigned long);
unsigned long Tulip_spintrylock(unsigned long *lock,unsigned long);
#endif


