#ifdef __GNUG__
#  pragma implementation
#endif

#include "Context.h"
#include <stdio.h>
#include <iostream.h>
#include <sys/asm_linkage.h>

extern "C" {
void Tulip_trampoline();
}
/*
 * A stack frame looks like:
 *
 * %fp->|				|
 *	|-------------------------------|
 *	|  Locals, temps, saved floats	|
 *	|-------------------------------|
 *	|  outgoing parameters past 6	|
 *	|-------------------------------|-\
 *	|  6 words for callee to dump	| |
 *	|  register arguments		| |
 *	|-------------------------------|  > minimum stack frame
 *	|  One word struct-ret address	| |
 *	|-------------------------------| |
 *	|  16 words to save IN and	| |
 * %sp->|  LOCAL register on overflow	| |
 *	|-------------------------------|-/
 */

void
Tulip_InitNP(Tulip_NonPremptiveContext *context,
	     Tulip_U64 *stack,
	     void *routine,
	     void *argument)
{
  
  /*
    *	To build a thread, we return to the first address of startOff,
    *	which will use the current FP & SP to build a local context,
    *	and then call the user main.
    *
    *	startOff needs to have a valid frame. The return address for this
    *	frame is NULL, since we never exit procedure startOff.
    */

  int i;
  Tulip_U64 *stacktemp,*In, *FP0, *FP1, *SP0, *SP1;
  char *pch = (char *)&stack[0];
  *pch++ = 'T'; *pch = '\0';
  stack[-1] = 0;
  stack[-2] = (Tulip_U64) &stack[-1];
  stack[-3] = (Tulip_U64) &stack[-2];


  stacktemp = stack-1;
  FP0 = stacktemp;
  stacktemp -= (MINFRAME/sizeof(Tulip_U64));
  SP0 = stacktemp;

  In = &SP0[8];
  In[6] = 0;
  In[7] = 0;

  // second frame
  FP1 = stacktemp;
  stacktemp -= (MINFRAME/sizeof(Tulip_U64));
  SP1 = stacktemp;
  In = &SP1[8];
  In[6] = (Tulip_U64)0;
  In[6] = 0;
  
  context -> R1 = (Tulip_U64) SP1;        /* sp */
  context -> R2 = (Tulip_U64) Tulip_trampoline-8;  /* pc */
  context -> R3 = (Tulip_U64) argument;          /* out0 */
  context -> R4 = (Tulip_U64) routine;           /* out1 */

  Tulip_SetNonPremptiveResume(context);

}
void 
Tulip_NonPremptiveContext::print_state()
{
  cerr << "\n" << flush;
}

extern void
Tulip_SetResume(Tulip_NonPremptiveContext *context,
		void * routine)
{
  context -> resume = (Tulip_voidfuncptr)routine;
}


