#ifndef _Tulip_Helpers_h_
#define _Tulip_Helpers_h_
#include <unistd.h>
#include <Tulip_Config.h>
#include <Tulip_Mutex.h>
#include <iostream.h>
#include <thread.h>
#include <Context.h>
#include "CUT_BinHeap.h"
#include <function.h>
#include <list>
#include <vector>

template <class T>
class LockedPriorityQueue  : public CUT_BinHeap<int, T, Higher>
{
 public:
  Tulip_Mutex qLock;
  void addUnLocked(T object) {
    insert(object->getPriority(),object);
  }
  void add(T object) {
    qLock.lock();
    insert(object->getPriority(),object);
    qLock.unlock();
  }

  T remove(){
    T object = NULL;
    qLock.lock();
    if (!empty()) {
      object = top();
      deq();
    }
    qLock.unlock();
    return object;
  }
  T removeUnLocked(){
    T object = NULL;
    if (!empty()) {
      object = top();
      deq();
    }
    return object;
  }
};


/*
template <class T>
class LockedPriorityQueue: public priority_queue<vector<T>,Higher<T> > {
 public:
  Tulip_Mutex qLock;
  void addUnLocked(T object) {
    push(object);
  }
  void add(T object) {
    qLock.lock();
    push(object);
    qLock.unlock();
  }

  T remove(){
    T object = NULL;
    qLock.lock();
    if (!empty()) {
      object = top();
      pop();
    }
    qLock.unlock();
    return object;
  }
  T removeUnLocked(){
    T object = NULL;
    if (!empty()) {
      object = top();
      pop();
    }
    return object;
  }
};
*/
extern "C" {
#include <sched.h>
#include <unistd.h>
}

inline int whoAmI() {
  return ((int) thr_self());
}
typedef greater_equal<int> Higher;

int Tulip_Fork(void *fArgs, voidPtrFuncP func);
class Tulip_Context;
typedef Tulip_Context * Tulip_ContextPtr;

class Tulip_PrivateContext {
  Tulip_ContextPtr contextP;
public:
  Tulip_PrivateContext();
  Tulip_ContextPtr operator->();
  void operator=(const Tulip_ContextPtr& context);
};


class Tulip_ForkArgs {
public:
  int id;
  int numcpus;
};

#endif
