/*=======================================================================
 $Log: Tulip_Context.C,v $
 Revision 1.1  1998/06/23 00:18:07  ejohnson
 moving directory hpcxx_rts up

 Revision 2.1  1998/06/15 13:11:26  suvas
 *** empty log message ***

 Revision 2.0  1998/05/22 16:57:31  suvas
 New Release (using main as a thread)

 Revision 1.5  1998/05/21 22:18:53  suvas
 Changed thread model to make "main" a thread. Fixed cpu affinity.

 Revision 1.2  1998/05/12 23:15:55  suvas
 This may look like C code, but it is really C++

 Copyright (c) 1998 Los Alamos National Laboratory
 Copyright (c) 1997 University of Colorado, Boulder

 Suvas Vajracharya and Dirk Grunwald

 $Id: Tulip_Context.C,v 1.1 1998/06/23 00:18:07 ejohnson Exp $
 $Date: 1998/06/23 00:18:07 $
 $Author: ejohnson $
 ========================================================================*/

#ifdef __GNUG__
#  pragma implementation
#endif

#include <iostream.h>
#include <stdio.h>
#include <assert.h>
#include <Tulip_iolock.h>
#include <unistd.h>
#include "Tulip_Context.h"
#include "Tulip_AcquireByUpcall.h"
#include "Tulip_Helpers.h"
#include "Tulip_SpinFetchAndOp.h"

#ifdef MIRACLE
#include "Tulip_Iterate.h"
#include "Tulip_Iterate2.h"
#endif

Tulip_PrivateContext Tulip_Context::loopback_pointer;
int Tulip_Context::waveFront=0;
int Tulip_Context::debug=0;
int Tulip_Context::nextID = 0;
Tulip_Mutex Tulip_Context::nextIDLock;
int Tulip_Context::numcpus = 0;
int Tulip_Context::GivingUpCounter=0;
int Tulip_Context::GivingUpGeneration=0;
Tulip_Mutex Tulip_Context::GivingUpLock;

Tulip_FetchAndOp Tulip_Context::GlobalThreadCounter(0);
Tulip_FetchAndOp Tulip_Context::GlobalEventsCounter(0);

Tulip_ContextPtr *Tulip_Context::Contexts = NULL;
Tulip_SpinBarrier Tulip_Context::barrier(1);

MainThread mainThread;
Tulip_Context systemContext;

RETURN_TYPE parastart(void *args)
{
  Tulip_ForkArgs *fArgs = (Tulip_ForkArgs *)args;   
  Tulip_Context *myContext = new Tulip_Context;
  Tulip_Context::Contexts[fArgs->id] = myContext;
  myContext->loopback_pointer = myContext;
  Tulip_Context::SetMe(fArgs->id);
  BEGIN_DEBUG_PRINT 
    cerr << "Worker " << fArgs->id << " is born\n" << flush;
  END_DEBUG_PRINT;
  delete fArgs;
  myContext->barrier.rendezvous();
  myContext->schedule_private();
#ifdef USE_PTHREADS
  return (RETURN_TYPE)NULL;
#else
  return;
#endif
}

void Tulip_Context::Initialize()
{
#ifdef __sgi
  usconfig(CONF_INITUSERS,Tulip_MAX_CPUS);
#endif
  Contexts = new Tulip_ContextPtr[Tulip_MAX_CPUS];
  for (int i =0; i < Tulip_MAX_CPUS; i++)
    Contexts[i] = NULL;
  assert(Contexts!=NULL);
  Contexts[0] = &systemContext;
  systemContext.me = 0;
  numcpus = 1;
  loopback_pointer = &systemContext;
  barrier.height(1);
  mainThread.affinity(0);
  mainThread.setPriority(9999);
  systemContext.setCurrentThread(&mainThread);  
  Tulip_U64 *stack_malloc_at = new Tulip_U64[SYSTEM_STACK_SIZE];
  Tulip_U64 *stack_end_at = &stack_malloc_at[SYSTEM_STACK_SIZE-1];  
  Tulip_InitNP(&syscontext, stack_end_at,&(Tulip_Context::_call_schedule), this);
}
Tulip_Context::Tulip_Context()
{
  pid = whoAmI();
  loopback_pointer = this;
  Tulip_SetNonPremptiveResume(&syscontext);
}
#ifdef __sgi
extern "C" {
#include <sched.h>
#include <unistd.h>
}
#endif
void Tulip_Context::_call_schedule(Tulip_ContextPtr sysContext)
{
  systemContext.schedule_private();
}
void Tulip_Context::addNCpus(int _numcpus)
{
  Tulip_ForkArgs *fArgs;
  if (_numcpus < 1) return;
  if ((numcpus+_numcpus) > Tulip_MAX_CPUS) {
    cerr << "Maximum cpus allowed is "<< Tulip_MAX_CPUS << "\n";
    cerr << "Spawning " << Tulip_MAX_CPUS << " cpus only.\n";
    _numcpus = Tulip_MAX_CPUS;
  }
  numcpus+= _numcpus;
  barrier.height(_numcpus+1);

  for (int i=0; i < _numcpus; i++) {
    fArgs = new Tulip_ForkArgs;
    nextIDLock.lock();
    fArgs->id = ++nextID;
    nextIDLock.unlock();
    int id = Tulip_Fork((void *)fArgs,parastart);
    if (id == -1) {
	perror("Tulip_Context.C: ");
	cerr << lock << "Error in creating a new worker. \n" << unlock;
	exit(-1);
    }
  }
  barrier.rendezvous();
}

void
Tulip_Context::suspend(Tulip_Thread *thread)
{
  BEGIN_DEBUG_PRINT 
    cerr << "Suspending Thread: " << thread -> getName() <<"\n";
  END_DEBUG_PRINT
  thread->state = SUSPENDED;
  thread_bag_lock.lock();
  thread->suspendedAt =  thread_bag.insert(thread_bag.begin(),thread);
  thread_bag_lock.unlock();
}
void
Tulip_Context::resume(list<Tulip_ThreadPtr>::iterator position)
{
  thread_bag_lock.lock();
  thread_bag.erase(position);
  thread_bag_lock.unlock();
  BEGIN_DEBUG_PRINT 
    cerr << "Resuming Thread: " << (*position) -> getName() <<"\n";
  END_DEBUG_PRINT

  add(*position);
}

Tulip_Thread *
Tulip_Context::removeThread_internal()
{
  Tulip_Thread *thread = NULL;

  thread = (Tulip_ThreadPtr) threadQueue.remove();

  if (thread) {
    if (thread->state != ALIVE)
      return removeThread_internal();
    else {
      BEGIN_DEBUG_PRINT 
      cerr <<"    Removed " << thread -> getName() << " from processor " << me <<"\n";
      END_DEBUG_PRINT;
      GlobalThreadCounter-=1;
      return thread;
    }
  }
  return NULL;

}


int
Tulip_Context::size_internal()
{
  return threadQueue.size();
}

Tulip_Thread *
Tulip_Context::removeThread()
{

  Tulip_Thread *thr=NULL;
  if (thr = Contexts[MYID]->affinityThreadRemove()) {
    BEGIN_DEBUG_PRINT 
      cerr << "Processor " << MYID << " found affinity thread " << thr->getName() <<"\n";
    END_DEBUG_PRINT;
    return thr;
  }
  int ask = (MYID)%numcpus;
  for (int i=0; i<numcpus; i++) {
     thr = NULL;
     BEGIN_DEBUG_PRINT 
       cerr << "Processor " << MYID << " looking for work in processor " << ask <<"\n";
     END_DEBUG_PRINT;
     if (thr = Contexts[ask]-> removeThread_internal()) {
       BEGIN_DEBUG_PRINT 
	 cerr << "Processor " << MYID << " found work in processor " << ask <<"\n";
       END_DEBUG_PRINT;
       return thr;
     }
     ask = (++ask)%numcpus;
   }
   return thr;
}


#include "Tulip_Mutex.h"
#include "Tulip_CSem.h"
Tulip_Mutex Mutex;
int val=0;
void Tulip_Context::schedule_private()
{
#ifdef MIRACLE
  Tulip_IteratePtr iter;
#endif
  BEGIN_DEBUG_PRINT
  cerr << "In private\n";
  END_DEBUG_PRINT
  while (1) {
    current_thread = NULL;
    if (GlobalThreadCounter.value() > 0) {
      if (current_thread = removeThread()) {
	BEGIN_DEBUG_PRINT 
	  cerr << "Processor " << me << " pulled " << current_thread->getName() << " from queue\n";
	END_DEBUG_PRINT;
	Tulip_SwitchTo(&(current_thread -> context), NULL, NULL, &syscontext);
	continue;
      }
    }
#ifdef MIRACLE    
    if (iter = removeNonRegistered()) {
      iter->registerMe();
    }
    if (iter = removeIterate()) {
      iter->execute();
      continue;
    }
#endif
    BEGIN_DEBUG_PRINT 
      cerr << " global events counter " << GivingUpCounter << ", " << GlobalEventsCounter.value() << "\n";
    END_DEBUG_PRINT;
    if ( GlobalEventsCounter.value() > 0 ) continue;
    BEGIN_DEBUG_PRINT
      cerr << "Checking if I should quit,numcpus = " << numcpus << "\n" ;
    END_DEBUG_PRINT
    GivingUpLock.lock();
    GivingUpCounter++;
    
    if ( GivingUpCounter == numcpus
	 && GlobalEventsCounter.value() == 0)
      {
	BEGIN_DEBUG_PRINT
	  cerr << "Quiting\n" ;
	END_DEBUG_PRINT
	GivingUpGeneration++;
	GivingUpCounter = 0;
	GivingUpLock.unlock();
	return;
      }
    else {
      int generation = GivingUpGeneration;
      VolatileInt *genp = &GivingUpGeneration;
      GivingUpLock.unlock();
      BEGIN_DEBUG_PRINT
	cerr << "Waiting for the other\n" ;
      END_DEBUG_PRINT
      while( generation == *genp
	     && (GlobalEventsCounter.value()) == 0);
      GivingUpLock.lock();
      if ( *genp != generation) {
	BEGIN_DEBUG_PRINT
	  cerr << "Giving up\n" ;
	END_DEBUG_PRINT
	GivingUpLock.unlock();
	return;
      }
      else {
	BEGIN_DEBUG_PRINT
	  cerr << "Waking up\n" ;
	END_DEBUG_PRINT
	GivingUpCounter--;
	GivingUpLock.unlock();
      }
    }
  }
}


void 
Tulip_Context::yield_internal(Tulip_Thread* th)
{
  //
  // Uniprocessor case - no data races - first, place this thread on the
  // thread list, then schedule the next available thread..
  //
  Tulip_Thread *them = NULL;

  // Get a thread to schedule..

  if (!(them = removeThread_internal())) {
    them = removeThread();
  }
  if (them) {
    current_thread = them;
    //
    // Switch to them with a call-back to add th to the scheduler..
    //
    //    cerr << "From: " << th->getName() << "\n";
    //    th->context.print_state();
    //    cerr << "To: " << them->getName() << "\n";
    //    them->context.print_state();
    BEGIN_DEBUG_PRINT 
      cerr << "About to yield from " << th -> getName() << " to " << them ->getName();
      cerr << " with " << threadQueue.size() << " jobs in the queue \n";
    END_DEBUG_PRINT;
    Tulip_SwitchTo(&(them -> context),
		   (Tulip_voidfuncptr*) Tulip_Context::add, th,
		   &(th -> context));
    BEGIN_DEBUG_PRINT 
      cerr << "Came back from yield " << th -> getName() << " to " << them ->getName();
      cerr << " with " << threadQueue.size() << " jobs in the queue \n";
    END_DEBUG_PRINT;
  }
}


void
Tulip_Context::kill_thread(Tulip_Thread* thread)
{
  if ( Tulip_Context::debug ) {
    BEGIN_DEBUG_PRINT 
      cerr << "Kill thread " << thread -> getName() << "\n";
    END_DEBUG_PRINT;
  }
  delete thread;
}


void
Tulip_Context::reservation_helper(Tulip_Context* sys)
{
  Tulip_Thread *thread = sys -> upcall_blocking_thread;
  assert(sys->upcall_blocking_thread->state == ALIVE);
  if ( sys -> upcall_resource -> acquire_by_upcall(thread) ) {
    //
    // That thread blocked & has been scheduled on some internal queue
    //
    if ( Tulip_Context::debug ) {
      BEGIN_DEBUG_PRINT 
	cerr << "I think " << thread -> getName() << " has blocked ..\n";
      END_DEBUG_PRINT;
    }

  } else {
    //
    // The thread didn't block - it's eligible for scheduling, although
    // it won't execute immediately..
    //
    sys -> add(thread);

    if ( Tulip_Context::debug ) {
      BEGIN_DEBUG_PRINT 
	cerr << "I think " << thread -> getName() << " did not block, scheduling..\n";
      END_DEBUG_PRINT;
    }
  }
}

void
Tulip_Context::reservation_upcall_internal(Tulip_AcquireByUpcall *resource)
{
  //
  // We may need to suspend the current thread, so
  // we need to locate a third party to avoid
  // race conditions.
  //
  Tulip_Thread *them = NULL;
  //  cerr << lock << "current thread " << current_thread << "\n" << flush << unlock;
  assert(current_thread != NULL);
  upcall_blocking_thread = current_thread;

  upcall_resource = resource;

  Tulip_NonPremptiveContext* context;

  // Get a thread to schedule..
  them = removeThread();
  if (them) {
    if ( Tulip_Context::debug ) {
      BEGIN_DEBUG_PRINT 
	cerr << "reservation_upcall from " << upcall_blocking_thread ->getName();
        cerr << " to " << them ->getName() << "\n";
      END_DEBUG_PRINT;
    }

    current_thread = them;
    context = &(current_thread -> context);
  }
  else {
    current_thread = NULL;
    context = &syscontext;
  }

  //
  // Now call handler to do the
  // dirty work..
  //
  Tulip_SwitchTo(context,
		 (Tulip_voidfuncptr*) Tulip_Context::reservation_helper,this,
		 &(upcall_blocking_thread->context));

  
}

void Tulip_Init()
{
  systemContext.Initialize();
}

__Tulip_Init_hack_t __Tulip_Init_hack_x; 
char __Tulip_Init_hack = 42; 
