/*=======================================================================
 $Log: Tulip_iolock.C,v $
 Revision 1.1  1998/06/23 00:18:20  ejohnson
 moving directory hpcxx_rts up

 Revision 2.0  1998/05/22 16:57:35  suvas
 New Release (using main as a thread)

 Revision 1.2  1998/05/12 23:16:02  suvas
 This may look like C code, but it is really C++

 Copyright (c) 1998 Los Alamos National Laboratory
 Copyright (c) 1997 University of Colorado, Boulder

 Suvas Vajracharya and Dirk Grunwald

 $Id: Tulip_iolock.C,v 1.1 1998/06/23 00:18:20 ejohnson Exp $
 $Date: 1998/06/23 00:18:20 $
 $Author: ejohnson $
 ========================================================================*/
#ifdef __GNUG__
#  pragma implementation
#endif

#include "Tulip_Mutex.h"
#include <iostream.h>
#include "Tulip_iolock.h"
#include "assert.h"



//
// Crude I/O locking. We keep a list of ios pointers, associate a
// semphore with each one and do the appropriate things.
//

class IosLock;

class IosLock {
public:
  Tulip_Mutex lock;
  ios *ptr;
  IosLock *next;

  IosLock(ios& p);
};

inline IosLock::IosLock(ios& p)
{
  ptr = &p;
  next = 0;
}

static IosLock *root = 0;

static IosLock *
lookup(ios& s) 
{
  ios *p = &s;
  IosLock* here = root;
  IosLock* follow = 0;

  while ( here ) {
    if ( here -> ptr == p ) {
      //
      // move to front, return pointer
      //
      if ( follow != 0 ) {
	//
	// unlink here
	//
	IosLock *tmp = here -> next;
	follow -> next = tmp;
	//
	// link here
	//
	here -> next = root;
	root = here;
      }
      return( here );
    }
    follow = here;
    here = here -> next;
  }
  IosLock* n = new IosLock(s);
  n -> next = root;
  root = n;
  return( n );
}

#if GNU_C_BROKEN
ios& lock(ios& os) 
{
  IosLock* p = lookup(os);
  assert(p != 0);
  p -> lock.lock();
  return( os );
}


ios& unlock(ios& os)
{
  IosLock* p = lookup(os);
  assert(p != 0);
  p -> lock.unlock();
  return( os );
}
#endif


ostream& lock(ostream& s) 
{
  ios& os = s;
  IosLock* p = lookup(os);
  assert(p != 0);
  p -> lock.lock();
  return( s );
}


ostream& unlock(ostream& s)
{
  ios& os = s;
  IosLock* p = lookup(os);
  assert(p != 0);
  p -> lock.unlock();
  return( s );
}


istream& lock(istream& s) 
{
  ios& os = s;
  IosLock* p = lookup(os);
  assert(p != 0);
  p -> lock.lock();
  return( s );
}


istream& unlock(istream& s)
{
  ios& os = s;
  IosLock* p = lookup(os);
  assert(p != 0);
  p -> lock.unlock();
  return( s );
}

