#include <stdio.h>
#include <limits.h>
#include <math.h>
#include <hpcxx_rts.h>

// include template definitions
#include <hpcxx_invoke.cc>
#include <HPCxx_GlobalPtr.cc>
#include <HPCxx_GlobalRefImpl.cc>

class BarrierTest : public HPCxx_Thread {
  int myId;
  HPCxx_Collective<int>& b;
  int myKey;
  
public:
  
  BarrierTest(int myId_, HPCxx_Collective<int>& b_) : myId(myId_),
  b(b_), myKey(0) { }
  
  virtual ~BarrierTest() { }
  
  void run()
  {
    double x = 1.0;
    myKey = b.acquireKey();

    for(int k = 0; k < 4; k++) {
      for(int i = 0; i < 1000000; i++)
	x = sin(x);

      printf("%d: waiting at barrier\n", myId);
      b.barrier(myKey);
      printf("%d: through barrier\n", myId);
    }
  }

};


int main(int argc, char** argv)
{
  int nthreads = 0;
  
  if ( argc < 2 ) {
    cerr << "Enter number of BarrierTest threads: ";
    cin >> nthreads;
  
  } else {
    nthreads = atoi(argv[1]);
  }
  
  HPCxx_Group* group;
  hpcxx_init(argc, argv, group);
  group->setNumThreads(nthreads + 1);
  
  HPCxx_Collective<int> b(group);
  
  BarrierTest** barrierTest = new BarrierTest* [nthreads];
  HPCxx_SyncQ<int>* isync = new HPCxx_SyncQ<int> [nthreads];
  
  int myKey = b.acquireKey();
  
  for(int i = 0; i < nthreads; i++)
    barrierTest[i] = new BarrierTest(i, b);
  
  for(int j = 0; j < nthreads; j++)
    barrierTest[j]->start();

  int k = 0;
  while(k++ < 4)
  {
    char c;
    printf("main at barrier. hit a key to continue\n");
    fscanf(stdin, "%c", &c);
    b.barrier(myKey);
  }

  for(int x = 0; x < nthreads; x++)
    barrierTest[x]->join();

  printf("DONE!\n");
  hpcxx_exit(group);
  return 0;
}

