//////////////////////////////////////////////

#include <stdio.h>

// Liz and Pete's invoke/minvoke stuff for heterogeneous machines

// simple class (that the user declared)
class foo {
public:
  void bar(int a) {
    printf("foo::bar %d\n",a);
  };
  void bar2(int a) {
    printf("foo::bar2 %d\n",a);
  };
};

void foobar(char *x) {
  int i;
  for (i=0; i<8; i++) {
    printf("%x ",x[i]);
  }
  printf("\n");
}

////////////////////////////////////////////////////////////////////////
// Simple hack to get standalone compile
template<class C>
class GlobalPtr {
public:
  C* localpointer;
  int context;
  int isLocal() { return 1; }  // always true for test
//========================================
//  hacked this so that we could pretend it is remote when called from
//  unpack_and_invoke
  int isLocal(int i) { return context==i; }   // for the test, 1 is local, 0 isremote

  GlobalPtr() : localpointer(0), context(1) {}
  GlobalPtr(C* p) : localpointer(p), context(0) {}
  GlobalPtr(C* p, int c) : localpointer(p), context(c) {}
//=====================================
};

////////////////////////////////////////////////////////////////////////

// The very scary implementation...

// The root of all evil... the base table-lookup class
class BaseLookup {
  int check_id;
public:
  static int next_id;
  BaseLookup() : check_id(next_id++) {}
  int get_id() { return check_id; }      // what is the unique id for this
  //===============================================
  virtual void unpack_and_invoke(char* buffer) = 0;
  //===============================================
};

int BaseLookup::next_id = 0;    // next thing...

// This should be a hash thing, but a quickie table array is ok...
BaseLookup *LUT[100];

//=====================================
BaseLookup* get_table_entry(int id) {
  printf("id = %d\n",id);
  return LUT[id];
}
//=====================================

template<class C, class A1>
class LookupTableEntry : public BaseLookup {
    void(C::*mfp)(A1);  // declare the mfp inside this entry
  public:

  typedef void(C::*P)(A1);
//=====================================
  struct mybuffer_t {
    int id;
    GlobalPtr<C> gp;
    A1 a1;
  };

  void unpack_and_invoke(char* buffer) {
    mybuffer_t* B=(mybuffer_t*) buffer;
    if ((B->gp).isLocal(1))
      (((B->gp).localpointer)->*mfp)(B->a1);
    else
      printf ("Error: pointer is not local to this context\n");
  }

  static char* pack(int C_RMF_id,GlobalPtr<C> gp,A1 a1) {
    mybuffer_t* B=new mybuffer_t;
    B->id=C_RMF_id;
    B->gp=gp;
    B->a1=a1;
    return (char*) B;
  }
//=====================================
  LookupTableEntry(void(C::*init_mfp)(A1)) : mfp(init_mfp), BaseLookup() {}

  // simple lookup
  P getmfp() {
    return mfp;
  }
};

// lookup particular mfp for a class
template<class C, class A1>
class lookupMFP {
public:
  typedef void(C::*P)(A1);

  static P lookup(int myC_RMF_id) {
    LookupTableEntry<C,A1> *LE = (LookupTableEntry<C,A1> *) LUT[myC_RMF_id];
    return LE->getmfp();
  }
};

// register it... (put it into the array, and return the id_key)
template<class C, class A1>
int register_MF( void (C::*mfp)(A1) ) {
  LookupTableEntry<C,A1> *LE = new LookupTableEntry<C,A1>(mfp);
  int id_x = LE->get_id();
  LUT[id_x] = LE;

  return id_x;
}

// applyer
template<class C, class A1>
void local_apply(C *mine, int C_RMF_id, A1 a1) {
  // mine is a pointer to a real local object

  typedef void(C::*P)(A1);
  P my_mfp;

  //  void (C::*mfp)(A1);   // declare a mfp to the specific object

  // "mine" and "&a1" are DUMMYS in the next statement, for templating
  foobar((char*) &my_mfp);

  my_mfp = lookupMFP<C,A1>::lookup(C_RMF_id);

  foobar((char*) &my_mfp);

  (mine->*my_mfp)(a1);  // apply it!  Cool.

}

//=======================================
template<class C, class A1>
void remote_apply(GlobalPtr<C> gp, int C_RMF_id, A1 a1) {
  char* b=LookupTableEntry<C,A1>::pack(C_RMF_id,gp,a1);
  send(gp.context,b);
  delete [] b;
}

template <class T>
void unpack(char* buffer,T& result) {
  result=*((T*) buffer);
}

// hack to test -- this will actually become the handler on the receiving side
void send(int, char* buffer) {
  int id;
  unpack(buffer,id);
  BaseLookup* entry=get_table_entry(id);
  entry->unpack_and_invoke(buffer);
}


//=======================================

template<class C, class A1>
void minvoke(GlobalPtr<C> RO, int C_RMF_id, A1 a1)
{
  // Simple, if it is local, just do it here...
  if (RO.isLocal()) {
    // things are easy, it is here
    C *mine;

    // should really be a cast, but for simple test we get it by hand
    mine = RO.localpointer;  // cast to get the real localpointer to the object
    local_apply(mine,C_RMF_id,a1);
  }
  else {
    // You must send a message to the remote context to do fancy stuff
    //    msgbuff B = packbuffer(RO, C_RMF_id, a1);
    //    send(RO.context, msgbuff);
    remote_apply(RO,C_RMF_id,a1);
  }
}

int main() {
  foo X,*x1;
  int mpid;

  x1 = &X;

  // "normal" way to do this:
  void (foo::*mfp)(int);
  void (foo::*mfp2)(int);

  mfp = &foo::bar;
  mfp2 = &foo::bar2;

  foobar( (char*) &mfp);
  foobar( (char*) &mfp2);

  X.bar(5);
  (x1->*mfp)(3);
  (x1->*mfp2)(6);


  // registration and remote way...
  mpid = register_MF(&foo::bar);
  printf("I just registered, and my id = %d\n",mpid);

  mpid = register_MF(&foo::bar2);
  printf("I just registered, and my id = %d\n",mpid);

//==================================================
  GlobalPtr<foo> F(x1,1);   // make a global pointer to the "local" object
  minvoke(F,mpid,99);

  GlobalPtr<foo> G(x1,0);   // make a  global pointer to the "remote" object
  minvoke(G,mpid,100);
//==================================================
}


