// Copyright (C) 1988 University of Illinois, Urbana, Illinois
// Copyright (C) 1989 University of Colorado, Boulder, Colorado
// Copyright (C) 1990 University of Colorado, Boulder, Colorado
// Copyright (C) 1994 University of Colorado, Boulder, Colorado
//
// written by Dirk Grunwald and Suvas Vajracharya
// {grunwald,suvas}@foobar.colorado.edu,
// adapted by Craig Rasmussen, rasmussn@lanl.gov (March, 1998)
//

#include <iostream.h>
#include <hpcxx_rts.h>
#include <HPCxx_Timers.h>

HPCxx_Barrier* barrier;

class GlobalData {
  public:  double *data;
  public:  int ghg;
};

GlobalData Data;

#define data(x,y) (*((Data.data)+(x)*Data.ghg+(y)))


class RedBlackSOR : public HPCxx_Thread
{
  private:  int chunk;
  private:  int iterations;
  private:  int low;
  private:  int high;
  private:  int globalhigh;
  private:  int globallow;
  private:  int stride;
  private:  int NumCpus;

  public:   RedBlackSOR( int lo_, int hi_, int glow_, int ghigh_,
		         int str_, int iter_, int chnk_, int num_ );
  public:   virtual void run(void);
};

//
// constructor for RedBlackSOR.
//
RedBlackSOR::RedBlackSOR( int _lo, int _hi, int _glo, int _ghi,
			  int _st, int _it, int _chnk, int _num )
            : HPCxx_Thread("RedBlackSOR")
{
  low = _lo;
  high = _hi;
  globallow = _glo;
  globalhigh = _ghi;
  stride = _st;
  iterations = _it;
  NumCpus = _num;
  chunk = _chnk;
}


void RedBlackSOR::run(void)
{
  int i, j;

  int key = barrier->getKey();

  // Computation assuming omega = 0
    
  cout << "    RedBlackSOR::run, beginning iterations\n" << flush;

  for (int c = 0; c < iterations; c++) { 
    // cout << lock << "it,me" << iterations<< "," << MYID << "\n";
    // cout << flush << unlock;

    ///////////////////////////////////////////////////////////////////////
    // Start computing the RED elements
    for (i = low; i <= high; i += 2) {
      for (j = 0; j <=  globalhigh; j += 2) {

	// Four Point Stencil Computation: 
	//  north, south, east, west are the values from neighbors 

        data(i,j) = (data(i-1,j) + data(i,j+1)
				 + data(i+1,j) + data(i,j-1)) / 4.0;
	data(i+1,j+1) = (data(i,j+1) + data(i+1,j+2)
				     + data(i+2,j+1) + data(i+1,j)) / 4.0;

      }
    }

    barrier->barrier(key);

    ///////////////////////////////////////////////////////////////////////
    // Start computing the BLACK elements 
    for (i = low; i <= high; i += 2) {
      for (j = 1;  j <= globalhigh; j += 2) {
	data(i,j) = (data(i-1,j) + data(i,j+1)
				 + data(i+1,j) + data(i,j-1)) / 4.0;
	data(i+1,j-1) = (data(i,j-1) + data(i+1,j)
				     + data(i+2,j-1) + data(i+1,j-2)) / 4.0;
      }
    }

    barrier->barrier(key);
  }
}


main(int argc, char* argv[])
{
  int i,j;

  HPCxx_Group* g;
  hpcxx_init(argc, argv, g);

  int cpus = (argc > 1) ? atoi(argv[1]) : 1;
  int debug = (argc > 2) ? atoi(argv[2]) : 0;
  int low = (argc > 3) ? atoi(argv[3]) : 0;
  Data.ghg = (argc > 4) ? atoi(argv[4]) : 1024;
  int stride = (argc > 5) ? atoi(argv[5]) : 1;
  int loop = (argc > 6) ? atoi(argv[6]) : 10;

  Data.data = new (double [(Data.ghg+4)*(Data.ghg+4)]);
  for (i = 0; i < (Data.ghg+4)*(Data.ghg+4); i++) {
    Data.data[i] = i%3;
  }

  // create barrier

  barrier = new HPCxx_Barrier(cpus);

  // create worker threads

  int high = Data.ghg;
  int grain = (high - low)/cpus;
  if (grain%2) grain++;
  int lo = low;

  RedBlackSOR** worker = new RedBlackSOR*[cpus];
  for (i = 0; i < cpus; i++, lo += grain) {
    worker[i] =
      new RedBlackSOR(lo, lo+grain-2, low, high-1, stride, loop, i, cpus);
  }
  cout << endl;

  // run worker tasks

  double start = HPCxx_StartTimer();
  for (i = 0; i < cpus; i++) {
    worker[i]->start();
  }

  // wait for workers to finish

  for (i = 0; i < cpus; i++) {
    worker[i]->join();
  }
  double eltime = HPCxx_StopTimer(start);

  cout << "    " << Data.ghg << " "<< cpus << " " << eltime << endl << flush;
  if (debug) {
    cout << "\n";
    for (i=0; i< Data.ghg; i++) {
      for(j=0; j<Data.ghg; j++) {
	cout << data(i,j) << " ";
      }
      cout << "\n";
    }
  }
  cout << endl << "Successfully completed RedBlackSOR" << endl;

  return hpcxx_exit(g);
}
