#include <stdlib.h>
#include <unistd.h>
#include <stream.h>
#include <signal.h>
#include <sys/time.h>

#include <nexus.h>
#include <hpcxx_rts.h>


#define ElapseTime(Start, Stop) \
    (  (  ( ((double)(Stop.tv_sec)) - ((double)(Start.tv_sec)) ) \
        + (((double)(Stop.tv_usec))/1000000.0) ) \
     - (((double)(Start.tv_usec))/1000000.0) )

void* _nx_st_rsh_info(void)
{
  return(NULL);
}

#include <stdio.h>
#include <math.h>
#include <strings.h>
#include <math.h>

/**************************************************************
  This is a sample program that is intended for multicontext
  sorting.  It does the binsort component as a "bitonic" binsort.
  this is not very efficient on a shared memory multiprocessor,
  but it does beat the most obvious sequiential bin sort.
*************************************************************/

typedef int hpcxx_buffer;

/***********************************************************
  The data is a vector of type P.  it is moved around in
  "Chunks" defined below.
************************************************************/

template <class P>
class Chunk {
public:
  P* data;
  int length;
  friend void hpcxx_pack(hpcxx_buffer* , Chunk<P>* , int);
};

void pack(hpcxx_buffer* buffer, int i, int l) { }

template<class P>
void hpcxx_pack(hpcxx_buffer* buffer, Chunk<P>* d, int l)
{
  for(int i = 0; i < l; i++){
    pack(buffer, d[i].length, 1);
    pack(buffer, d[i].data, d[i].length);
  }
}

/*
template<class P>
void hpcxx_unpack(hpcxx_buffer* buffer, Chunk<P>* d, int l)
{
   for(int i = 0; i < l; i++){
       unpack(buffer, &d[i].length, 1);
       unpack(buffer, d[i].data, d[i].length);
       }
}
*/

template <class P>
class test {
  int key;
  int flg; // if 0 return reverse of condition

public:
  test(int flag, int i): key(i), flg(flag) { }

  int operator()(P& x){
    if(flg) return (x < key);
    else return (x >= key);
  }
};

/********************************************************************
  The basic Bin object holds an array of P type objects.
  To facilitate parallel communication there is also a
  "temp" array of the same size.

  The Bin needs an array of pointers to the other bins and it needs
  to know which bin it is in the list.  The array of bin pointers
  is
     Bin<P> **bin_list;
  which will become
     HPCxx_GlobalPtr<Bin<P> > *bin_list;
  in the distributed version.

*********************************************************************/       

template <class P>
class Bin: public HPCxx_Runnable {
public:
  P* list;
  P* tmp;
  int s;
  int max_s;
  int my_id;
  int num_bins;
  Bin<P>** bin_list;
  HPCxx_CSem* semptr;
  HPCxx_Barrier* col;
  int key;
  HPCxx_Mutex m;
  Chunk<P> chnk;

  Bin() { }
  Bin(int x, int id, HPCxx_CSem* sp, Bin<P>** bp, int n, HPCxx_Barrier* cl)
  {
    semptr = sp;
    bin_list = bp;
    num_bins = n;
    my_id = id;
    s = 0;
    list = new P[x];
    tmp = new P[x];
    max_s = x;
    col = cl;
    key = col->getKey();
  }

  int size() { return s; }
  P& operator [](int i) { return list[i]; }
  void append(Chunk<P>& c)
  {
    m.lock();
    if(c.length + s > max_s)
      cout << "bin overflow" << endl;
    else {
      bcopy(c.data, &(list[s]), c.length*sizeof(P));
      s = s+c.length;
    }
    m.unlock();
  }
  
  void swap(int i, int j)
  {
    P tmp = list[i];
    list[i] = list[j];
    list[j] = tmp;
  }
  Chunk<P>* remove(test<P> cond){ // remove all that satisfy condition.
    m.lock();
    int i = 0;
    int j = s-1;

    while(i < j) {
      while(i < s && !cond(list[i]) ) i++;
      while(j > 0 && cond(list[j]) ) j--;
      if(i<j) swap(i,j);
    }

    Chunk<P>* q = &chnk;
    q->data = tmp;
    q->length = s-i;
    bcopy(&(list[i]), q->data, q->length*sizeof(P));
    q->length = s-i;
    s = i;
    m.unlock();
    return q;
  }
  
  void exchange(test<P> cond, Bin* other)
  {
    Chunk<P>* p = remove(cond);
    col->barrier(key);
    other->append(*p);
    col->barrier(key);
  }
  void run(){
    int n = num_bins;
    int hn = num_bins/2;
    int k = n;
    int base = 0;

    while(k > 1) {
      int side = (my_id >= hn)?1:0;
      int partner = base +((my_id + hn - 2*base) % k);
      test<P> cond(side, hn);
      exchange(cond, bin_list[partner]);
      k = k/2;
      base = base + side*(hn-base);
      hn = base + k/2;
    }
    semptr->incr();
  }

  void print()
  {
    cout << "bin["<<my_id<<"]=";
    for(int i = 0; i < s; i++)
      cout << " " << list[i];
    cout <<endl;
  }
  
  int check(int k)
  {
    if(k < 0) return s;
    for(int i = 0; i < s; i++)
      if(list[i] != k) cout << "error!" ;
    return s;
  }
};

int cmp(const void* x, const void* y)
{
  int* px = (int*) x;
  int* py = (int*) y;
  return *px < *py;
}

#define NUMBINS 8
#define BINSIZE 400000

/* the following is a trivial program to do the same
   thing sequentially as fast as possible */


int bigarray[BINSIZE*NUMBINS];
int bins[NUMBINS][BINSIZE];
int binsize[NUMBINS];

void qbinsort()
{
  int i,  p;
  for(i = 0; i < NUMBINS; i++) binsize[i] = 0;
  for(i = 0; i < BINSIZE*NUMBINS; i++){
    p = bigarray[i];
    bins[p][binsize[p]++] = p;
  }
}


typedef Bin<int>*  BinPtr;

int main(int argc, char** argv)
{
  timeval start_time, stop_time;
  float work_time;
  int total;

  HPCxx_Group* g;
  hpcxx_init(argc, argv, g);
  
  printf(" up and running!\n");
  
  HPCxx_CSem sem(NUMBINS);
  HPCxx_Barrier col(NUMBINS);
  
  Bin<int>** bl = new BinPtr[NUMBINS];
  int i, j;
  for(i = 0; i < NUMBINS; i++)
    bl[i]= new Bin<int>(3*BINSIZE, i, &sem, bl, NUMBINS, &col);
  
  srand(time(NULL)); // initialize the random number generator
  int v, bk = 0;
  for(i = 0; i < NUMBINS; i++) {
    for(j = 0; j < BINSIZE; j++) {
      v = rand() % NUMBINS; 
      (*bl[i])[j] = v;
      bigarray[bk++] = v;
    }
    bl[i]->s = BINSIZE;
  }
  total = 0;
  for(i = 0; i < NUMBINS; i++)
    total +=  bl[i]->check(-1);
  
  cout << "start sorting !! total = " << total << endl;
  gettimeofday(&start_time, NULL);
  HPCxx_Thread** t = new HPCxx_Thread*[NUMBINS];
  for(i = 0; i < NUMBINS; i++) {
    t[i] = new HPCxx_Thread(bl[i]);
    t[i]->start();
  }
  sem.wait();
  for(i = 0; i < NUMBINS; i++)
    t[i]->join();
  
  gettimeofday(&stop_time, NULL);
  work_time = ElapseTime(start_time, stop_time);
  cout << "parallel work done.  time =" << work_time << endl;
  cout << "now doing sequential version" << endl;
  gettimeofday(&start_time, NULL);
  qbinsort();
  gettimeofday(&stop_time, NULL);
  work_time = ElapseTime(start_time, stop_time);
  cout << "sequential version done. time =" << work_time << endl;

  total = 0;
  for(i = 0; i < NUMBINS; i++)
    total +=  bl[i]->check(i);

  cout << "total = " << total << endl;
  hpcxx_exit(g);
  return 0;
}

     
