
#include <stdlib.h>
#include <unistd.h>
#include <stream.h>
#include <signal.h>

#include <nexus.h>
#include <hpcxx_rts.h>
#include "invoke.h"  // should not be needed.
void hpcxx_barrier();

class C{
  public:
    int i;
    float f;
    C(int i_, float f_): i(i_), f(f_){}
    C(){ i = 0; f = 0.0; }
    static int registrationID(){ return 345; }
    float foo(int i_){ return f+i_; }
};

void hpcxx_pack(HPCxx_Buffer &b, C *p, int count){
  for(int j = 0; j < count; j++){
    hpcxx_pack(b, &(p[j].i), 1);
    hpcxx_pack(b, &(p[j].f), 1);
  }
}

void hpcxx_unpack(HPCxx_Buffer &b, C *p, int count){
  for(int j = 0; j < count; j++){
     hpcxx_unpack(b, &(p[j].i), 1);
     hpcxx_unpack(b, &(p[j].f), 1);
   }
}

hpcxx_id_t foo_id;

class gpC: public HPCxx_GlobalPtr<C>{
  public:
     //static hpcxx_id_t foo_id;
     gpC(HPCxx_GlobalPtr<C> &p): HPCxx_GlobalPtr<C>(p){}
     gpC *operator ->(){
        return this;
       }
     float foo(int i){
        float t;
        hpcxx_minvoke((HPCxx_GlobalPtr<C> *) this, t, foo_id,  i);
        return t;
       }
     static void set_foo_id(int i){ foo_id = i; }
};

     
HPCxx_GlobalPtr<C> fetch_C(int i){
     C *p = new C(7+i, 3.14);
     HPCxx_GlobalPtr<C> gp(p);
     printf("fetch_C called\n");
     return gp;
}

HPCxx_GlobalPtr<int> fetch_int(int i){
     int *p = new int[i];
     for(int j = 0; j < i; j++) p[j] = 100+j;
     HPCxx_GlobalPtr<int> gp(p);
     printf("fetch_int called\n");
     return gp;
}

class floatAdd{
  public:
    floatAdd(){}
    float operator ()(float x, float y){ return x+y; }
};
  
class MyThread: public HPCxx_Thread{
  int id;
  int key;
  HPCxx_Collective<int> &c;
  int *result;
  HPCxx_CSem &s;
 public:
    MyThread(int i, HPCxx_Collective<int> &cr, HPCxx_CSem &so): 
           id(i), c(cr), s(so){
          key = cr.acquireKey();
       }
    void run(){
       int size = 1;
       result = c.gather(key, &id, size);
       if(id == 0){
         for(int i = 0; i < size; i++)
               printf("result[%d] = %d\n", i, result[i]);
         }
       s.incr();
     }
};

       
int main(int argc, char **argv)
{

  HPCxx_Group *group;
  hpcxx_registerClass((C *) NULL);
  hpcxx_id_t fetch_id = hpcxx_register(fetch_int, 22);
  hpcxx_id_t fetch_C_id = hpcxx_register(fetch_C, 23);
  gpC::set_foo_id(hpcxx_register(C::foo, 24));
  int j;
  hpcxx_init(argc, argv, group);
  group->setNumThreads(3);
  HPCxx_Collective<int> bar(group);
  HPCxx_Collective<HPCxx_GlobalPtr<int> > dalloc(group);
  HPCxx_Reduct<float, floatAdd> add(group, floatAdd());
  add.synchronousSetUp();
  bar.synchronousSetUp();
  dalloc.synchronousSetUp();
  /*
  if(group->getLocalIndex() == 0){
    // get the context id of the other context
    HPCxx_ContextID *othercid = group->getContextID(1);
    
    cout << "HPC++ Initialized.." << endl;
    HPCxx_GlobalPtr<int> gp;
    int size = 1000;
    hpcxx_invoke(*othercid, gp, fetch_id,  size);
    printf("got the gp\n");
    int *z = new int[size];
    HPCxx_GlobalPtr<int> gp3(gp);
    HPCxx_GlobalPtr<int> gp4(gp);
    gp.read(z, size);
    for(j = 0; j < 10; j++)
      printf("z[%d] = %d \n", j, z[j]);
    int w = 2999;
    int *wp = &w;
    *gp4++ = w;
    gp3++;
    gp3++;
    for(j = 0; j < 5; j++) *gp3++ = 70+j;
    gp[6] = 999;
    gp4 = gp+5;
    w = *gp4;
    printf("w = %d\n", w);
    HPCxx_GlobalPtr<int> gp5;
    //gp5 = &(gp[7]);
    // *gp5 = 54321;
    w = gp[7];
    printf("gp[7] = %d\n", w);
    gp.read(z, size);
    for(j = 0; j < 10; j++)
      printf("z[%d] = %d \n", j, z[j]);
    
    HPCxx_GlobalPtr<C> gp2;
    int d = 1;
    hpcxx_invoke(*othercid, gp2, fetch_C_id,  d);
    printf("got the C gp\n");
    C cp(0,0.0);
    cp = *gp2;
    printf(" i = %d  f = %f\n", cp.i, cp.f);
    gpC p(gp2);
    float numb = p->foo(10);
    printf("numb = %f\n", numb);
    printf("all done!\n");
   
  } // end of master block
 */
 int xx, size;
 
   
 // now test the allocator idea.

 int *fp = new int[100];
 HPCxx_GlobalPtr<int> fgp(fp);
 xx = group->getLocalIndex()*1000;
 for(j = 0; j < 100; j++) fp[j] = xx+j*100+7;
 size = 1;
 printf("[%d]doing the reduct .. \n", group->getLocalIndex());
 HPCxx_GlobalPtr<int> *fp_array = NULL;
 fp_array =  dalloc.gather(&fgp, size);
 if(fp_array == NULL) printf("[%d] ERROR \n", group->getLocalIndex());
 printf("[%d]reduct finished\n", group->getLocalIndex());
 //if(group->getLocalIndex() == 0)
   for(j = 0; j < size; j++){
 	int f = fp_array[j][j];
        //int f = *(fp_array[j]);
 	printf("[%d] f= %d\n", group->getLocalIndex(), f);
        }
 printf("[%d] in barrier\n", group->getLocalIndex());
 dalloc.barrier();
 printf("[%d] outof barrier\n", group->getLocalIndex());
xx = group->getLocalIndex()+100;
 if(group->getLocalIndex()==0) size = 1; else size = 0;
 int *outarr = bar.gather(&xx, size);
 for(j = 0; j < size; j++) printf("[%d] outarr[%d] = %d\n", 
       group->getLocalIndex(), j, outarr[j]);

 HPCxx_CSem cs(3);
  for(j = 0; j < 3; j++){
     MyThread *thread = new MyThread(j+10*group->getLocalIndex(), bar,cs);
     thread->start();
     }
 cs.wait();

 float xf = 3.14;
 float *yf = add(&xf, 1);
 printf("[%d] yf = %f\n", group->getLocalIndex(), *yf);
 printf("[%d] addrof(xf) = %xl yf = %xl  xf=%f\n",
    group->getLocalIndex(), &xf, yf, xf);
 hpcxx_barrier();
 printf("[%d] out of barrier\n", group->getLocalIndex());
 hpcxx_exit(group);
  return(0);
}
