//
// hello.C is a simple program which demonstrates the basics of the
// HPCxx_Thread class. It creates a number of threads, each of which print
// a hello message, then die.
//

#include <iostream.h>
#include <stdlib.h>
#include <hpcxx_rts.h>

int nCompleted = 0;
int running = 0;
int error = 0;
int loser = 0;
HPCxx_Mutex mutex;

class HelloThread : public HPCxx_Thread 
{ 
  private: char* msg;
  private: int   id;

  public:  HelloThread(char* msg_, int id_): msg(msg_), id(id_), HPCxx_Thread()
    {
    }

  public:  void run()
    {
      mutex.lock();		// protect critical section with a lock

      //
      // check to see if lock holds
      //
      if (running != 0) {
	cout << "ERROR, lock failed" << endl;
	error = 2;
      } else {
	running = 1;
      }

      cout << "    Thread " << id << ": " << msg << endl << flush;

      nCompleted += 1;		// equals # of threads when finished
      loser = id;		// loser is the last thread to complete

      running = 0;
      mutex.unlock();
    }
};


int main(int argc, char* argv[])
{
  int i;
  HPCxx_Group* g;
  hpcxx_init(argc, argv, g);

  char* msg = "Hello world!";
  int nThreads = (argc > 1) ? atoi(argv[1]) : 2;
  cout << endl;

  //
  // create threads
  //
  HelloThread** worker = new HelloThread*[nThreads];
  for (i = 0; i < nThreads; i++) {
    worker[i] = new HelloThread(msg, i+1);
  }
  
  if (loser != 0 || nCompleted != 0) {
    cout << "ERROR, threads started prematurely" << endl;
    return 1;
  }

  //
  // start threads
  // 
  for (i = 0; i < nThreads; i++) {
    worker[i]->start();
  }

  //
  // wait for threads to finish
  //
  for (i = 0; i < nThreads; i++) {
    worker[i]->join();
  }

  cout << endl;
  if (loser == 0 || nCompleted != nThreads || error != 0) {
    cout << "ERROR, hello threads did not complete correctly" << endl;
    if (error == 0) error = 1;
    return error;
  } else {
    cout << "Successfully completed hello, loser is " << loser << endl;
  }

  //
  // clean up
  //
  for (i = 0; i < nThreads; i++) {
    delete worker[i];
  }
  delete[] worker;

  return hpcxx_exit(g);
}
