#include <stdio.h>
#include <limits.h>
#include <hpcxx_rts.h>
#include <hpcxx_hack.h>
#include <math.h>

class BarrierTest : public HPCxx_Thread {
    
    int myId;
    HPCxx_Collective<int> &b;
    HPCxx_SyncQ<int> *isync;
    int myKey;
public:

    BarrierTest( int myId_,
                 HPCxx_Collective<int> &b_,
                 HPCxx_SyncQ<int> *isync_ ) : myId(myId_), b(b_),
                                              isync( isync_ ) {
        myKey = 0;
    }

    virtual ~BarrierTest() {}

    void run( void ) {
        int irun = 0;
        int i = 0;
        double x = 1.0;
        myKey = b.acquireKey();
        int k = 0;
        //while (( k++ < 10) && ((irun = isync[myId]) > 0)  ) {
        while(k++ < 10){
            for(i = 0; i < 1000000; i++) x = sin(x);
            printf("%d: waiting at barrier\n", myId);
            b.barrier( myKey );
            printf("%d: through barrier\n", myId);
        }
    }

};

// Driver
int main( int argc, char *argv[] ) {
    int i = 0;
    int nthreads = 0;
    int myKey = 0;
    int threadRun = 1;
    HPCxx_Group *group = NULL;
    BarrierTest **barrierTest;
    HPCxx_SyncQ<int> *isync;

    if ( argc < 2 ) {
        fprintf( stderr, "Enter number of BarrierTest threads:  " );
        fscanf( stdin, "%d", &nthreads );
    }
    else {
        sscanf( argv[1],  "%d", &nthreads );
    }

    hpcxx_init( argc, argv, group );
    group->setNumThreads(nthreads+1);

    HPCxx_Collective<int> b(group);

    barrierTest = new BarrierTest *[nthreads];
    isync = new HPCxx_SyncQ<int> [nthreads];


    myKey = b.acquireKey();

    for ( i = 0; i < nthreads; i++ ) {
        barrierTest[i] = new BarrierTest( i, b, isync );
    }

    for ( i = 0; i < nthreads; i++ ) {
        barrierTest[i]->start();
    }

    for(int j = 0; j < 10; j++){

        for ( i = 0; i < nthreads; i++ ) {
            isync[i] = threadRun;
        }
        char c;
        printf("main at barrier. hit a key to continue\n");
        fscanf(stdin, "%c", &c);
        b.barrier( myKey );

    }
    for(i = 0; i < nthreads; i++) barrierTest[i]->join();

    (void) hpcxx_exit( group );

    return( 0 );
}
