#include <stdlib.h>
#include <unistd.h>
#include <stream.h>
#include <signal.h>

#include <nexus.h>
#include <hpcxx_rts.h>

#define ElapseTime(Start, Stop) \
    (  (  ( ((double)(Stop.tv_sec)) - ((double)(Start.tv_sec)) ) \
        + (((double)(Stop.tv_usec))/1000000.0) ) \
     - (((double)(Start.tv_usec))/1000000.0) )

void *_nx_st_rsh_info(void)
{
  return(NULL);
} 
#include <stdio.h>
#include <math.h>
#include <string.h>

// three parameters: G = grav. constant, DeltaT = time step size
// WORKERS is number of threads.
#define G 1.0
#define DeltaT 0.01
#define WORKERS 4

template <class T>
class vector{
   public:
      int size;
      T *data;
      vector(int n): size(n){
          data = new T[n];
         };
      T & operator [](int i){ return data[i]; }
};

class vect{
   public:
      double d[3];
      vect(){ d[0] = d[1] = d[2] = 0.0; }
      vect(double x,double y,double z){ d[0] = x; d[1] = y; d[2] = z; }
      double &operator [](int i){ return d[i]; }
      vect operator +(const vect &y){
          vect t(d[0]+y.d[0], d[1]+y.d[1], d[2]+y.d[2]);
          return t;
          }
      vect operator -( vect &y){
          vect t(d[0]-y.d[0], d[1]-y.d[1], d[2]-y.d[2]);
          return t;
          }
      vect operator *( double x){
          vect t(d[0]*x, d[1]*x, d[2]*x);
          return t;
          }
      double norm(){
          return sqrt(d[0]*d[0]+d[1]*d[1]+d[2]*d[2]);
          }
      double max(){
          double x;
          x = (d[0]>0)?d[0]: -d[0];
          double y;
          y = (d[1]>0)?d[1]: -d[1];
          double z;
          z = (d[2]>0)?d[2]: -d[2];
          if(x > y)
             return (x>z)? x:z;
          else
             return (y>z)? y:z;
          }
      vect &operator =(double x){ 
            d[0]= d[1]= d[2] = x; 
          return *this;
         }
      vect &operator =(const vect& x){ 
            d[0]=x.d[0];  d[1]=x.d[1];  d[2] = x.d[2]; 
          return *this;
         }
      void print(){
          cout << "[ " << d[0] << ", " << d[1] << ", " << d[2] << "]" << endl;
          }
};

class Mass{
  public:
   double mass;
   vect position;
   Mass(){ mass = 0; position[0] = 0; position[1] = 0; position[2] = 0;}
   Mass(double m, double x, double y, double z): mass(m){
          position[0] = x; position[1] = y; position[2] = z;
      }
   Mass &operator +=(Mass &rhs){
     for(int i = 0; i < 3; i++)
        position[i] += rhs.mass*rhs.position[i];
     return *this;
   }
   Mass &operator /=(const double d){
     for(int i = 0; i < 3; i++)
        position[i] /= d;
     return *this;
   }
   double distance(Mass &M){
      double dx = position[0]-M.position[0];
      double dy = position[1]-M.position[1];
      double dz = position[2]-M.position[2];
      return sqrt((double) (dx*dx + dy*dy + dz*dz));
   }
};

class Particle{
 public:
    Mass m;
    vect vel;
    // vect vel2; used for error check
    double key;
    Particle(){}
    Particle(double M, double x, double y, double z,
             double vx, double vy, double vz){
       m.mass = M; 
       m.position[0] = x; m.position[1] = y; m.position[2] = z;
       vel[0] = vx; vel[1] = vy; vel[2] = vz;
       // vel2[0] = vx; vel2[1] = vy; vel2[2] = vz;
       key = 0;
     }
    void set(double M, double x, double y, double z,
             double vx, double vy, double vz){
       m.mass = M; 
       m.position[0] = x; m.position[1] = y; m.position[2] = z;
       vel[0] = vx; vel[1] = vy; vel[2] = vz;
       // vel2[0] = vx; vel2[1] = vy; vel2[2] = vz;
       key = m.position[0];
     }
    void set_key(int i){ key = m.position[i%3]; }
    int operator >(Particle &p){ return key > p.key; }
    void update_vel(Mass &M){
        double distance = m.distance(M);
        if(distance > 0){ // not me!
           double factor = G*m.mass*(M.mass)/(distance*distance*distance);
           vect force = (M.position - m.position)*factor;
           vel = vel+ force*DeltaT;
          }
        }
    void update_pos(){
        m.position = m.position + vel*DeltaT;
       }
    /* used for error check
    void update_vel2(Mass &M){
        double distance = m.distance(M);
        if(distance > 0){ // not me!
           double factor = G*m.mass*(M.mass)/(distance*distance*distance);
           vect force = (M.position - m.position)*factor;
           vel2 = vel2 + force*DeltaT;
          }
        }
    */

};

int comp( const void *l, const void *r){
     Particle *lm = (Particle *) l;
     Particle *rm = (Particle *) r;
     return *lm > *rm;
}

int sort(vector<Particle> &v, int i, int j, double *mi, double *ma, 
	 double *av){
   double ave = 0;
   int t;
   double max = v[j-1].key;
   double min = v[i].key;
   for(t = i; t < j; t++){ 
         if(v[t].key > max) max = v[t].key;
         if(v[t].key < min) min = v[t].key;
         ave += v[t].key;
         }
   ave /= j-i;
   *mi = min; *ma = max; *av = ave;
   j--;
   int swap_count = 0;
   while( j > i){
      while( (v[i].key < ave) && (i < j)) i++;
      while( (v[j].key > ave) && (i < j)) j--;
      if(i < j){
        Particle p = v[i];
        v[i] = v[j];
        v[j] = p;
        swap_count++;
        }
      }
   //cout << "ave = " << ave << " moved = " << 2*swap_count << endl;
   return i;
}

void init(vector<Particle> &v){
    int s = v.size;
    int i;
    // place particles on a sphere of unit radius moving out at
    // unit velocity away from the origin.
    srand(time(NULL)); // initialize the random number generator
    double h = 1.0;
    for(i = 1; i < 15; i++) h = 2*h;
    for(i = 0; i < s; i++){
         double x, y, z;
         x = rand()-h; y = rand()-h; z = rand()-h;
         double t = sqrt(x*x + y*y + z*z);
         x = x/t;  y = y/t; z = z/t;
         v[i].set(1.0/v.size,x,y,z, x,y,z);
         //cout << "----- particle " << i << " --------------\n";
         //cout << "mass = " << v[i].m.mass << endl; 
         //cout << "position[0]=" << v[i].m.position[0] << endl;
         //cout << "position[1]=" << v[i].m.position[1] << endl;
         //cout << "position[2]=" << v[i].m.position[2] << endl;
         //cout << endl;
        }
}

class Tree{
   public:
      Mass m;
      double width;
      vect size;
      Mass get_mass(){ return m; } 
      virtual void update_tree_mass(){}
      virtual void update_velocity(Particle &p){}
      Tree():m(), size(0,0,0){ width = 0;}
      virtual void free(){};
};

class TreeNode: public Tree{
   public:
      static TreeNode *free_list;
      TreeNode *next_free;
      Tree *left, *right;
      static TreeNode *alloc(){
          if(free_list == NULL) return new TreeNode;
          else{
             TreeNode *p = free_list;
             free_list = free_list->next_free;
             return p;
            }
        }
      void free(){ next_free = free_list; 
                   free_list = this; 
                   left->free();
                   right->free();
                   left = right = NULL;
                 }
      void update_tree_mass(){
         left->update_tree_mass();
         right->update_tree_mass();
         m.mass = left->m.mass + right->m.mass;
         m.position = 0.0;
         m +=  left->m;
         m += right->m;
         m /= m.mass;
         }
     void update_velocity(Particle &p){
        double distance = p.m.distance(m);
        if(distance > 0.0){ // not me!
            if(width/distance < 1.0){
               double factor = G*p.m.mass*(m.mass)/(distance*distance*distance);
                 vect force = (m.position - p.m.position)*factor;
                 p.vel = p.vel+ force*DeltaT;
                 }
            else{
               left->update_velocity(p);
               right->update_velocity(p);
            }
        }
     }
};

class Leaf: public Tree{
  public:
      static Leaf *free_list;
      Leaf *next_free;
      Particle * particle;
      Leaf(): Tree(){ particle = NULL;}
      static Leaf *alloc(){
          if(free_list == NULL) return new Leaf;
          else{
             Leaf *p = free_list;
             free_list = free_list->next_free;
             p->particle = NULL;
             return p;
            }
      }
      void free(){ next_free = free_list; free_list = this; }

      void update_tree_mass(){
         if(particle)
            m = particle->m;
      }
     void update_velocity(Particle &p){
       if(particle)
        p.update_vel(m);
     }
};

Leaf * Leaf::free_list = NULL;
TreeNode * TreeNode::free_list = NULL;
int node_count = 0;
Tree * balance(vector<Particle> &v, int depth, int l, int r, vect size){
   int i;
   TreeNode *tp;
   // cout << "depth =" << depth << " l=" << l << " r=" << r << endl;
   if(r == l){ // empty segment
      Leaf *lp = Leaf::alloc();
      lp->particle = NULL;
      return lp;
      }
   if((r-l) == 1){ // single particle
      Leaf *lp = Leaf::alloc();
      lp->particle = &(v[l]);
      lp->update_tree_mass();
      return lp;
      }
   for(i = l; i < r; i++) v[i].set_key(depth);
   
   double x_min, x_max, x_mid;
   int h = sort(v, l, r, &x_min, &x_max, &x_mid);
   tp = TreeNode::alloc();
   vect left_half(size[0], size[1], size[2]);
   vect right_half(size[0], size[1], size[2]);
   int k = (depth+1)%3;
   left_half[k] = x_mid-x_min; 
   right_half[k] = x_max-x_mid;
   tp->left = balance(v, depth+1, l, h, left_half);
   tp->right = balance(v, depth+1, h, r, right_half);
   tp->width = size.max();
   tp->size = size;
   return tp;
}

// generated thread subclass for parallelized loop below
class LoopThread: public HPCxx_Thread{
   vector<Particle> &v;
   Tree *t;
   int base;
   HPCxx_CSem *c;
  public:
    LoopThread(HPCxx_Runnable *r, HPCxx_CSem * s, Tree * T, vector<Particle> &V, 
               int cnt):
                 HPCxx_Thread(r), c(s), t(T), v(V), base(cnt){}
    void run(){
       int j;
       for(j = base; j < base+v.size/WORKERS; j++){
          t->update_velocity(v[j]);
          } 
       //printf("worker %8d done\n", base);
       c->incr();
    }
};


void calculate_force(vector<Particle> &v, Tree *t){
       // for(int i = 0; i < v.size; i++)
       //   t->update_velocity(v[i]);
       // generated code for parallelized loop
     
       LoopThread *th[WORKERS];
       HPCxx_CSem *c = new HPCxx_CSem(WORKERS);
       for(int cnt = 0; cnt < WORKERS; cnt++){
          th[cnt] = new LoopThread(NULL,c,t, v,v.size/WORKERS*cnt);
          th[cnt]->start();
          }
       c->wait();
     
      //
}

void take_step(vector<Particle> &v, Tree **t){
    for(int i = 0; i < v.size; i++)
         v[i].update_pos();
   (*t)->free();
   vect init_size(10,10,10); //box around origin [-5,5]
   node_count = 0;
   *t = balance(v, 0, 0, v.size, init_size);
   (*t)->update_tree_mass();

}

/* used for error check   
void n_square_method(vector<Particle> &v){
  int i, j; 
    for(i = 0; i < v.size; i++){
       for(j = 0; j < v.size; j++)
              v[i].update_vel2(v[j].m);
       }
}
*/

int main(int argc, char **argv){
   int i,s;
   timeval start_time, stop_time;
   float work_time;
   HPCxx_Group *g;
   hpcxx_init(argc, argv, g);
   
   printf(" up and running!\n");
   cout << "input number of particles (in K) (suggest 32):";
   cin >> s; s = s*1024;
   cout << "size = " << s << endl;
   vector<Particle> *vp = new vector<Particle>(s);
   vector<Particle> &v = *vp;
   init(v);
   double total = 0.0;
   for(i = 0; i < v.size; i++)
      total += v[i].m.position[0];
   // cout << " position 0 average = " << total/v.size << endl;
   total = 0;
   for(i = 0; i < v.size; i++)
      total += v[i].m.position[1];
   // cout << " position 1 average = " << total/v.size << endl;
   total = 0;
   for(i = 0; i < v.size; i++)
      total += v[i].m.position[2];
   // cout << " position 2 average = " << total/v.size << endl;

   vect init_size(10,10,10); //box around origin [-5,5]
   gettimeofday(&start_time, NULL);
   Tree *root = balance(v, 0, 0, v.size, init_size);
   root->update_tree_mass();
   gettimeofday(&stop_time, NULL);
   work_time = ElapseTime(start_time, stop_time);
   cout << "done. build tree time = " << work_time << endl;

   // cout << "mass = " << root->m.mass << endl; 
   // cout << "position[0]=" << root->m.position[0] << endl;
   // cout << "position[1]=" << root->m.position[1] << endl;
   // cout << "position[2]=" << root->m.position[2] << endl;
   double total_time = 0.0;
   for(int step = 0; step < 10; step++){
   	gettimeofday(&start_time, NULL);
   	calculate_force(v, root);
   	take_step(v, &root);
   	gettimeofday(&stop_time, NULL);
   	work_time = ElapseTime(start_time, stop_time);
   	cout << "step  time = " << work_time << endl;
        total_time += work_time;
      }
   cout << "total time = " << total_time << endl;
  /* test of O(n^2) method. skip this one.
   gettimeofday(&start_time, NULL);
   n_square_method(v);
   gettimeofday(&stop_time, NULL);
   work_time = ElapseTime(start_time, stop_time);
   cout << "done.  time = " << work_time << endl;
   vect error(0,0,0);
   double nrm = 0;
   for(i = 0; i < v.size; i++){
       cout << "vel =";  v[i].vel.print();
       cout << "vel2=";  v[i].vel2.print();
       error = v[i].vel-v[i].vel2;
       double t = error.norm();
       nrm += t*t;
       }
   nrm = sqrt(nrm)/v.size;
   cout << "error = " << nrm << endl;
   */
   hpcxx_exit(g);
   return 0;
}

   
   
   
