#include <stdlib.h>
#include <unistd.h>
#include <stream.h>
#include <signal.h>

#include <nexus.h>
#include <hpcxx_rts.h>

#define ElapseTime(Start, Stop) \
    (  (  ( ((double)(Stop.tv_sec)) - ((double)(Start.tv_sec)) ) \
        + (((double)(Stop.tv_usec))/1000000.0) ) \
     - (((double)(Start.tv_usec))/1000000.0) )

void *_nx_st_rsh_info(void)
{
  return(NULL);
} 
#include <stdio.h>
#include <strings.h>
#include <math.h>



typedef int hpcxx_buffer;

/**********************************************************
  This program tests the HPCxx_Reduct1 class
  by doing reductions on big blocks of data.
**********************************************************/

/***********************************************************
  A block class used to move a block of data around
  int the multicontext case.
***********************************************************/
template <class P>
class Block{
  public:
     P *data;
     int length;
     Block(int l, P zero){
        length = l;
        data = new P[l];
        for(int i = 0; i < l; i++) data[i] = zero;
        }
     Block(){ data = NULL; length = 0; }
     Block &operator =(const Block &b){ 
         length = b.length; 
         if(data == NULL) data = new P[length];
         bcopy(b.data, data, length*sizeof(P));
         return *this;}
     P &operator [](int i){ return data[i]; }
     friend void hpcxx_pack(hpcxx_buffer *, Block<P> *, int);
};

void hpcxx_pack(hpcxx_buffer *buffer, int i, int l){}

template<class P>
void hpcxx_pack(hpcxx_buffer *buffer, Block<P> *d, int l){
   hpcxx_pack(buffer, l, 1);
   for(int i = 0; i < l; i++){
       hpcxx_pack(buffer, d[i].length, 1);
       hpcxx_pack(buffer, d[i].data, d[i].length);
       }
}

template<class P>
void hpcxx_unpack(hpcxx_buffer *buffer, Block<P> *d, int &l){
   hpcxx_unpack(buffer, l, 1);
   for(int i = 0; i < l; i++){
       hpcxx_unpack(buffer, &d[i].length, 1);
       hpcxx_unpack(buffer, d[i].data, d[i].length);
       }
}

// *************************************************************
// user code here
// *************************************************************

class intAdd{
  public:
   void operator ()(int &x, int &y){  x = x+y; }
};

class blockAdd{
  public:
    void operator()( Block<float> &a, Block<float> &b){
         int s = a.length;
         float *p = a.data;
         float *q = b.data;
         for(int i = 0; i < s; i++) p[i] = p[i]+ q[i];
    }
};
#define BLOCKSIZE 1000000
#define REPS 2
#define NUMWORKERS 16

/**************************************************
  Class R below tests the speed of the hpc++ reduction.
  each R object has a block of 1 million floats and it does
  both a distructive and non-destructive reductions. (2 repetitions
  each.) and computes the speed in mflops.
***************************************************/


class R: public HPCxx_Runnable{
   HPCxx_Reduct1<Block<float>, blockAdd > &r;
   int my_id, my_key;
   Block<float> *myBlock;
   HPCxx_CSem *s;
  public:
     R(int id, HPCxx_Reduct1<Block<float>, blockAdd> &red, HPCxx_CSem *t ): 
        my_id(id), r(red), s(t){
                my_key = r.getKey();
                myBlock = new Block<float>(BLOCKSIZE, id);
        }
     void run(){
        int i,j;
        Block<float> total(BLOCKSIZE, 0);
        timeval start_time, stop_time;
        float work_time;
        for(j = 0; j < 10; j++){
           for(i = 0; i < myBlock->length; i++)
              myBlock->data[i] = j+i;
	   }
        r.barrier(my_key);
        //printf("befor reduct for id %d\n",my_id);
        if(my_id==0) gettimeofday(&start_time, NULL);
        for(i = 0; i < REPS; i++)
           r.destructive(my_key, myBlock, blockAdd());
        if(my_id == 0){
                gettimeofday(&stop_time, NULL);
		work_time = ElapseTime(start_time, stop_time);
                printf("destructive version reduction speed (Mflops)= %f\n", 
		    (2.0*REPS*BLOCKSIZE*NUMWORKERS)/(work_time*1.0e6));
	        }
        r.barrier(my_key);
        //printf("befor reduct2 for id %d\n",my_id);
        if(my_id==0) gettimeofday(&start_time, NULL);
        for(i = 0; i < REPS; i++)
           total = r(my_key, *myBlock, blockAdd());
        if(my_id == 0){
                gettimeofday(&stop_time, NULL);
		work_time = ElapseTime(start_time, stop_time);
                printf("standard version reduction speed (Mflops)= %f\n", 
		    (2.0*REPS*BLOCKSIZE*NUMWORKERS)/(work_time*1.0e6));
	        }
       s->incr();
       }
};
typedef R* Rptr;
        


int main(int argc, char *argv[]){
    timeval start_time, stop_time;
    float work_time;
    HPCxx_Group *g;
    hpcxx_init(argc, argv, g);

    printf(" up and running!\n");
    blockAdd sum;
    g->setNumThreads(NUMWORKERS);
    HPCxx_CSem sem(NUMWORKERS);
    HPCxx_Reduct1<Block<float>, blockAdd> red(*g);
     int i;
    R **r = new Rptr[NUMWORKERS];
    for(i = 0; i < NUMWORKERS; i++) 
            r[i] = new R(i, red, &sem);

     gettimeofday(&start_time, NULL);
     for(i = 0; i < NUMWORKERS; i++){
         HPCxx_Thread *t = new HPCxx_Thread(r[i]);
         t->start();
         }
     sem.wait();
     gettimeofday(&stop_time, NULL);
     work_time = ElapseTime(start_time, stop_time);
     cout << " total time =" << work_time << endl;
     hpcxx_exit(g);
     return 0;
}


    

