
#include "HPCxx_Profile.h"
#include "HPCxx_Barrier.h"
#include <iostream.h>

//#--------------------------------------------------------------------------
// These methods construct a new barrier object given either an integer
// representing the number of participating threads in the barrier, or
// an HPCxx_Group object from which the number of threads can be obtained.
// <ul>
// <li><b>height</b> -- the height of the barrier (number of participating
// threads)
// <li><b>group</b> -- the group of participating threads
// </ul>
//#--------------------------------------------------------------------------
// <group>

HPCxx_Barrier::HPCxx_Barrier(int height)
{
	TAU_PROFILE("HPCxx_Barrier::HPCxx_Barrier()", "void (int)", TAU_HPCXX);
	size = height;
	key_counter = 0;
	inside = 0;
	sync = new HPCxx_SyncQ<int>[height];
}

HPCxx_Barrier::HPCxx_Barrier(HPCxx_Group& group)
{
	TAU_PROFILE("HPCxx_Barrier::HPCxx_Barrier()", "void (HPCxx_Group)", 
 	  TAU_HPCXX);
	key_counter = 0;
	inside = 0;
	size = group.getNumThreads();
	sync = new HPCxx_SyncQ<int>[size];
	if(!sync){ 
		printf(" barrier alloc error\n");
		exit(0);
	}
}

//#--------------------------------------------------------------------------
// Returns a key for the requesting thread.
// <note role=warning> This method must be called only once per thread.
// </note>
//#--------------------------------------------------------------------------

int HPCxx_Barrier::getKey(void)
{
	TAU_PROFILE("HPCxx_Barrier::getKey()", "int ()", TAU_HPCXX);
	m.lock();
	int t = key_counter++;
	m.unlock();
	if(t >= size) { 
		printf("error: too many requesting threads\n");
		exit(0);
	}
	return t;
}

//#--------------------------------------------------------------------------
// Causes the calling thread to block until all participating threads have
// called one of these two barrier functions.
// <ul>
// <li><b>key</b> -- the key that has been previously assigned to the
// calling thread (obtained via the getKey() method).
// </ul>
//#--------------------------------------------------------------------------
// <group>

void HPCxx_Barrier::barrier(int key) 
{
	TAU_PROFILE("HPCxx_Barrier::barrier()", "void (int)", TAU_HPCXX);
	int i, j = 1;
	m.lock();
	inside++;
	if(inside == size){
		inside = 0;
		m.unlock();
		for(i = 0; i < size; i++)
		if(i != key) sync[i].write(j);
	}
	else{
		m.unlock();
		sync[key].read(j);
	}
}
