//##########################################################################
//HPC++Lib: An object-oriented class library for high performance
//parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_Buffer.C,v 1.2 1998/08/07 20:24:49 sameer Exp $
//##########################################################################

// $Id: HPCxx_Buffer.C,v 1.2 1998/08/07 20:24:49 sameer Exp $

#include <HPCxx_Profile.h>
#include <HPCxx_Buffer.h>

//
// This file includes the HPCxx_Buffer member function definitions and
// also the hpcxx_pack/unpack() routines as they are just wrappers
// into the HPCxx_Buffer class.
//


HPCxx_Buffer::HPCxx_Buffer(int bufSize) : impl_(bufSize)
{
}

HPCxx_Buffer::~HPCxx_Buffer()
{
}

int HPCxx_Buffer::pack(const float* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::pack()", "int (float *, int)", TAU_HPCXX);
  return(impl_.pack(data, count));
}

int HPCxx_Buffer::pack(const double* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::pack()", "int (double *, int)", TAU_HPCXX);
  return(impl_.pack(data, count));
}

int HPCxx_Buffer::pack(const short* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::unpack()", "int (short *, int)", TAU_HPCXX);
  return(impl_.pack(data, count));
}

int HPCxx_Buffer::pack(const u_short* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::pack()", "int (u_short *, int)", TAU_HPCXX);
  return(impl_.pack(data, count));
}

int HPCxx_Buffer::pack(const int* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::pack()", "int (int *, int)", TAU_HPCXX);
  return(impl_.pack(data, count));
}

int HPCxx_Buffer::pack(const u_int* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::pack()", "int (u_int *, int)", TAU_HPCXX);
  return(impl_.pack(data, count));
}

int HPCxx_Buffer::pack(const long* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::pack()", "int (long *, int)", TAU_HPCXX);
  return(impl_.pack(data, count));
}

int HPCxx_Buffer::pack(const u_long* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::pack()", "int (u_long *, int)", TAU_HPCXX);
  return(impl_.pack(data, count));
}

int HPCxx_Buffer::pack(const char* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::pack()", "int (char *, int)", TAU_HPCXX);
  return(impl_.pack(data, count));
}

int HPCxx_Buffer::pack(const u_char* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::pack()", "int (u_char *, int)", TAU_HPCXX);
  return(impl_.pack(data, count));
}

#ifndef CONTEXT_ID_AS_INT
//
// If HPCxx_ContextID is an int, then there will be two integer
// packing/unpacking routines, so one of them need to be removed. 
//
int HPCxx_Buffer::pack(const HPCxx_ContextID* data, int count){
  TAU_PROFILE("HPCxx_Buffer::pack()", "int (HPCxx_ContextID *, int)", TAU_HPCXX);
  return(impl_.pack(data, count));
}

void HPCxx_Buffer::unpack(HPCxx_ContextID* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::unpack()", "void (HPCxx_ContextID *, int)", TAU_HPCXX);
  impl_.unpack(data, count);
}
#endif

void HPCxx_Buffer::unpack(float* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::unpack()", "void (float *, int)", TAU_HPCXX);
  impl_.unpack(data, count);
}

void HPCxx_Buffer::unpack(double* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::unpack()", "void (double *, int)", TAU_HPCXX);
  impl_.unpack(data, count);
}

void HPCxx_Buffer::unpack(short* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::unpack()", "void (short *, int)", TAU_HPCXX);
  impl_.unpack(data, count);
}

void HPCxx_Buffer::unpack(u_short* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::unpack()", "void (u_short*, int)", TAU_HPCXX);
  impl_.unpack(data, count);
}

void HPCxx_Buffer::unpack(int* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::unpack()", "void (int *, int)", TAU_HPCXX);
  impl_.unpack(data, count);
}

void HPCxx_Buffer::unpack(u_int* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::unpack()", "void (u_int*, int)", TAU_HPCXX);
  impl_.unpack(data, count);
}

void HPCxx_Buffer::unpack(long* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::unpack()", "void (long *, int)", TAU_HPCXX);
  impl_.unpack(data, count);
}

void HPCxx_Buffer::unpack(u_long* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::unpack()", "void (u_long*, int)", TAU_HPCXX);
  impl_.unpack(data, count);
}

void HPCxx_Buffer::unpack(char* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::unpack()", "void (char *, int)", TAU_HPCXX);
  impl_.unpack(data, count);
}

void HPCxx_Buffer::unpack(u_char* data, int count)
{
  TAU_PROFILE("HPCxx_Buffer::unpack()", "void (u_char *, int)", TAU_HPCXX);
  impl_.unpack(data, count);
}


//
// The hpcxx_pack and hpcxx_unpack routines.
//


void hpcxx_pack(HPCxx_Buffer& buffer, float* data, int count)
{
  TAU_PROFILE("hpcxx_pack()", "void (HPCxx_Buffer, float*, int)", TAU_HPCXX);
  buffer.pack(data, count);
}

void hpcxx_pack(HPCxx_Buffer& buffer, double* data, int count)
{
  TAU_PROFILE("hpcxx_pack()", "void (HPCxx_Buffer, double*, int)", TAU_HPCXX);
  buffer.pack(data, count);
}

void hpcxx_pack(HPCxx_Buffer& buffer, short* data, int count)
{
  TAU_PROFILE("hpcxx_pack()", "void (HPCxx_Buffer, short *, int)", TAU_HPCXX);
  buffer.pack(data, count);
}

void hpcxx_pack(HPCxx_Buffer& buffer, u_short* data, int count)
{
  TAU_PROFILE("hpcxx_pack()", "void (HPCxx_Buffer, u_short*, int)", TAU_HPCXX);
  buffer.pack(data, count);
}

void hpcxx_pack(HPCxx_Buffer& buffer, int* data, int count)
{
  TAU_PROFILE("hpcxx_pack()", "void (HPCxx_Buffer, int *, int)", TAU_HPCXX);
  buffer.pack(data, count);
}

void hpcxx_pack(HPCxx_Buffer& buffer, u_int* data, int count)
{
  TAU_PROFILE("hpcxx_pack()", "void (HPCxx_Buffer, u_int*, int)", TAU_HPCXX);
  buffer.pack(data, count);
}

void hpcxx_pack(HPCxx_Buffer& buffer, long* data, int count)
{
  TAU_PROFILE("hpcxx_pack()", "void (HPCxx_Buffer, long *, int)", TAU_HPCXX);
  buffer.pack(data, count);
}

void hpcxx_pack(HPCxx_Buffer& buffer, u_long* data, int count)
{
  TAU_PROFILE("hpcxx_pack()", "void (HPCxx_Buffer, u_long*, int)", TAU_HPCXX);
  buffer.pack(data, count);
}

void hpcxx_pack(HPCxx_Buffer& buffer, char* data, int count)
{
  TAU_PROFILE("hpcxx_pack()", "void (HPCxx_Buffer, char *, int)", TAU_HPCXX);
  buffer.pack(data, count);
}

void hpcxx_pack(HPCxx_Buffer& buffer, u_char* data, int count)
{
  TAU_PROFILE("hpcxx_pack()", "void (HPCxx_Buffer, u_char *, int)", TAU_HPCXX);
  buffer.pack(data, count);
}

void hpcxx_unpack(HPCxx_Buffer& buffer, float* data, int count)
{
  TAU_PROFILE("hpcxx_unpack()", "void (HPCxx_Buffer, float *, int)", TAU_HPCXX);
  buffer.unpack(data, count);
}

void hpcxx_unpack(HPCxx_Buffer& buffer, double* data, int count)
{
  TAU_PROFILE("hpcxx_unpack()", "void (HPCxx_Buffer, double *, int)", TAU_HPCXX);
  buffer.unpack(data, count);
}

void hpcxx_unpack(HPCxx_Buffer& buffer, short* data, int count)
{
  TAU_PROFILE("hpcxx_unpack()", "void (HPCxx_Buffer, short *, int)", TAU_HPCXX);
  buffer.unpack(data, count);
}

void hpcxx_unpack(HPCxx_Buffer& buffer, u_short* data, int count)
{
  TAU_PROFILE("hpcxx_unpack()", "void (HPCxx_Buffer, u_short *, int)", 
    TAU_HPCXX); 
  buffer.unpack(data, count); 
} 

void hpcxx_unpack(HPCxx_Buffer& buffer, int* data, int count)
{
  TAU_PROFILE("hpcxx_unpack()", "void (HPCxx_Buffer, int *, int)", TAU_HPCXX);
  buffer.unpack(data, count);
}

void hpcxx_unpack(HPCxx_Buffer& buffer, u_int* data, int count)
{
  TAU_PROFILE("hpcxx_unpack()", "void (HPCxx_Buffer, u_int *, int)", TAU_HPCXX);
  buffer.unpack(data, count);
}

void hpcxx_unpack(HPCxx_Buffer& buffer, long* data, int count)
{
  TAU_PROFILE("hpcxx_unpack()", "void (HPCxx_Buffer, long *, int)", TAU_HPCXX);
  buffer.unpack(data, count);
}

void hpcxx_unpack(HPCxx_Buffer& buffer, u_long* data, int count)
{
  TAU_PROFILE("hpcxx_unpack()", "void (HPCxx_Buffer, u_long *, int)", TAU_HPCXX);
  buffer.unpack(data, count);
}

void hpcxx_unpack(HPCxx_Buffer& buffer, char* data, int count)
{
  TAU_PROFILE("hpcxx_unpack()", "void (HPCxx_Buffer, char *, int)", TAU_HPCXX);
  buffer.unpack(data, count);
}

void hpcxx_unpack(HPCxx_Buffer& buffer, u_char* data, int count)
{
  TAU_PROFILE("hpcxx_unpack()", "void (HPCxx_Buffer, u_char *, int)", TAU_HPCXX);
  buffer.unpack(data, count);
}
