//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_CSem.C,v 1.2 1998/08/07 20:24:50 sameer Exp $
//##########################################################################

// $Id: HPCxx_CSem.C,v 1.2 1998/08/07 20:24:50 sameer Exp $

#include <HPCxx_Profile.h>
#include <HPCxx_CSem.h>


HPCxx_CSem::HPCxx_CSem() : val_(0), counter_(0)
{
}

HPCxx_CSem::HPCxx_CSem(int val) : val_(val), counter_(0)
{
}

HPCxx_CSem::~HPCxx_CSem()
{
}

void HPCxx_CSem::wait()
{
  TAU_PROFILE("HPCxx_CSem::wait()", "void ()", TAU_HPCXX);
  mutex_.lock();
    if (counter_ < val_) cond_.wait(mutex_);
    counter_ = 0;
  mutex_.unlock();
}

void HPCxx_CSem::incr()
{
  TAU_PROFILE("HPCxx_CSem::incr()", "void ()", TAU_HPCXX);
  mutex_.lock();
  counter_++;
  if (counter_ == val_) {
	mutex_.unlock();
	cond_.signal();
  } else {
	mutex_.unlock();
  }
}

void HPCxx_CSem::operator++()
{
  TAU_PROFILE("HPCxx_CSem::operator++()", "void ()", TAU_HPCXX);
  incr();
}

int HPCxx_CSem::getCount()
{
  TAU_PROFILE("HPCxx_CSem::getCount()", "int ()", TAU_HPCXX);
  return(val_);
}

int HPCxx_CSem::setCount(int val)
{
  TAU_PROFILE("HPCxx_CSem::setCount()", "int (int)", TAU_HPCXX);
  int tmp = val_;

  val_ = val;
  return(tmp);
}
