//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_Context.C,v 1.1 1998/06/23 00:55:28 ejohnson Exp $
//##########################################################################

// $Id: HPCxx_Context.C,v 1.1 1998/06/23 00:55:28 ejohnson Exp $

#include <string.h>
#include <HPCxx_Context.h>

HPCxx_Context::HPCxx_Context(HPCxx_ContextID *id,
							 bool isMaster,
							 const char *name)
  : id_(id), isMaster_(isMaster)
{
  if(name) strcpy(name_, name);
}

HPCxx_Context::HPCxx_Context(const HPCxx_Context &context)
{
  id_ = context.id_;
  isMaster_ = context.isMaster_;
  if(context.name_) strcpy(name_, context.name_);
}

HPCxx_Context::~HPCxx_Context()
{
}

// Return the unique ID for this context;
HPCxx_ContextID *HPCxx_Context::getContextID()
{
  return(id_);
}

bool HPCxx_Context::isMasterContext()
{
  return(isMaster_);
}

// Return the number of threads in this context
int HPCxx_Context::getNumThreads()
{
  // FIXME - not yet implemented
  return(0);
}

// Null terminated list of the threads in this node
// The caller of this function must 'delete []' the memory
// caveat - This list may not coincide with the number of
// nodes returned by getNumThreads()
HPCxx_Thread *HPCxx_Context::getThreads()
{
  // FIXME - not yet implemented
  return((HPCxx_Thread *)0);
}


// Return the number of groups of which this context is
// a member.
int HPCxx_Context::getNumGroups()
{
  // FIXME - not yet implemented
  return(0);
}

// Return a list of the groups of which this context is
// a member.
// User must 'delete []' the array
// caveat - This list may not coincide with the number of
// groups returned by getNumGroups()
HPCxx_Group *HPCxx_Context::getGroups()
{
  // FIXME - not yet implemented
  return((HPCxx_Group *)0);
}

// Set the name of the current context
void HPCxx_Context::setName(const char *name)
{
  strcpy(name_, name);
}

// Retuns the name of the current context
const char *HPCxx_Context::getName()
{
  return(name_);
}
