//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_Context.h,v 1.1 1998/06/23 00:55:29 ejohnson Exp $
//##########################################################################

// $Id: HPCxx_Context.h,v 1.1 1998/06/23 00:55:29 ejohnson Exp $

#ifndef __HPCXXCONTEXT_H_
#define __HPCXXCONTEXT_H_

//# include files
#include <hpcxx_types.h>
#include <HPCxx_Thread.h>

class HPCxx_Group;

//#*************************************************************************/
//#CLASS|STRUCT|FUNCTION itemname
// <summary> One-line summary of this class </summary>
// <keywords> container iterator etc </keywords>
//
// <motivation>
//   Why the class was written
// </motivation>
// <prerequisite>
//   <li> list of classes which should be understood before this one,
//   <li> using the form on the next line:
//#   <li> <linkto class=ClassName> ClassName </linkto>
// </prerequisite>
// If the class is templated, add the following list of requirements
// for each template param 'arg':
// <templating arg=T> Arbitrary : What this arg is for
//   <li> Copy constructor (a requirement)
//   <li> Function 'apply' (another requirement)
// </templating>
// Use <templating arg=T> Arbitrary </templating> for args with no reqs.
//
// <synopsis> [ <user> ]
//   Complete description of this class.  All comments should precede the
//   item they are describing, and be in C++-style comments.  If this class
//   is to be used by POOMA users, bracket the text which provides a synopsis
//   of the class for the user in  <user> ... </user> statements.
//   This description can include:
//     regular text, including regular HTML formatting tags like <i> </i>
//     <srcblock> source code </srcblock>
// [ </user> ] </synopsis>
//
// <method name=methodname>
//   Description of this method
// </method
// <method name=secondname>
//   Another description
// </method>
//
// <inheritance>
//   WHY this class is inherited from what it does
// </inheritance>
// <limitations>
//   <li> Know limitations in this class, if any.  If none, congratulations
//   <li> you can skip this section.
// </limitations>
// <bugs>
//   <li> Known or suspected bugs, one per <li>
// </bugs>
// <todo asof="MM/DD/YYYY">
//   <li> item 1
// </todo>
//
//#*************************************************************************/
class HPCxx_Context {
private:
  HPCxx_ContextID  *id_;
  bool              isMaster_;
  char              name_[HPCXX_MAX_NAME_LEN];
  
public:
  HPCxx_Context(HPCxx_ContextID *id = (HPCxx_ContextID*) 0, 
				bool isMaster = false, 
				const char *name = (const char*) 0);
  HPCxx_Context(const HPCxx_Context &context);

  ~HPCxx_Context(void);

  // Return the unique ID for this context;
  HPCxx_ContextID *getContextID(void);

  bool isMasterContext(void);

  // Return the number of threads in this context
  int getNumThreads(void);

  // Null terminated list of the threads in this node
  // The caller of this function must 'delete []' the memory
  // caveat - This list may not coincide with the number of
  // nodes returned by getNumThreads()
  HPCxx_Thread *getThreads();

  // Return the number of groups of which this context is a member.
  int getNumGroups(void);

  // Return a list of the groups of which this context is a member.
  // User must 'delete []' the array
  // caveat - This list may not coincide with the number of
  // groups returned by getNumGroups()
  HPCxx_Group *getGroups(void);

  // Set the name of the current context
  void setName(const char *name);

  // Returns the name of the current context
  const char *getName(void);
};

#endif
