//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_GlobalPtr.cc,v 1.4 1998/08/08 00:04:44 rasmussn Exp $
//##########################################################################

#include <HPCxx_GlobalPtr.h>

template<class T>
HPCxx_GlobalPtr<T>::HPCxx_GlobalPtr()
  : globalRef_()
{
}

template<class T> HPCxx_GlobalPtr<T>::HPCxx_GlobalPtr(T *ptr)
  : globalRef_(ptr)
{
}

template<class T> HPCxx_GlobalPtr<T>::HPCxx_GlobalPtr(T *ptr, HPCxx_ContextID *id)
  : globalRef_(ptr, id)
{
}

template<class T> HPCxx_GlobalPtr<T>::HPCxx_GlobalPtr(const HPCxx_GlobalPtr<T> &gp)
  : globalRef_(gp.globalRef_)
{
}

template<class T> HPCxx_GlobalPtr<T>::~HPCxx_GlobalPtr()
{
}

// Deref a GlobalPtr to a GlobalRefImpl
template<class T> HPCxx_GlobalRefImpl<T>& HPCxx_GlobalPtr<T>::operator*()
{
  return(globalRef_);
}

/*
template<class T> const HPCxx_GlobalRefImpl<T>& HPCxx_GlobalPtr<T>::operator*() const
{
  return(globalRef_);
}
*/

template<class T> HPCxx_GlobalPtr<T>::operator T *() const
{
  return((T *)globalRef_.getPointer());
} 

template<class T> HPCxx_GlobalPtr<T>::operator HPCxx_ContextID &() const
{
  return((HPCxx_ContextID &) globalRef_);
}

template<class T> HPCxx_ContextID * HPCxx_GlobalPtr<T>::getContextID() const
{
  return((HPCxx_ContextID *) globalRef_.contextID_);
}

template<class T> HPCxx_GlobalPtr<T>& HPCxx_GlobalPtr<T>::operator=(const HPCxx_GlobalPtr<T> &gp)
{
  //if (*this == gp) return(*this); there is a problem with == .
  globalRef_ = gp.globalRef_;
  return(*this);
}

template<class T> HPCxx_GlobalPtr<T>& HPCxx_GlobalPtr<T>::operator=(const T *ptr)
{
  //printf("big undone bit here\n");
  return(*this);
}

// Test the equality of two GlobalPtr's
template<class T> int HPCxx_GlobalPtr<T>::operator==(const HPCxx_GlobalPtr<T>& gp)
{
  return(globalRef_ == gp.globalRef_);
}

template<class T> int HPCxx_GlobalPtr<T>::operator==(const HPCxx_GlobalPtr<T>& gp) const
{
  return(globalRef_ == gp.globalRef_);
}

template<class T> int HPCxx_GlobalPtr<T>::operator!=(const HPCxx_GlobalPtr<T>& gp)
{
  return(globalRef_ != gp.globalRef_);
}

template<class T> int HPCxx_GlobalPtr<T>::operator!=(const HPCxx_GlobalPtr<T>& gp) const
{
  return(globalRef_ != gp.globalRef_);
}

// GlobalPtr Arithmetic
template<class T> HPCxx_GlobalPtr<T>& HPCxx_GlobalPtr<T>::operator++() // prefix ++
{
  ++globalRef_;
  return *this;
}

template<class T> HPCxx_GlobalPtr<T> HPCxx_GlobalPtr<T>::operator++(int) // postfix ++
{
  HPCxx_GlobalPtr<T> tmp(*this);
  ++globalRef_;
  return tmp; 
}

template<class T> HPCxx_GlobalPtr<T>& HPCxx_GlobalPtr<T>::operator--() // prefix --
{
  --globalRef_;
  return *this;
}         

template<class T> HPCxx_GlobalPtr<T> HPCxx_GlobalPtr<T>::operator--(int) // postfix --
{
  HPCxx_GlobalPtr<T> tmp(*this);
  --globalRef_;
  return tmp;  
} 

template<class T> HPCxx_GlobalPtr<T>& HPCxx_GlobalPtr<T>::operator+=(int inc)
{
  globalRef_ += inc;
  return *this;
}

template<class T> HPCxx_GlobalPtr<T>& HPCxx_GlobalPtr<T>::operator-=(int dec)
{
  globalRef_ -= dec;
  return *this;
}

template<class T> HPCxx_GlobalPtr<T> HPCxx_GlobalPtr<T>::operator+(int inc)
{
  HPCxx_GlobalPtr<T> tmp(*this);
  tmp+=inc;
  return tmp;  
}

template<class T> HPCxx_GlobalPtr<T> HPCxx_GlobalPtr<T>::operator-(int dec)
{
  HPCxx_GlobalPtr<T> tmp(*this);
  tmp-=dec;
  return tmp;  
}
  
template<class T> int HPCxx_GlobalPtr<T>::operator<(const HPCxx_GlobalPtr<T>& gp)
{
  return(globalRef_ < gp.globalRef_);
}

template<class T> int HPCxx_GlobalPtr<T>::operator<(const HPCxx_GlobalPtr<T>& gp) const
{
  return(globalRef_ < gp.globalRef_);
}

template<class T> int HPCxx_GlobalPtr<T>::operator>(const HPCxx_GlobalPtr<T>& gp)
{
  return(globalRef_ > gp.globalRef_);
}

template<class T> int HPCxx_GlobalPtr<T>::operator>(const HPCxx_GlobalPtr<T>& gp) const
{
  return(globalRef_ > gp.globalRef_);
}

template<class T> int HPCxx_GlobalPtr<T>::operator<=(const HPCxx_GlobalPtr<T>& gp)
{
  return(globalRef_ <= gp.globalRef_);
}

template<class T> int HPCxx_GlobalPtr<T>::operator<=(const HPCxx_GlobalPtr<T>& gp) const
{
  return(globalRef_ <= gp.globalRef_);
}

template<class T> int HPCxx_GlobalPtr<T>::operator>=(const HPCxx_GlobalPtr<T>& gp)
{
  return(globalRef_ >= gp.globalRef_);
}

template<class T> int HPCxx_GlobalPtr<T>::operator>=(const HPCxx_GlobalPtr<T>& gp) const
{
  return(globalRef_ >= gp.globalRef_);
}


//-----------------------------------------------------------------------------
// Returns the type of this class, e.g., HPCXX_TYPE_INT (for integer type)
//-----------------------------------------------------------------------------
template<class T> hpcxx_type_t HPCxx_GlobalPtr<T>::type(void) const
{
fprintf(stderr, "<T>::type, call to default (unknown) type\n");
  return HPCXX_TYPE_UNKNOWN;
}


template<class T> HPCxx_GlobalRefImpl<T> HPCxx_GlobalPtr<T>::operator[](int index)
{
  HPCxx_GlobalRefImpl<T> tmp;
  tmp = globalRef_.incr(index);
  return(tmp);
}

/*
template<class T> const HPCxx_GlobalRefImpl<T> HPCxx_GlobalPtr<T>::operator[](int index) const
{
  HPCxx_GlobalRefImpl<T> tmp;
  tmp = globalRef_.incr(index);
  return(tmp);
}
*/

// Block data transfer
template<class T> void HPCxx_GlobalPtr<T>::read(T *ptr, int count, int offset)
{
  globalRef_.read(ptr, count, offset);
}

template<class T> void HPCxx_GlobalPtr<T>::write(T *ptr, int count, int offset)
{
  globalRef_.write(ptr, count, offset);
}
  
// Test to see if the pointer is local to our context
template<class T> bool HPCxx_GlobalPtr<T>::isLocal() const
{
  // cast to bool needed for AIX
  return((bool)globalRef_.isLocal());
}

// Test to see if the pointer is local to our context
// ptr is a valid reference only if it is local
template<class T> bool HPCxx_GlobalPtr<T>::isLocal(T *&ptr)
{
  // cast to bool needed for AIX
  return((bool)globalRef_.isLocal(ptr));
}
