//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_Group.h,v 1.2 1998/07/02 16:07:20 rasmussn Exp $
//##########################################################################

// $Id: HPCxx_Group.h,v 1.2 1998/07/02 16:07:20 rasmussn Exp $

#ifndef __HPCXX_GROUP_H_
#define __HPCXX_GROUP_H_

#include <unistd.h>			// for NULL
#include <HPCxx_Context.h>
#include <hpcxx_types.h>

//#*************************************************************************/
//#CLASS|STRUCT|FUNCTION itemname
// <summary> One-line summary of this class </summary>
// <keywords> container iterator etc </keywords>
//
// <motivation>
//   Why the class was written
// </motivation>
// <prerequisite>
//   <li> list of classes which should be understood before this one,
//   <li> using the form on the next line:
//#   <li> <linkto class=ClassName> ClassName </linkto>
// </prerequisite>
// If the class is templated, add the following list of requirements
// for each template param 'arg':
// <templating arg=T> Arbitrary : What this arg is for
//   <li> Copy constructor (a requirement)
//   <li> Function 'apply' (another requirement)
// </templating>
// Use <templating arg=T> Arbitrary </templating> for args with no reqs.
//
// <synopsis> [ <user> ]
//   Complete description of this class.  All comments should precede the
//   item they are describing, and be in C++-style comments.  If this class
//   is to be used by POOMA users, bracket the text which provides a synopsis
//   of the class for the user in  <user> ... </user> statements.
//   This description can include:
//     regular text, including regular HTML formatting tags like <i> </i>
//     <srcblock> source code </srcblock>
// [ </user> ] </synopsis>
//
// <method name=methodname>
//   Description of this method
// </method
// <method name=secondname>
//   Another description
// </method>
//
// <inheritance>
//   WHY this class is inherited from what it does
// </inheritance>
// <limitations>
//   <li> Know limitations in this class, if any.  If none, congratulations
//   <li> you can skip this section.
// </limitations>
// <bugs>
//   <li> Known or suspected bugs, one per <li>
// </bugs>
// <todo asof="MM/DD/YYYY">
//   <li> item 1
// </todo>
//
//#*************************************************************************/
class HPCxx_Group {
private:
  HPCxx_ContextID  **contextIDs_;
  int               numContextIDs_;
  int               numThreads_;
  hpcxx_id_t        id_;
  hpcxx_id_t        local_index_;
  int               isMasterGroup_;
  char              name_[HPCXX_MAX_NAME_LEN];

public:
  // Create a new group with an id obtained by hpcxx_getGlobalID()
  HPCxx_Group(hpcxx_id_t groupID = HPCXX_GEN_LOCAL_GROUPID,
			  const char *name = NULL);

  // Create a group for the context id and the passed in group id
  HPCxx_Group(HPCxx_ContextID *&id,
			  hpcxx_id_t groupID = HPCXX_GEN_LOCAL_GROUPID,
			  const char *name = NULL);

  // Create a group with a list of context ids and group id
  HPCxx_Group(HPCxx_ContextID **id, 
			  int count,
			  hpcxx_id_t groupID = HPCXX_GEN_LOCAL_GROUPID,
			  const char *name = NULL);

  ~HPCxx_Group();

  // Obtain the unique identifier for the group
  hpcxx_id_t &getGroupID();

  // Get the number of contexts that are participating
  // in this group
  int getNumContexts();


  // Return an ordered array of context IDs in
  // this group.  This array is identical for every member
  // of the group.
  // Caller must 'delete ' the memory for the GlobalPtrs 
  // caveat - getNumContexts may not equal the number of
  // contexts in this array
  HPCxx_ContextID **getContextIDs();

  // Return the context id for zero-based context <n> where
  // <n> is less than the current number of contexts
  HPCxx_ContextID *getContextID(int context);

  int getLocalIndex(){ return local_index_;}
  void setLocalIndex(int i){ local_index_ = i; }
  // Set the number of threads for this group in *this*
  // context.
  void setNumThreads(int count);

  // Return the number of threads for this group in *this* context
  int getNumThreads();

  // Set the name of the current group
  void setName(const char *name);

  // Retuns the name of the current group
  const char *getName();
 
};

#endif



