//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_Node.C,v 1.2 1998/08/07 20:24:54 sameer Exp $
//##########################################################################

// $Id: HPCxx_Node.C,v 1.2 1998/08/07 20:24:54 sameer Exp $

#include <HPCxx_Profile.h>
#include <string.h>
#include <HPCxx_Node.h>

HPCxx_Node::HPCxx_Node(const char *name)
{
  strcpy(name_, name);
}

HPCxx_Node::HPCxx_Node(const HPCxx_Node &node)
  : impl_(node.impl_)
{
  strcpy(name_, node.name_);
}

HPCxx_Node::~HPCxx_Node()
{
}

// Return true if the context is local to this node
bool HPCxx_Node::contextIsLocal(const HPCxx_ContextID &id)
{
  TAU_PROFILE("HPCxx_Node::contextIsLocal()", "bool (HPCxx_ContextID )", 
    TAU_HPCXX);
  return(impl_.contextIsLocal(id));
}

// Return the number of contexts in this node 
int HPCxx_Node::getNumContexts()
{
  TAU_PROFILE("HPCxx_Node::getNumContexts()", "int ()", TAU_HPCXX);
  return(impl_.getNumContexts());
}

// Null terminated list of the contexts in this node
// The caller of this function must 'delete []' the memory
HPCxx_ContextID *HPCxx_Node::getContexts()
{
  TAU_PROFILE("HPCxx_Node::getContexts()", "HPCxx_ContextID * ()", TAU_HPCXX);
  return(impl_.getContexts());
}

// Create a new context and add it to this node
int HPCxx_Node::addContext()
{
  TAU_PROFILE("HPCxx_Node::addContext()", "int ()", TAU_HPCXX);
  return(impl_.addContext());
}

// Create a new context and run the specified executable
int HPCxx_Node::addContext(const char *exePath)
{
  TAU_PROFILE("HPCxx_Node::addContext()", "int (char *)", TAU_HPCXX);
  return(impl_.addContext(exePath));
}

// Set the name of the current node
void HPCxx_Node::setName(const char *name)
{
  TAU_PROFILE("HPCxx_Node::setName()", "void (char *)", TAU_HPCXX);
  strcpy(name_, name);
}

// Retuns the name of the current node
const char *HPCxx_Node::getName()
{
  TAU_PROFILE("HPCxx_Node::getName()", "char * ()", TAU_HPCXX);
  return(name_);
}
