//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_Node.h,v 1.2 1998/07/02 16:05:30 rasmussn Exp $
//##########################################################################

#ifndef __HPCXX_NODE_H_
#define __HPCXX_NODE_H_

//# include files
#include <hpcxx_types.h>
#include <HPCxx_Context.h>
#include <HPCxx_NodeImpl.h>

//#*************************************************************************/
//#CLASS|STRUCT|FUNCTION itemname
// <summary> One-line summary of this class </summary>
// <keywords> container iterator etc </keywords>
//
// <motivation>
//   Why the class was written
// </motivation>
// <prerequisite>
//   <li> list of classes which should be understood before this one,
//   <li> using the form on the next line:
//#   <li> <linkto class=ClassName> ClassName </linkto>
// </prerequisite>
// If the class is templated, add the following list of requirements
// for each template param 'arg':
// <templating arg=T> Arbitrary : What this arg is for
//   <li> Copy constructor (a requirement)
//   <li> Function 'apply' (another requirement)
// </templating>
// Use <templating arg=T> Arbitrary </templating> for args with no reqs.
//
// <synopsis> [ <user> ]
//   Complete description of this class.  All comments should precede the
//   item they are describing, and be in C++-style comments.  If this class
//   is to be used by POOMA users, bracket the text which provides a synopsis
//   of the class for the user in  <user> ... </user> statements.
//   This description can include:
//     regular text, including regular HTML formatting tags like <i> </i>
//     <srcblock> source code </srcblock>
// [ </user> ] </synopsis>
//
// <method name=methodname>
//   Description of this method
// </method
// <method name=secondname>
//   Another description
// </method>
//
// <inheritance>
//   WHY this class is inherited from what it does
// </inheritance>
// <limitations>
//   <li> Know limitations in this class, if any.  If none, congratulations
//   <li> you can skip this section.
// </limitations>
// <bugs>
//   <li> Known or suspected bugs, one per <li>
// </bugs>
// <todo asof="MM/DD/YYYY">
//   <li> item 1
// </todo>
//
//#*************************************************************************/
class HPCxx_Node {
private:
  HPCxx_NodeImpl impl_;
  char           name_[HPCXX_MAX_NAME_LEN];

public:

  HPCxx_Node(const char *name = NULL);
  HPCxx_Node(const HPCxx_Node &node);
  ~HPCxx_Node();

  // Return true if the context is local to this node
  bool contextIsLocal(const HPCxx_ContextID &id);

  // Return the number of contexts in this node 
  int getNumContexts();

  // Null terminated list of the contexts in this node
  // The caller of this function must 'delete []' the memory
  // caveat - This list may not coincide with the number of
  // contexts returned by getNumGroups()
  HPCxx_ContextID *getContexts();

  // Create a new context and add it to this node
  // FIXME - currently not implemented
  int addContext();

  // Create a new context and run the specified executable
  // FIXME - currently not implemented
  int addContext(const char *exePath);

  // Set the name of the current node
  void setName(const char *name);

  // Retuns the name of the current node
  const char *getName();
};

#endif
