//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_RFITable.h,v 1.1 1998/06/23 00:55:42 ejohnson Exp $
//##########################################################################

#ifndef __HPCXX_RFITABLE_H_
#define __HPCXX_RFITABLE_H_

#include <HPCxx_Buffer.h>
#include <iostream.h>

// global remote function table, instantiated in hpcxx_register.C
class HPCxx_RFITable_;
extern HPCxx_RFITable_ gHPCxx_RFITable;

//
// Base class to support registration and to provide
// the sentinel node in the registration table
//
class HPCxx_RFITableEntry {
public:
  HPCxx_RFITableEntry *next;
  hpcxx_id_t           id;

  HPCxx_RFITableEntry() : next(NULL), id(HPCXX_INVALID_ID) {}
  virtual void execute(void *, void *, HPCxx_Buffer &){
     cout << "execute: wrong member called!" << endl;
     }
  virtual void returnValue(void * , HPCxx_Buffer &){}
  virtual void read(void *, void *, int , int ){}
  virtual void write(void *, void *, int , int , HPCxx_Buffer &){}
};

//
// The Remote Function Table used to store the pointers to
// global and member fuctions.  There should only ever be
// one of these instantiated.
//
class HPCxx_RFITable_ {
  HPCxx_RFITableEntry *head;
  HPCxx_RFITableEntry *tail;

public:
  HPCxx_RFITable_() {
	head = tail = new HPCxx_RFITableEntry;
  }

  ~HPCxx_RFITable_() {
	HPCxx_RFITableEntry* tmp = head;
	while (tmp != NULL) {
	  HPCxx_RFITableEntry* next = tmp->next;
	  delete tmp;
	  tmp = next;
	}
  }

  hpcxx_id_t addEntry(HPCxx_RFITableEntry *entry, hpcxx_id_t id) {
	// id's may not be duplicated, if an attempt is made to
	// register an id more than once HPCXX_INVALID_ID is returned

	HPCxx_RFITableEntry *tmp = getEntry(id);
	if (tmp != NULL) return(HPCXX_INVALID_ID);
        entry->id = id;
        entry->next = head;
	head = entry;
	return(id);
  }

  HPCxx_RFITableEntry *getEntry(int id) {
	HPCxx_RFITableEntry *tmp = head;
	while (tmp != tail) {
	  if (tmp->id == id) return(tmp);
	  tmp = tmp->next;
	}
	return((HPCxx_RFITableEntry *)0);
  }
};

#endif	// __HPCXX_RFITABLE_H_

