//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_Runnable.h,v 1.1 1998/06/23 00:55:44 ejohnson Exp $
//##########################################################################

// $Id: HPCxx_Runnable.h,v 1.1 1998/06/23 00:55:44 ejohnson Exp $

#ifndef __HPCXX_RUNNABLE_H_
#define __HPCXX_RUNNABLE_H_

//#----------------------------------------------------------------------------
// <summary>Interface for an object that is runnable by a thread.</summary>
//
// One way of creating threads is to inherit from the HPCxx_Thread class 
// and override the run method to specify work to do.  Another way is to
// create a class which mixes in the HPCxx_Runnable interface, thus providing
// a run method for the thread to call.  In this later case, an instance of
// a thread is created by passing the runnable object to the constructor of the
// HPCxx_Thread class.
//
// As an example, to create a runnable class one needs to implement the
// run method as shown below:
// <srcblock>
// class RunnableObject : public HPCxx_Runnable {
//     void run(void)
//     {
//         // do something useful
//     }
// };
// </srcblock>
// Then a thread can be created and run by the following:
// <srcblock>
// RunnableObject runnable;
// HPCxx_Thread thread(&runnable);
// thread.start();
// </srcblock>
// <prerequisite>
//   <li> <a href="HPCxx_Thread.html">HPCxx_Thread</a>
// </prerequisite>
// <hr>
//#----------------------------------------------------------------------------
class HPCxx_Runnable {
public:
  //#--------------------------------------------------------------------------
  // Pure virtual function that must be overridden to provide work for the
  // thread to do.  This function is called by its thread when the thread is
  // started.
  //#--------------------------------------------------------------------------
  virtual void run(void) = 0;
};

#endif
