//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_Sync.cc,v 1.1 1998/06/23 00:55:45 ejohnson Exp $
//##########################################################################

#include <HPCxx_Sync.h>

template<class T> HPCxx_Sync<T>::HPCxx_Sync(void) : virgin_(1)
{
  buffer_= (void *) &val_;
}

template<class T> HPCxx_Sync<T>::~HPCxx_Sync(void)
{
}

template<class T> void HPCxx_Sync<T>::setBufferAddr(void * p)
{
   buffer_= p;
}

template<class T> void * HPCxx_Sync<T>::getBufferAddr(void)
{
   return(buffer_);
}

template<class T> T& HPCxx_Sync<T>::operator=(T& val)
{
  write(val);
  return(val);
}

template<class T> HPCxx_Sync<T>::operator T ()
{
  T val;
  
  read(val);
  return(val);
}
  
template<class T> bool HPCxx_Sync<T>::peek(T &val)
{
  bool status = (bool)0;
  mutex_.lock();
    if (!virgin_) {
	  val = val_;
	  status = (bool)1;
	}
  mutex_.unlock();
  return(status);
}

template<class T> void HPCxx_Sync<T>::read(T &val)
{
  mutex_.lock();

  // While should be used here, because, according to the pthreads standard,
  // wait may awaken prematurely.
  while (virgin_) cond_.wait(mutex_);
  val = val_;
  mutex_.unlock();
}

template<class T> void HPCxx_Sync<T>::write(T& val)
{
  mutex_.lock();
    if (virgin_) {
	  virgin_ = 0;
	  val_ = val;
	  cond_.broadcast();
	}
  mutex_.unlock();
}
