//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_SyncQ.cc,v 1.1 1998/06/23 00:55:48 ejohnson Exp $
//##########################################################################

#include <HPCxx_SyncQ.h>

template<class T> HPCxx_SyncQ<T>::HPCxx_SyncQ(void) : count_(0)
{
  head_ = tail_ = NULL;
}

template<class T> HPCxx_SyncQ<T>::~HPCxx_SyncQ(void)
{
  HPCxx_SyncCell<T> *tmp;

  while (head_ != NULL) {
	tmp = head_;
	head_ = head_->next_;
	delete tmp;
  }
}

template<class T> void HPCxx_SyncQ<T>::read(T& data)
{
  HPCxx_SyncCell<T> *cell;

  mutex_.lock();
    count_--;	
    if (head_ == NULL) cond_.wait(mutex_);
	cell  = head_;
        head_ = head_->next_;
	data  = cell->data_;
        if (head_ == NULL) tail_ = NULL;
	delete cell;
	count_++;
  mutex_.unlock();
}
 
template<class T> void HPCxx_SyncQ<T>::write(T& data)
{
  HPCxx_SyncCell<T> *cell;
  mutex_.lock();
        cell = new HPCxx_SyncCell<T>(data);
        if (tail_) tail_->next_ = cell;
        tail_ = cell;
	cell->next_ = NULL;
        if (head_ == NULL) head_ = tail_;
	if (count_ < 0) cond_.signal();
  mutex_.unlock();
}
  
// Read a value from the queue;  equiv read()
template<class T> HPCxx_SyncQ<T>::operator T(void)
{
  T data;
  read(data);
  return(data);
}
  
// Write *to* the queue, not copy it; equiv write()
template<class T> T& HPCxx_SyncQ<T>::operator= (T& data)
{
  write(data);
  return(data);
}

template<class T> int HPCxx_SyncQ<T>::length(void)
{
  int count;

  mutex_.lock();
    count = count_;
  mutex_.unlock();

  return(count);
}

template<class T> bool HPCxx_SyncQ<T>::peek(T& data)
{
  bool status = (bool)0; // assume the queue is empty

  mutex_.lock();
    if (head_ != NULL) {
	  data = head_->data_;
	  status = (bool)1;
	}
  mutex_.unlock();
  return(status);
}

//
// Formally was blocking "peek"
//  -- not sure about this... dbg.
template<class T> void HPCxx_SyncQ<T>::waitAndCopy(T& data)
{
  mutex_.lock();
  if (head_ == NULL) cond_.wait(mutex_);
  else data = tail_->data_;
  mutex_.unlock();
}
