//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: HPCxx_Thread.C,v 1.2 1998/07/09 23:48:59 sameer Exp $
//##########################################################################

// $Id: HPCxx_Thread.C,v 1.2 1998/07/09 23:48:59 sameer Exp $

#include <HPCxx_Profile.h>
#include <string.h>
#include <HPCxx_Thread.h>

HPCxx_Thread::HPCxx_Thread(const char* name)
{
  TAU_PROFILE("HPCxx_Thread::HPCxx_Thread()", "void (char *)", TAU_HPCXX);
  if(name) strcpy(name_, name);
}

HPCxx_Thread::HPCxx_Thread(const HPCxx_Thread& thread) : impl_(thread.impl_)
{
  TAU_PROFILE("HPCxx_Thread::HPCxx_Thread()", "void (HPCxx_Thread&)", TAU_HPCXX);
  if(thread.name_) strcpy(name_, thread.name_);
}

HPCxx_Thread::HPCxx_Thread(HPCxx_Runnable* target, const char* name)
  : impl_(target)
{
  TAU_PROFILE("HPCxx_Thread::HPCxx_Thread()", "void (HPCxx_Runnable*, char *)",
    TAU_HPCXX);
  if(name) strcpy(name_, name);
}

HPCxx_Thread::~HPCxx_Thread(void)
{
  TAU_PROFILE("HPCxx_Thread::~HPCxx_Thread()", "void ()", TAU_HPCXX);
}

HPCxx_Thread& HPCxx_Thread::operator=(const HPCxx_Thread& thread)
{
  TAU_PROFILE("HPCxx_Thread::operator=()", "HPCxx_Thread& (HPCxx_Thread& )",
    TAU_HPCXX);
  if (&thread == this) return *this; // self-assignment

  impl_ = thread.impl_;
  if(thread.name_) strcpy(name_, thread.name_);
  return *this;
}

void HPCxx_Thread::run(void)
{
  TAU_PROFILE("HPCxx_Thread::run()", "void ()", TAU_HPCXX);
  impl_.run();
}

void HPCxx_Thread::start(void)
{
  TAU_PROFILE("HPCxx_Thread::start()", "void ()", TAU_HPCXX);
  impl_.start(*this);
}

void HPCxx_Thread::stop(void *status)
{
  TAU_PROFILE("HPCxx_Thread::stop()", "void (void *)", TAU_HPCXX);
  impl_.stop(status);
}

void HPCxx_Thread::yield(void)
{
  TAU_PROFILE("HPCxx_Thread::yield()", "void ()", TAU_HPCXX);
  HPCxx_ThreadImpl::yield();
}

void HPCxx_Thread::sleep(long milliseconds, long nanoseconds)
{
  TAU_PROFILE("HPCxx_Thread::sleep()", "void (long, long)", TAU_HPCXX);
  HPCxx_ThreadImpl::sleep(milliseconds, nanoseconds);
}

void HPCxx_Thread::suspend(void)
{
  TAU_PROFILE("HPCxx_Thread::suspend()", "void ()", TAU_HPCXX);
  impl_.suspend();
}

void HPCxx_Thread::resume(void)
{
  TAU_PROFILE("HPCxx_Thread::resume()", "void ()", TAU_HPCXX);
  impl_.resume();
}

bool HPCxx_Thread::isAlive(void)
{
  TAU_PROFILE("HPCxx_Thread::isAlive()", "bool ()", TAU_HPCXX);
  return(impl_.isAlive());
}

HPCxx_Thread *HPCxx_Thread::currentThread(void)
{
  TAU_PROFILE("HPCxx_Thread::currentThread()", "HPCxx_Thread * ()", TAU_HPCXX);
  return(HPCxx_ThreadImpl::currentThread());
}

int HPCxx_Thread::getThreadID(void)
{
  TAU_PROFILE("HPCxx_Thread::getThreadID()", "int ()", TAU_HPCXX);
  return(HPCxx_ThreadImpl::getThreadID());
}

void HPCxx_Thread::join(long milliseconds, long nanoseconds)
{
  TAU_PROFILE("HPCxx_Thread::join()", "void (long, long)", TAU_HPCXX);
  impl_.join(milliseconds, nanoseconds);
}

void HPCxx_Thread::setName(const char* name)
{
  TAU_PROFILE("HPCxx_Thread::setName()", "void (char *)", TAU_HPCXX);
  if(name) strcpy(name_, name);
}

const char* HPCxx_Thread::getName(void)
{
  TAU_PROFILE("HPCxx_Thread::getName()", "char * ()", TAU_HPCXX);
  return(name_);
}

int HPCxx_Thread::getPriority(void)
{
  TAU_PROFILE("HPCxx_Thread::getPriority()", "int ()", TAU_HPCXX);
  return(impl_.getPriority());
}

int HPCxx_Thread::setPriority(int priority)
{
  TAU_PROFILE("HPCxx_Thread::setPriority()", "int (int )", TAU_HPCXX);
  return(impl_.setPriority(priority));
}
