//-----------------------------------------------------------------------------
// hpcxx_invoke.cc -- contains template declarations for invoking remote
// member functions
//-----------------------------------------------------------------------------

#include <HPCxx_Sync.h>
#include <HPCxx_SyncQ.h>
#include <HPCxx_GlobalPtr.h>
#include <HPCxx_Buffer.h>
#include <hpcxx_register.h>
#include <iostream.h>

extern int barrierSetUp_id;
extern HPCxx_SyncQ<int> barrier_set_up_lock[2];
extern HPCxx_GlobalPtr<HPCxx_BufferBlock> *barrier_ptr;


//
// Synchronous invocation of remote global functions
//

//---- hpcxx_invoke -----------------------------------------------------------
// Invokes a remote function that takes no args and returns a value.
//-----------------------------------------------------------------------------
template<class R>
int hpcxx_invoke(HPCxx_ContextID& target, R& result, int funcID)
{
  HPCxx_Buffer b(sizeof(int) + sizeof(HPCxx_ContextID) + 1000);

  // allocate a sync var that will be used to wait for the
  // return value from the remote function call 
  //HPCxx_Sync<R> *result_sync = new (HPCXX_NEW) HPCxx_Sync<R>;
  HPCxx_Sync<R> result_sync;

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R > > gp_to_res_sync( &result_sync);
  
  // Build the packet, first pack the global pointer to the result,
  // then the function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  // wait for the result
  result_sync.read(result);

  return(0);
}


//---- hpcxx_invoke -----------------------------------------------------------
// Invokes a remote function that takes one arg and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1>
int hpcxx_invoke(HPCxx_ContextID& target, R& result, int funcID, A1& a1)
{
  HPCxx_Buffer b(sizeof(int) + sizeof(A1) + sizeof(HPCxx_ContextID) + 32);

  // allocate a sync var that will be used to wait for the
  // return value from the remote function call 
  HPCxx_Sync<R> result_sync;

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R > > gp_to_res_sync( &result_sync);
  
  // Build the packet, first pack the global pointer to the result,
  // then the function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  // wait for the result
  result_sync.read(result);

  return(0);
}


//---- hpcxx_invoke -----------------------------------------------------------
// Invokes a remote function that takes two args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2>
int hpcxx_invoke( HPCxx_ContextID& target, R& result, int funcID,
		  A1& a1, A2& a2 )
{
  HPCxx_Buffer b(sizeof(int) + sizeof(A1) + sizeof(A2)
		             + sizeof(HPCxx_ContextID) + 32);

  // allocate a sync var that will be used to wait for the
  // return value from the remote function call 
  //HPCxx_Sync<R> *result_sync = new (HPCXX_NEW) HPCxx_Sync<R>;
  HPCxx_Sync<R> result_sync;

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R > > gp_to_res_sync( &result_sync);
  
  // Build the packet, first pack the global pointer to the result,
  // then the function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  // wait for the result
  result_sync.read(result);

  return(0);
}


//---- hpcxx_invoke -----------------------------------------------------------
// Invokes a remote function that takes three args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3>
int hpcxx_invoke( HPCxx_ContextID& target, R& result, int funcID,
		  A1& a1, A2& a2, A3& a3 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3);
  HPCxx_Buffer b(sizeof(int) + sz + sizeof(HPCxx_ContextID) + 32);

  // allocate a sync var that will be used to wait for the
  // return value from the remote function call 
  //HPCxx_Sync<R> *result_sync = new (HPCXX_NEW) HPCxx_Sync<R>;
  HPCxx_Sync<R> result_sync;

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R > > gp_to_res_sync( &result_sync);
  
  // Build the packet, first pack the global pointer to the result,
  // then the function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  // wait for the result
  result_sync.read(result);

  return(0);
}


//---- hpcxx_invoke -----------------------------------------------------------
// Invokes a remote function that takes four args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4>
int hpcxx_invoke( HPCxx_ContextID& target, R& result, int funcID,
		  A1& a1, A2& a2, A3& a3, A4& a4 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3)+ sizeof(A4);
  HPCxx_Buffer b(sizeof(int) + sz + sizeof(HPCxx_ContextID) + 32);

  // allocate a sync var that will be used to wait for the
  // return value from the remote function call 
  //HPCxx_Sync<R> *result_sync = new (HPCXX_NEW) HPCxx_Sync<R>;
  HPCxx_Sync<R> result_sync;

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R > > gp_to_res_sync( &result_sync);
  
  // Build the packet, first pack the global pointer to the result,
  // then the function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);
  hpcxx_pack(b, &a4, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  // wait for the result
  result_sync.read(result);

  return(0);
}


//---- hpcxx_invoke -----------------------------------------------------------
// Invokes a remote function that takes five args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4, class A5>
int hpcxx_invoke( HPCxx_ContextID& target, R& result, int funcID,
		  A1& a1, A2& a2, A3& a3, A4& a4, A5& a5 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3)+ sizeof(A4)+sizeof(A5);
  HPCxx_Buffer b(sizeof(int) + sz + sizeof(HPCxx_ContextID) + 32);

  // allocate a sync var that will be used to wait for the
  // return value from the remote function call 
  //HPCxx_Sync<R> *result_sync = new (HPCXX_NEW) HPCxx_Sync<R>;
  HPCxx_Sync<R> result_sync;

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R > > gp_to_res_sync( &result_sync);
  
  // Build the packet, first pack the global pointer to the result,
  // then the function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);
  hpcxx_pack(b, &a4, 1);
  hpcxx_pack(b, &a5, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  // wait for the result
  result_sync.read(result);

  return(0);
}


//---- hpcxx_invoke -----------------------------------------------------------
// Invokes a remote function that takes six args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4, class A5, class A6>
int hpcxx_invoke( HPCxx_ContextID& target, R& result, int funcID,
		  A1& a1, A2& a2, A3& a3, A4& a4, A5& a5, A6& a6 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3)+ sizeof(A4) + sizeof(A5)
		      + sizeof(A6);
  HPCxx_Buffer b(sizeof(int) + sz + sizeof(HPCxx_ContextID) + 32);

  // allocate a sync var that will be used to wait for the
  // return value from the remote function call 
  //HPCxx_Sync<R> *result_sync = new (HPCXX_NEW) HPCxx_Sync<R>;
  HPCxx_Sync<R> result_sync;

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R > > gp_to_res_sync( &result_sync);
  
  // Build the packet, first pack the global pointer to the result,
  // then the function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);
  hpcxx_pack(b, &a4, 1);
  hpcxx_pack(b, &a5, 1);
  hpcxx_pack(b, &a6, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  // wait for the result
  result_sync.read(result);

  return(0);
}

//
// Asynchronous invocation of remote global functions
//

//---- hpcxx_ainvoke ----------------------------------------------------------
// Invokes a remote function that takes no args and returns a value.
//-----------------------------------------------------------------------------
template<class R>
int hpcxx_ainvoke(HPCxx_ContextID& target, HPCxx_Sync<R>& result, int funcID)
{
  HPCxx_Buffer b(sizeof(int) + sizeof(HPCxx_ContextID) + 1000);

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R > > gp_to_res_sync( &result);
  
  // Build the packet, first pack the global pointer to the result,
  // then the function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  return(0);
}


//---- hpcxx_ainvoke ----------------------------------------------------------
// Invokes a remote function that takes one arg and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1>
int hpcxx_ainvoke( HPCxx_ContextID& target, HPCxx_Sync<R>& result, int funcID,
		   A1& a1 )
{
  HPCxx_Buffer b(sizeof(int) + sizeof(A1) + sizeof(HPCxx_ContextID) + 32);

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R > > gp_to_res_sync( &result);
  
  // build the packet. first pack the global pointer to the result,
  // then the function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  return(0);
}


//---- hpcxx_ainvoke ----------------------------------------------------------
// Invokes a remote function that takes two args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2>
int hpcxx_ainvoke( HPCxx_ContextID& target, HPCxx_Sync<R>& result, int funcID,
		   A1& a1, A2& a2 )
{
  HPCxx_Buffer b(sizeof(int) + sizeof(A1) + sizeof(A2)
		             + sizeof(HPCxx_ContextID) + 32);
  HPCxx_GlobalPtr<HPCxx_Sync<R > > gp_to_res_sync( &result);
  
  // Build the packet, first pack the global pointer to the result,
  // then the function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  return(0);
}


//---- hpcxx_ainvoke ----------------------------------------------------------
// Invokes a remote function that takes three args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3>
int hpcxx_ainvoke( HPCxx_ContextID& target, HPCxx_Sync<R>& result, int funcID,
		   A1& a1, A2& a2, A3& a3 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3);
  HPCxx_Buffer b(sizeof(int) + sz + sizeof(HPCxx_ContextID) + 32);

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R > > gp_to_res_sync( &result);
  
  // Build the packet, first pack the global pointer to the result,
  // then the function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  return(0);
}


//---- hpcxx_ainvoke ----------------------------------------------------------
// Invokes a remote function that takes four args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4>
int hpcxx_ainvoke( HPCxx_ContextID& target, HPCxx_Sync<R>& result, int funcID,
		   A1& a1, A2& a2, A3& a3, A4& a4 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3) + sizeof(A4);
  HPCxx_Buffer b(sizeof(int) + sz + sizeof(HPCxx_ContextID) + 32);

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R > > gp_to_res_sync( &result);
  
  // Build the packet, first pack the global pointer to the result,
  // then the function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);
  hpcxx_pack(b, &a4, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  return(0);
}


//---- hpcxx_ainvoke ----------------------------------------------------------
// Invokes a remote function that takes five args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4, class A5>
int hpcxx_ainvoke( HPCxx_ContextID& target, HPCxx_Sync<R>& result, int funcID,
		   A1& a1, A2& a2, A3& a3, A4& a4, A5& a5 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3) + sizeof(A4) + sizeof(A5);
  HPCxx_Buffer b(sizeof(int) + sz + sizeof(HPCxx_ContextID) + 32);

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R > > gp_to_res_sync( &result);
  
  // Build the packet, first pack the global pointer to the result,
  // then the function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);
  hpcxx_pack(b, &a4, 1);
  hpcxx_pack(b, &a5, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  return(0);
}


//---- hpcxx_ainvoke ----------------------------------------------------------
// Invokes a remote function that takes six args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4, class A5, class A6>
int hpcxx_ainvoke( HPCxx_ContextID& target, HPCxx_Sync<R>& result, int funcID,
		   A1& a1, A2& a2, A3& a3, A4& a4, A5& a5, A6& a6 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3) + sizeof(A4) + sizeof(A5)
		      + sizeof(A6);
  HPCxx_Buffer b(sizeof(int) + sz + sizeof(HPCxx_ContextID) + 32);

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R > > gp_to_res_sync( &result);
  
  // build the packet. first pack the global pointer to the result,
  // then the function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);
  hpcxx_pack(b, &a4, 1);
  hpcxx_pack(b, &a5, 1);
  hpcxx_pack(b, &a6, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  return(0);
}


//
// Synchronous invocation of remote member functions
//

//---- hpcxx_minvoke ----------------------------------------------------------
// Invokes a remote member function that takes no args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R>
int hpcxx_minvoke(HPCxx_GlobalPtr<C>* gp, R& result, int funcID)
{
  HPCxx_Buffer b(sizeof(int) + 2*sizeof(HPCxx_ContextID) + 32);
  HPCxx_ContextID& target = *(gp->getContextID());

  // allocate a sync var that will be used to wait for the
  // return value from the remote function call 
  //HPCxx_Sync<R> *result_sync = new (HPCXX_NEW) HPCxx_Sync<R>;
  HPCxx_Sync<R> result_sync;
  
  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R> > gp_to_res_sync( &result_sync);
  
  // Build the packet; first pack the global pointer to the result, then the
  // gp to the object then the member function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  
  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 
  
  // wait for the result
  result_sync.read(result);

  return(0);
}


//---- hpcxx_minvoke ----------------------------------------------------------
// Invokes a remote member function that takes one arg and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1>
int hpcxx_minvoke(HPCxx_GlobalPtr<C>* gp, R& result, int funcID, A1& a1)
{
  int sz = sizeof(A1);
  HPCxx_Buffer b(sizeof(int) + sz + 2*sizeof(HPCxx_ContextID) + 32);

  HPCxx_ContextID& target = *(gp->getContextID());

  // allocate a sync var that will be used to wait for the
  // return value from the remote function call 

  HPCxx_Sync<R> result_sync;

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R> > gp_to_res_sync( &result_sync);

  // Build the packet; first pack the global pointer to the result, then the
  // gp to the object then the member function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  // wait for the result
  result_sync.read(result);

  return(0);
}


//---- hpcxx_minvoke ----------------------------------------------------------
// Invokes a remote member function that takes two args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2>
int hpcxx_minvoke( HPCxx_GlobalPtr<C>* gp, R& result, int funcID,
		   A1& a1, A2& a2 )
{
  int sz = sizeof(A1) + sizeof(A2);
  HPCxx_Buffer b(sizeof(int) + sz + 2*sizeof(HPCxx_ContextID) + 32);

  HPCxx_ContextID& target = *(gp->getContextID());

  // allocate a sync var that will be used to wait for the
  // return value from the remote function call 
  HPCxx_Sync<R> result_sync ;

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R> > gp_to_res_sync( &result_sync);

  // Build the packet; first pack the global pointer to the result, then the
  // gp to the object then the member function id, and then the args.

  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  // wait for the result
  result_sync.read(result);

  return(0);
}


//---- hpcxx_minvoke ----------------------------------------------------------
// Invokes a remote member function that takes three args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2, class A3>
int hpcxx_minvoke( HPCxx_GlobalPtr<C>* gp, R& result, int funcID,
		   A1& a1, A2& a2, A3& a3 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3);
  HPCxx_Buffer b(sizeof(int) + sz + 2*sizeof(HPCxx_ContextID) + 32);

  HPCxx_ContextID& target = *(gp->getContextID());

  // allocate a sync var that will be used to wait for the
  // return value from the remote function call 
  HPCxx_Sync<R> result_sync ;

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R> > gp_to_res_sync( &result_sync);

  // Build the packet; first pack the global pointer to the result, then the
  // gp to the object then the member function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  // wait for the result
  result_sync.read(result);

  return(0);
}


//---- hpcxx_minvoke ----------------------------------------------------------
// Invokes a remote member function that takes four args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2, class A3, class A4>
int hpcxx_minvoke( HPCxx_GlobalPtr<C>* gp, R& result, int funcID,
		   A1& a1, A2& a2, A3& a3, A4& a4 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3) + sizeof(A4);
  HPCxx_Buffer b(sizeof(int) + sz + 2*sizeof(HPCxx_ContextID) + 32);

  HPCxx_ContextID& target = *(gp->getContextID());

  // allocate a sync var that will be used to wait for the
  // return value from the remote function call 
  HPCxx_Sync<R> result_sync ;

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R> > gp_to_res_sync( &result_sync);

  // Build the packet; first pack the global pointer to the result, then the
  // gp to the object then the member function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);
  hpcxx_pack(b, &a4, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  // wait for the result
  result_sync.read(result);

   return(0);
}


//---- hpcxx_minvoke ----------------------------------------------------------
// Invokes a remote member function that takes five args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2, class A3, class A4, class A5>
int hpcxx_minvoke( HPCxx_GlobalPtr<C>* gp, R& result, int funcID,
		   A1& a1, A2& a2, A3& a3, A4& a4, A5& a5 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3) + sizeof(A4) + sizeof(A5);
  HPCxx_Buffer b(sizeof(int) + sz + 2*sizeof(HPCxx_ContextID) + 32);

  HPCxx_ContextID& target = *(gp->getContextID());

  // allocate a sync var that will be used to wait for the
  // return value from the remote function call 
  HPCxx_Sync<R> result_sync;

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R> > gp_to_res_sync( &result_sync);

  // Build the packet; first pack the global pointer to the result, then the
  // gp to the object then the member function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);
  hpcxx_pack(b, &a4, 1);
  hpcxx_pack(b, &a5, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  // wait for the result
  result_sync.read(result);

  return(0);
}


//---- hpcxx_minvoke ----------------------------------------------------------
// Invokes a remote member function that takes six args and returns a value.
//-----------------------------------------------------------------------------
template< class C, class R, class A1, class A2, class A3, class A4, class A5,
          class A6 >
int hpcxx_minvoke( HPCxx_GlobalPtr<C>* gp, R& result, int funcID,
		   A1& a1, A2& a2, A3& a3, A4& a4, A5& a5, A6& a6 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3) + sizeof(A4) + sizeof(A5)
                      + sizeof(A6);
  HPCxx_Buffer b(sizeof(int) + sz + 2*sizeof(HPCxx_ContextID) + 32);

  HPCxx_ContextID& target = *(gp->getContextID());

  // allocate a sync var that will be used to wait for the
  // return value from the remote function call 
  HPCxx_Sync<R> result_sync ;

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R> > gp_to_res_sync(&result_sync);

  // Build the packet; first pack the global pointer to the result, then the
  // gp to the object then the member function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);
  hpcxx_pack(b, &a4, 1);
  hpcxx_pack(b, &a5, 1);
  hpcxx_pack(b, &a6, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  // wait for the result
  result_sync.read(result);

  return(0);
}


//
// Asynchronous invocation of remote member functions
//

//---- hpcxx_aminvoke ---------------------------------------------------------
// Invokes a remote member function that takes no args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R>
int hpcxx_aminvoke(HPCxx_GlobalPtr<C>* gp, HPCxx_Sync<R>& result, int funcID)
{
  HPCxx_Buffer b(sizeof(int) + 2*sizeof(HPCxx_ContextID) + 32);

  HPCxx_ContextID& target = *(gp->getContextID());

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R> > gp_to_res_sync(&result);

  // Build the packet; first pack the global pointer to the result, then the
  // gp to the object then the member function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  return(0);
}


//---- hpcxx_aminvoke ---------------------------------------------------------
// Invokes a remote member function that takes one arg and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1>
int hpcxx_aminvoke( HPCxx_GlobalPtr<C>* gp, HPCxx_Sync<R>& result, int funcID,
		    A1& a1 )
{
  int sz = sizeof(A1);
  HPCxx_Buffer b(sizeof(int) + sz + 2*sizeof(HPCxx_ContextID) + 32);

  HPCxx_ContextID& target = *(gp->getContextID());

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R> > gp_to_res_sync(&result);

  // Build the packet; first pack the global pointer to the result, then the
  // gp to the object then the member function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  return(0);
}


//---- hpcxx_aminvoke ---------------------------------------------------------
// Invokes a remote member function that takes two args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2>
int hpcxx_aminvoke( HPCxx_GlobalPtr<C>* gp, HPCxx_Sync<R>& result, int funcID,
		    A1& a1, A2& a2 )
{
  int sz = sizeof(A1) + sizeof(A2);
  HPCxx_Buffer b(sizeof(int) + sz + 2*sizeof(HPCxx_ContextID) + 32);

  HPCxx_ContextID& target = *(gp->getContextID());

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R> > gp_to_res_sync( &result);

  // Build the packet; first pack the global pointer to the result, then the
  // gp to the object then the member function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  return(0);
}


//---- hpcxx_aminvoke ---------------------------------------------------------
// Invokes a remote member function that takes three args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2, class A3>
int hpcxx_aminvoke( HPCxx_GlobalPtr<C>* gp, HPCxx_Sync<R>& result, int funcID,
		    A1& a1, A2& a2, A3& a3 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3);
  HPCxx_Buffer b(sizeof(int) + sz + 2*sizeof(HPCxx_ContextID) + 32);

  HPCxx_ContextID& target = *(gp->getContextID());

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R> > gp_to_res_sync(&result);

  // Build the packet; first pack the global pointer to the result, then the
  // gp to the object then the member function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  return(0);
}


//---- hpcxx_aminvoke ---------------------------------------------------------
// Invokes a remote member function that takes four args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2, class A3, class A4>
int hpcxx_aminvoke( HPCxx_GlobalPtr<C>* gp, HPCxx_Sync<R>& result, int funcID,
		    A1& a1, A2& a2, A3& a3, A4& a4 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3) + sizeof(A4);
  HPCxx_Buffer b(sizeof(int) + sz + 2*sizeof(HPCxx_ContextID) + 32);

  HPCxx_ContextID& target = *(gp->getContextID());

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R> > gp_to_res_sync(&result);

  // Build the packet; first pack the global pointer to the result, then the
  // gp to the object then the member function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);
  hpcxx_pack(b, &a4, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

   return(0);
}


//---- hpcxx_aminvoke ---------------------------------------------------------
// Invokes a remote member function that takes five args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2, class A3, class A4, class A5>
int hpcxx_aminvoke( HPCxx_GlobalPtr<C>* gp, HPCxx_Sync<R>& result, int funcID,
		    A1& a1, A2& a2, A3& a3, A4& a4, A5& a5 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3) + sizeof(A4) + sizeof(A5);
  HPCxx_Buffer b(sizeof(int) + sz + 2*sizeof(HPCxx_ContextID) + 32);

  HPCxx_ContextID& target = *(gp->getContextID());

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R> > gp_to_res_sync( &result);

  // Build the packet; first pack the global pointer to the result, then the
  // gp to the object then the member function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);
  hpcxx_pack(b, &a4, 1);
  hpcxx_pack(b, &a5, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

  return(0);
}


//---- hpcxx_aminvoke ---------------------------------------------------------
// Invokes a remote member function that takes six args and returns a value.
//-----------------------------------------------------------------------------
template< class C, class R, class A1, class A2, class A3, class A4, class A5,
          class A6 >
int hpcxx_aminvoke( HPCxx_GlobalPtr<C>* gp, HPCxx_Sync<R>& result, int funcID,
		    A1& a1, A2& a2, A3& a3, A4& a4, A5& a5, A6& a6 )
{
  int sz = sizeof(A1) + sizeof(A2) + sizeof(A3) + sizeof(A4) + sizeof(A5)
		      + sizeof(A6);
  HPCxx_Buffer b(sizeof(int) + sz + 2*sizeof(HPCxx_ContextID) + 32);

  HPCxx_ContextID& target = *(gp->getContextID());

  // allocate a global pointer to this sync variable.
  HPCxx_GlobalPtr<HPCxx_Sync<R> > gp_to_res_sync(&result);

  // Build the packet; first pack the global pointer to the result, then the
  // gp to the object then the member function id, and then the args.
  hpcxx_pack(b, &gp_to_res_sync, 1);
  hpcxx_pack(b, &funcID, 1);
  hpcxx_pack(b, &a1, 1);
  hpcxx_pack(b, &a2, 1);
  hpcxx_pack(b, &a3, 1);
  hpcxx_pack(b, &a4, 1);
  hpcxx_pack(b, &a5, 1);
  hpcxx_pack(b, &a6, 1);

  // send the message to the target context (call to the low level protocol)
  hpcxx_sendGlobalFunction(b, target); 

   return(0);
}


template <class T>
class HPCxx_Collective:public HPCxx_BufferBlock {
 protected:
  HPCxx_Group* g;
  int myKey;
  int isRoot;
  int rChild, lChild, parentIndex;
  HPCxx_GlobalPtr<HPCxx_BufferBlock>* lc, *rc, *parent;
  HPCxx_GlobalPtr<HPCxx_BufferBlock> *This;
  typedef T*  Tptr;
  Tptr* thread_vector;
  T* thread_group;
  HPCxx_SyncQ<int>* thread_queue;
  HPCxx_Mutex thLock;
  int num_threads_here;
  int num_threads_needed;
  int thread_total_size;
  int* thread_sizes;
  int threads_registered;

  int tree_count(int x, int max_depth, int depth, int index, int size,
		 int& this_root)
  {
    int l_child, r_child;
    if(index >= size) {
      this_root = -1;
      return index;
    }
    if(depth == max_depth){
      if(x == index){ 
	lChild = -1; rChild = -1; 
	if(depth==0) isRoot = 1;
      }
      this_root = index;
      return index+1;
    }
    else{
      if(index < size-1)
	index = tree_count(x, max_depth, depth+1, index, size-1, l_child);
      else l_child = -1;
      this_root = index++; 
      if(index < size)
	index = tree_count(x, max_depth, depth+1, index,size, r_child);
      else r_child = -1;
      if(this_root == x) { 
	lChild = l_child; 
	rChild = r_child; 
	if(depth==0) isRoot = 1;
      }
      return index;
    }
  }
  
  void calculate_tree_neighbors(int size, int x) {
    int maxd = 0;
    int k = 1;
    while(k-1 < size){ k = k*2; maxd++; }
    maxd--;
    tree_count(x, maxd, 0, 0, size, k);
  } 
  
 public:
  
  HPCxx_Collective(HPCxx_Group* gr):HPCxx_BufferBlock(), g(gr) {
    lc=NULL;
    rc=NULL;
    parent=NULL;
    num_threads_here = 0;
    threads_registered = 0;
    num_threads_needed = g->getNumThreads();
    thread_sizes = new int[num_threads_needed];
    thread_vector = new Tptr[num_threads_needed];
    thread_queue = new HPCxx_SyncQ<int>[num_threads_needed];
    myKey = g->getLocalIndex();
    int N = g->getNumContexts();
    isRoot = 0; parentIndex = -1;
    calculate_tree_neighbors(N, myKey);
    //  printf("[%d] has children %d and %d and isRoot=%d\n",
	       //  myKey, lChild, rChild, isRoot);
    This = new HPCxx_GlobalPtr<HPCxx_BufferBlock>(this);
  }

  ~HPCxx_Collective(){
    delete [] thread_vector;
    delete [] thread_group;
    delete [] thread_queue;
    if(This) delete This;
    if(parent) delete parent;
    if(lc) delete lc;
    if(rc) delete rc;
  }
  // synchronousSetUp contains a barrier op and must
       // be executed by each context in the group aftere the initialization.
       void synchronousSetUp(){
    HPCxx_ContextID* cid;
    if(!isRoot){
      barrier_ptr = This;
      int i = 1;
      barrier_set_up_lock[0] = i;
      parent = new HPCxx_GlobalPtr<HPCxx_BufferBlock>;
      //printf("[%d] is waiting for parent\n", myKey);
      parentIndex = barrier_set_up_lock[1];
      //printf("[%d] has parent %d\n", myKey, parentIndex);
      *parent = *barrier_ptr;
    }
    if(lChild > -1){
      cid = g->getContextID(lChild);
      lc = new HPCxx_GlobalPtr<HPCxx_BufferBlock>;
      //printf("[%d] is sending to lc %d\n", myKey, lChild);
      hpcxx_invoke(*cid, *lc, barrierSetUp_id,  *This, myKey); 
    }
    if(rChild > -1){
      cid = g->getContextID(rChild);
      rc = new HPCxx_GlobalPtr<HPCxx_BufferBlock>;
      //printf("[%d] is sending to rc %d\n", myKey, rChild);
      hpcxx_invoke(*cid, *rc, barrierSetUp_id,  *This, myKey); 
    }
    barrier();
  }
   T* extract(HPCxx_Buffer* b, int&  size){
     T* p = NULL;
     hpcxx_unpack(*b, &size, 1);
     if(size > 0) p = new T[size];
     for(int i = 0; i < size; i++) hpcxx_unpack(*b, &(p[i]), 1);
     return p;
   }
   virtual HPCxx_Buffer* combine(T* a1, int s1, T* a2, int s2, T* a3, int s3,
				 T*& t_array, int& t_size){
     t_size = s1+s2+s3;
     //printf("[%d] in combine with t_size=%d\n", myKey, t_size);
     if(t_size > 0 ) t_array = new T[t_size];
     else t_array = NULL;
     int i,j;
     j = 0;
     for(i = 0; i < s1; i++, j++)t_array[j] = a1[i];
     for(i = 0; i < s2; i++, j++)t_array[j] = a2[i];
     for(i = 0; i < s3; i++, j++)t_array[j] = a3[i];
     HPCxx_Buffer* buf = new HPCxx_Buffer( (t_size)*sizeof(T)+sizeof(int)+32 );
     hpcxx_pack(*buf, &t_size, 1);
     for(i = 0; i < t_size; i++) hpcxx_pack(*buf, &(t_array[i]), 1);
     return buf;
   }
   // the member gather collects vectors from each member and concatinates
   // them togheter to a large array. Gather can also be used to broadcast
   // a vector from one to all other members.  members with nothing to send
   // set the size value to 0
  
  T* gather(T* x, int& size) {
     HPCxx_Buffer* b;
     int l, r, t;
     T *lcarray, *rcarray, *parray;
     int lcsize, rcsize, psize;
     // first propogate up the tree. 
     // each node waits for its children
     // printf("[%d] in.  lChild=%d rChild=%d\n", myKey, lChild, rChild);
     if(lChild > -1 ){
       l = lcLock; 
       lcarray = extract(lcBuffer, lcsize);
       lcLock2.write(l);
       //printf("[%d] got left child\n", myKey);
     }
     else lcsize = 0;

     if(rChild > -1 ) {
       r = rcLock; 
       rcarray = extract(rcBuffer, rcsize);
       rcLock2.write(r);         
       //printf("[%d] got right child\n", myKey);
     }
     else rcsize = 0;

     //send result up to parent.
	 int i_am_a_lc = (myKey < parentIndex)?1:0;
     int i_am_a_rc = (myKey > parentIndex)?1:0;
     if (isRoot) i_am_a_lc = i_am_a_rc = 0;
     int mysize = size;
     b = combine(lcarray, lcsize, x, mysize, rcarray, rcsize, parray, size);

     if(i_am_a_lc) { 
       hpcxx_sendToBufferBlockLC(*b, *(parent->getContextID()));
     }
     else if(i_am_a_rc) { 
       hpcxx_sendToBufferBlockRC(*b, *(parent->getContextID()));
     }
     //printf("[%d] waiting for parent\n", myKey);
     
     /* now wait for parent and propogate down*/
     if(!isRoot){ 
       l = pLock; 
       parray = extract(pBuffer, size); 
       pLock2.write(l);
       if((lChild> -1) || (rChild > -1)){ 
	 b = new HPCxx_Buffer(size*(sizeof(T)+sizeof(int))+32);
	 hpcxx_pack(*b, &size, 1);
	 if(size > 0) hpcxx_pack(*b, parray, size);
       }
     }
     //printf("[%d] is sending down. size = %d\n", myKey, size);
     if (lChild > -1 )
       hpcxx_sendToBufferBlockP(*b, *(lc->getContextID()));  
     if (rChild > -1 )
       hpcxx_sendToBufferBlockP(*b, *(rc->getContextID()));  
     //printf("[%d] out\n", myKey);
     return parray;
   }

   int acquireKey() {
     thLock.lock();
     int i = threads_registered++;
     thLock.unlock();
     if(i > num_threads_needed-1) {
       cout << "error: too many thread keys!" << endl;
     }
     return i;
   }
   T* gather(int threadKey, T* thread_data, int& size){
     int i, j, k;
     thread_vector[threadKey] = thread_data;
     thread_sizes[threadKey] = size;
     thLock.lock();
     k = ++num_threads_here;
     thLock.unlock();
     if(k == num_threads_needed){
       thread_total_size = 0;
       for(i = 0; i < num_threads_here; i++){ 
	 int lsize = thread_sizes[i];
	 thread_total_size += lsize;   
       }
       thread_group = new T[thread_total_size];
       k = 0;
       for(i = 0; i < num_threads_here; i++) 
	 for(j = 0; j < thread_sizes[i]; j++)
	   thread_group[k++] = thread_vector[i][j];
       thread_group = gather(thread_group, thread_total_size);
       
       num_threads_here = 0;
       for(i = 0; i < num_threads_needed; i++){
	 if(i != threadKey) thread_queue[i].write(i);
       }
     }
     else i = thread_queue[threadKey];
     size = thread_total_size;
     return thread_group;
   }
   
   void barrier(void){
     HPCxx_Buffer b(sizeof(int));
     int l, r;
     l = 0;
     hpcxx_pack(b,&l,1);
     // first propogate up the tree. 
     // each node waits for its children
     // printf("[%d] in.  lChild=%d rChild=%d\n", myKey, lChild, rChild);
     if(lChild > -1 ){
       l = lcLock; 
       lcLock2.write(l);
       //printf("[%d] got left child\n", myKey);
     }
     if(rChild > -1 ){
       r = rcLock; 
       rcLock2.write(r);         
       //printf("[%d] got right child\n", myKey);
     }
     //send result up to parent.
	 int i_am_a_lc = (myKey < parentIndex)?1:0;
     int i_am_a_rc = (myKey > parentIndex)?1:0;
     if (isRoot) i_am_a_lc = i_am_a_rc = 0;
     
     if(i_am_a_lc){ 
       hpcxx_sendToBufferBlockLC(b, *(parent->getContextID()));
     }
     else if(i_am_a_rc){ 
       hpcxx_sendToBufferBlockRC(b, *(parent->getContextID()));
     }
     //printf("[%d] waiting for parent\n", myKey);
     
     // now wait for parent and propogate down
			      if(!isRoot){ 
				l = pLock; 
				pLock2.write(l);
			      }
     //printf("[%d] is sending down. size = %d\n", myKey, size);
     if (lChild > -1 )
       hpcxx_sendToBufferBlockP(b, *(lc->getContextID()));  
     if (rChild > -1 )
       hpcxx_sendToBufferBlockP(b, *(rc->getContextID()));  
     //printf("[%d] out\n", myKey);
   }
   void barrier(int threadKey){
     int i;
     thLock.lock();
     num_threads_here++;
     thLock.unlock();
     if(num_threads_here == num_threads_needed){
       if(g->getNumContexts()> 1) barrier();
       int k = num_threads_here;
       num_threads_here = 0;
       for(i = 0; i < k; i++){
	 int c = 1;
	 if(i != threadKey) thread_queue[i] = c;
       }
     }
     else i = thread_queue[threadKey];
     
   }
};


template<class T, class Op> 
class HPCxx_Reduct: public HPCxx_Collective<T> {
  Op op;
 public:
  HPCxx_Reduct(HPCxx_Group *g, Op oper): HPCxx_Collective<T>(g), op(oper){}
  HPCxx_Buffer* combine(T* a1, int s1, T* a2, int s2, T* a3, int s3,
			 T*& t_array, int& t_size){
    int i;
    int error = 0;
    if(s2 == 1){
      if(s1 == s2) *a2 = op(*a1, *a2);
      else if (s1 != 0) error=1 ;
      if(s2 == s3) *a2 = op(*a2, *a3);
      else if (s3 != 0) error=1 ;
    }
    else if(s2 > 1){
      if(s1 == s2)
	for(i = 0; i < s2; i++) a2[i] = op(a1[i],a2[i]);
      else if (s1 != 0) error=1 ;
      if(s2 == s3)
	for(i = 0; i < s2; i++) a2[i] = op(a2[i],a3[i]);
      else if (s3 != 0) error=1 ;
    }
    else error = 1;
    if(error) cout << "error in reduct: size missmatch" << endl;
    t_size = s2;
    t_array = a2;
    HPCxx_Buffer *buf = new HPCxx_Buffer(s2*sizeof(T)+sizeof(int));
    hpcxx_pack(*buf, &t_size, 1);
    hpcxx_pack(*buf, t_array, t_size);
    return buf;
  }
  
  T * operator()(T* data, int size){
    return gather(data, size);
  }
  T operator()(T dataval){
    T x = dataval;
    int v = 1;
    return *gather(&x, v);
  }
  T operator()(int threadKey, T* data, int size){
    int i, j;
    thread_vector[threadKey] = data;
    thread_sizes[threadKey] = size;
    thLock.lock();
    num_threads_here++;
    thLock.unlock();
    if(num_threads_here == num_threads_needed){
      for(i = 0; i < num_threads_here; i++){ 
	if(size != thread_sizes[i])
	  cout << "reduct size error" << endl;
      }
      thread_group = new T[size];
      /* set the first val of the array to the first val of each array to
	 be concatenated */
      for(j = 0; j < num_threads_here; j++) thread_group[j] = thread_vector[j][0];

      /* collapse the array of arrays into an array of one value each */
      for(i = 0; i < num_threads_here; i++) {
	for(j = 1; j < size; j++) {
	  thread_group[i] = op(thread_group[i],thread_vector[i][j]);
	}
      }

      /* collapse this array into one value */
      for(i=1; i<num_threads_here; i++) {
	op(thread_group[0], thread_group[i]);
      }
      /* thread_group = gather(thread_group, size); */
      
      num_threads_here = 0;
      for(i = 0; i < num_threads_needed; i++){
	if(i != threadKey) thread_queue[i].write(i);
      }
    }
    else i = thread_queue[threadKey];
    return *thread_group;
  }
  T operator()(int threadKey, T dataval){
    T x = dataval;
    
    return operator()(threadKey, &x, 1);
  }
};
