//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: hpcxx_invoke.h,v 1.1 1998/06/23 00:55:57 ejohnson Exp $
//##########################################################################

// $Id: hpcxx_invoke.h,v 1.1 1998/06/23 00:55:57 ejohnson Exp $

#ifndef __HPCXX_INVOKE_H_
#define __HPCXX_INVOKE_H_

#include <HPCxx_GlobalPtr.h>
#include <HPCxx_Context.h>
#include <HPCxx_Sync.h>


template <class C>
class HPCxx_RegisteredType {
  int key;
 public:
  HPCxx_RegisteredType() {
    C *p = NULL;
    key = hpcxx_registerClass(p);
  }
};


//---- hpcxx_invoke -----------------------------------------------------------
// Invokes a remote function that takes no args and returns a value.
//-----------------------------------------------------------------------------
template<class R>
int hpcxx_invoke(HPCxx_ContextID& target, R& result, int funcID);

//---- hpcxx_invoke -----------------------------------------------------------
// Invokes a remote function that takes one arg and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1>
int hpcxx_invoke(HPCxx_ContextID& target, R& result, int funcID, A1 &a1);

//---- hpcxx_invoke -----------------------------------------------------------
// Invokes a remote function that takes two args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2>
int hpcxx_invoke( HPCxx_ContextID& target, R& result, int funcID,
		  A1& a1, A2& a2 );

//---- hpcxx_invoke -----------------------------------------------------------
// Invokes a remote function that takes three args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3>
int hpcxx_invoke( HPCxx_ContextID& target, R& result, int funcID,
		  A1& a1, A2& a2, A3& a3 );

//---- hpcxx_invoke -----------------------------------------------------------
// Invokes a remote function that takes four args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4>
int hpcxx_invoke( HPCxx_ContextID& target, R& result, int funcID,
		  A1& a1, A2& a2, A3& a3, A4& a4 );

//---- hpcxx_invoke -----------------------------------------------------------
// Invokes a remote function that takes five args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4, class A5>
int hpcxx_invoke( HPCxx_ContextID& target, R& result, int funcID,
		  A1& a1, A2& a2, A3& a3, A4& a4, A5& a5 );

//---- hpcxx_invoke -----------------------------------------------------------
// Invokes a remote function that takes six args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4, class A5, class A6>
int hpcxx_invoke( HPCxx_ContextID& target, R& result, int funcID,
		  A1& a1, A2& a2, A3& a3, A4& a4, A5& a5, A6& a6 );


//
// Asynchronous invocation of remote global functions
//

//---- hpcxx_ainvoke ----------------------------------------------------------
// Invokes a remote function that takes no args and returns a value.
//-----------------------------------------------------------------------------
template<class R>
int hpcxx_ainvoke(HPCxx_ContextID& target, HPCxx_Sync<R>& result, int funcID);

//---- hpcxx_ainvoke ----------------------------------------------------------
// Invokes a remote function that takes one arg and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1>
int hpcxx_ainvoke( HPCxx_ContextID& target, HPCxx_Sync<R>& result, int funcID,
		   A1& a1 );

//---- hpcxx_ainvoke ----------------------------------------------------------
// Invokes a remote function that takes two args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2>
int hpcxx_ainvoke( HPCxx_ContextID& target, HPCxx_Sync<R>& result, int funcID,
		   A1& a1, A2& a2 );

//---- hpcxx_ainvoke ----------------------------------------------------------
// Invokes a remote function that takes three args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3>
int hpcxx_ainvoke( HPCxx_ContextID& target, HPCxx_Sync<R>& result, int funcID,
		   A1& a1, A2& a2, A3& a3 );

//---- hpcxx_ainvoke ----------------------------------------------------------
// Invokes a remote function that takes four args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4>
int hpcxx_ainvoke( HPCxx_ContextID& target, HPCxx_Sync<R>& result, int funcID,
		   A1& a1, A2& a2, A3& a3, A4& a4 );

//---- hpcxx_ainvoke ----------------------------------------------------------
// Invokes a remote function that takes five args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4, class A5>
int hpcxx_ainvoke( HPCxx_ContextID& target, HPCxx_Sync<R>& result, int funcID,
		   A1& a1, A2& a2, A3& a3, A4& a4, A5& a5 );

//---- hpcxx_ainvoke ----------------------------------------------------------
// Invokes a remote function that takes six args and returns a value.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4, class A5, class A6>
int hpcxx_ainvoke( HPCxx_ContextID& target, HPCxx_Sync<R>& result, int funcID,
		   A1& a1, A2& a2, A3& a3, A4& a4, A5& a5, A6& a6 );


//
//
// Remote member function invocation
//
//

//---- hpcxx_minvoke ----------------------------------------------------------
// Invokes a remote member function that takes no args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R>
int hpcxx_minvoke(HPCxx_GlobalPtr<C>* gp, R& result, int funcID);

//---- hpcxx_minvoke ----------------------------------------------------------
// Invokes a remote member function that takes one arg and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1>
int hpcxx_minvoke(HPCxx_GlobalPtr<C>* gp, R& result, int funcID, A1& a1);

//---- hpcxx_minvoke ----------------------------------------------------------
// Invokes a remote member function that takes two args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2>
int hpcxx_minvoke( HPCxx_GlobalPtr<C>* gp, R& result, int funcID,
		   A1& a1, A2& a2 );

//---- hpcxx_minvoke ----------------------------------------------------------
// Invokes a remote member function that takes three args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2, class A3>
int hpcxx_minvoke( HPCxx_GlobalPtr<C>* gp, R& result, int funcID,
		   A1& a1, A2& a2, A3& a3 );

//---- hpcxx_minvoke ----------------------------------------------------------
// Invokes a remote member function that takes four args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2, class A3, class A4>
int hpcxx_minvoke( HPCxx_GlobalPtr<C>* gp, R& result, int funcID,
		   A1& a1, A2& a2, A3& a3, A4& a4 );

//---- hpcxx_minvoke ----------------------------------------------------------
// Invokes a remote member function that takes five args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2, class A3, class A4, class A5>
int hpcxx_minvoke( HPCxx_GlobalPtr<C>* gp, R& result, int funcID,
		   A1& a1, A2& a2, A3& a3, A4& a4, A5& a5 );

//---- hpcxx_minvoke ----------------------------------------------------------
// Invokes a remote member function that takes six args and returns a value.
//-----------------------------------------------------------------------------
template< class C, class R, class A1, class A2, class A3, class A4, class A5,
          class A6 >
int hpcxx_minvoke( HPCxx_GlobalPtr<C>* gp, R& result, int funcID,
		   A1& a1, A2& a2, A3& a3, A4& a4, A5& a5, A6& a6 );


//
// Asynchronous invocation of remote member functions
//

//---- hpcxx_aminvoke ---------------------------------------------------------
// Invokes a remote member function that takes no args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R>
int hpcxx_aminvoke(HPCxx_GlobalPtr<C>* gp, HPCxx_Sync<R>& result, int funcID);

//---- hpcxx_aminvoke ---------------------------------------------------------
// Invokes a remote member function that takes one arg and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1>
int hpcxx_aminvoke( HPCxx_GlobalPtr<C>* gp, HPCxx_Sync<R>& result, int funcID,
		    A1& a1 );

//---- hpcxx_aminvoke ---------------------------------------------------------
// Invokes a remote member function that takes two args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2>
int hpcxx_aminvoke( HPCxx_GlobalPtr<C>* gp, HPCxx_Sync<R>& result, int funcID,
		    A1& a1, A2& a2 );

//---- hpcxx_aminvoke ---------------------------------------------------------
// Invokes a remote member function that takes three args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2, class A3>
int hpcxx_aminvoke( HPCxx_GlobalPtr<C>* gp, HPCxx_Sync<R>& result, int funcID,
		    A1& a1, A2& a2, A3& a3 );

//---- hpcxx_aminvoke ---------------------------------------------------------
// Invokes a remote member function that takes four args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2, class A3, class A4>
int hpcxx_aminvoke( HPCxx_GlobalPtr<C>* gp, HPCxx_Sync<R>& result, int funcID,
		    A1& a1, A2& a2, A3& a3, A4& a4 );

//---- hpcxx_aminvoke ---------------------------------------------------------
// Invokes a remote member function that takes five args and returns a value.
//-----------------------------------------------------------------------------
template<class C, class R, class A1, class A2, class A3, class A4, class A5>
int hpcxx_aminvoke( HPCxx_GlobalPtr<C>* gp, HPCxx_Sync<R>& result, int funcID,
		    A1& a1, A2& a2, A3& a3, A4& a4, A5& a5 );

//---- hpcxx_aminvoke ---------------------------------------------------------
// Invokes a remote member function that takes six args and returns a value.
//-----------------------------------------------------------------------------
template< class C, class R, class A1, class A2, class A3, class A4, class A5,
          class A6 >
int hpcxx_aminvoke( HPCxx_GlobalPtr<C>* gp, HPCxx_Sync<R>& result, int funcID,
		    A1& a1, A2& a2, A3& a3, A4& a4, A5& a5, A6& a6 );

#endif
