//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: hpcxx_register.C,v 1.2 1998/08/07 20:24:56 sameer Exp $
//##########################################################################

#include <HPCxx_Profile.h>
#include <HPCxx_RFITable.h>
#include <HPCxx_Buffer.h>
#include <iostream.h>

HPCxx_RFITable_ gHPCxx_RFITable;


int hpcxx_runRF(void* sp, void* objp, int funcID, HPCxx_Buffer& b)
{
   TAU_PROFILE("hpcxx_runRF()", "int (void *, void *, int, HPCxx_Buffer)", 
     TAU_HPCXX);
   HPCxx_RFITableEntry *e = gHPCxx_RFITable.getEntry(funcID);
   if(!e){
       cout << "failed to find " << funcID << endl;
       return -1;
       }
   e->execute(sp, objp, b); 
   return 0;
}


int hpcxx_runResult(void* sync, int funcID, HPCxx_Buffer& b)
{
   TAU_PROFILE("hpcxx_runResult()", "int (void *, int, HPCxx_Buffer)", TAU_HPCXX);
   HPCxx_RFITableEntry *e = gHPCxx_RFITable.getEntry(funcID);
   if(!e){
       cout << "failed to find " << funcID << endl;
       return -1;
       }
   e->returnValue(sync, b); 
   return 0;
}


int hpcxx_runRead(void* sync, void* q, int typeID, int count, int offset)
{
   TAU_PROFILE("hpcxx_runRead()", "int (void *, void *, int, int, int)",
     TAU_HPCXX);
   HPCxx_RFITableEntry *e = gHPCxx_RFITable.getEntry(typeID);
   if(!e){
       cout << "failed to find type " << typeID << endl;
       return -1;
       }
   e->read(sync, q, count, offset); 
   return 0;
}


int hpcxx_runWrite( void* sync, void* q, int typeID, int count, int offset,
                    HPCxx_Buffer& b )
{ 
   TAU_PROFILE("hpcxx_runWrite()", 
     "int (void *, void *, int, int, int, HPCxx_Buffer)", TAU_HPCXX);
   HPCxx_RFITableEntry *e = gHPCxx_RFITable.getEntry(typeID);
   if(!e){
       cout << "failed to find type " << typeID << endl;
       return -1;
       }
   e->write(sync, q, count, offset, b); 
   return 0;
}


