//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: hpcxx_register.cc,v 1.2 1998/07/02 16:12:50 rasmussn Exp $
//##########################################################################

#include <hpcxx_rts.h>
#include <hpcxx_register.h>
#include <hpcxx_rtsImpl.h>
#include <HPCxx_Buffer.h>
#include <HPCxx_RFITable.h>


//
// Remote global function pointer class
//
template <class R>
class HPCxx_RFPointerZeroArgs : public HPCxx_RFITableEntry {
private:
  R (*fp_)();
public:
  HPCxx_RFPointerZeroArgs(R fp(void)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer& b) {
      R r;
      r = (*fp_)();
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1); 
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void* sync, HPCxx_Buffer& b) {
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R>* rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
   }
};

template <class R, class A1>
class HPCxx_RFPointerOneArg : public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1);

public:
  HPCxx_RFPointerOneArg(R fp(A1)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b) {
      A1 a1;
      R r;
      hpcxx_unpack(b, &a1, 1);
      r = (*fp_)(a1);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1); 
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

template <class R, class A1, class A2>
class HPCxx_RFPointerTwoArgs: public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1, A2);

public:
  HPCxx_RFPointerTwoArgs(R fp(A1, A2)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b ){
       A1 a1;
       A2 a2;
      R r;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      r = (*fp_)(a1, a2);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

template <class R, class A1, class A2, class A3>
class HPCxx_RFPointerThreeArgs: public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1, A2, A3);

public:
  HPCxx_RFPointerThreeArgs(R fp(A1, A2, A3)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b ){
       A1 a1;
       A2 a2;
       A3 a3;
      R r;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      r = (*fp_)(a1, a2, a3);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

template <class R, class A1, class A2, class A3, class A4>
class HPCxx_RFPointerFourArgs: public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1, A2, A3, A4);

public:
  HPCxx_RFPointerFourArgs(R fp(A1, A2, A3, A4)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b ){
       A1 a1;
       A2 a2;
       A3 a3;
       A4 a4;
      R r;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      hpcxx_unpack(b, &a4, 1);
      r = (*fp_)(a1, a2, a3, a4);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

template <class R, class A1, class A2, class A3, class A4, class A5>
class HPCxx_RFPointerFiveArgs: public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1, A2, A3, A4, A5);

public:
  HPCxx_RFPointerFiveArgs(R fp(A1, A2, A3, A4, A5)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b ){
       A1 a1;
       A2 a2;
       A3 a3;
       A4 a4;
       A5 a5;
      R r;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      hpcxx_unpack(b, &a4, 1);
      hpcxx_unpack(b, &a5, 1);
      r = (*fp_)(a1, a2, a3, a4, a5);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

template <class R, class A1, class A2, class A3, class A4, class A5, class A6>
class HPCxx_RFPointerSixArgs: public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1, A2, A3, A4, A5, A6);

public:
  HPCxx_RFPointerSixArgs(R fp(A1, A2, A3, A4, A5, A6)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b ){
       A1 a1;
       A2 a2;
       A3 a3;
       A4 a4;
       A5 a5;
       A6 a6;
      R r;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      hpcxx_unpack(b, &a4, 1);
      hpcxx_unpack(b, &a5, 1);
      hpcxx_unpack(b, &a6, 1);
      r = (*fp_)(a1, a2, a3, a4, a5, a6);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

template <class R, class A1, class A2, class A3, class A4, class A5, class A6,
          class A7>
class HPCxx_RFPointerSevenArgs: public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1, A2, A3, A4, A5, A6, A7);

public:
  HPCxx_RFPointerSevenArgs(R fp(A1, A2, A3, A4, A5, A6, A7)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b ){
       A1 a1;
       A2 a2;
       A3 a3;
       A4 a4;
       A5 a5;
       A6 a6;
       A7 a7;
      R r;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      hpcxx_unpack(b, &a4, 1);
      hpcxx_unpack(b, &a5, 1);
      hpcxx_unpack(b, &a6, 1);
      hpcxx_unpack(b, &a7, 1);
      r = (*fp_)(a1, a2, a3, a4, a5, a6, a7);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

template <class R, class A1, class A2, class A3, class A4, class A5, class A6,
          class A7, class A8>
class HPCxx_RFPointerEightArgs: public HPCxx_RFITableEntry {
private:
  R (*fp_)(A1, A2, A3, A4, A5, A6, A7, A8);

public:
  HPCxx_RFPointerEightArgs(R fp(A1, A2, A3, A4, A5, A6, A7, A8)) : fp_(fp) {}
  operator void *() { return(fp_); }
  void execute(void * callerSP, void *v, HPCxx_Buffer &b ){
       A1 a1;
       A2 a2;
       A3 a3;
       A4 a4;
       A5 a5;
       A6 a6;
       A7 a7;
       A8 a8;
      R r;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      hpcxx_unpack(b, &a4, 1);
      hpcxx_unpack(b, &a5, 1);
      hpcxx_unpack(b, &a6, 1);
      hpcxx_unpack(b, &a7, 1);
      hpcxx_unpack(b, &a8, 1);
      r = (*fp_)(a1, a2, a3, a4, a5, a6, a7, a8);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

//
// Remote Class Pointer class
//


template<class C>
class HPCxx_RCPointer : public HPCxx_RFITableEntry {
private:
  C *c_;

public:
  HPCxx_RCPointer(C *c) : c_(c) {}
  operator C *() { return(c_); }
  void read(void *callerSP, void *p, int count, int offset){
     HPCxx_Buffer br( 3*sizeof(int)+count*sizeof(C) + 32);
     C *cp = (C *) p;
     cp = cp+offset;
     hpcxx_pack(br, &id, 1);
     hpcxx_pack(br, &count, 1);
     hpcxx_pack(br, cp, count);
     hpcxx_returnGlobalFunction(br, callerSP);
   }
  void returnValue(void *sync, HPCxx_Buffer &b){
     int count;
     hpcxx_unpack(b, &count, 1);
     HPCxx_Sync<int> *rp = (HPCxx_Sync<int> *) sync;
     C * user_buff = (C *) rp->getBufferAddr();
     if(user_buff) hpcxx_unpack(b, user_buff, count);
     rp->write(count);
   }
  void write(void *callerSP, void *p, int count, int offset, HPCxx_Buffer &b){
     HPCxx_Buffer br( 3*sizeof(int) + 32);
     C *cp = (C *) p;
     cp = cp+offset;
     hpcxx_unpack(b, cp, count);
     hpcxx_pack(br, &id, 1);
     hpcxx_pack(br, &count, 1);
     hpcxx_returnGlobalFunction(br, callerSP);
   }
};

// Remote Member Function Entry with no args
template<class R, class C>
class HPCxx_RMFPointerZeroArgs : public HPCxx_RFITableEntry {
private:
  typedef R (C::*MFP)();
  MFP mfp_;

public:
  HPCxx_RMFPointerZeroArgs(MFP mfp) : mfp_(mfp) {}
  operator MFP () { return(mfp_); }
  void execute(void * callerSP, void *pv,  HPCxx_Buffer &b ){
      R r;
      C *p = (C *)pv;
      r = (p->*mfp_)();
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

// Remote Member Function Entry with one arg
template<class R, class C, class A1>
class HPCxx_RMFPointerOneArg : public HPCxx_RFITableEntry {
private:
   //R (C::*mfp_)(A1);
  typedef R (C::*MFP)(A1);
  MFP mfp_;
public:
  HPCxx_RMFPointerOneArg(MFP mfp) : mfp_(mfp) {}
  operator MFP () { return(mfp_); }
  void execute(void * callerSP, void *pv,  HPCxx_Buffer &b ){
       A1 a1;
      R r;
      C *p = (C *)pv;
      hpcxx_unpack(b, &a1, 1);
      r = (p->*mfp_)(a1);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1); 
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

// Remote Member Function Entry with two args
template<class R, class C, class A1, class A2>
class HPCxx_RMFPointerTwoArgs : public HPCxx_RFITableEntry {
private:
  typedef R (C::*MFP)(A1, A2);
  MFP mfp_;

public:
  HPCxx_RMFPointerTwoArgs(MFP mfp) : mfp_(mfp) {}
  operator MFP () { return(mfp_); }
  void execute(void * callerSP, void *pv,  HPCxx_Buffer &b ){
       A1 a1;
       A2 a2;
      R r;
      C *p = (C *)pv;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      r = (p->*mfp_)(a1,a2);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

// Remote Member Function Entry with three args
template<class R, class C, class A1, class A2, class A3>
class HPCxx_RMFPointerThreeArgs : public HPCxx_RFITableEntry {
private:
  typedef R (C::*MFP)(A1, A2, A3);
  MFP mfp_;

public:
  HPCxx_RMFPointerThreeArgs(MFP mfp) : mfp_(mfp) {}
  operator MFP () { return(mfp_); }
  void execute(void * callerSP, void *pv,  HPCxx_Buffer &b ){
       A1 a1;
       A2 a2;
       A3 a3;
      R r;
      C *p = (C *)pv;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      r = (p->*mfp_)(a1,a2,a3);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

// Remote Member Function Entry with four args
template<class R, class C, class A1, class A2, class A3, class A4>
class HPCxx_RMFPointerFourArgs : public HPCxx_RFITableEntry {
private:
  typedef R (C::*MFP)(A1, A2, A3, A4);
  MFP mfp_;

public:
  HPCxx_RMFPointerFourArgs(MFP mfp) : mfp_(mfp) {}
  operator MFP () { return(mfp_); }
  void execute(void * callerSP, void *pv,  HPCxx_Buffer &b ){
       A1 a1;
       A2 a2;
       A3 a3;
       A4 a4;
      R r;
      C *p = (C *)pv;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      hpcxx_unpack(b, &a4, 1);
      r = (p->*mfp_)(a1,a2,a3,a4);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

// Remote Member Function Entry with five args
template<class R, class C, class A1, class A2, class A3, class A4, class A5>
class HPCxx_RMFPointerFiveArgs : public HPCxx_RFITableEntry {
private:
  typedef R (C::*MFP)(A1, A2, A3, A4, A5);
  MFP mfp_;

public:
  HPCxx_RMFPointerFiveArgs(MFP mfp) : mfp_(mfp) {}
  operator MFP () { return(mfp_); }
  void execute(void * callerSP, void *pv,  HPCxx_Buffer &b ){
       A1 a1;
       A2 a2;
       A3 a3;
       A4 a4;
       A5 a5;
      R r;
      C *p = (C *)pv;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      hpcxx_unpack(b, &a4, 1);
      hpcxx_unpack(b, &a5, 1);
      r = (p->*mfp_)(a1,a2,a3,a4,a5);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

// Remote Member Function Entry with six args
template<class R, class C, class A1, class A2, class A3, class A4, class A5, class A6>
class HPCxx_RMFPointerSixArgs : public HPCxx_RFITableEntry {
private:
  typedef R (C::*MFP)(A1, A2, A3, A4, A5, A6);
  MFP mfp_;

public:
  HPCxx_RMFPointerSixArgs(MFP mfp) : mfp_(mfp) {}
  operator MFP () { return(mfp_); }
  void execute(void * callerSP, void *pv,  HPCxx_Buffer &b ){
       A1 a1;
       A2 a2;
       A3 a3;
       A4 a4;
       A5 a5;
       A6 a6;
      R r;
      C *p = (C *)pv;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      hpcxx_unpack(b, &a4, 1);
      hpcxx_unpack(b, &a5, 1);
      hpcxx_unpack(b, &a6, 1);
      r = (p->*mfp_)(a1,a2,a3,a4,a5,a6);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

// Remote Member Function Entry with seven args
template<class R, class C, class A1, class A2, class A3, class A4, class A5, class A6, class A7>
class HPCxx_RMFPointerSevenArgs : public HPCxx_RFITableEntry {
private:
  typedef R (C::*MFP)(A1, A2, A3, A4, A5, A6, A7);
  MFP mfp_;

public:
  HPCxx_RMFPointerSevenArgs(MFP mfp) : mfp_(mfp) {}
  operator MFP () { return(mfp_); }
  void execute(void * callerSP, void *pv,  HPCxx_Buffer &b ){
       A1 a1;
       A2 a2;
       A3 a3;
       A4 a4;
       A5 a5;
       A6 a6;
       A7 a7;
      R r;
      C *p = (C *)pv;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      hpcxx_unpack(b, &a4, 1);
      hpcxx_unpack(b, &a5, 1);
      hpcxx_unpack(b, &a6, 1);
      hpcxx_unpack(b, &a7, 1);
      r = (p->*mfp_)(a1,a2,a3,a4,a5,a6,a7);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};

// Remote Member Function Entry with eight args
template<class R, class C, class A1, class A2, class A3, class A4, class A5, class A6, class A7, class A8>
class HPCxx_RMFPointerEightArgs : public HPCxx_RFITableEntry {
private:
  typedef R (C::*MFP)(A1, A2, A3, A4, A5, A6, A7, A8);
  MFP mfp_;

public:
  HPCxx_RMFPointerEightArgs(MFP mfp) : mfp_(mfp) {}
  operator MFP () { return(mfp_); }
  void execute(void * callerSP, void *pv,  HPCxx_Buffer &b ){
       A1 a1;
       A2 a2;
       A3 a3;
       A4 a4;
       A5 a5;
       A6 a6;
       A7 a7;
       A8 a8;
      R r;
      C *p = (C *)pv;
      hpcxx_unpack(b, &a1, 1);
      hpcxx_unpack(b, &a2, 1);
      hpcxx_unpack(b, &a3, 1);
      hpcxx_unpack(b, &a4, 1);
      hpcxx_unpack(b, &a5, 1);
      hpcxx_unpack(b, &a6, 1);
      hpcxx_unpack(b, &a7, 1);
      hpcxx_unpack(b, &a8, 1);
      r = (p->*mfp_)(a1,a2,a3,a4,a5,a6,a7,a8);
      HPCxx_Buffer br(sizeof(R)+sizeof(int));
      hpcxx_pack(br, &id, 1);
      hpcxx_pack(br, &r, 1);
      hpcxx_returnGlobalFunction(br, callerSP);
  }
  void returnValue(void *sync, HPCxx_Buffer &b){
      R r;
      hpcxx_unpack(b, &r, 1);  // get the return value
      HPCxx_Sync<R> *rp = (HPCxx_Sync<R> *) sync;
      rp->write(r);
  }
};


//-----------------------------------------------------------------------------
// Register a class.
//-----------------------------------------------------------------------------
template<class C>
hpcxx_id_t hpcxx_registerClass(C *c)
{
  hpcxx_id_t registerID = HPCxx_Type<C>::registrationID();
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(new HPCxx_RCPointer<C>(c), registerID);
}


//-----------------------------------------------------------------------------
// Register a global function that takes no arguments.
//-----------------------------------------------------------------------------
template<class R>
hpcxx_id_t hpcxx_register(R (*f) (void), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
	        new HPCxx_RFPointerZeroArgs<R>(f), registerID);
}


//-----------------------------------------------------------------------------
// Register a global function that takes one argument.
//-----------------------------------------------------------------------------
template<class R, class A1>
hpcxx_id_t hpcxx_register(R (*f) (A1), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
                new HPCxx_RFPointerOneArg<R,A1>(f), registerID);
}


//-----------------------------------------------------------------------------
// Register a global function that takes two arguments.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2>
hpcxx_id_t hpcxx_register(R (*f) (A1, A2), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
                new HPCxx_RFPointerTwoArgs<R,A1,A2>(f), registerID);
}


//-----------------------------------------------------------------------------
// Register a global function that takes three arguments.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3>
hpcxx_id_t hpcxx_register(R (*f) (A1, A2, A3), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
                new HPCxx_RFPointerThreeArgs<R,A1,A2,A3>(f), registerID);
}


//-----------------------------------------------------------------------------
// Register a global function that takes four arguments.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4>
hpcxx_id_t hpcxx_register(R (*f) (A1, A2, A3, A4), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
                new HPCxx_RFPointerFourArgs<R,A1,A2,A3,A4>(f), registerID);
}

//-----------------------------------------------------------------------------
// Register a global function that takes five arguments.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4, class A5>
hpcxx_id_t hpcxx_register(R (*f) (A1, A2, A3, A4, A5), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
                new HPCxx_RFPointerFiveArgs<R,A1,A2,A3,A4,A5>(f), registerID);
}


//-----------------------------------------------------------------------------
// Register a global function that takes six arguments.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2, class A3, class A4, class A5, class A6>
hpcxx_id_t hpcxx_register(R (*f) (A1, A2, A3, A4, A5, A6), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
               new HPCxx_RFPointerSixArgs<R,A1,A2,A3,A4,A5,A6>(f), registerID);
}


//-----------------------------------------------------------------------------
// Register a global function that takes seven arguments.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2,
         class A3, class A4, class A5, class A6, class A7>
hpcxx_id_t hpcxx_register( R (*f) (A1, A2, A3, A4, A5, A6, A7),
			   hpcxx_id_t registerID )
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
     new HPCxx_RFPointerSevenArgs<R,A1,A2,A3,A4,A5,A6,A7>(f), registerID);
}


//-----------------------------------------------------------------------------
// Register a global function that takes eight arguments.
//-----------------------------------------------------------------------------
template<class R, class A1, class A2,
         class A3, class A4, class A5, class A6, class A7, class A8>
hpcxx_id_t hpcxx_register( R (*f) (A1, A2, A3, A4, A5, A6, A7, A8),
			   hpcxx_id_t registerID )
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
     new HPCxx_RFPointerEightArgs<R,A1,A2,A3,A4,A5,A6,A7,A8>(f), registerID);
}


//-----------------------------------------------------------------------------
// Register a member function that takes no arguments.
//-----------------------------------------------------------------------------
template<class R, class C>
hpcxx_id_t hpcxx_register(R (C::*mf) (void), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
		new HPCxx_RMFPointerZeroArgs<R,C>(mf), registerID);
}


//-----------------------------------------------------------------------------
// Register a member function that takes one arguments.
//-----------------------------------------------------------------------------
template<class R, class C, class A1>
hpcxx_id_t hpcxx_register(R (C::*mf) (A1), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
		new HPCxx_RMFPointerOneArg<R,C,A1>(mf), registerID);
}


//-----------------------------------------------------------------------------
// Register a member function that takes two arguments.
//-----------------------------------------------------------------------------
template<class R, class C, class A1, class A2>
hpcxx_id_t hpcxx_register(R (C::*mf) (A1, A2), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
                new HPCxx_RMFPointerTwoArgs<R,C,A1,A2>(mf), registerID);
}


//-----------------------------------------------------------------------------
// Register a member function that takes three arguments.
//-----------------------------------------------------------------------------
template<class R, class C, class A1, class A2, class A3>
hpcxx_id_t hpcxx_register(R (C::*mf) (A1, A2, A3), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
                new HPCxx_RMFPointerThreeArgs<R,C,A1,A2,A3>(mf), registerID);
}


//-----------------------------------------------------------------------------
// Register a member function that takes four arguments.
//-----------------------------------------------------------------------------
template<class R, class C, class A1, class A2, class A3, class A4>
hpcxx_id_t hpcxx_register(R (C::*mf) (A1, A2, A3, A4), hpcxx_id_t registerID)
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
                new HPCxx_RMFPointerFourArgs<R,C,A1,A2,A3,A4>(mf), registerID);
}


//-----------------------------------------------------------------------------
// Register a member function that takes five arguments.
//-----------------------------------------------------------------------------
template<class R, class C, class A1, class A2, class A3, class A4, class A5>
hpcxx_id_t hpcxx_register( R (C::*mf) (A1, A2, A3, A4, A5),
			   hpcxx_id_t registerID )
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
     new HPCxx_RMFPointerFiveArgs<R,C,A1,A2,A3,A4,A5>(mf), registerID);
}


//-----------------------------------------------------------------------------
// Register a member function that takes six arguments.
//-----------------------------------------------------------------------------
template<class R, class C, class A1, class A2,
         class A3, class A4, class A5, class A6>
hpcxx_id_t hpcxx_register( R (C::*mf) (A1, A2, A3, A4, A5, A6),
			   hpcxx_id_t registerID )
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
     new HPCxx_RMFPointerSixArgs<R,C,A1,A2,A3,A4,A5,A6>(mf), registerID);
}


//-----------------------------------------------------------------------------
// Register a member function that takes seven arguments.
//-----------------------------------------------------------------------------
template<class R, class C, class A1, class A2,
         class A3, class A4, class A5, class A6, class A7>
hpcxx_id_t hpcxx_register( R (C::*mf) (A1, A2, A3, A4, A5, A6, A7),
			   hpcxx_id_t registerID )
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(
     new HPCxx_RMFPointerSevenArgs<R,C,A1,A2,A3,A4,A5,A6,A7>(mf), registerID);
}


//-----------------------------------------------------------------------------
// Register a member function that takes eight arguments.
//-----------------------------------------------------------------------------
template<class R, class C, class A1, class A2,
         class A3, class A4, class A5, class A6, class A7, class A8>
hpcxx_id_t hpcxx_register( R (C::*mf) (A1, A2, A3, A4, A5, A6, A7, A8),
			   hpcxx_id_t registerID )
{
  if (registerID == HPCXX_GEN_LOCAL_REGISTERID) {
    registerID = hpcxx_getLocalID(HPCXX_REGISTER_IDSPACE);
    if (registerID == HPCXX_INVALID_ID) {
      return HPCXX_INVALID_ID;
    }
  }
  return gHPCxx_RFITable.addEntry(new
     HPCxx_RMFPointerEightArgs<R,C,A1,A2,A3,A4,A5,A6,A7,A8>(mf), registerID);
}
