//##########################################################################
// HPC++Lib: An object-oriented class library for high performance 
//           parallel programming.
//
// Copyright (c) 1997
//	Indiana University
//	Los Alamos National Laboratory
// 
// http://www.acl.lanl.gov/Tulip
// http://www.extreme.indiana.edu/HPC++
// 
// Pete Beckman, Dennis Gannon, Todd Green, Liz Johnson
//
// HPC++_VER: $Id: hpcxx_rts.C,v 1.2 1998/08/07 20:24:57 sameer Exp $
//##########################################################################


// $Id: hpcxx_rts.C,v 1.2 1998/08/07 20:24:57 sameer Exp $


#include <HPCxx_Profile.h>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include <unistd.h>

#include <hpcxx_rts.h>

// Globals
int hpcxx_debug_level = 0;
int hpcxx_trace_level = 0;

// Struct for the internal representation of an idspace.
// Local keys are returned in the range of 0 - 0x7FFFFFFF
// Global keys are returned in the range of -2 - 0x80000000
// -1 is reserved for HPCXX_INVALID_ID
//
struct hpcxx_idspace_t {
  hpcxx_id_t local_id;
  hpcxx_id_t global_id;

  hpcxx_idspace_t() : local_id(HPCXX_MIN_LOCAL_ID), global_id(HPCXX_MIN_GLOBAL_ID) {}
  hpcxx_id_t getNextAvailLocalID() {
	return((local_id == INT_MAX) ? HPCXX_INVALID_ID : local_id++);
  }
  hpcxx_id_t getNextAvailGlobalID() {
	return((global_id == INT_MIN) ? HPCXX_INVALID_ID : global_id--);
  }
};

// Allocate an identifier space.  Two spaces are allocated and
// reserved by HPC++ Lib:
//     HPCXX_REGISTER_IDSPACE
//     HPCXX_GROUP_IDSPACE

static int HPCXX_NEXT_AVAIL_IDSPACE = HPCXX_GROUP_IDSPACE + 1;
static const int HPCXX_MAX_IDSPACE        = 32;

static hpcxx_idspace_t HPCXX_IDSPACE_TABLE[HPCXX_MAX_IDSPACE];

// Return the next available idspace or HPCXX_INVALID_IDSPACE
// if no spaces are left
//
int hpcxx_createIDSpace()
{
  return((HPCXX_NEXT_AVAIL_IDSPACE == HPCXX_MAX_IDSPACE) ?
		 HPCXX_INVALID_IDSPACE : HPCXX_NEXT_AVAIL_IDSPACE++);
}

// After having allocated an id space, id's may be obtained
// via the two functions.
//
// hpcxx_getGlobalID implies a global communication and guarantee's
// synchronization across contexts
//
// hpcxx_getLocalID implies no communication, but it is the user's
// responsibility to insure that keys are synchronized across contexts
// (SPMD mode)
//
hpcxx_id_t hpcxx_getGlobalID(int idSpace)
{
  // FIXME
  // not yet implemented

  // if (isn't local) {
  //   do RMI to master context
  // } else {
  //   return(HPCXX_IDSPACE_TABLE[idSpace].getNextAvailGlobalID());
  // }
  return(HPCXX_INVALID_ID);
}

hpcxx_id_t hpcxx_getLocalID(int idSpace)
{
  return(HPCXX_IDSPACE_TABLE[idSpace].getNextAvailLocalID());
}

// Debug string
void hpcxx_debug(const char *msg, int debugLevel)
{
  if (hpcxx_debug_level >= debugLevel) hpcxx_msg(msg);
}

// Abort the hpcxx computation.  hpcxx_abort() will call
// hpcxx_exit()
void hpcxx_abort(int status);

// Overloaded operator new to allocate memory from shared space
void *operator new(size_t s, const hpcxx_shmallocator_t &)
{
  TAU_PROFILE("operator new ()", "void *(size_t, hpcxx_shmallocator_t )", 
    TAU_HPCXX);
  return(hpcxx_malloc(s));
}

// Overloaded operator new[] to allocate memory from shared space
#ifdef HAVE_OP_NEW_ARRAY
void* operator new[] (size_t s, const hpcxx_shmallocator_t &)
{
  TAU_PROFILE("operator new[] ()", "void * (size_t, hpcxx_shmallocator_t )",
    TAU_HPCXX);
  return hpcxx_malloc(s);
}
#endif
