//-----------------------------------------------------------------------------
// hpcxx_rtsTmpl.cc - Templated functions for the Run Time System Library
//
// Copyright (c) 1997 Los Alamos National Laboratory
//-----------------------------------------------------------------------------
// $Id: hpcxx_rts.cc,v 1.1 1998/06/23 00:56:03 ejohnson Exp $
//-----------------------------------------------------------------------------


//---- hpcxx_sizeof -----------------------------------------------------------
// Returns the packed size of type T in bytes.
//-----------------------------------------------------------------------------
template<class T> size_t hpcxx_sizeof(T& t)
{
  return sizeof(t);
}


//---- hpcxx_delete -----------------------------------------------------------
// Deletes an array of objects by explicitly calling the destructor for each
// of the objects in the array and then freeing the memory for the array.
// This function must be called if the array of objects has been allocated
// using the (HPCXX_NEW) placement allocator.
//
// aT        the array to delete
// nObjects  the number of objects in the array to delete
//-----------------------------------------------------------------------------
template<class T> void hpcxx_delete(T* aT, int nObjects)
{
  for (int i = 0; i < nObjects; i++) {
    aT[i].~T();
  }
  hpcxx_free(aT);
}


//---- hpcxx_delete -----------------------------------------------------------
// Deletes an object by explicitly calling its destructor and then freeing
// the memory for the object.  This function must be called if the object
// has been allocated using the (HPCXX_NEW) placement allocator.
//
// pT  pointer to the object that is to be deleted
//-----------------------------------------------------------------------------
template<class T> void hpcxx_delete(T* pT)
{
  pT->~T();
  hpcxx_free(pT);
}


template<class T> void hpcxx_delete(const HPCxx_GlobalPtr<T> &gp)
{
  if (!(gp.isLocal())) {
    hpcxx_msg("WARNING: hpcxx_delete() of non-local global pointer ignored\n");
  } else {
    T *localPtr = gp;
    localPtr->T::~T();
    gp.globalRef_.destroy();
    hpcxx_free(localPtr);
  }
}
