#include "Complex.h"


Complex& Complex::operator= (Complex& rhs)
{
  if (*this == rhs) return *this;

  re = rhs.re;
  im = rhs.im;
  return *this;
}

Complex& Complex::operator+ (Complex& rhs)
{
  re += rhs.re;
  im += rhs.im;
  return *this;
}

Complex& Complex::operator- (Complex& rhs)
{
  re -= rhs.re;
  im -= rhs.im;
  return *this;
}

Complex& Complex::operator* (Complex& rhs)
{
  Complex temp = *this;
  re = re * rhs.re - im * rhs.im;
  im = temp.im * rhs.re + temp.re * rhs.im;
  return *this;
}

Complex& Complex::operator* (const double& rhs)
{
  re *= rhs;
  im *= rhs;
  return *this;
}

inline int Complex::operator== (const Complex& rhs)
{
  if ((fabs(re - rhs.re) < .00001) && (fabs(im - rhs.im) < .000001))
    return 1;
  
  return 0;
}
  

ostream& operator<< (ostream& s, Complex z)
{
  s.precision(6);
  s.setf(ios::fixed);
  return s << "(" << z.re << ", " << z.im << ")";
}


// This new version of the pack function operates on Complex objects.
void hpcxx_pack (HPCxx_Buffer &buffer, Complex* data, int count) 
{ 
  for (int i=0; i<count; i++)
    {
      hpcxx_pack(buffer, &data[i].re); 
      hpcxx_pack(buffer, &data[i].im);
    }
}


// This new version of the unpack function operates on Complex objects.
void hpcxx_unpack (HPCxx_Buffer &buffer, Complex* data, int count) 
{ 
  for (int i=0; i<count; i++) 
    {
      hpcxx_unpack(buffer, &data[i].re); 
      hpcxx_unpack(buffer, &data[i].im);
    }
}

size_t hpcxx_sizeof (Complex& object)
{
  return (2 * sizeof(double));
}

// These are the ComplexArray methods:

int ComplexArray::operator== (const ComplexArray& rhs)
{
  int i = 0;
  while (A[i] == rhs.A[i]) {
    if (i >= sz) return 0;
    i++;
  }
  return 1;
}

Complex& ComplexArray::operator[] (const unsigned int index)
{
  if (index > sz) {
    cerr << "Error: attempting to access invalid index value\n";
    exit(1);
  }
  return A[index];
}


void ComplexArray::out ()
{
  for (int i=0; i<sz; i++)
    cout << A[i] << "\n" << flush;
}


// This new version of the pack function operates on Complex objects.
void hpcxx_pack (HPCxx_Buffer &buffer, ComplexArray* data, int count) 
{ 
  for (int i=0; i<count; i++)
    {
      hpcxx_pack(buffer, data[i].A, data[i].sz); 
    }
}


// This new version of the unpack function operates on Complex objects.
void hpcxx_unpack (HPCxx_Buffer &buffer, ComplexArray* data, int count) 
{
  for (int i=0; i<count; i++) 
    {
      hpcxx_unpack(buffer, data[i].A, data[i].sz); 
    }
}


size_t hpcxx_sizeof (ComplexArray& object)
{
  return (sizeof(int) + object.sz * hpcxx_sizeof(object.A[0]));
}










