#include <stdio.h>
#include <iostream.h>
#include <HPCxx_Buffer.h>


/* class Complex
 * -------------
 * A simple class consisting of the set (real, imaginary).  This class was 
 * chosen for simplicity, but it illustrates the techniques for using user
 * defined types with the HPCxx_Buffer.
 */

class Complex {
private:
  double re;
  double im;

public:
  inline Complex (double r = 0, double i = 0): re(r), im(i) { }
  virtual ~Complex() { }

  Complex& operator= (Complex& rhs);
  Complex& operator+ (Complex& rhs);
  Complex& operator- (Complex& rhs);
  Complex& operator* (Complex& rhs);
  Complex& operator* (const double& rhs);
  inline int operator== (const Complex& rhs);
  
  friend ostream& operator<< (ostream& s, Complex z);
  friend void hpcxx_pack (HPCxx_Buffer &buffer, Complex* data, int count = 1);
  friend void hpcxx_unpack (HPCxx_Buffer &buffer, Complex* data, int count = 1); 
  friend size_t hpcxx_sizeof (Complex& object);
};



/*
 * class ComplexArray
 * ------------------
 * This class makes arrays of Complex objects.  The size of each array is 
 * random (see the instatiation of the class) and each element's data has a 
 * random number multiplied by a common base resulting in psudo random data.
 */

class ComplexArray {
private:
  Complex* A;
  int sz;

public:
  ComplexArray(int size = 1, double base_re = 0, double base_im = 0) 
  { 
    A = new Complex[sz = size];
    
    Complex tmp(base_re, base_im);
    for (int i = 0; i < sz; i++)
      A[i] = tmp;
  }

  ~ComplexArray() { delete [] A; }

  static int registrationID() { return 932; }

  Complex* getArray_Ptr() { return A; }
  int getsize() { return sz; }

  int operator== (const ComplexArray& rhs);
  Complex& operator[] (const unsigned int index);
  void out();
  
  friend void hpcxx_pack (HPCxx_Buffer &buffer, ComplexArray* data, int count = 1);
  friend void hpcxx_unpack (HPCxx_Buffer &buffer, ComplexArray* data, int count = 1);
  friend size_t hpcxx_sizeof (ComplexArray& object);
  
};




