cc-----------------------------------------------------------------------------
cc TestGP_F
cc --------
cc This file contains code for testing the Fortran interface to global
cc pointers.  It also makes one call to a remote (Fortran) function.
cc-----------------------------------------------------------------------------
      program main

      integer i, gp_id, val, error, myContext, nContexts, hello_id
      integer int_array(10), buf(10)
      integer group(2), map(2), context0(2), int_gp(2)
cc
cc initialize variables
cc
      gp_id = 33
      hello_id = 1
cc
cc initialize HPC++
cc
      call hpcxx_init(group, error)
cc
cc Register the remote functions with HPC++.  The code to register the
cc functions was automatically generated with the register-functions
cc functions in the the util directory.  The default location for this
cc automatically generated code is the file register.C.
cc
      call hpcxx_register_functions(map)
cc
cc get information about the number of contexts and the local context index
cc
      call Group_getNumContexts(group, nContexts, error)
      call Group_getLocalIndex(group, myContext, error)
cc
cc get a reference to context 0
cc
      call Group_getContextID(group, context0, 0, error)
      call hpcxx_barrier(error)
cc
cc call the remote function HelloWorld (registered above)
cc
      call hpcxx_invoke_rmfi(context0, hello_id, error, val, myContext)
      if (val .ne. myContext) write(6,*) "ERROR, invoking HelloWorld"

      if (myContext .eq. 0) then

         do i = 1, 10
            int_array(i) = i
         end do
cc
cc Create a new global pointer in the local context, pointing to int_array,
cc to be known by the id, gp_id.
cc
         call GlobalPtr_new(int_gp, 3, int_array, gp_id, error)
cc
cc As a test, retrieve the same global pointer by using the gp_id handle.
cc
         call GlobalPtr_find(int_gp, context0, gp_id, error)
cc
cc Read from the global pointer into the local buffer, buf.
cc
         call GlobalPtr_read(int_gp, buf, 10, 0, error)
         do i = 1, 10
            write(6,*) "    buf(", i, ") = ", buf(i)
            if (buf(i) .ne. i) then
               write(6,*) "ERROR, global pointer read error,", buf(i)
            end if
         end do
         write(6,*)
cc
cc Add 10 to all of the elements and write back to the global pointer.
cc
         do i = 1, 10
            buf(i) = i + 10
         end do
         call GlobalPtr_write(int_gp, buf, 10, 0, error)
         do i = 1, 10
            write(6,*) "    int_array(", i, ") = ", int_array(i)
            if (int_array(i) .ne. i + 10) then
               write(6,*) "ERROR, global pointer read error,", int_array(i)
            end if
         end do
         write(6,*)

      end if

      call hpcxx_barrier(error)

      if (myContext .eq. 1) then
cc
cc Retrieve the global pointer from context 0 by using the gp_id handle and
cc read from the remote global pointer into the local buffer, buf.
cc
         call GlobalPtr_find(int_gp, context0, gp_id, error)
         call GlobalPtr_read(int_gp, buf, 10, 0, error)
cc
cc modify the array
cc
         do i = 1, 10
            if (buf(i) .ne. i + 10) then
               write(6,*) "ERROR, global pointer read error,", buf(i)
            end if
            buf(i) = 11 - i
         end do
cc
cc write the modified array back to its owner
cc
         call GlobalPtr_write(int_gp, buf, 10, 0, error)

      end if

      call hpcxx_barrier(error)
cc
cc test local write
cc
      if (myContext .eq. 0) then
         do i = 1, 10
            write(6,*) "    int_array(", i, ") = ", int_array(i)
            if (int_array(i) .ne. 11 - i .and. nContexts .eq. 2) then
               write(6,*) "ERROR, global pointer read error,", int_array(i)
            end if
            if (int_array(i) .ne. i + 10 .and. nContexts .eq. 1) then
               write(6,*) "ERROR, global pointer read error,", int_array(i)
            end if
         end do
         write(6,*)
      end if
cc
cc Delete the global pointer.  Both contexts must do this.  The remote
cc context must delete the global pointer because it was returned a pointer
cc to a local proxy by the GlobalPtr_find call, which must be deleted.
cc      
      call GlobalPtr_delete(int_gp, error)
cc
cc exit the HPC++ runtime
cc
      call hpcxx_exit(group, error)

        stop 
      end

      integer function HelloWorld(iVal)
        integer iVal
cc
cc printing from non-master context creates weird timing dependent problems
cc with Nexus
cc
cc        print *, "Hello from context", iVal
       HelloWorld = iVal
      end

