#include <HPCxx_Profile.h>
#include <math.h>
#include <assert.h>
#include <iostream.h>
#include <hpcxx_rts.h>
#include <HPCxx_Sync.cc>
#include <hpcxx_invoke.cc>
#include <hpcxx_register.cc>
#include <HPCxx_GlobalPtr.cc>
#include <hpcxx_invokeImpl.cc>
#include "HPCxx_GlobalRefImpl.cc"


// declare function the test functions

int zero(void)
{
  TAU_PROFILE("zero()", "int ()", TAU_USER);
  return 1;
}

int first(const char* cmd)
{
  TAU_PROFILE("first()", "int (char *)", TAU_USER);
  if (strcmp("Da man", cmd) == 0)
    return 1;
  return 0;
}

int second(const char* cmd, float pi)
{
  TAU_PROFILE("second()", "int (char *, float)", TAU_USER);
  if ((strcmp("Da man", cmd) == 0) && (fabs(pi - 3.14159) < 0.000003))
    return 1;
  return 0;
}

int third(const char* cmd, float pi, int age)
{
  TAU_PROFILE("third()", "int (char *, float, int)", TAU_USER);
  if ((strcmp("Da man", cmd) == 0) && (fabs(pi - 3.14159) < 0.000003) && 
      (age == 21))
    return 1;
  return 0;
}

int fourth(const char* cmd, float pi, int age, char letter)
{
  TAU_PROFILE("fourth()", "int (char *, float, int, char )", TAU_USER);
  if ((strcmp("Da man", cmd) == 0) && (fabs(pi - 3.14159) < 0.000003) && 
      (age == 21) && (letter == 'M'))
    return 1;
  return 0;
}

int fifth(const char* cmd, float pi, int age, char letter, short aShort)
{
  TAU_PROFILE("fifth()", "int (char *, float, int, char, short)", TAU_USER);
  if ((strcmp("Da man", cmd) == 0) && (fabs(pi - 3.14159) < 0.000003) &&
      (age == 21) && (letter == 'M') && (aShort == 86))
    return 1;
  return 0;
}

int sixth(const char* cmd, float pi, int age, 
	  char letter, short aShort, u_short anotherShort)
{
  TAU_PROFILE("sixth()", "int (char *, float, int, char, short, u_short)",
    TAU_USER);
  if ((strcmp("Da man", cmd) == 0) && (fabs(pi - 3.14159) < 0.000003) &&
      (age == 21) && (letter == 'M') && (aShort == 86) && (anotherShort == 65))
    return 1;
  return 0;
}

void check_rtn_val(int return_val, char* func_name)
{
  TAU_PROFILE("check_rtn_val()", "void (int, char * )", TAU_USER);
  if (return_val != 1)
    cerr << "ERROR in function -> " << func_name << "\n";
}


// remote function identifiers

static int cmd_ID0;
static int cmd_ID1;
static int cmd_ID2;
static int cmd_ID3;
static int cmd_ID4;
static int cmd_ID5;
static int cmd_ID6;

// return varables

static int result0;
static int result1;
  
int main (int argc, char *argv[])
{
  TAU_PROFILE("main()", "int (int, char **)", TAU_DEFAULT);
  HPCxx_Group* group;
  hpcxx_init(argc, argv, group);

  //
  // all processes are executing in SPMD mode at this point
  //
  
  int nc = group->getNumContexts() - 1;
  int context = group->getLocalIndex();
  
  //
  // register remote function
  //

  cmd_ID0 = hpcxx_register(zero, 32);
  cmd_ID1 = hpcxx_register(first, 33);
  cmd_ID2 = hpcxx_register(second, 34);
  cmd_ID3 = hpcxx_register(third, 35);
  cmd_ID4 = hpcxx_register(fourth, 36);
  cmd_ID5 = hpcxx_register(fifth, 37);
  cmd_ID6 = hpcxx_register(sixth, 38);

  const char* msg = "Da man";
  float pi = 3.1415927;
  int age = 21;
  char letter = 'M';
  short aShort = 86;
  u_short anotherShort = 65;

  hpcxx_barrier();

  if (context == 0) {
    
    // send the command
    
    int rtn[7];
    HPCxx_ContextID context_id = *group->getContextID(1);
    
    hpcxx_invoke(context_id, rtn[0], cmd_ID0);
#ifdef DEBUG
    cerr << "\nFinished invoke of zero, returned " << rtn[0] << endl;
#endif //DEBUG
    check_rtn_val(rtn[0], (char*) "zero");

    hpcxx_invoke(context_id, rtn[1], cmd_ID1, msg);
#ifdef DEBUG
    cerr << "\nFinished invoke of first, returned " << rtn[1] << endl;
#endif //DEBUG
    check_rtn_val(rtn[1], (char*) "first");

    hpcxx_invoke(context_id, rtn[2], cmd_ID2, msg, pi);
#ifdef DEBUG
    cerr << "\nFinished invoke of second, returned " << rtn[2] << endl;
#endif //DEBUG
    check_rtn_val(rtn[2], (char*) "second");

    hpcxx_invoke(context_id, rtn[3], cmd_ID3, msg, pi, age);
#ifdef DEBUG
    cerr << "\nFinished invoke of third, returned " << rtn[3] << endl;
#endif //DEBUG
    check_rtn_val(rtn[3], (char*) "third");

    hpcxx_invoke(context_id, rtn[4], cmd_ID4, msg, pi, age, letter);
#ifdef DEBUG
    cerr << "\nFinished invoke of fourth, returned " << rtn[4] << endl;
#endif //DEBUG
    check_rtn_val(rtn[4], (char*) "fourth");

    hpcxx_invoke(context_id, rtn[5], cmd_ID5, msg, pi, age, letter, aShort);
#ifdef DEBUG
    cerr << "\nFinished invoke of fifth, returned " << rtn[5] << endl;
#endif //DEBUG
    check_rtn_val(rtn[5], (char*) "fifth");

    hpcxx_invoke(context_id, rtn[6], cmd_ID6, msg, pi, age, 
		 letter, aShort, anotherShort);
#ifdef DEBUG
    cerr << "\nFinished invoke of sixth, returned " << rtn[6] << endl;
#endif //DEBUG
    check_rtn_val(rtn[6], (char*) "sixth");

    int sum = 0;
    for (int i=0; i<7; i++) sum += rtn[i];

#ifdef DEBUG
    cerr << "\nThe result of context 0 is " << sum << endl;
#endif //DEBUG

    result0 = sum % 7;
  }
  
  if (context == 1) {

    HPCxx_Sync<int> rtn[7];
    HPCxx_ContextID context_id = *group->getContextID(0);

    hpcxx_ainvoke(context_id, rtn[0], cmd_ID0);
#ifdef DEBUG
    cerr << "\nFinished invoke of zero" << endl;
#endif //DEBUG

    hpcxx_ainvoke(context_id, rtn[1], cmd_ID1, msg);
#ifdef DEBUG
    cerr << "\nFinished invoke of first" << endl;
#endif //DEBUG

    hpcxx_ainvoke(context_id, rtn[2], cmd_ID2, msg, pi);
#ifdef DEBUG
    cerr << "\nFinished invoke of second" << endl;
#endif //DEBUG

    hpcxx_ainvoke(context_id, rtn[3], cmd_ID3, msg, pi, age);
#ifdef DEBUG
    cerr << "\nFinished invoke of third" << endl;
#endif //DEBUG

    hpcxx_ainvoke(context_id, rtn[4], cmd_ID4, msg, pi, age, letter);
#ifdef DEBUG
    cerr << "\nFinished invoke of fourth" << endl;
#endif //DEBUG

    hpcxx_ainvoke(context_id, rtn[5], cmd_ID5, msg, pi, age, letter, aShort);
#ifdef DEBUG
    cerr << "\nFinished invoke of fifth" << endl;
#endif //DEBUG

    hpcxx_ainvoke(context_id, rtn[6], cmd_ID6, msg, pi, age, 
		  letter, aShort, anotherShort);
#ifdef DEBUG
    cerr << "\nFinished invoke of sixth" << endl;
#endif //DEBUG

    int out[7];

    rtn[0].read(out[0]);
    check_rtn_val(out[0], (char*) "zero");

    rtn[1].read(out[1]);
    check_rtn_val(out[1], (char*) "first");

    rtn[2].read(out[2]);
    check_rtn_val(out[2], (char*) "second");

    rtn[3].read(out[3]);
    check_rtn_val(out[3], (char*) "third");

    rtn[4].read(out[4]);
    check_rtn_val(out[4], (char*) "fourth");

    rtn[5].read(out[5]);
    check_rtn_val(out[5], (char*) "fifth");

    rtn[6].read(out[6]);
    check_rtn_val(out[6], (char*) "sixth");

    int sum = 0;
    for (int k=0; k<7; k++) sum += out[k];

#ifdef DEBUG
    cerr << "\nThe result of context 1 is " << sum << endl;
#endif //DEBUG

    result1 = sum % 7;

    }

  //
  // clean up
  //

  hpcxx_barrier();
  hpcxx_exit(group);

#ifdef DEBUG
  cerr << "\nContext " << context << " called hpcxx_exit\n";
  cerr << "\nThe final result is " << result0 + result1 << "\n";
#endif //DEBUG

  return result0 + result1;
}


void hpcxx_pack(HPCxx_Buffer& buffer, const char** strArray, int count)
{
  for (int i = 0; i < count; i++) {
    int length = strlen(strArray[i]);
    buffer.pack(&length, 1);
    buffer.pack((char*) strArray[i], length);
  }
}

void hpcxx_unpack(HPCxx_Buffer& buffer, const char** strArray, int count)
{
  for (int i = 0; i < count; i++) {
    int length;
    buffer.unpack(&length, 1);
    
    //
    // strArray is not really const. It must be declared that way because
    // of the symetry required in the way the pack and unpack functions are
    // used.
    //
    
    char** str = (char**) strArray;
    str[i] = new char[length + 1];
    str[i][length] = '\0';
    buffer.unpack(str[i], length);
  }
}

