#include <HPCxx_Profile.h>
#include <math.h>
#include <assert.h>
#include <iostream.h>
#include <hpcxx_rts.h>
#include <HPCxx_Sync.cc>
#include <hpcxx_invoke.cc>
#include <hpcxx_register.cc>
#include <HPCxx_GlobalPtr.cc>
#include <hpcxx_invokeImpl.cc>
#include <HPCxx_GlobalRefImpl.cc>

// class Bar
// -----------
// This is a simple class which has methods with between zero and six
// parameters.  The methods are used to test the different versions of the 
// hpcxx_minvoke and hpcxx_aminvoke which can call member functions with up to
// six parameters. 
  
class Bar {
private:
  int anInt;
  int id_;

public:
  Bar(int id = 0, int AnInt = 0) : id_(id), anInt(AnInt) { }

  static int registrationID() { return 645; }

  friend void hpcxx_pack (HPCxx_Buffer &buffer, Bar *data, int size = 1);
  friend void hpcxx_unpack (HPCxx_Buffer &buffer, Bar *data, int size = 1);

  int foo0()
  {
    TAU_PROFILE("foo0()", "int ()", TAU_USER);
    return id_;
  }

  int foo1(int bar)
  {
    TAU_PROFILE("foo1()", "int (int )", TAU_USER);
    if (bar == 13)
      return id_;

    cerr << "ERROR in execution of function -> foo1\n";
    return 0;
  }

  int foo2(int bar, char ch)
  {
    TAU_PROFILE("foo2()", "int (int, char )", TAU_USER);
    if ((bar == 13) && (ch == 'a'))
      return id_;

    cerr << "ERROR in execution of function -> foo2\n";
    return 0;
  }

  int foo3(int bar, char ch, short sh)
  { 
    TAU_PROFILE("foo3()", "int (int, char, short )", TAU_USER);
    if ((bar == 13) && (ch == 'a') && (sh == 41))
      return id_;

    cerr << "ERROR in execution of function -> foo3\n";
    return 0;
  }

  int foo4(int bar, char ch, short sh, float fl)
  {
    TAU_PROFILE("foo4()", "int (int, char, short, float )", TAU_USER);
    if ((bar == 13) && (ch == 'a') && (fabs(fl - 2.7182818) < .0000002) &&
	(sh == 41))
      return id_;

    cerr << "ERROR in execution of function -> foo4\n";
    return 0;
  }

  int foo5(int bar, char ch, short sh, float fl, long ln)
  {
    TAU_PROFILE("foo5()", "int (int, char, short, float, long )", TAU_USER);
    if ((bar == 13) && (sh == 41) && (fabs(fl - 2.7182818) < .0000002) &&
	(ch == 'a') && (ln == 235463246))
      return id_;
    
    cerr << "ERROR in execution of function -> foo5\n";
    return 0;
  }

  int foo6(int bar, char ch, short sh, float fl, long ln, u_long ul)
  {
    TAU_PROFILE("foo5()", "int (int, char, short, float, long, u_long )", 
      TAU_USER);
    if ((bar == 13) && (sh == 41) && (fabs(fl - 2.7182818) < .0000002) &&
	(ch == 'a') && (ln == 235463246) && (ul == 546367))
      return id_;
    
    cerr << "ERROR in execution of function -> foo6\n";
    return 0;
  }

};

// Here are the pack and unpack functions for class Bar.

void hpcxx_pack(HPCxx_Buffer &buffer, Bar *data, int size)
{
  TAU_PROFILE("hpcxx_pack()", "void (HPCxx_Buffer &, Bar *, int )", TAU_USER);
  for (int i=0; i<size; i++) {
    hpcxx_pack(buffer, &data[i].id_);
    hpcxx_pack(buffer, &data[i].anInt);
  }
}


void hpcxx_unpack (HPCxx_Buffer &buffer, Bar *data, int size)
{
  TAU_PROFILE("hpcxx_unpack()", "void (HPCxx_Buffer, Bar *, int )", TAU_USER);
  for (int i=0; i<size; i++) {
    hpcxx_unpack(buffer, &data[i].id_);
    hpcxx_unpack(buffer, &data[i].anInt);
  }
}


// A sync variable has been defined to facilitate passing global pointers to
// other contexts.
typedef HPCxx_GlobalPtr<Bar> gptr_to_Bar;
HPCxx_Sync<gptr_to_Bar> a_truly_global_ptr;

// This function passes a global pointer between contexts.
int put_gp(gptr_to_Bar ptr)
{
  TAU_PROFILE("put_gp()", "int (gptr_to_Bar  )", TAU_USER);
  a_truly_global_ptr = ptr;
  return 0;
}

static int result1;
static int result2;

int
main (int argc, char* argv[])
{
  TAU_PROFILE("main()", "int (int, char * )", TAU_DEFAULT);
  HPCxx_Group* group;
  hpcxx_init(argc, argv, group);
  
  int nc = group->getNumContexts() - 1;
  int context = group->getLocalIndex();

  //  nexus_printf("nc = %d, context = %d\n", nc, context);

  // register the member functions
  static int foo_id0 = hpcxx_register(&Bar::foo0, 30);
  static int foo_id1 = hpcxx_register(&Bar::foo1, 31);
  static int foo_id2 = hpcxx_register(&Bar::foo2, 32);
  static int foo_id3 = hpcxx_register(&Bar::foo3, 33);
  static int foo_id4 = hpcxx_register(&Bar::foo4, 34);
  static int foo_id5 = hpcxx_register(&Bar::foo5, 35);
  static int foo_id6 = hpcxx_register(&Bar::foo6, 36);
  static int put_gpid = hpcxx_register(&put_gp, 37);

  int anint = 13;
  char achar = 'a';
  short ashort = 41;
  float afloat = 2.718282;
  long along = 235463246;
  u_long au_long = 546367;
  
  //  nexus_printf("registered functions\n");
  
  hpcxx_barrier();

  if (context == 0)
    {
      int res;

      Bar* x = new Bar(5, 37);
      HPCxx_GlobalPtr<Bar> Bar_ptr(x);

      Bar* y = new Bar(7, 37);
      HPCxx_GlobalPtr<Bar> Bar_ptr2(y);
  
      // pass the global pointers to another context
      hpcxx_invoke(*group->getContextID(1), res, put_gpid, Bar_ptr);
      hpcxx_invoke(*group->getContextID(2), res, put_gpid, Bar_ptr2);
      
    } else if (context == 2) {
      
      HPCxx_Sync<int> result[7];
      gptr_to_Bar Bar_gp = a_truly_global_ptr;
      
      hpcxx_aminvoke(&Bar_gp, result[0], foo_id0);
#ifdef DEBUG
    cerr << "\nFinished invoke of foo0" << endl;
#endif //DEBUG

      hpcxx_aminvoke(&Bar_gp, result[1], foo_id1, anint);
#ifdef DEBUG
    cerr << "\nFinished invoke of foo1" << endl;
#endif //DEBUG

      hpcxx_aminvoke(&Bar_gp, result[2], foo_id2, anint, achar);
#ifdef DEBUG
    cerr << "\nFinished invoke of foo2" << endl;
#endif //DEBUG

      hpcxx_aminvoke(&Bar_gp, result[3], foo_id3, anint, achar, ashort);
#ifdef DEBUG
    cerr << "\nFinished invoke of foo3" << endl;
#endif //DEBUG

      hpcxx_aminvoke(&Bar_gp, result[4], foo_id4, anint, achar, ashort, 
		     afloat);
#ifdef DEBUG
    cerr << "\nFinished invoke of foo4" << endl;
#endif //DEBUG

      hpcxx_aminvoke(&Bar_gp, result[5], foo_id5, anint, achar, ashort, 
		     afloat, along);
#ifdef DEBUG
    cerr << "\nFinished invoke of foo5" << endl;
#endif //DEBUG

      hpcxx_aminvoke(&Bar_gp, result[6], foo_id6, anint, achar, ashort, 
		     afloat, along, au_long);
#ifdef DEBUG
    cerr << "\nFinished invoke of foo6" << endl;
#endif //DEBUG

      int values[7];
      
      // wait for the methods to finish executing
      for (int k=0; k<7; k++) result[k].read(values[k]);
      
      // Add up the returned values to determine if each method was executed.
      int sum = 0;
      for (int i=0; i<7; i++) sum += values[i];
      
#ifdef DEBUG
    cerr << "\nThe result of context 2 is " << sum << endl;
#endif //DEBUG

      // we want to return 0 (for success)
      result2 = sum % (7*7);

    } else { // context 1
      
      int result[7];
      gptr_to_Bar Bar_gp = a_truly_global_ptr;

      hpcxx_minvoke(&Bar_gp, result[0], foo_id0);
#ifdef DEBUG
    cerr << "\nFinished invoke of foo0, returned " << result[0] << endl;
#endif //DEBUG

      hpcxx_minvoke(&Bar_gp, result[1], foo_id1, anint);
#ifdef DEBUG
    cerr << "\nFinished invoke of foo1, returned " << result[1] << endl;
#endif //DEBUG

      hpcxx_minvoke(&Bar_gp, result[2], foo_id2, anint, achar);
#ifdef DEBUG
    cerr << "\nFinished invoke of foo2, returned " << result[2] << endl;
#endif //DEBUG

      hpcxx_minvoke(&Bar_gp, result[3], foo_id3, anint, achar, ashort);
#ifdef DEBUG
    cerr << "\nFinished invoke of foo3, returned " << result[3] << endl;
#endif //DEBUG

      hpcxx_minvoke(&Bar_gp, result[4], foo_id4, anint, achar, ashort, 
		    afloat);
#ifdef DEBUG
    cerr << "\nFinished invoke of foo4, returned " << result[4] << endl;
#endif //DEBUG

      hpcxx_minvoke(&Bar_gp, result[5], foo_id5, anint, achar, ashort, 
		    afloat, along);
#ifdef DEBUG
    cerr << "\nFinished invoke of foo5, returned " << result[5] << endl;
#endif //DEBUG

      hpcxx_minvoke(&Bar_gp, result[6], foo_id6, anint, achar, ashort, 
		    afloat, along, au_long);
#ifdef DEBUG
    cerr << "\nFinished invoke of foo6, returned " << result[6] << endl;
#endif //DEBUG

      // Add up the returned values to determine if each method was executed.
      int sum = 0;
      for (int i=0; i<7; i++) sum += result[i];

#ifdef DEBUG
    cerr << "\nThe result of context 1 is " << sum << endl;
#endif //DEBUG

      result1 = sum % (7*5);

    }

  // synchronize execution of the 3 contexts.
  hpcxx_barrier();
  hpcxx_exit(group);
  
#ifdef DEBUG
  cerr << "\nContext " << context << " called hpcxx_exit\n";
  cerr << "\nThe final result is " << result1 + result2 + 0 << "\n";
#endif //DEBUG

  return (result1 + result2 + 0);
}

template void hpcxx_unpack(HPCxx_Buffer &, HPCxx_GlobalPtr<Bar> *, int);

