#include <hpcxx_rts.h>
#include <ORB.h>

#include "Remote.h"
#include "RemoteStub.h"

FILE* openDebugFile(int context, HPCxx_Group*);	// for writing debug info.

static void registerMemberFunctions(void)
{
  hpcxx_curID_s = RemoteStub::registerMemberFunctions(hpcxx_curID_s);
}

int main(int argc, char* argv[])
{
  float sum0 = 0.;
  float sum1 = 0.;
  RemoteIntr* remote  = NULL;
  RemoteIntr* remote0 = NULL;
  RemoteIntr* remote1 = NULL;

  registerMemberFunctions();

  //
  // initialize HPC++
  //
  HPCxx_Group* group;
  hpcxx_init(argc, argv, group);

  //
  // all processes are executing in SPMD mode at this point
  //

  int nc = group->getNumContexts() - 1;
  int context = group->getLocalIndex();
  HPCxx_ContextID* contextID = group->getContextID(context);
  FILE* fd = openDebugFile(context, group);

  //
  // obtain a reference to the global name server
  //
  NameContext_ptr nameServer = ORB::resolve_initial_references(group);

  //
  // Do context specific initialization
  //
  if (context == 0) {
    remote = new Remote(1, 9.);			// create local object
    nameServer -> bind("Remote0", remote);	// and provide remote access
  } else {
    remote = new Remote(2, 3.);			// create local object
    nameServer -> bind("Remote1", remote);	// and provide remote access
  }

  // wait for all contexts to finish initialization
  hpcxx_barrier();

  //
  // obtain references to the remote objects
  //
  remote0 = RemoteStub::_narrow(nameServer->resolve("Remote0"));
  remote1 = RemoteStub::_narrow(nameServer->resolve("Remote1"));

  //
  // invoke a method on the remote objects
  //
  sum0 = remote0->mixedSum(1, 3.);
  sum1 = remote1->mixedSum(2, 4.);

  fprintf(fd, "sum0 = %f,  sum1 = %f\n", sum0, sum1);

  //
  // clean up
  //

  delete remote;
  delete remote0;
  delete remote1;
  delete nameServer;

  fclose(fd);
  hpcxx_barrier();

  if (context == 0) hpcxx_exit(group);
  return 0;
}


FILE* openDebugFile(int context, HPCxx_Group* group)
{
  char file[64];
  sprintf(file, "/tmp/debug_%d", context);
  FILE* fd = fopen(file, "w");
  if (fd == NULL) {
    printf("context[%d], failed to open debug file\n", context);
    hpcxx_exit(group);
    exit(1);
  }
  return fd;
}

