//
// testSync tests the functionality of the HPCxx_Sync class.
//

#include <iostream.h>
#include <math.h>
#include <hpcxx_rts.h>
#include <HPCxx_Sync.cc>

int             error = 0;	// set to non-zero if error occurs
HPCxx_Sync<int> syncVar;
HPCxx_Mutex     mutex;

//
// This class writes to the sync variable
//
class WriteThread : public HPCxx_Thread 
{ 
  public:  WriteThread(void): HPCxx_Thread("Writer Thread")
    {
    }
  public:  void run()
    {
      yield();

      mutex.lock();
      cerr << "    WriteThread::run(): about to write " << endl;
      mutex.unlock();

      int iVal = 3;
      syncVar.write(iVal);

      mutex.lock();
      cerr << "    WriteThread::run(): did write " << endl;
      mutex.unlock();

    }
};


//
// This class watches for the counter to reach a threshold
//
class ReadThread : public HPCxx_Thread 
{ 
  public:  ReadThread(void): HPCxx_Thread("Reader Thread")
    {
    }
  public:  void run()
    {
      mutex.lock();
      cerr << "    ReadThread::run(): about to read " << endl;
      mutex.unlock();

      int iVal;
      syncVar.read(iVal);
      if (iVal != 3) error = 3;

      mutex.lock();
      cerr << "    ReadThread::run(): read value = " << iVal << endl;
      mutex.unlock();
    }
};


int main(int argc, char* argv[])
{
  int i;
  HPCxx_Group* g;
  hpcxx_init(argc, argv, g);

  // lets just try reading and writing from one thread first

  HPCxx_Sync<float> sync;
  float rVal = 3.3;
  sync.write(rVal);
  rVal = 0.;
  sync.read(rVal);
  if (fabs(rVal - 3.3) > 0.0000001) error = 33;

  cerr << "    main: read value = " << rVal << endl;

  int nThreads = 2;		// will want multiple readers eventually
  HPCxx_Thread* thread[2];

  //
  // create threads
  //
  thread[0] = new ReadThread();
  thread[1] = new WriteThread();

  //
  // start threads
  //
  for (i = 0; i < nThreads; i++) {
    thread[i]->start();
  }

  //
  // wait for threads to finish
  //
  for (i = 0; i < nThreads; i++) {
    thread[i]->join();
  }

  //
  // clean up
  //
  for (i = 0; i < nThreads; i++) {
    delete thread[i];
  }

  cerr << endl;
  if (error != 0) {
    cerr << "ERROR, HPCxx_Sync test did not complete correctly, error = "
	 << error << endl;
    return error;
  } else {
    cout << "Successfully completed HPCxx_Sync test" << endl;
  }

  return hpcxx_exit(g);
}
